//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __CODIMENSIONAL_GEOMETRY_SPATIAL_PARTITION__
#define __CODIMENSIONAL_GEOMETRY_SPATIAL_PARTITION__
#include <PhysBAM_Tools/Data_Structures/HASHTABLE.h>
#include <PhysBAM_Tools/Arrays/INDIRECT_ARRAY.h>
#include <PhysBAM_Tools/Grids_Uniform/GRID.h>
#include <PhysBAM_Tools/Grids_Uniform/UNIFORM_GRID_ITERATOR_CELL.h>
#include <PhysBAM_Dynamics/Particles/CODIMENSIONAL_FLUID_PARTICLES.h>
#include "../Meshing/CODIMENSIONAL_MESH.h"

namespace PhysBAM
{
class CODIMENSIONAL_ELEMENT_INDEX
{public:
    int index;int type;int object_id;
    CODIMENSIONAL_ELEMENT_INDEX(int index_input=0,int type_input=0,int object_id_input=0):index(index_input),type(type_input),object_id(object_id_input){}
    bool operator == (const CODIMENSIONAL_ELEMENT_INDEX& e) const {return index==e.index&&type==e.type&&object_id==e.object_id;}
};

template<class TV>
class CODIMENSIONAL_GEOMETRY_SPATIAL_PARTITION
{
    typedef typename TV::SCALAR T;typedef typename TV::template REBIND<int>::TYPE TV_INT;
    typedef VECTOR<T,2> TV2;typedef VECTOR<T,3> TV3;typedef VECTOR<int,2> TV_INT2;typedef VECTOR<int,3> TV_INT3;typedef VECTOR<int,4> TV_INT4;
public:
    T voxel_size,one_over_voxel_size;
    HASHTABLE<TV_INT,ARRAY<CODIMENSIONAL_ELEMENT_INDEX> > spatial_hashtable;

    void Clear(){spatial_hashtable.Remove_All();}
    void Set_Voxel_Size(const T voxel_size_input){voxel_size=voxel_size_input;one_over_voxel_size=(T)1/voxel_size;}
    TV_INT Voxel(const TV& location) const {return TV_INT(floor(location*one_over_voxel_size));}
    RANGE<TV_INT> Voxel(const RANGE<TV>& box) const {return RANGE<TV_INT>(Voxel(box.min_corner),Voxel(box.max_corner));}
    RANGE<TV_INT> Voxel_With_One_Ring_Neighbors(const TV& location) const {TV_INT voxel=Voxel(location);return RANGE<TV_INT>(voxel-TV_INT::Constant_Vector(1),voxel+TV_INT::Constant_Vector(1));}

    void Add_To_Cell(const TV_INT& voxel,const CODIMENSIONAL_ELEMENT_INDEX& e)
    {ARRAY<CODIMENSIONAL_ELEMENT_INDEX>& elements=spatial_hashtable.Get_Or_Insert(voxel,ARRAY<CODIMENSIONAL_ELEMENT_INDEX>());elements.Append(e);}
    void Add_To_Cell(const RANGE<TV_INT>& voxel_range,const CODIMENSIONAL_ELEMENT_INDEX& e)
    {GRID<TV> unused;for(UNIFORM_GRID_ITERATOR_CELL<TV> iter(unused,voxel_range);iter.Valid();iter.Next()){Add_To_Cell(iter.Cell_Index(),e);}}

    void Get_From_Cell(const TV_INT& voxel,ARRAY<CODIMENSIONAL_ELEMENT_INDEX>& elements)
    {ARRAY<CODIMENSIONAL_ELEMENT_INDEX>* elements_in_voxel=spatial_hashtable.Get_Pointer(voxel);
     if(elements_in_voxel){for(int i=1;i<=(*elements_in_voxel).m;i++)elements.Append_Unique((*elements_in_voxel)(i));}}
    void Get_From_Cell(const RANGE<TV_INT>& voxel_range,ARRAY<CODIMENSIONAL_ELEMENT_INDEX>& elements)
    {GRID<TV> unused;for(UNIFORM_GRID_ITERATOR_CELL<TV> iter(unused,voxel_range);iter.Valid();iter.Next()){Get_From_Cell(iter.Cell_Index(),elements);}}
    
    void Get_From_Cell(const TV_INT& voxel,const int object_id,ARRAY<CODIMENSIONAL_ELEMENT_INDEX>& elements)
    {ARRAY<CODIMENSIONAL_ELEMENT_INDEX>* elements_in_voxel=spatial_hashtable.Get_Pointer(voxel);
    if(elements_in_voxel){for(int i=1;i<=(*elements_in_voxel).m;i++)if((*elements_in_voxel)(i).object_id==object_id)elements.Append_Unique((*elements_in_voxel)(i));}}
    void Get_From_Cell(const RANGE<TV_INT>& voxel_range,const int object_id,ARRAY<CODIMENSIONAL_ELEMENT_INDEX>& elements)
    {GRID<TV> unused;for(UNIFORM_GRID_ITERATOR_CELL<TV> iter(unused,voxel_range);iter.Valid();iter.Next()){Get_From_Cell(iter.Cell_Index(),object_id,elements);}}

    void Get_From_Cell_With_One_Ring_Neighbors(const TV& pos,ARRAY<CODIMENSIONAL_ELEMENT_INDEX>& elements)
    {RANGE<TV_INT> voxel_range=Voxel_With_One_Ring_Neighbors(pos);Get_From_Cell(voxel_range,elements);}
    void Get_From_Cell_With_One_Ring_Neighbors(const TV& pos,const int object_id,ARRAY<CODIMENSIONAL_ELEMENT_INDEX>& elements)
    {RANGE<TV_INT> voxel_range=Voxel_With_One_Ring_Neighbors(pos);Get_From_Cell(voxel_range,object_id,elements);}

    bool Remove_From_Cell(const TV_INT& voxel,const CODIMENSIONAL_ELEMENT_INDEX& e)
    {ARRAY<CODIMENSIONAL_ELEMENT_INDEX>* element_in_voxel=spatial_hashtable.Get_Pointer(voxel);
    if(element_in_voxel){if(int found=(*element_in_voxel).Find(e)){element_in_voxel->Remove_Index_Lazy(found);return true;}}return false;}
    bool Remove_From_Cell(const RANGE<TV_INT>& voxel_range,const CODIMENSIONAL_ELEMENT_INDEX& e)
    {GRID<TV> unused;bool removed=false;for(UNIFORM_GRID_ITERATOR_CELL<TV> iter(unused,voxel_range);iter.Valid();iter.Next()){if(Remove_From_Cell(iter.Cell_Index(),e))removed=true;}return removed;}

    template<class TV_INTN> void Insert_Into_Hashtable(ARRAY_VIEW<TV>& X,const TV_INTN& e,const int index,const int object_id,/*result*/RANGE<TV_INT>& voxel_range,const T thickness=(T)0)
    {INDIRECT_ARRAY<ARRAY_VIEW<TV>,TV_INTN&> indir_X(X,e);RANGE<TV> box=RANGE<TV>::Bounding_Box(indir_X).Thickened((T).5*thickness);voxel_range=Voxel(box);Add_To_Cell(voxel_range,CODIMENSIONAL_ELEMENT_INDEX(index,(SIMPLEX_TYPE)TV_INTN::m,object_id));}
};
}

#endif
