//#####################################################################
// Copyright 2014, Ed Quigley
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################

#include <PhysBAM_Geometry/Basic_Geometry/BOX.h>
#include <PhysBAM_Geometry/Basic_Geometry/CYLINDER.h>
#include <PhysBAM_Geometry/Basic_Geometry/RAY.h>
#include <PhysBAM_Geometry/Basic_Geometry/SEGMENT_3D.h>
#include <PhysBAM_Geometry/Basic_Geometry/TETRAHEDRON.h>
#include <PhysBAM_Geometry/Basic_Geometry/TRIANGLE_3D.h>
#include <PhysBAM_Geometry/Basic_Geometry_Intersections/SEGMENT_3D_TRIANGLE_3D_INTERSECTION.h>
#include <PhysBAM_Geometry/Basic_Geometry_Intersections/TRIANGLE_3D_TRIANGLE_3D_INTERSECTION.h>
#include <PhysBAM_Geometry/Basic_Geometry_Intersections/CYLINDER_TRIANGLE_3D_INTERSECTION.h>
#include <PhysBAM_Geometry/Basic_Geometry_Intersections/RAY_TRIANGLE_3D_INTERSECTION.h>
#include <PhysBAM_Tools/Data_Structures/HASHTABLE.h>
#include <PhysBAM_Tools/Data_Structures/PAIR.h>
#include <PhysBAM_Tools/Interpolation/LINEAR_INTERPOLATION.h>

#include <algorithm>
#include <map>
#include <vector>
using std::map;
using std::pair;

#include "CODIMENSIONAL_INTERSECTION.h"

using namespace PhysBAM;

template<class T>
bool PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Cylinder_Triangle_Thickness_Intersection(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT2 simplex1,TV_INT3 simplex2)
{
    CYLINDER<T> cylinder(X1(simplex1(1)),X1(simplex1(2)),(T)0.5*((T)0.5*t1(simplex1(1))+(T)0.5*t1(simplex1(2))));//0.5s because cylinder radius is half of average thickness

    const int& tri1=simplex2(1),tri2=simplex2(2),tri3=simplex2(3);
    TV triangle_normal=TV::Cross_Product(X2(tri2)-X2(tri1),X2(tri3)-X2(tri1)).Normalized();
    TV t1_upper(X2(tri1)+(T)0.5*t2(tri1)*triangle_normal),
       t1_lower(X2(tri1)-(T)0.5*t2(tri1)*triangle_normal),
       t2_upper(X2(tri2)+(T)0.5*t2(tri2)*triangle_normal),
       t2_lower(X2(tri2)-(T)0.5*t2(tri2)*triangle_normal),
       t3_upper(X2(tri3)+(T)0.5*t2(tri3)*triangle_normal),
       t3_lower(X2(tri3)-(T)0.5*t2(tri3)*triangle_normal);
    ARRAY<TRIANGLE_3D<T> > tris;
    tris.Append(TRIANGLE_3D<T>(t1_upper,t2_upper,t3_upper));
    tris.Append(TRIANGLE_3D<T>(t1_lower,t2_lower,t3_lower));
    tris.Append(TRIANGLE_3D<T>(t1_upper,t1_lower,t2_lower));
    tris.Append(TRIANGLE_3D<T>(t1_upper,t2_lower,t2_upper));
    tris.Append(TRIANGLE_3D<T>(t2_upper,t2_lower,t3_lower));
    tris.Append(TRIANGLE_3D<T>(t2_upper,t3_lower,t3_upper));
    tris.Append(TRIANGLE_3D<T>(t3_upper,t3_lower,t1_upper));
    tris.Append(TRIANGLE_3D<T>(t3_upper,t1_lower,t1_upper));

    bool intersection=false;
    for(int i=1;i<=tris.m&&!intersection;i++){
        intersection=intersection||INTERSECTION::Intersects(cylinder,tris(i));
    }

    return intersection;
}

template<class T>
bool PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Cylinder_Tetrahedron_Intersection(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT2 simplex1,TV_INT4 simplex2)
{
    CYLINDER<T> cylinder(X1(simplex1(1)),X1(simplex1(2)),(T)0.5*(t1(simplex1(1))+t1(simplex1(2))));
    const int& tet1=simplex2(1),tet2=simplex2(2),tet3=simplex2(3),tet4=simplex2(4);

    ARRAY<TRIANGLE_3D<T> > triangles;
    triangles.Append(TRIANGLE_3D<T>(X2(tet1),X2(tet2),X2(tet3)));
    triangles.Append(TRIANGLE_3D<T>(X2(tet1),X2(tet2),X2(tet4)));
    triangles.Append(TRIANGLE_3D<T>(X2(tet1),X2(tet3),X2(tet4)));
    triangles.Append(TRIANGLE_3D<T>(X2(tet2),X2(tet3),X2(tet4)));
    bool intersection=false;
    for(int i=1;i<=triangles.m&&!intersection;i++){
        intersection=intersection||INTERSECTION::Intersects(cylinder,triangles(i));
    }
    return intersection;
}

template<class T>
bool PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Segment_Capsule_Intersection(const TV& s1,const TV& s2,const ARRAY_VIEW<TV>& X,const ARRAY_VIEW<T>& t,TV_INT2 simplex)
{
    T distance=Distance_Segment_To_Segment(s1,s2,X(simplex(1)),X(simplex(2)));
    T thickness=(T)0.5*(t(simplex(1)),t(simplex(2)));
    distance-=(T)0.5*thickness;
    return (distance<=(T)0);
}

template<class T>
bool PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Prism_Prism_Intersection(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT3 simplex1,TV_INT3 simplex2)
{
    TRIANGLE_3D<T> tri1(X1(simplex1(1)),X1(simplex1(2)),X1(simplex1(3))),tri2(X2(simplex2(1)),X2(simplex2(2)),X2(simplex2(3)));

    bool intersect=false;
    if(INTERSECTION::Intersects(tri1,tri2)){intersect=true;}
    else{
        T distance=Distance_Triangle_To_Triangle(tri1,tri2);
        T max_t1=max(t1(simplex1(1)),max(t1(simplex1(2)),t1(simplex1(3))));
        T max_t2=max(t2(simplex2(1)),max(t2(simplex2(2)),t2(simplex2(3))));
        if(distance<=(T)0.5*max_t1+(T)0.5*max_t2){//early exit test
            const TV& r_normal=tri1.normal;
            const TV& s_normal=tri2.normal;
            int r1=simplex1(1),r2=simplex1(2),r3=simplex1(3);
            int s1=simplex2(1),s2=simplex2(2),s3=simplex2(3);
            TV r1_upper(X1(r1)+(T)0.5*t1(r1)*r_normal);
            TV r1_lower(X1(r1)-(T)0.5*t1(r1)*r_normal);
            TV r2_upper(X1(r2)+(T)0.5*t1(r2)*r_normal);
            TV r2_lower(X1(r2)-(T)0.5*t1(r2)*r_normal);
            TV r3_upper(X1(r3)+(T)0.5*t1(r3)*r_normal);
            TV r3_lower(X1(r3)-(T)0.5*t1(r3)*r_normal);

            TV s1_upper(X2(s1)+(T)0.5*t2(s1)*s_normal);
            TV s1_lower(X2(s1)-(T)0.5*t2(s1)*s_normal);
            TV s2_upper(X2(s2)+(T)0.5*t2(s2)*s_normal);
            TV s2_lower(X2(s2)-(T)0.5*t2(s2)*s_normal);
            TV s3_upper(X2(s3)+(T)0.5*t2(s3)*s_normal);
            TV s3_lower(X2(s3)-(T)0.5*t2(s3)*s_normal);

            ARRAY<TRIANGLE_3D<T> > tris1,tris2;
            tris1.Append(TRIANGLE_3D<T>(r1_upper,r2_upper,r3_upper));
            tris1.Append(TRIANGLE_3D<T>(r1_lower,r2_lower,r3_lower));
            tris1.Append(TRIANGLE_3D<T>(r1_upper,r1_lower,r2_lower));
            tris1.Append(TRIANGLE_3D<T>(r1_upper,r2_lower,r2_upper));
            tris1.Append(TRIANGLE_3D<T>(r2_upper,r2_lower,r3_lower));
            tris1.Append(TRIANGLE_3D<T>(r2_upper,r3_lower,r3_upper));
            tris1.Append(TRIANGLE_3D<T>(r3_upper,r3_lower,r1_upper));
            tris1.Append(TRIANGLE_3D<T>(r3_upper,r1_lower,r1_upper));

            tris2.Append(TRIANGLE_3D<T>(s1_upper,s2_upper,s3_upper));
            tris2.Append(TRIANGLE_3D<T>(s1_lower,s2_lower,s3_lower));
            tris2.Append(TRIANGLE_3D<T>(s1_upper,s1_lower,s2_lower));
            tris2.Append(TRIANGLE_3D<T>(s1_upper,s2_lower,s2_upper));
            tris2.Append(TRIANGLE_3D<T>(s2_upper,s2_lower,s3_lower));
            tris2.Append(TRIANGLE_3D<T>(s2_upper,s3_lower,s3_upper));
            tris2.Append(TRIANGLE_3D<T>(s3_upper,s3_lower,s1_upper));
            tris2.Append(TRIANGLE_3D<T>(s3_upper,s1_lower,s1_upper));

            for(int i=1;i<=tris1.m&&!intersect;i++){
                for(int j=1;j<=tris2.m&&!intersect;j++){
                    intersect=INTERSECTION::Intersects(tris1(i),tris2(j));
                }
            }
        }
    }
    return intersect;
}

template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,int simplex1,int simplex2)
{
    return Distance_Point_To_Point(X1,X2,t1,t2,simplex1,simplex2);
}

template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,int simplex1,TV_INT2 simplex2)
{
    return Distance_Point_To_Capsule(X1,X2,t1,t2,simplex1,simplex2);
}

template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,int simplex1,TV_INT3 simplex2)
{
    return Distance_Point_To_Triangle(X1,X2,t1,t2,simplex1,simplex2);
}

template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,int simplex1,TV_INT4 simplex2)
{
    return Distance_Point_To_Tetrahedron(X1,X2,t1,t2,simplex1,simplex2);
}

template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT2 simplex1,int simplex2)
{
    return Distance_Capsule_To_Point(X1,X2,t1,t2,simplex1,simplex2);
}

template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT2 simplex1,TV_INT2 simplex2)
{
    return Distance_Capsule_To_Capsule(X1,X2,t1,t2,simplex1,simplex2);
}

template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT2 simplex1,TV_INT3 simplex2)
{
    return Distance_Capsule_To_Triangle(X1,X2,t1,t2,simplex1,simplex2);
}

template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT2 simplex1,TV_INT4 simplex2)
{
    return Distance_Capsule_To_Tetrahedron(X1,X2,t1,t2,simplex1,simplex2);
}

template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT3 simplex1,int simplex2)
{
    return Distance_Triangle_To_Point(X1,X2,t1,t2,simplex1,simplex2);
}

template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT3 simplex1,TV_INT2 simplex2)
{
    return Distance_Triangle_To_Capsule(X1,X2,t1,t2,simplex1,simplex2);
}

template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT3 simplex1,TV_INT3 simplex2)
{
    return Distance_Triangle_To_Triangle_Thickness(X1,X2,t1,t2,simplex1,simplex2);
}

template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT3 simplex1,TV_INT4 simplex2)
{
    return Distance_Triangle_To_Tetrahedron_Thickness(X1,X2,t1,t2,simplex1,simplex2);
}

template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT4 simplex1,int simplex2)
{
    return Distance_Tetrahedron_To_Point(X1,X2,t1,t2,simplex1,simplex2);
}

template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT4 simplex1,TV_INT2 simplex2)
{
    return Distance_Tetrahedron_To_Capsule(X1,X2,t1,t2,simplex1,simplex2);
}

template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT4 simplex1,TV_INT3 simplex2)
{
    return Distance_Tetrahedron_To_Triangle_Thickness(X1,X2,t1,t2,simplex1,simplex2);
}

template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT4 simplex1,TV_INT4 simplex2)
{
    return Distance_Tetrahedron_To_Tetrahedron(X1,X2,t1,t2,simplex1,simplex2);
}


template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Point_To_Point(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,int point_element1,int point_element2)
{
    return (X1(point_element1)-X2(point_element2)).Magnitude()-(T)0.5*t1(point_element1)-(T)0.5*t2(point_element2);
}

////////////////////////////////////////////////////////////
// Point intersection functions
////////////////////////////////////////////////////////////
template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Point_To_Capsule(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,int point_element,TV_INT2 segment_element)
{
    int segment_particle1=segment_element(1),segment_particle2=segment_element(2);
    T t=(T)0;
    T distance;
    TV closest=Closest_Point_On_Line_Segment(X1(point_element),X2(segment_particle1),X2(segment_particle2),t);
    if(t<(T)0||t>(T)1){
        TV plane_normal=(t<(T)0?(T)-1:(T)1)*(X2(segment_particle2)-X2(segment_particle1)).Normalized();
        TV circle_center=X2(t<(T)0?segment_particle1:segment_particle2);
        T circle_radius=(T)0.5*t2(t<(T)0?segment_particle1:segment_particle2);
        closest=Closest_Point_On_Plane(X1(point_element),closest,plane_normal);
        if((closest-circle_center).Magnitude()>circle_radius)
            closest=circle_center+circle_radius*(closest-circle_center).Normalized();
        distance=(X1(point_element)-closest).Magnitude();
        distance-=(T)0.5*t1(point_element);
    }
    else{
        distance=(X1(point_element)-closest).Magnitude();
        distance-=(T)0.5*t1(point_element);
        distance-=(T)0.5*LINEAR_INTERPOLATION<T,T>::Linear(t2(segment_particle1),t2(segment_particle2),t);
    }
    return distance;
}

template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Point_To_Triangle(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,int point_element,TV_INT3 triangle_element)
{
    int p1=triangle_element(1),
        p2=triangle_element(2),
        p3=triangle_element(3);
    TRIANGLE_3D<T> triangle(X2(p1),X2(p2),X2(p3));
    TV point_on_plane=Closest_Point_On_Plane(X1(point_element),X2(p1),triangle.normal.Normalized());
    TV weights;
    TV closest_point=triangle.Closest_Point(X1(point_element),weights);
    //TODO: verify correctness
    T triangle_thickness=(T)0.5*(weights(1)*t2(p1)+weights(2)*t2(p2)+weights(3)*t2(p3));
    T distance_to_plane=(X1(point_element)-point_on_plane).Magnitude();
    triangle_thickness=min(triangle_thickness,distance_to_plane);
    T sign=(TV::Dot_Product(X1(point_element)-point_on_plane,triangle.normal)>(T)0?(T)1:(T)-1);
    closest_point+=sign*triangle_thickness*triangle.normal;
    T distance=(X1(point_element)-closest_point).Magnitude();
    distance-=(T)0.5*t1(point_element);
    return distance;
}

template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Point_To_Tetrahedron(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,int point_element,TV_INT4 tetrahedron_element)
{
    int p1=tetrahedron_element(1),
        p2=tetrahedron_element(2),
        p3=tetrahedron_element(3),
        p4=tetrahedron_element(4);
    TETRAHEDRON<T> tet(X2(p1),X2(p2),X2(p3),X2(p4));
    TV weights;//unused--thickness is not considered for tetrahedra
    TV closest_point=tet.Closest_Point(X1(point_element),weights);
    T distance=(X1(point_element)-closest_point).Magnitude();
    distance-=(T)0.5*t1(point_element);
    return distance;
}

////////////////////////////////////////////////////////////
// Segment intersection functions
////////////////////////////////////////////////////////////
template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Capsule_To_Point(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT2 segment_element,int point_element)
{
    return Distance_Point_To_Capsule(X2,X1,t2,t1,point_element,segment_element);//note order of array views must be switched
}

//assumes constant thickness (uses the average thickness of each capsule)
template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Capsule_To_Capsule(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT2 segment_element1,TV_INT2 segment_element2)
{
    int s1a=segment_element1(1),
        s1b=segment_element1(2),
        s2a=segment_element2(1),
        s2b=segment_element2(2);
    T distance=Distance_Segment_To_Segment(X1(s1a),X1(s1b),X2(s2a),X2(s2b));
    T thickness1=(T)0.5*(t1(s1a)+t1(s1b));
    T thickness2=(T)0.5*(t2(s2a)+t2(s2b));
    distance-=(T)0.5*(thickness1+thickness2);
    return distance;
}

//assumes constant thickness for capsule
template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Capsule_To_Triangle(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT2 segment_element,TV_INT3 triangle_element)
{
    int s1=segment_element(1),
        s2=segment_element(2),
        tri1=triangle_element(1),
        tri2=triangle_element(2),
        tri3=triangle_element(3);
    SEGMENT_3D<T> segment(X1(s1),X1(s2));
    TV triangle_normal=TV::Cross_Product(X2(tri2)-X2(tri1),X2(tri3)-X2(tri1)).Normalized();

    TV t1_upper(X2(tri1)+(T)0.5*t2(tri1)*triangle_normal),
       t1_lower(X2(tri1)-(T)0.5*t2(tri1)*triangle_normal),
       t2_upper(X2(tri2)+(T)0.5*t2(tri2)*triangle_normal),
       t2_lower(X2(tri2)-(T)0.5*t2(tri2)*triangle_normal),
       t3_upper(X2(tri3)+(T)0.5*t2(tri3)*triangle_normal),
       t3_lower(X2(tri3)-(T)0.5*t2(tri3)*triangle_normal);
    ARRAY<TRIANGLE_3D<T> > tris;
    tris.Append(TRIANGLE_3D<T>(t1_upper,t2_upper,t3_upper));
    tris.Append(TRIANGLE_3D<T>(t1_lower,t2_lower,t3_lower));
    tris.Append(TRIANGLE_3D<T>(t1_upper,t1_lower,t2_lower));
    tris.Append(TRIANGLE_3D<T>(t1_upper,t2_lower,t2_upper));
    tris.Append(TRIANGLE_3D<T>(t2_upper,t2_lower,t3_lower));
    tris.Append(TRIANGLE_3D<T>(t2_upper,t3_lower,t3_upper));
    tris.Append(TRIANGLE_3D<T>(t3_upper,t3_lower,t1_upper));
    tris.Append(TRIANGLE_3D<T>(t3_upper,t1_lower,t1_upper));
    T min_distance=Distance_Segment_To_Triangle(segment,tris(1));
    for(int i=2;i<=tris.m;i++)min_distance=min(min_distance,Distance_Segment_To_Triangle(segment,tris(i)));
    T thickness=(T)0.5*(t1(s1)+t1(s2));
    min_distance-=(T)0.5*thickness;
    return min_distance;
}

//only expands triangles whose dot product with the expansion_normal parameter is positive
template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Capsule_To_Triangle(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT2 segment_element,TV_INT3 triangle_element,const TV& expansion_normal)
{
    int s1=segment_element(1),
        s2=segment_element(2),
        tri1=triangle_element(1),
        tri2=triangle_element(2),
        tri3=triangle_element(3);
    SEGMENT_3D<T> segment(X1(s1),X1(s2));
    TV triangle_normal=TV::Cross_Product(X2(tri2)-X2(tri1),X2(tri3)-X2(tri1)).Normalized();
    T dot=TV::Dot_Product(triangle_normal,expansion_normal);
    T upper_factor=(dot>(T)0?(T)1:(T)0);
    T lower_factor=(dot<(T)0?(T)-1:(T)0);
    TV t1_upper(X2(tri1)+upper_factor*t2(tri1)*triangle_normal),
       t1_lower(X2(tri1)+lower_factor*t2(tri1)*triangle_normal),
       t2_upper(X2(tri2)+upper_factor*t2(tri2)*triangle_normal),
       t2_lower(X2(tri2)+lower_factor*t2(tri2)*triangle_normal),
       t3_upper(X2(tri3)+upper_factor*t2(tri3)*triangle_normal),
       t3_lower(X2(tri3)+lower_factor*t2(tri3)*triangle_normal);
    ARRAY<TRIANGLE_3D<T> > tris;
    tris.Append(TRIANGLE_3D<T>(t1_upper,t2_upper,t3_upper));
    tris.Append(TRIANGLE_3D<T>(t1_lower,t2_lower,t3_lower));
    tris.Append(TRIANGLE_3D<T>(t1_upper,t1_lower,t2_lower));
    tris.Append(TRIANGLE_3D<T>(t1_upper,t2_lower,t2_upper));
    tris.Append(TRIANGLE_3D<T>(t2_upper,t2_lower,t3_lower));
    tris.Append(TRIANGLE_3D<T>(t2_upper,t3_lower,t3_upper));
    tris.Append(TRIANGLE_3D<T>(t3_upper,t3_lower,t1_upper));
    tris.Append(TRIANGLE_3D<T>(t3_upper,t1_lower,t1_upper));
    T min_distance=Distance_Segment_To_Triangle(segment,tris(1));
    for(int i=2;i<=tris.m;i++)min_distance=min(min_distance,Distance_Segment_To_Triangle(segment,tris(i)));
    T thickness=(T)0.5*(t1(s1)+t1(s2));
    min_distance-=(T)0.5*thickness;
    return min_distance;
}

//assumes constant thickness for capsule
template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Capsule_To_Tetrahedron(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT2 segment_element,TV_INT4 tetrahedron_element)
{
    int s1=segment_element(1),
        s2=segment_element(2);
    T distance=Distance_Segment_To_Tetrahedron(X1,X2,t1,t2,segment_element,tetrahedron_element);
    T thickness=(T)0.5*(t1(s1)+t1(s2));
    distance-=(T)0.5*thickness;
    return distance;
}

////////////////////////////////////////////////////////////
// Triangle intersection functions
////////////////////////////////////////////////////////////

template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Triangle_To_Point(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT3 triangle_element,int point_element)
{
    return Distance_Point_To_Triangle(X2,X1,t2,t1,point_element,triangle_element);//note order of array views must be switched
}

template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Triangle_To_Capsule(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT3 triangle_element,TV_INT2 segment_element)
{
    return Distance_Capsule_To_Triangle(X2,X1,t2,t1,segment_element,triangle_element);//note order of array views must be switched
}

template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Triangle_To_Triangle_Thickness(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT3 triangle_element1,TV_INT3 triangle_element2)
{
    int s1=triangle_element1(1),
        s2=triangle_element1(2),
        s3=triangle_element1(3),
        r1=triangle_element2(1),
        r2=triangle_element2(2),
        r3=triangle_element2(3);
    TV s_normal=TV::Cross_Product(X1(s2)-X1(s1),X1(s3)-X1(s1)).Normalized();
    TV t_normal=TV::Cross_Product(X2(r2)-X2(r1),X2(r3)-X2(r1)).Normalized();
    TV s1_upper(X1(s1)+(T)0.5*t1(s1)*s_normal),
       s1_lower(X1(s1)-(T)0.5*t1(s1)*s_normal),
       s2_upper(X1(s2)+(T)0.5*t1(s2)*s_normal),
       s2_lower(X1(s2)-(T)0.5*t1(s2)*s_normal),
       s3_upper(X1(s3)+(T)0.5*t1(s3)*s_normal),
       s3_lower(X1(s3)-(T)0.5*t1(s3)*s_normal),
       t1_upper(X2(r1)+(T)0.5*t2(r1)*s_normal),
       t1_lower(X2(r1)-(T)0.5*t2(r1)*s_normal),
       t2_upper(X2(r2)+(T)0.5*t2(r2)*s_normal),
       t2_lower(X2(r2)-(T)0.5*t2(r2)*s_normal),
       t3_upper(X2(r3)+(T)0.5*t2(r3)*s_normal),
       t3_lower(X2(r3)-(T)0.5*t2(r3)*s_normal);
    ARRAY<TRIANGLE_3D<T> > tris1;
    tris1.Append(TRIANGLE_3D<T>(s1_upper,s2_upper,s3_upper));
    tris1.Append(TRIANGLE_3D<T>(s1_lower,s2_lower,s3_lower));
    tris1.Append(TRIANGLE_3D<T>(s1_upper,s1_lower,s2_lower));
    tris1.Append(TRIANGLE_3D<T>(s1_upper,s2_lower,s2_upper));
    tris1.Append(TRIANGLE_3D<T>(s2_upper,s2_lower,s3_lower));
    tris1.Append(TRIANGLE_3D<T>(s2_upper,s3_lower,s3_upper));
    tris1.Append(TRIANGLE_3D<T>(s3_upper,s3_lower,s1_upper));
    tris1.Append(TRIANGLE_3D<T>(s3_upper,s1_lower,s1_upper));
    ARRAY<TRIANGLE_3D<T> > tris2;
    tris2.Append(TRIANGLE_3D<T>(t1_upper,t2_upper,t3_upper));
    tris2.Append(TRIANGLE_3D<T>(t1_lower,t2_lower,t3_lower));
    tris2.Append(TRIANGLE_3D<T>(t1_upper,t1_lower,t2_lower));
    tris2.Append(TRIANGLE_3D<T>(t1_upper,t2_lower,t2_upper));
    tris2.Append(TRIANGLE_3D<T>(t2_upper,t2_lower,t3_lower));
    tris2.Append(TRIANGLE_3D<T>(t2_upper,t3_lower,t3_upper));
    tris2.Append(TRIANGLE_3D<T>(t3_upper,t3_lower,t1_upper));
    tris2.Append(TRIANGLE_3D<T>(t3_upper,t1_lower,t1_upper));
    T min_distance=Distance_Triangle_To_Triangle(tris1(1),tris2(1));
    for(int i=1;i<=tris1.m;i++){
        for(int j=1;j<=tris2.m;j++){
            if(i==1&&j==1)continue;
            min_distance=min(min_distance,Distance_Triangle_To_Triangle(tris1(i),tris2(j)));
        }
    }
    return min_distance;
}

template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Triangle_To_Tetrahedron_Thickness(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT3 triangle_element,TV_INT4 tetrahedron_element)
{
    int tri1=triangle_element(1),
        tri2=triangle_element(2),
        tri3=triangle_element(3),
        tet1=tetrahedron_element(1),
        tet2=tetrahedron_element(2),
        tet3=tetrahedron_element(3),
        tet4=tetrahedron_element(4);
    TV triangle_normal=TV::Cross_Product(X1(tri2)-X1(tri1),X1(tri3)-X1(tri1)).Normalized();
    TV v1_upper(X1(tri1)+(T)0.5*t1(tri1)*triangle_normal),
       v1_lower(X1(tri1)-(T)0.5*t1(tri1)*triangle_normal),
       v2_upper(X1(tri2)+(T)0.5*t1(tri2)*triangle_normal),
       v2_lower(X1(tri2)-(T)0.5*t1(tri2)*triangle_normal),
       v3_upper(X1(tri3)+(T)0.5*t1(tri3)*triangle_normal),
       v3_lower(X1(tri3)-(T)0.5*t1(tri3)*triangle_normal);
    ARRAY<TRIANGLE_3D<T> > tris;
    tris.Append(TRIANGLE_3D<T>(v1_upper,v2_upper,v3_upper));
    tris.Append(TRIANGLE_3D<T>(v1_lower,v2_lower,v3_lower));
    tris.Append(TRIANGLE_3D<T>(v1_upper,v1_lower,v2_lower));
    tris.Append(TRIANGLE_3D<T>(v1_upper,v2_lower,v2_upper));
    tris.Append(TRIANGLE_3D<T>(v2_upper,v2_lower,v3_lower));
    tris.Append(TRIANGLE_3D<T>(v2_upper,v3_lower,v3_upper));
    tris.Append(TRIANGLE_3D<T>(v3_upper,v3_lower,v1_upper));
    tris.Append(TRIANGLE_3D<T>(v3_upper,v1_lower,v1_upper));
    ARRAY<TRIANGLE_3D<T> > tet_faces;
    tet_faces.Append(TRIANGLE_3D<T>(X2(tet1),X2(tet2),X2(tet3)));
    tet_faces.Append(TRIANGLE_3D<T>(X2(tet1),X2(tet2),X2(tet4)));
    tet_faces.Append(TRIANGLE_3D<T>(X2(tet1),X2(tet3),X2(tet4)));
    tet_faces.Append(TRIANGLE_3D<T>(X2(tet2),X2(tet3),X2(tet4)));
    T min_distance=Distance_Triangle_To_Triangle(tris(1),tet_faces(1));
    for(int i=1;i<=tris.m;i++){
        for(int j=1;j<=tet_faces.m;j++){
            if(i==1&&j==1)continue;
            min_distance=min(min_distance,Distance_Triangle_To_Triangle(tris(i),tet_faces(j)));
        }
    }
    return min_distance;
}

////////////////////////////////////////////////////////////
// Tetrahedron intersection functions
////////////////////////////////////////////////////////////
template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Tetrahedron_To_Point(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT4 tetrahedron_element,int point_element)
{
    return Distance_Point_To_Tetrahedron(X2,X1,t2,t1,point_element,tetrahedron_element);//note order of array views must be switched
}

template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Tetrahedron_To_Capsule(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT4 tetrahedron_element,TV_INT2 segment_element)
{
    return Distance_Capsule_To_Tetrahedron(X2,X1,t2,t1,segment_element,tetrahedron_element);//note order of array views must be switched
}

template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Tetrahedron_To_Triangle_Thickness(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT4 tetrahedron_element,TV_INT3 triangle_element)
{
    return Distance_Triangle_To_Tetrahedron_Thickness(X2,X1,t2,t1,triangle_element,tetrahedron_element);//note order of array views must be switched
}

template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Tetrahedron_To_Tetrahedron(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT4 tetrahedron_element1,TV_INT4 tetrahedron_element2)
{
    int s1=tetrahedron_element1(1),
        s2=tetrahedron_element1(2),
        s3=tetrahedron_element1(3),
        s4=tetrahedron_element1(4),
        r1=tetrahedron_element2(1),
        r2=tetrahedron_element2(2),
        r3=tetrahedron_element2(3),
        r4=tetrahedron_element2(4);
    ARRAY<TRIANGLE_3D<T> > faces1,faces2;
    faces1.Append(TRIANGLE_3D<T>(X1(s1),X1(s2),X1(s3)));
    faces1.Append(TRIANGLE_3D<T>(X1(s1),X1(s2),X1(s4)));
    faces1.Append(TRIANGLE_3D<T>(X1(s1),X1(s3),X1(s4)));
    faces1.Append(TRIANGLE_3D<T>(X1(s2),X1(s3),X1(s4)));
    faces2.Append(TRIANGLE_3D<T>(X2(r1),X2(r2),X2(r3)));
    faces2.Append(TRIANGLE_3D<T>(X2(r1),X2(r2),X2(r4)));
    faces2.Append(TRIANGLE_3D<T>(X2(r1),X2(r3),X2(r4)));
    faces2.Append(TRIANGLE_3D<T>(X2(r2),X2(r3),X2(r4)));
    T min_distance=Distance_Triangle_To_Triangle(faces1(1),faces2(1));
    for(int i=1;i<=faces1.m;i++){
        for(int j=1;j<=faces2.m;j++){
            if(i==1&&j==1)continue;
            min_distance=min(min_distance,Distance_Triangle_To_Triangle(faces1(i),faces2(j)));
        }
    }
    return min_distance;
}

////////////////////////////////////////////////////////////
// Support functions
////////////////////////////////////////////////////////////
template<class T>
VECTOR<T,3> PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Closest_Point_On_Line_Segment(const TV& point,const TV& point_a,const TV& point_b,T& t)
{
    TV segment_vector=(point_b-point_a);
    TV vector=(point-point_a);
    T segment_length=segment_vector.Magnitude();
    t=TV::Dot_Product(segment_vector.Normalized(),vector);
    TV closest_point;
    if(t<(T)0)closest_point=point_a;
    //else if(t>(T)1)closest_point=point_b;
    else if(t>segment_length)closest_point=point_b;
    else closest_point=point_a+t*segment_vector.Normalized();
    return closest_point;
}

template<class T>
VECTOR<T,3> PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Closest_Point_On_Plane(const TV& point,const TV& point_on_plane,const TV& plane_normal)
{//closest point on a plane: http://stackoverflow.com/questions/9605556/how-to-project-a-3d-point-to-a-3d-plane
    TV plane_to_point=(point-point_on_plane);
    TV plane_projection=point-(TV::Dot_Product(plane_normal,plane_to_point)*plane_normal);
    return plane_projection;
}

//NOTE: assumes zero thickness
template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Segment_To_Segment(const TV& p1,const TV& q1,const TV& p2,const TV& q2)
{//Code from Real-Time Collision Detection - section 5.1.9
    //Computes closest points C1 and C2 of S1(s)=P1+s*(Q1-P1) and
    //S2(t)=P2+t*(Q2-P2), returning the distance between S1(s) and S2(t)
    static T epsilon=(T)1e-9;
    T s,t;
    TV c1,c2;
    TV d1=q1-p1;//Direction vector of segment S1
    TV d2=q2-p2;//Direction vector of segment S2
    TV r=p1-p2;
    T a=TV::Dot_Product(d1,d1);//Squared length of segment S1, always nonnegative
    T e=TV::Dot_Product(d2,d2);//Squared length of segment S2, always nonnegative
    T f=TV::Dot_Product(d2,r);
    //Check if either or both segments degenerate into points
    if(a<=epsilon&&e<=epsilon){
        //Both segments degenerate into points
        s=(T)0;
        t=(T)0;
        c1=p1;
        c2=p2;
        return (c1-c2).Magnitude();
    }
    if(a<=epsilon){
        //First segment degenerates into a point
        s=(T)0;
        t=f/e; //s=0 => t = (b*s + f) / e = f / e
        t=clamp(t,(T)0,(T)1);
    }
    else{
        T c=TV::Dot_Product(d1,r);
        if(e<=epsilon){
            //Second segment degenerates into a point
            t=(T)0;
            s=clamp((T)-1*c/a,(T)0,(T)1);//t=0 => s = (b*t-c)/a = -c/a
        }
        else{
            //The general nondegenerate case starts here
            T b=TV::Dot_Product(d1,d2);
            T denom=a*e-b*b;//Always nonnegative
            //If segments are not parallel, compute closest point on L1 to L2 and clamp to segment S1. Else pick arbitrary s (here 0)
            if(denom!=(T)0){
                s=clamp((b*f-c*e)/denom,(T)0,(T)1);
            }
            else{
                s=(T)0;
            }
            //Compute point on L2 closest to S1(s) using t=Dot((P1+D1*s)-P2,D2) / Dot(D2,D2) = (b*s+f)/e
            t=(b*s+f)/e;
            //If t in [0,1] done. Else clamp t, recompute s for the new value of t using s=Dot((P2+D2*t)-P1,D1) / Dot(D1,D1)= (t*b-c) / a and clamp s to [0, 1]
            if(t<(T)0){
                t=(T)0;
                s=clamp((T)-1*c/a,(T)0,(T)1);
            }
            else if(t>(T)1){
                t=(T)1;
                s=clamp((b-c)/a,(T)0,(T)1);
            }
        }
    }
    c1=p1+d1*s;
    c2=p2+d2*t;
    //LOG::cout<<"c1: "<<c1<<", c2: "<<c2<<std::endl;
    return (c1-c2).Magnitude();
}

//NOTE: assumes zero thickness
template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Segment_To_Triangle(const SEGMENT_3D<T>& segment,const TRIANGLE_3D<T>& triangle)
{//From Real-Time Collision Detection - section 5.1.10
    SEGMENT_3D<T> ab(triangle.x1,triangle.x2),
                  bc(triangle.x2,triangle.x3),
                  ca(triangle.x3,triangle.x1);
    RAY<TV> ray(segment);
    bool intersects=INTERSECTION::Intersects(ray,triangle);
    T min_distance;
    bool early_termination=false;
    if(intersects){
        T t=ray.t_max;
        T t_end=(segment.x1-segment.x2).Magnitude();
        if(t<t_end){
            min_distance=(T)0;//segment_intersects_triangle
            early_termination=true;
        }
    }
    if(!early_termination){
        ARRAY<T> distances;
        distances.Append(Distance_Segment_To_Segment(segment.x1,segment.x2,ab.x1,ab.x2));
        distances.Append(Distance_Segment_To_Segment(segment.x1,segment.x2,bc.x1,bc.x2));
        distances.Append(Distance_Segment_To_Segment(segment.x1,segment.x2,ca.x1,ca.x2));
        distances.Append(triangle.Distance_To_Triangle(segment.x1));
        distances.Append(triangle.Distance_To_Triangle(segment.x2));
        min_distance=distances(1);
        for(int i=1;i<=distances.m;i++)if(distances(i)<min_distance)min_distance=distances(i);
    }
    return min_distance;
}

//NOTE: assumes zero thickness
template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Segment_To_Tetrahedron(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT2 segment_element,TV_INT4 tetrahedron_element)
{
    int s1=segment_element(1),
        s2=segment_element(2),
        r1=tetrahedron_element(1),
        r2=tetrahedron_element(2),
        r3=tetrahedron_element(3),
        r4=tetrahedron_element(4);
    SEGMENT_3D<T> segment(X1(s1),X1(s2));
    ARRAY<TRIANGLE_3D<T> > triangles;
    //does ccw vs cw ordering matter here?
    triangles.Append(TRIANGLE_3D<T>(X2(r1),X2(r2),X2(r3)));
    triangles.Append(TRIANGLE_3D<T>(X2(r1),X2(r2),X2(r4)));
    triangles.Append(TRIANGLE_3D<T>(X2(r1),X2(r3),X2(r4)));
    triangles.Append(TRIANGLE_3D<T>(X2(r2),X2(r3),X2(r4)));
    T min_distance=Distance_Segment_To_Triangle(segment,triangles(1));
    for(int i=2;i<=4;i++)min_distance=min(min_distance,Distance_Segment_To_Triangle(segment,triangles(i)));
    return min_distance;
}

//NOTE: assumes zero thickness
template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Distance_Triangle_To_Triangle(const TRIANGLE_3D<T>& tri1,const TRIANGLE_3D<T>& tri2)
{//from Real-Time Collision Detection - section 5.1.11

    //hack -TODO: clean up
    if(INTERSECTION::Intersects(tri1,tri2))
        return (T)-1;
    /*
    //NOTE: this algorithm assumes that triangles are not intersecting
    ARRAY<T> distances;
    ARRAY<SEGMENT_3D<T> > segments1,segments2;
    segments1.Append(SEGMENT_3D<T>(tri1.x1,tri1.x2));
    segments1.Append(SEGMENT_3D<T>(tri1.x2,tri1.x3));
    segments1.Append(SEGMENT_3D<T>(tri1.x3,tri1.x1));
    segments2.Append(SEGMENT_3D<T>(tri2.x1,tri2.x2));
    segments2.Append(SEGMENT_3D<T>(tri2.x2,tri2.x3));
    segments2.Append(SEGMENT_3D<T>(tri2.x3,tri2.x1));
    for(int i=1;i<=3;i++){
        distances.Append(Distance_Segment_To_Triangle(segments1(i),tri2));
        distances.Append(Distance_Segment_To_Triangle(segments2(i),tri1));
    }
    T min_distance=distances(1);
    for(int i=2;i<=distances.m;i++)if(distances(i)<min_distance)min_distance=distances(i);
    for(int i=1;i<=distances.m;i++){
        LOG::cout<<"distance "<<i<<": "<<distances(i)<<std::endl;
    }
    return min_distance;
    */
    //NOTE: this algorithm assumes that triangles are not intersecting
    ARRAY<T> distances;
    ARRAY<SEGMENT_3D<T> > segments1,segments2;
    segments1.Append(SEGMENT_3D<T>(tri1.x1,tri1.x2));
    segments1.Append(SEGMENT_3D<T>(tri1.x2,tri1.x3));
    segments1.Append(SEGMENT_3D<T>(tri1.x3,tri1.x1));
    segments2.Append(SEGMENT_3D<T>(tri2.x1,tri2.x2));
    segments2.Append(SEGMENT_3D<T>(tri2.x2,tri2.x3));
    segments2.Append(SEGMENT_3D<T>(tri2.x3,tri2.x1));
    for(int i=1;i<=3;i++){
        for(int j=1;j<=3;j++){
            distances.Append(Distance_Segment_To_Segment(segments1(i).x1,segments1(i).x2,segments2(j).x1,segments2(j).x2));
        }
    }
    distances.Append(tri1.Distance_To_Triangle(tri2.x1));
    distances.Append(tri1.Distance_To_Triangle(tri2.x2));
    distances.Append(tri1.Distance_To_Triangle(tri2.x3));
    distances.Append(tri2.Distance_To_Triangle(tri1.x1));
    distances.Append(tri2.Distance_To_Triangle(tri1.x2));
    distances.Append(tri2.Distance_To_Triangle(tri1.x3));
    T min_distance=distances(1);
    for(int i=2;i<=distances.m;i++)if(distances(i)<min_distance)min_distance=distances(i);
    return min_distance;
}

template<class T>
T PhysBAM::CODIMENSIONAL_INTERSECTION<T>::
Fraction_Of_Segment_In_Sphere(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,int point,TV_INT2 segment)
{
    T dist1=(X1(point)-X2(segment(1))).Magnitude_Squared();
    T dist2=(X1(point)-X2(segment(2))).Magnitude_Squared();
    T radius2=sqr((T)0.5*t1(point));

    if(dist1>radius2&&dist2>radius2)return (T)0;

    int p1=(dist1<dist2?segment(1):segment(2));
    int p2=(dist1<dist2?segment(2):segment(1));

    TV v=(X2(p2)-X2(p1));
    TV center=X1(point);

    T a=TV::Dot_Product(v,v);
    T b=(T)2*TV::Dot_Product(center,v);
    T c=TV::Dot_Product(center,center)-radius2;

    //solve quadratic
    T disc=b*b-(T)4*a*c;
    if(a==(T)0||disc<(T)0)return (T)0;

    //T x1=(-b-sqrt(disc))/((T)2*a);
    T x2=(-b+sqrt(disc))/((T)2*a);

    T weight=x2; //unconditionally?

    return weight;
}

template class CODIMENSIONAL_INTERSECTION<float>;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
template class CODIMENSIONAL_INTERSECTION<double>;
#endif
