//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <PhysBAM_Tools/Data_Structures/QUEUE.h>
#include <PhysBAM_Geometry/Surface_Mesh_Level_Sets/FAST_MARCHING_METHOD_SURFACE_MESH.h>
#include <PhysBAM_Geometry/Level_Sets/FAST_MARCHING.h>
#include <PhysBAM_Geometry/Level_Sets/LEVELSET_UTILITIES.h>
#include <PhysBAM_Geometry/Basic_Geometry/SEGMENT_3D.h>
#include <PhysBAM_Tools/Data_Structures/HASHTABLE.h>
#include <PhysBAM_Tools/Data_Structures/HASHTABLE_ITERATOR.h>
#include "../Meshing/CODIMENSIONAL_HELPER.h"
#include "../Levelset/FAST_MARCHING_METHOD_CODIMENSIONAL.h"
#include "../Levelset/LEVELSET_MULTIPLE_CODIMENSIONAL.h"

using namespace PhysBAM;

template<class TV>
PhysBAM::FAST_MARCHING_METHOD_CODIMENSIONAL<TV>::FAST_MARCHING_METHOD_CODIMENSIONAL(CODIMENSIONAL_MESH& mesh_input,
                                                                                    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles_input,
                                                                                    bool update_slow_interface_particles_/*=true*/,
                                                                                    bool ignore_tetrahedron_interior_/*=false*/,
                                                                                    bool only_update_interface_/*=false*/,
                                                                                    bool use_narrow_band_/*=false*/,
                                                                                    int narrow_band_width_/*=1*/,
                                                                                    T speed_threshold_/*=(T)1e-5*/)
    :mesh(mesh_input),particles(particles_input),large_phi((T)1e10),update_slow_interface_particles(update_slow_interface_particles_),speed_threshold(speed_threshold_),use_non_drifting_near_interface(false),ignore_tetrahedron_interior(ignore_tetrahedron_interior_),active(particles_input.X.m),
    only_update_interface(only_update_interface_),use_narrow_band(use_narrow_band_),narrow_band_width(narrow_band_width_),distance_to_interface(particles_input.X.m)
{
}

template<class TV>
void PhysBAM::FAST_MARCHING_METHOD_CODIMENSIONAL<TV>::Fast_Marching_Method(ARRAY_VIEW<T>& phi)
{
    if(ignore_tetrahedron_interior){
        active.Fill(false);
        for(int i=1;i<=mesh.segment_mesh.elements.m;i++){
            const TV_INT2& v=mesh.segment_mesh.elements(i);
            for(int j=1;j<=v.m;j++)active(v(j))=true;
        }
        for(int i=1;i<=mesh.triangle_mesh.elements.m;i++){
            const TV_INT3& v=mesh.triangle_mesh.elements(i);
            for(int j=1;j<=v.m;j++)active(v(j))=true;
        }
        for(int i=1;i<=mesh.tetrahedron_mesh.boundary_mesh->elements.m;i++){
            const TV_INT3& v=mesh.tetrahedron_mesh.boundary_mesh->elements(i);
            for(int j=1;j<=v.m;j++)active(v(j))=true;
        }
    }
    else{
        active.Fill(true);
    }

    HASHTABLE<int> segment_particles,triangle_particles,tetrahedron_particles;
    for(int i=1;i<=mesh.segment_mesh.elements.m;i++){
        const TV_INT2& segment=mesh.segment_mesh.elements(i);
        for(int j=1;j<=segment.m;j++)segment_particles.Set(segment(j));
    }
    for(int i=1;i<=mesh.triangle_mesh.elements.m;i++){
        const TV_INT3& triangle=mesh.triangle_mesh.elements(i);
        for(int j=1;j<=triangle.m;j++)triangle_particles.Set(triangle(j));
    }
    for(int i=1;i<=mesh.tetrahedron_mesh.elements.m;i++){
        const TV_INT4& tetrahedron=mesh.tetrahedron_mesh.elements(i);
        for(int j=1;j<=tetrahedron.m;j++)tetrahedron_particles.Set(tetrahedron(j));
    }

    ARRAY<int> codim_0x_particles;//particles on codimension 0 and codimension 1 or 2
    for(HASHTABLE_ITERATOR<int> iter(tetrahedron_particles);iter.Valid();iter.Next()){
        int p=iter.Key();
        if(segment_particles.Contains(p)||triangle_particles.Contains(p))codim_0x_particles.Append(p);
    }

    int heap_length=0;
    ARRAY<int> close_k(phi.m);close_k.Fill(0);
    ARRAY<int> heap(phi.m,false);
    ARRAY<bool> done(phi.m);done.Fill(false);
    Initialize_Interface(phi,done,close_k,heap,heap_length,codim_0x_particles);

    if(only_update_interface)return;

    if(use_narrow_band)Set_Graph_Distance_To_Interface(done);

    while(heap_length!=0){
        int index=heap(1);done(index)=true;close_k(index)=0;//add to done, remove from close
        FAST_MARCHING<T>::Down_Heap(phi,close_k,heap,heap_length);heap_length--;
        const ARRAY<int>& nbs=(*mesh.neighbor_nodes)(index);
        for(int i=1;i<=nbs.m;i++){
            if(ignore_tetrahedron_interior&&!active(nbs(i)))continue;
            if(use_narrow_band&&distance_to_interface(nbs(i))>narrow_band_width)continue;
            if(!done(nbs(i))){
                Update_Or_Add_Neighbor(phi,done,close_k,heap,heap_length,nbs(i),codim_0x_particles);
            }
        }
    }
}

template<class TV>
void PhysBAM::FAST_MARCHING_METHOD_CODIMENSIONAL<TV>::Initialize_Interface(ARRAY_VIEW<T>& phi,ARRAY<bool>& done,ARRAY<int>& close_k,ARRAY<int>& heap,int& heap_length,ARRAY<int>& codim_0x_particles)
{
    for(int i=1;i<=mesh.segment_mesh.elements.m;i++){
        const TV_INT2& v=mesh.segment_mesh.elements(i);
        if(phi(v(1))*phi(v(2))<(T)0)for(int d=1;d<=2;d++)Add_To_Initial(done,close_k,v(d));
    }
    for(int i=1;i<=mesh.triangle_mesh.elements.m;i++){
        const TV_INT3& v=mesh.triangle_mesh.elements(i);
        if(phi(v(1))*phi(v(2))<(T)0||phi(v(2))*phi(v(3))<(T)0){
            for(int d=1;d<=3;d++)Add_To_Initial(done,close_k,v(d));
        }
    }
    if(ignore_tetrahedron_interior){
        TRIANGLE_MESH& tet_boundary=*(mesh.tetrahedron_mesh.boundary_mesh);
        for(int i=1;i<=tet_boundary.elements.m;i++){
            const TV_INT3& v=tet_boundary.elements(i);
            if(phi(v(1))*phi(v(2))<(T)0||phi(v(2))*phi(v(3))<(T)0){
                for(int d=1;d<=3;d++)Add_To_Initial(done,close_k,v(d));
            }
        }
    }
    else{
        for(int i=1;i<=mesh.tetrahedron_mesh.elements.m;i++){
            const TV_INT4& v=mesh.tetrahedron_mesh.elements(i);
            if(phi(v(1))*phi(v(2))<(T)0||phi(v(2))*phi(v(3))<(T)0||phi(v(3))*phi(v(4))<(T)0){
                for(int d=1;d<=4;d++){
                    Add_To_Initial(done,close_k,v(d));
                }
            }
        }
    }

    ////update phi close to interface
    Initialize_Phi_Near_Interface(phi,done);
    if(!ignore_tetrahedron_interior)Get_Tetrahedron_Interface_Faces(phi);

    //set non-interface phi values to infinitely far away
    for(int i=1;i<=phi.m;i++){
        if(active(i)&&!done(i)) phi(i)=LEVELSET_UTILITIES<T>::Sign(phi(i)) * large_phi;
    }

    //if(!ignore_tetrahedron_interior)Get_Tetrahedron_Interface_Faces(phi);
    
    //// initialize close points
    for(int index=1;index<=phi.m;index++) if(active(index)&&close_k(index)){
        Update_Close_Point(phi,done,index,codim_0x_particles);
        heap_length++;heap(heap_length)=index;
        FAST_MARCHING<T>::Up_Heap(phi,close_k,heap,heap_length);
    }
}

template<class TV>
void PhysBAM::FAST_MARCHING_METHOD_CODIMENSIONAL<TV>::Add_To_Initial(ARRAY<bool>& done,ARRAY<int>& close_k,const int& p)
{
    done(p)=true;close_k(p)=0; // add to done, remove from close
    const ARRAY<int>& nbs=(*mesh.neighbor_nodes)(p);
    for(int i=1;i<=nbs.m;i++){if(!done(nbs(i))) close_k(nbs(i))=1;}
}

template<class TV>
void PhysBAM::FAST_MARCHING_METHOD_CODIMENSIONAL<TV>::Initialize_Phi_Near_Interface(ARRAY_VIEW<T>& phi,ARRAY<bool>& done)
{
    const T tolerance=(T)1e-7;
    ARRAY<T> phi_new(phi.m);
    for(int p=1;p<=phi.m;p++)if(done(p)){
        T closest_distance(large_phi);T current_phi=phi(p);if(abs(current_phi)<tolerance){phi_new(p)=current_phi;continue;}

        ////segment
        ARRAY<int>& nb_segs=(*mesh.segment_mesh.incident_elements)(p);
        for(int i=1;i<=nb_segs.m;i++){
            int s=nb_segs(i);const TV_INT2 v=Reordered(mesh.segment_mesh.elements(s),p);
            ARRAY<TV> added_points;ARRAY<TV_INT2> added_edges;Find_Edge_Interface(v,phi,tolerance,added_points,added_edges);T distance=large_phi;
            if(added_points.m==1) distance=(particles.X(p)-added_points(1)).Magnitude();
            closest_distance=min(closest_distance,distance);
        }

        if(!use_non_drifting_near_interface){
            ////triangles
            ARRAY<int>& nb_tris=(*mesh.triangle_mesh.incident_elements)(p);
            for(int i=1;i<=nb_tris.m;i++){
                int t=nb_tris(i);const TV_INT3 v=Reordered(mesh.triangle_mesh.elements(t),p);
                ARRAY<TV> added_points;ARRAY<TV_INT2> added_edges;Find_Edge_Interface(v,phi,tolerance,added_points,added_edges);T distance=large_phi;
                if(added_points.m==1) distance=(particles.X(p)-added_points(1)).Magnitude();
                else if(added_points.m==2) distance=SEGMENT_3D<T>(added_points(1),added_points(2)).Distance_From_Point_To_Segment/*Distance_From_Point_To_Line*/(particles.X(p));
                closest_distance=min(closest_distance,distance);
            }

            ////tetrahedra
            if(ignore_tetrahedron_interior){
                ARRAY<int>& nb_tris=(*mesh.tetrahedron_mesh.boundary_mesh->incident_elements)(p);
                for(int i=1;i<=nb_tris.m;i++){
                    int t=nb_tris(i);const TV_INT3& v=Reordered(mesh.tetrahedron_mesh.boundary_mesh->elements(t),p);
                    ARRAY<TV> added_points;ARRAY<TV_INT2> added_edges;Find_Edge_Interface(v,phi,tolerance,added_points,added_edges);T distance=large_phi;
                    if(added_points.m==1) distance=(particles.X(p)-added_points(1)).Magnitude();
                    else if(added_points.m==2) distance=SEGMENT_3D<T>(added_points(1),added_points(2)).Distance_From_Point_To_Segment(particles.X(p));
                    closest_distance=min(closest_distance,distance);
                }
            }
            else{
                ARRAY<int>& nb_tets=(*mesh.tetrahedron_mesh.incident_elements)(p);
                for(int i=1;i<=nb_tets.m;i++){
                    int t=nb_tets(i);const TV_INT4 v=Reordered(mesh.tetrahedron_mesh.elements(t),p);
                    ARRAY<TV> added_points;ARRAY<TV_INT2> added_edges;int incident_interface_edge_num=Find_Edge_Interface(v,phi,tolerance,added_points,added_edges);T distance=large_phi;
                    if(incident_interface_edge_num==1) distance=(particles.X(p)-added_points(1)).Magnitude();
                    else if(incident_interface_edge_num==2) distance=SEGMENT_3D<T>(added_points(1),added_points(2)).Distance_From_Point_To_Segment(particles.X(p));
                    else if(incident_interface_edge_num==3) distance=TRIANGLE_3D<T>(added_points(1),added_points(2),added_points(3)).Distance_To_Triangle(particles.X(p));
                    closest_distance=min(closest_distance,distance);
                }
            }
        }
        else{
            int element_n=0;T distance_sum=(T)0;

            ////triangles
            ARRAY<int>& nb_tris=(*mesh.triangle_mesh.incident_elements)(p);
            for(int i=1;i<=nb_tris.m;i++){
                int t=nb_tris(i);const TV_INT3 v=Reordered(mesh.triangle_mesh.elements(t),p);
                ARRAY<TV> added_points;ARRAY<TV_INT2> added_edges;Find_Edge_Interface(v,phi,tolerance,added_points,added_edges);T distance=large_phi;
                if(added_points.m>=2){
                    distance=SEGMENT_3D<T>(added_points(1),added_points(2)).Distance_From_Point_To_Line(particles.X(p));
                    //distance=SEGMENT_3D<T>(added_points(1),added_points(2)).Distance_From_Point_To_Segment(particles.X(p));
                    distance_sum+=distance;element_n++;
                }
            }

            ////tetrahedra
            if(ignore_tetrahedron_interior){
                TRIANGLE_MESH& tet_boundary=*mesh.tetrahedron_mesh.boundary_mesh;
                ARRAY<int>& nb_tris=(*tet_boundary.incident_elements)(p);
                for(int i=1;i<=nb_tris.m;i++){
                    int t=nb_tris(i);const TV_INT3 v=Reordered(tet_boundary.elements(t),p);
                    ARRAY<TV> added_points;ARRAY<TV_INT2> added_edges;Find_Edge_Interface(v,phi,tolerance,added_points,added_edges);T distance=large_phi;
                    if(added_points.m>=2){
                        distance=SEGMENT_3D<T>(added_points(1),added_points(2)).Distance_From_Point_To_Line(particles.X(p));
                        distance_sum+=distance;element_n++;
                    }
                }
            }
            /*
            else{
                //int min_type=0;int closest_tet=0;
                ARRAY<int>& nb_tets=(*mesh.tetrahedron_mesh.incident_elements)(p);
                for(int i=1;i<=nb_tets.m;i++){
                    int t=nb_tets(i);const TV_INT4 v=Reordered(mesh.tetrahedron_mesh.elements(t),p);
                    ARRAY<TV> added_points;ARRAY<TV_INT2> added_edges;//int incident_interface_edge_num=Find_Edge_Interface(v,phi,tolerance,added_points,added_edges);
                    T distance=large_phi;
                    if(added_points.m>=3){
                        distance=abs(PLANE<T>(added_points(1),added_points(2),added_points(3)).Signed_Distance(particles.X(p)));
                        //distance=TRIANGLE_3D<T>(added_points(1),added_points(2),added_points(3)).Distance_To_Triangle(particles.X(p));
                        distance_sum+=distance;element_n++;
                    }
                }
            }//ignore_tetrahedron_interior
            */
            if(element_n>0)closest_distance=distance_sum/element_n;
            //if(p==12)LOG::cout<<"p: "<<p<<" guid "<<particles.guid(p)<<" type "<<min_type<<", tet "<<min_tet<<", phi: "<<current_phi<<", dis: "<<closest_distance<<std::endl;
        }

        phi_new(p)=closest_distance*LEVELSET_UTILITIES<T>::Sign(current_phi);
    }
    for(int p=1;p<=phi.m;p++){
        if(done(p)){
            if(update_slow_interface_particles)phi(p)=phi_new(p);
            else if(particles.V(p).Magnitude()>speed_threshold)phi(p)=phi_new(p);//only update phi for particles moving faster than speed_threshold
        }
    }
}

template<class TV>
void PhysBAM::FAST_MARCHING_METHOD_CODIMENSIONAL<TV>::Update_Close_Point(ARRAY_VIEW<T>& phi,ARRAY<bool>& done,const int& p,ARRAY<int>& codim_0x_particles)
{
    //phi(p)=(T)FLT_MAX;
    //segments
    ARRAY<int>& nb_segs=(*mesh.segment_mesh.incident_elements)(p);
    for(int i=1;i<=nb_segs.m;i++){
        int neighbor=nb_segs(i);
        if(done(neighbor)){
            T signed_distance=phi(neighbor)+LEVELSET_UTILITIES<T>::Sign(phi(p))*(particles.X(p)-particles.X(neighbor)).Magnitude();
            if(abs(phi(p))>abs(signed_distance)){
                phi(p)=signed_distance;
            }
        }
    }

    //triangles
    ARRAY<TV> tmp_normals(particles.X.m);tmp_normals.Fill(TV::Axis_Vector(2));ARRAY_VIEW<TV> tmp_normals_array_view(tmp_normals);
    ////TODO: fix this with correct normals
    ARRAY<int>& nb_tris=(*mesh.triangle_mesh.incident_elements)(p);
    for(int i=1;i<=nb_tris.m;i++){
        const TV_INT3& v=mesh.triangle_mesh.elements(nb_tris(i));
        FAST_MARCHING_METHOD_SURFACE_MESH<TV>::Update_Close_Point_Helper(v,particles.X,tmp_normals_array_view,mesh.triangle_mesh,phi,done,p);
        //FAST_MARCHING_METHOD_SURFACE_MESH<TV>::Update_Close_Point_Acute_Triangle_Only_Helper(v,particles.X,tmp_normals_array_view,mesh.triangle_mesh,phi,done,p);        
    }
    
    //tetrahedra
    if(ignore_tetrahedron_interior){
        TRIANGLE_MESH& tet_boundary=*mesh.tetrahedron_mesh.boundary_mesh;
        ARRAY<TV> tmp_normals(particles.X.m);tmp_normals.Fill(TV::Axis_Vector(2));ARRAY_VIEW<TV> tmp_normals_array_view(tmp_normals);
        ////TODO: fix this with correct normals? (it looks like it's unused...)
        ARRAY<int>& nb_tris=(*tet_boundary.incident_elements)(p);
        for(int i=1;i<=nb_tris.m;i++){
            const TV_INT3& v=tet_boundary.elements(nb_tris(i));
            FAST_MARCHING_METHOD_SURFACE_MESH<TV>::Update_Close_Point_Helper(v,particles.X,tmp_normals_array_view,tet_boundary,phi,done,p);
        }
    }
    else{
        Update_Close_Point_Using_Tetrahedra(particles.X,phi,mesh.tetrahedron_mesh,done,p,codim_0x_particles);
    }
}

template<class TV>
void PhysBAM::FAST_MARCHING_METHOD_CODIMENSIONAL<TV>::Update_Or_Add_Neighbor(ARRAY_VIEW<T>& phi,ARRAY<bool>& done,ARRAY<int>& close_k,ARRAY<int>& heap,int& heap_length,const int& neighbor,ARRAY<int>& codim_0x_particles)
{
    if(close_k(neighbor)){
        Update_Close_Point(phi,done,neighbor,codim_0x_particles);
        FAST_MARCHING<T>::Up_Heap(phi,close_k,heap,close_k(neighbor));}
    else{
        heap_length++;heap(heap_length)=neighbor;close_k(neighbor)=heap_length;
        Update_Close_Point(phi,done,neighbor,codim_0x_particles);
        FAST_MARCHING<T>::Up_Heap(phi,close_k,heap,heap_length);}
}

template<class TV> template<int N>
int PhysBAM::FAST_MARCHING_METHOD_CODIMENSIONAL<TV>::Find_Edge_Interface(const VECTOR<int,N>& v,ARRAY_VIEW<T>& phi,const T tolerance,ARRAY<TV>& added_points,ARRAY<TV_INT2>& added_edges)
{
    //for(int i=2;i<=N;i++)if((phi(v(1))>(T)0&&phi(v(i))<tolerance)||(phi(v(1))<(T)0&&phi(v(i))>-tolerance)){
    int c=0;
    for(int p=1;p<=N-1;p++){
        //for(int i=p+1;i<=N;i++)if((phi(v(p))>(T)0&&phi(v(i))<=tolerance)||(phi(v(p))<=(T)0&&phi(v(i))>-tolerance)){
        for(int i=p+1;i<=N;i++)if((phi(v(p))>(T)0&&phi(v(i))<=(T)0)||(phi(v(p))<=(T)0&&phi(v(i))>(T)0)){
            TV point=LEVELSET_MULTIPLE_CODIMENSIONAL<TV>::Convex_Average(phi(v(p)),particles.X(v(p)),phi(v(i)),particles.X(v(i)));
            added_points.Append(point);added_edges.Append(TV_INT2(v(p),v(i)));if(p==1)c++;
        }
    }
    return c;
}

template<class TV>
void PhysBAM::FAST_MARCHING_METHOD_CODIMENSIONAL<TV>::Set_Graph_Distance_To_Interface(const ARRAY<bool>& done)
{
    int large_int=1000000;//this will be the graph distance for particles in regions not connected to any interface particles
    distance_to_interface.Fill(large_int);
    ARRAY<bool> visited(particles.X.m);visited.Fill(false);
    QUEUE<int> front(particles.X.m);
    for(int i=1;i<=particles.X.m;i++){
        if(done(i)){
            distance_to_interface(i)=0;
            front.Enqueue(i);
            visited(i)=true;
        }
    }
    while(!front.Empty()){
        int p=front.Dequeue();
        int graph_distance=distance_to_interface(p);
        const ARRAY<int>& nbs=(*mesh.neighbor_nodes)(p);
        for(int i=1;i<=nbs.m;i++){
            if(visited(nbs(i)))continue;
            visited(nbs(i))=true;
            distance_to_interface(nbs(i))=graph_distance+1;
            front.Enqueue(nbs(i));
        }
    }
}

template<class TV>
void PhysBAM::FAST_MARCHING_METHOD_CODIMENSIONAL<TV>::Get_Tetrahedron_Interface_Faces(const ARRAY_VIEW<T>& phi)
{
    tet_interface_faces.Remove_All();
    tet_interface_indices.Remove_All();

    TETRAHEDRON_MESH& tet_mesh=mesh.tetrahedron_mesh;
    const ARRAY_VIEW<TV>& X=particles.X;
    for(int i=1;i<=tet_mesh.elements.m;i++){
        //bool has_pos=false,has_neg=false;
        int pos_count=0,neg_count=0;
        ARRAY<int> pos_indices,neg_indices;
        const TV_INT4& v=tet_mesh.elements(i);
        for(int j=1;j<=4;j++){
            if(phi(v(j))>(T)0){pos_count++;pos_indices.Append(v(j));}
            else {neg_count++;neg_indices.Append(v(j));}

            }
        if(pos_count>0&&neg_count>0){
            //LOG::cout<<"interface tet "<<i<<": "<<v<<std::endl;
            //for(int j=1;j<=4;j++){LOG::cout<<phi(v(j))<<", ";}LOG::cout<<std::endl;

            PHYSBAM_ASSERT(pos_count+neg_count==4);
            switch(neg_count){
                case 1:{
                    T theta1=LEVELSET_UTILITIES<T>::Theta(phi(neg_indices(1)),phi(pos_indices(1)));
                    T theta2=LEVELSET_UTILITIES<T>::Theta(phi(neg_indices(1)),phi(pos_indices(2)));
                    T theta3=LEVELSET_UTILITIES<T>::Theta(phi(neg_indices(1)),phi(pos_indices(3)));

                    TV v01=(X(pos_indices(1))-X(neg_indices(1)));
                    TV v02=(X(pos_indices(2))-X(neg_indices(1)));
                    TV v03=(X(pos_indices(3))-X(neg_indices(1)));

                    TV p1=X(neg_indices(1))+theta1*v01;
                    TV p2=X(neg_indices(1))+theta2*v02;
                    TV p3=X(neg_indices(1))+theta3*v03;
                    tet_interface_faces.Append(TRIANGLE_3D<T>(p1,p2,p3));
                    tet_interface_indices.Append(v);
                    }break;
                case 2:{
                    T theta1=LEVELSET_UTILITIES<T>::Theta(phi(neg_indices(1)),phi(pos_indices(1)));
                    T theta2=LEVELSET_UTILITIES<T>::Theta(phi(neg_indices(1)),phi(pos_indices(1)));
                    T theta3=LEVELSET_UTILITIES<T>::Theta(phi(neg_indices(1)),phi(pos_indices(1)));
                    T theta4=LEVELSET_UTILITIES<T>::Theta(phi(neg_indices(1)),phi(pos_indices(1)));

                    TV v1=(X(pos_indices(1))-X(neg_indices(1)));
                    TV v2=(X(pos_indices(2))-X(neg_indices(1)));
                    TV v3=(X(pos_indices(1))-X(neg_indices(2)));
                    TV v4=(X(pos_indices(2))-X(neg_indices(2)));

                    TV p1=X(neg_indices(1))+theta1*v1;
                    TV p2=X(neg_indices(1))+theta2*v2;
                    TV p3=X(neg_indices(2))+theta3*v3;
                    TV p4=X(neg_indices(2))+theta4*v4;
                    tet_interface_faces.Append(TRIANGLE_3D<T>(p1,p2,p3));
                    tet_interface_indices.Append(v);
                    tet_interface_faces.Append(TRIANGLE_3D<T>(p2,p3,p4));
                    tet_interface_indices.Append(v);
                    }break;
                case 3:{
                    T theta1=LEVELSET_UTILITIES<T>::Theta(phi(pos_indices(1)),phi(neg_indices(1)));
                    T theta2=LEVELSET_UTILITIES<T>::Theta(phi(pos_indices(1)),phi(neg_indices(2)));
                    T theta3=LEVELSET_UTILITIES<T>::Theta(phi(pos_indices(1)),phi(neg_indices(3)));

                    TV v01=(X(neg_indices(1))-X(pos_indices(1)));
                    TV v02=(X(neg_indices(2))-X(pos_indices(1)));
                    TV v03=(X(neg_indices(3))-X(pos_indices(1)));

                    TV p1=X(pos_indices(1))+theta1*v01;
                    TV p2=X(pos_indices(1))+theta2*v02;
                    TV p3=X(pos_indices(1))+theta3*v03;
                    tet_interface_faces.Append(TRIANGLE_3D<T>(p1,p2,p3));
                    tet_interface_indices.Append(v);
                    }break;
            }
        }
    }
}

template<class TV>
void PhysBAM::FAST_MARCHING_METHOD_CODIMENSIONAL<TV>::Update_Close_Point_Using_Tetrahedra(const ARRAY_VIEW<TV>& X,ARRAY_VIEW<T>& phi,const TETRAHEDRON_MESH& tet_mesh,ARRAY<bool>& done,int index,ARRAY<int>& codim_0x_particles)
{
    //the closest interface point for a tet particle can come from either an interior tet, or a particle that's also part of another codimension
    for(int i=1;i<=tet_interface_faces.m;i++){
        const TRIANGLE_3D<T>& tri=tet_interface_faces(i);
        T dist=tri.Distance_To_Triangle(X(index));
        if(abs(phi(index))>dist){
            phi(index)=LEVELSET_UTILITIES<T>::Sign(phi(index))*dist;
        }
    }

    for(int i=1;i<=codim_0x_particles.m;i++){
        int p=codim_0x_particles(i);
        T dist=phi(p)+LEVELSET_UTILITIES<T>::Sign(phi(index))*(X(p)-X(index)).Magnitude();
        if(abs(phi(index))>abs(dist)){
            phi(index)=dist;
            //LOG::cout<<"new min from surface point: "<<phi(index)<<std::endl;
        }
    }
}

#define INSTANTIATION_HELPER_FAST_MARCHING_METHOD_CODIMENSIONAL(T,d)           \
    template class FAST_MARCHING_METHOD_CODIMENSIONAL<VECTOR<T,d> >;     \
    template int FAST_MARCHING_METHOD_CODIMENSIONAL<VECTOR<T,d> >::Find_Edge_Interface(const VECTOR<int,2>& v,ARRAY_VIEW<T>& phi,const T tolerance,ARRAY<VECTOR<T,d> >& added_points,ARRAY<VECTOR<int,2> >& interface_edges); \
    template int FAST_MARCHING_METHOD_CODIMENSIONAL<VECTOR<T,d> >::Find_Edge_Interface(const VECTOR<int,3>& v,ARRAY_VIEW<T>& phi,const T tolerance,ARRAY<VECTOR<T,d> >& added_points,ARRAY<VECTOR<int,2> >& interface_edges); \
    template int FAST_MARCHING_METHOD_CODIMENSIONAL<VECTOR<T,d> >::Find_Edge_Interface(const VECTOR<int,4>& v,ARRAY_VIEW<T>& phi,const T tolerance,ARRAY<VECTOR<T,d> >& added_points,ARRAY<VECTOR<int,2> >& interface_edges); 

INSTANTIATION_HELPER_FAST_MARCHING_METHOD_CODIMENSIONAL(float,3)
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
INSTANTIATION_HELPER_FAST_MARCHING_METHOD_CODIMENSIONAL(double,3)
#endif
