//#####################################################################
// Copyright 2013, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <PhysBAM_Geometry/Basic_Geometry/TRIANGLE_3D.h>
#include <PhysBAM_Geometry/Basic_Geometry/SPHERE.h>
#include "../Meshing/BALL_PIVOTING_TRIANGULATION.h"
#include "../Meshing/CODIMENSIONAL_HELPER.h"
using namespace PhysBAM;

template<class T>
void PhysBAM::BALL_PIVOTING_TRIANGULATION<T>::Initialize_Length_Scale(const T length_scale_input)
{
    length_scale=length_scale_input;
    T delaunay_circumradius=(TV::m==2?(T)(sqrt(3.)/3.)*((T)2*length_scale):(T)(/*(T)1.2**/sqrt(6.)/4.)*((T)2*length_scale));
    alpha_ball_radius=(T)1.2*delaunay_circumradius;
    max_segment_length=(T)1.5*length_scale;
    epsilon=(T)1e-20;
}

template<class T>
void PhysBAM::BALL_PIVOTING_TRIANGULATION<T>::Construct_Mesh()
{
    Advancing_Front_Triangulation();
    LOG::cout<<"ball pivoting triangulation: p: "<<particles.X.m<<", tri: "<<triangle_mesh.elements.m<<std::endl;
}

template<class T>
void PhysBAM::BALL_PIVOTING_TRIANGULATION<T>::Advancing_Front_Triangulation()
{
    if(use_dbg){/*no perturb update*/}
    else{
        perturb.Resize(particles.X.m);
        T epsilon=(T)1e-10;
        for(int i=1;i<=particles.X.m;i++){
            perturb(i)=random.Get_Uniform_Vector(-TV::Constant_Vector(epsilon),TV::Constant_Vector(epsilon));
        }
    }
    pX.Resize(particles.X.m);
    for(int i=1;i<=particles.X.m;i++){pX(i)=particles.X(i)+perturb(i);}

    ////update low dimensional particle indices
    triangulation_candidate_particles.Remove_All();
    for(int i=1;i<=particles.X.m;i++){triangulation_candidate_particles.Append(i);}

    front_particle_states.Resize(triangulation_candidate_particles.m);front_particle_states.Fill(AVAILABLE);
    front_particle_regions.Resize(triangulation_candidate_particles.m);front_particle_regions.Fill(0);
    //for(int i=1;i<=triangulation_candidate_particles.m;i++){
    //    int p=triangulation_candidate_particles(i);
    //}

    particle_bounding_box=Get_Particle_Bounding_Box();
    particle_grid=Get_Particle_Grid(particle_bounding_box,(T)2*alpha_ball_radius*(T)1.01);
    Resize_Elements_In_Cell_With_Grid(particles_in_cell,particle_grid);

    for(int i=1;i<=triangulation_candidate_particles.m;i++){
        int p=triangulation_candidate_particles(i);
        TV_INT cell_index=particle_grid.Cell(pX(p));if(!particle_grid.Inside_Domain(cell_index,0))continue;
        particles_in_cell(cell_index).Append(i);
    }

    int region=0;
    front_edge_hashtable.Remove_All();front_edge_hashtable.Clean_Memory();
    interior_edge_hashtable.Remove_All();interior_edge_hashtable.Clean_Memory();

    while(true){
        while(FRONT_EDGE* e=Find_Edge_From_Front()){
            int k=Find_Next_Triangle_Particle(e);
            //int k=Find_Next_Triangle_Particle_Ball_Pivoting(e);
            if(k!=0){
                int p_i=triangulation_candidate_particles(e->i);
                int p_j=triangulation_candidate_particles(e->j);
                int p_k=triangulation_candidate_particles(k);
                front_particle_regions(k)=region;
                triangle_mesh.elements.Append(VECTOR<int,3>(p_j,p_i,p_k));
                //LOG::cout<<"add triangle: "<<e->j<<", "<<e->i<<", "<<k<<": "<<VECTOR<int,3>(p_j,p_i,p_k)<<std::endl;
                Add_Particle_To_Front(e,k);
                if(use_dbg&&triangle_mesh.elements.m>=dbg_number){
                    //FRONT_EDGE* e=HASHTABLE_ITERATOR<VECTOR<int,2>,FRONT_EDGE*>(front_edge_hashtable).Data();
                    //Check_Front(e);

                    //LOG::cout<<"interior: ";
                    //for(HASHTABLE_ITERATOR<VECTOR<int,2> > iter(interior_edge_hashtable);iter.Valid();iter.Next()){
                    //    LOG::cout<<iter.Key()<<",";
                    //}LOG::cout<<std::endl;
                    return;
                }
            }
            else{
                e->active=false;
            }
        }

        VECTOR<int,3> seed_triangle=Find_Seed_Triangle();
        if(seed_triangle!=VECTOR<int,3>()){
            region++;
            int i=seed_triangle(1);int j=seed_triangle(2);int k=seed_triangle(3);
            int p_i=triangulation_candidate_particles(i);
            int p_j=triangulation_candidate_particles(j);
            int p_k=triangulation_candidate_particles(k);
            front_particle_regions(i)=front_particle_regions(j)=front_particle_regions(k)=region;
            triangle_mesh.elements.Append(VECTOR<int,3>(p_i,p_j,p_k));
            //LOG::cout<<"seed triangle: "<<i<<", "<<j<<", "<<k<<": "<<VECTOR<int,3>(p_i,p_j,p_k)<<std::endl;
            Add_Triangle_Edges_To_Front(seed_triangle);
        }
        else{////cannot find new seed triangle
            //LOG::cout<<"break"<<std::endl;
            break;
        }
    }

    //////add segments between boundary particles
    //Boundary_Particle_Segmentation();
}

template<class T>
void PhysBAM::BALL_PIVOTING_TRIANGULATION<T>::Check_Front(FRONT_EDGE* e)
{
    LOG::cout<<"Front: ";
    FRONT_EDGE* head=e;
    while(e){
        LOG::cout<<"["<<e->i<<", "<<e->j<<"]";
        e=e->next;
        if(e==0||e==head){LOG::cout<<std::endl;return;}
    }
}

template<class T>
FRONT_EDGE* PhysBAM::BALL_PIVOTING_TRIANGULATION<T>::Find_Edge_From_Front()
{   
    for(HASHTABLE_ITERATOR<VECTOR<int,2>,FRONT_EDGE*> iter(front_edge_hashtable);iter.Valid();iter.Next()){
        FRONT_EDGE* e=iter.Data();
        if(e->active)return e;
    }
    return 0;
}

template<class T>
bool PhysBAM::BALL_PIVOTING_TRIANGULATION<T>::Circle_Circle_Intersection(const TV2& c1,const T r1,const TV2& c2,const T r2,TV2& p1,TV2& p2)
{
    T d=(c2-c1).Magnitude();
    if(d>r1+r2||d<abs(r1-r2))return false;

    T L1=(pow<2>(r1)-pow<2>(r2)+pow<2>(d))/((T)2*d);
    T h1=(T)sqrt(pow<2>(r1)-pow<2>(L1));
    TV2 n1=(c2-c1).Normalized();
    p1=c1+n1*L1+n1.Perpendicular()*h1;
    p2=c1+n1*L1-n1.Perpendicular()*h1;
    return true;
}

template<class T>
int PhysBAM::BALL_PIVOTING_TRIANGULATION<T>::Find_Next_Triangle_Particle_Ball_Pivoting(const FRONT_EDGE* e)
{
    //////
    //TV2 c1=TV2();TV2 c2=TV2((T)1,(T)0);TV2 t1,t2;
    //LOG::cout<<Circle_Circle_Intersection(c1,(T)1,c2,(T)1,t1,t2)<<": "<<t1<<", "<<t2<<std::endl;

    int i=e->i;
    int p_i=triangulation_candidate_particles(i);

    int j=e->j;
    int p_j=triangulation_candidate_particles(j);

    int v=e->v;
    int p_v=triangulation_candidate_particles(v);
    TV n_ijv=(TV::Cross_Product(pX(p_j)-pX(p_i),pX(p_v)-pX(p_i))).Normalized();

    ARRAY<int> nb_particles;
    TV mid=(T).5*(pX(p_i)+pX(p_j));
    TV_INT cell_index_mid=particle_grid.Cell(mid);
    if(!particle_grid.Inside_Domain(cell_index_mid,0))return 0;
    for(int nb_c=1;nb_c<=pow<TV::m>(3);nb_c++){
        TV_INT nb_offset=Codimensional::CELL_NEIGHBOR_ITERATOR<TV::m>(nb_c);
        TV_INT nb_cell_idx=cell_index_mid+nb_offset;
        for(int nb_k=1;nb_k<=particles_in_cell(nb_cell_idx).m;nb_k++){
            int nb_i=particles_in_cell(nb_cell_idx)(nb_k);
            int nb_p=triangulation_candidate_particles(nb_i);
            if(nb_i==i||nb_i==j||nb_i==v||(particles.X(nb_p)-mid).Magnitude()>alpha_ball_radius*(T)2)continue;

            ////skip the particle if: 1. not avaliable or front or terminated; 2. ==i or j; 4. too big between angle; 
            ////5. has non-manifold edge; 6. incorrect order on front 
            if(front_particle_states(nb_i)!=AVAILABLE&&front_particle_states(nb_i)!=TERMINATED&&front_particle_states(nb_i)!=FRONT)continue;

            TV n_kji=(TV::Cross_Product(pX(p_j)-pX(nb_p),pX(p_i)-pX(nb_p))).Normalized(); ////triangle kji		
            if(TV::Angle_Between(n_ijv,n_kji)>(T).25*(T)pi){continue;}
            if(front_edge_hashtable.Contains(VECTOR<int,2>(i,nb_i))||front_edge_hashtable.Contains(VECTOR<int,2>(nb_i,j))){continue;}
            if(interior_edge_hashtable.Contains(VECTOR<int,2>(i,nb_i).Sorted())||interior_edge_hashtable.Contains(VECTOR<int,2>(j,nb_i).Sorted())){continue;}

            nb_particles.Append(nb_i);
        }
    }

    TV circumcircle_center=TRIANGLE_3D<T>::Circumcenter(pX(p_i),pX(p_j),pX(p_v));
    T circumcircle_radius=(circumcircle_center-pX(p_i)).Magnitude();
    if(circumcircle_radius>alpha_ball_radius){LOG::cout<<"Invalid triangle int ball pivoting"<<std::endl;return 0;}
    T h=sqrt(pow<2>(alpha_ball_radius)-pow<2>(circumcircle_radius));
    TV circumsphere_center=circumcircle_center+n_ijv*h;

    TV n_ij=(pX(p_j)-pX(p_i)).Normalized();
    TV ny=(TV::Cross_Product(n_ij,circumsphere_center-pX(p_i))).Normalized();
    TV nx=(TV::Cross_Product(ny,n_ij)).Normalized();
    MATRIX<T,3,2> proj_T=MATRIX<T,3,2>(nx,ny);
    MATRIX<T,2,3> proj=proj_T.Transposed();


    TV2 c_on_plane=proj*(circumsphere_center-mid);
    T rc_on_plane=c_on_plane.Magnitude();
    //T a_start=TV2::Oriented_Angle_Between(TV2::Axis_Vector(1),c_on_plane);if(a_start<(T)0)a_start+=(T)two_pi;
    T a_start=(T)0;

    ////debug
    if(use_dbg&&triangle_mesh.elements.m>=dbg_number-1){
        LOG::cout<<"i, j, v: "<<triangulation_candidate_particles(i)<<", "<<triangulation_candidate_particles(j)<<", "<<triangulation_candidate_particles(v)<<std::endl;
        LOG::cout<<"nbs: ";
        for(int i=1;i<=nb_particles.m;i++){
            LOG::cout<<triangulation_candidate_particles(nb_particles(i))<<" ";
        }LOG::cout<<std::endl;
        LOG::cout<<"a_start: "<<a_start<<std::endl;
        LOG::cout<<"rc: "<<rc_on_plane<<std::endl;
        LOG::cout<<"nx,ny,nz: "<<nx<<", "<<ny<<", "<<n_ij<<std::endl;
    }

    T angle_min=FLT_MAX;int i_min=0;
    for(int nb_c=1;nb_c<=nb_particles.m;nb_c++){
        int k=nb_particles(nb_c);
        int p_k=triangulation_candidate_particles(k);
        TV2 k_on_plane=proj*(pX(p_k)-mid);
        TV proj_k=mid+proj_T*k_on_plane;
        T Lk2=(proj_k-pX(p_k)).Magnitude_Squared();
        if(use_dbg&&triangle_mesh.elements.m>=dbg_number-1){
            LOG::cout<<"p_k: "<<pX(p_k)<<", proj_pk: "<<proj_k<<std::endl;
        }
        if(Lk2>pow<2>(alpha_ball_radius))continue;
        T rk_on_plane=(T)sqrt(pow<2>(alpha_ball_radius)-Lk2);

        if(use_dbg&&triangle_mesh.elements.m>=dbg_number-1){
            LOG::cout<<"nb on plane: "<<k_on_plane<<", "<<rk_on_plane<<std::endl;
        }

        TV2 p1,p2;
        if(Circle_Circle_Intersection(k_on_plane,rk_on_plane,TV2(),rc_on_plane,p1,p2)){
            T a1=TV2::Oriented_Angle_Between(TV2::Axis_Vector(1),p1);if(a1<(T)0)a1+=(T)two_pi;
            //a1-=a_start;if(a1<(T)0)a1+=(T)two_pi;
            T a2=TV2::Oriented_Angle_Between(TV2::Axis_Vector(1),p2);if(a2<(T)0)a2+=(T)two_pi;
            //a2-=a_start;if(a2<(T)0)a2+=(T)two_pi;
            TV2 ins_p=a1<=a2?p1:p2;T ins_a=a1<=a2?a1:a2;
            if(use_dbg&&triangle_mesh.elements.m>=dbg_number-1){
                LOG::cout<<"test: "<<triangulation_candidate_particles(k)<<": angle: "<<ins_a<<std::endl;
            }
            if(ins_a<angle_min){
                angle_min=ins_a;
                i_min=k;
            }
        }
    }

    if(use_dbg&&triangle_mesh.elements.m>=dbg_number-1){
        LOG::cout<<"find k: "<<triangulation_candidate_particles(i_min)<<std::endl;
    }

    return i_min;
}


template<class T>
int PhysBAM::BALL_PIVOTING_TRIANGULATION<T>::Find_Next_Triangle_Particle(const FRONT_EDGE* e)
{
    int i=e->i;
    int p_i=triangulation_candidate_particles(i);

    int j=e->j;
    int p_j=triangulation_candidate_particles(j);

    int v=e->v;
    int p_v=triangulation_candidate_particles(v);
    TV n1=(TV::Cross_Product(pX(p_i)-pX(p_v),pX(p_j)-pX(p_v))).Normalized();	////triangle vij

    //LOG::cout<<"v, i, j: "<<p_v<<", "<<p_i<<", "<<p_j<<std::endl;

    ARRAY<int> nb_particles;
    TV mid=(T).5*(pX(p_i)+pX(p_j));
    TV_INT cell_index_mid=particle_grid.Cell(mid);
    if(!particle_grid.Inside_Domain(cell_index_mid,0))return 0;
    for(int nb_c=1;nb_c<=pow<TV::m>(3);nb_c++){
        TV_INT nb_offset=Codimensional::CELL_NEIGHBOR_ITERATOR<TV::m>(nb_c);
        TV_INT nb_cell_idx=cell_index_mid+nb_offset;
        for(int nb_k=1;nb_k<=particles_in_cell(nb_cell_idx).m;nb_k++){
            int nb_i=particles_in_cell(nb_cell_idx)(nb_k);
            int nb_p=triangulation_candidate_particles(nb_i);
            if(nb_i==i||nb_i==j||(particles.X(nb_p)-mid).Magnitude()>alpha_ball_radius*(T)2)continue;
            nb_particles.Append(nb_i);
        }
    }

    int k=0;
    ARRAY<int> candidate_k;
    for(int nb_k=1;nb_k<=nb_particles.m;nb_k++){
        int c=nb_particles(nb_k);
        int p_c=triangulation_candidate_particles(c);

        ////fill hole on the mesh first
        if(c!=v&&e->next->j==c&&e->next->next->j==i){return c;}

        ////skip the particle if: 1. not avaliable or front or terminated; 2. ==i or j; 4. too big between angle; 
        ////5. has non-manifold edge; 6. incorrect order on front 
        if((front_particle_states(c)!=AVAILABLE&&front_particle_states(c)!=TERMINATED&&front_particle_states(c)!=FRONT)||c==i||c==j||c==v){continue;}

        TV n2=(TV::Cross_Product(pX(p_j)-pX(p_c),pX(p_i)-pX(p_c))).Normalized(); ////triangle kji		
        if(TV::Angle_Between(n1,n2)>(T).25*(T)pi){continue;}
        if(front_edge_hashtable.Contains(VECTOR<int,2>(i,c))||front_edge_hashtable.Contains(VECTOR<int,2>(c,j))){continue;}
        if(interior_edge_hashtable.Contains(VECTOR<int,2>(i,c).Sorted())||interior_edge_hashtable.Contains(VECTOR<int,2>(j,c).Sorted())){continue;}

        candidate_k.Append(c);
    }

    ////debug
    if(use_dbg&&triangle_mesh.elements.m>=dbg_number-1){
        LOG::cout<<"candidate k: ";
        for(int i=1;i<=candidate_k.m;i++){
            LOG::cout<<triangulation_candidate_particles(candidate_k(i))<<" ";
        }LOG::cout<<std::endl;
        LOG::cout<<"nb p: ";
        for(int i=1;i<=nb_particles.m;i++){
            LOG::cout<<triangulation_candidate_particles(nb_particles(i))<<" ";
        }LOG::cout<<std::endl;
        //LOG::cout<<"tri p: ";
        //for(int i=1;i<=triangulation_candidate_particles.m;i++){
        //    LOG::cout<<triangulation_candidate_particles(i)<<" ";
        //}LOG::cout<<std::endl;
    }

    for(int nb_k=1;nb_k<=candidate_k.m;nb_k++){
        int c=candidate_k(nb_k);
        int p_c=triangulation_candidate_particles(c);
        TV cj_x_ci=TV::Cross_Product(pX(p_j)-pX(p_c),pX(p_i)-pX(p_c));
        if(cj_x_ci.Magnitude()<epsilon)continue;    ////k,j,i are colinear
        TV n2=(cj_x_ci).Normalized(); ////normal for triangle kji		
        TV circumcircle_center=TRIANGLE_3D<T>::Circumcenter(pX(p_i),pX(p_j),pX(p_c));
        T circumcircle_radius=(circumcircle_center-pX(p_i)).Magnitude();
        if(circumcircle_radius>alpha_ball_radius)continue;
        T h=sqrt(pow<2>(alpha_ball_radius)-pow<2>(circumcircle_radius));

        TV ij=pX(p_j)-pX(p_i);
        //T dir=TV::Dot_Product(TV::Cross_Product(n1,n2),ij)>=(T)0?(T)1:(T)-1;
        T dir=(T)1;
        TV circumsphere_center=circumcircle_center+dir*n2*h;
        bool has_inside_p=false;
        int inside_i=0;
        for(int nb_j=1;nb_j<=nb_particles.m;nb_j++){
            if(c==nb_particles(nb_j)||i==nb_particles(nb_j)||j==nb_particles(nb_j))continue;
            int p_k=triangulation_candidate_particles(nb_particles(nb_j));
            if(SPHERE<TV>(circumsphere_center,alpha_ball_radius).Lazy_Inside(pX(p_k))){
                has_inside_p=true;inside_i=nb_particles(nb_j);break;
            }
        }

        ////debug
        if(use_dbg&&triangle_mesh.elements.m>=dbg_number-1){
            if(has_inside_p){
                LOG::cout<<"inside: "<<triangulation_candidate_particles(c)<<": "<<triangulation_candidate_particles(inside_i)<<": sc: "
                    <<circumsphere_center<<", dir: "<<dir<<", n2: "<<n2<<", cc: "<<circumcircle_center<<", p: "<<pX(triangulation_candidate_particles(inside_i))<<", h: "<<h<<", sr: "<<alpha_ball_radius<<", cr: "<<circumcircle_radius<<std::endl;
            }
            else{
                LOG::cout<<"find : "<<triangulation_candidate_particles(c)<<": sc: "
                    <<circumsphere_center<<", dir: "<<dir<<", cc: "<<circumcircle_center<<", h: "<<h<<", sr: "<<alpha_ball_radius<<", cr: "<<circumcircle_radius<<std::endl;
            }
        }

        if(!has_inside_p){
            ////debug
            if(k!=0)LOG::cout<<"change k "<<triangulation_candidate_particles(k)<<" : "<<triangulation_candidate_particles(c)<<std::endl;
            else if(use_dbg&&triangle_mesh.elements.m>=dbg_number-1){LOG::cout<<"initialize k "<<c<<": "<<triangulation_candidate_particles(c)<<std::endl;}

            k=c;
        }
    }

    ////debug
    if(use_dbg&&triangle_mesh.elements.m>=dbg_number-1){
        if(k!=0)LOG::cout<<"i, j: "<<p_i<<", "<<p_j<<", k: "<<k<<", "<<triangulation_candidate_particles(k)<<std::endl;
    }

    return k;
}

template<class T>
VECTOR<int,3> PhysBAM::BALL_PIVOTING_TRIANGULATION<T>::Find_Seed_Triangle()
{
    int i,j,k,p_i,p_j;
    for(i=1;i<=front_particle_states.m;i++){
        if(front_particle_states(i)==AVAILABLE){
            ////find the first vertex
            //LOG::cout<<"i: "<<i<<std::endl;
            p_i=triangulation_candidate_particles(i);
            TV_INT cell_index_i=particle_grid.Cell(pX(p_i));
            if(!particle_grid.Inside_Domain(cell_index_i,0))continue;

            ////find the second vertex
            ARRAY<int> nb_particles;
            for(int nb_c=1;nb_c<=pow<TV::m>(3);nb_c++){
                TV_INT nb_offset=Codimensional::CELL_NEIGHBOR_ITERATOR<TV::m>(nb_c);
                TV_INT nb_cell_idx=cell_index_i+nb_offset;
                for(int nb_k=1;nb_k<=particles_in_cell(nb_cell_idx).m;nb_k++){
                    int nb_i=particles_in_cell(nb_cell_idx)(nb_k);
                    nb_particles.Append(nb_i);
                }
            }
            T nearest_distance=(T)FLT_MAX;
            int nearest_i=0;
            for(int nb_k=1;nb_k<=nb_particles.m;nb_k++){
                int c=nb_particles(nb_k);
                int p_c=triangulation_candidate_particles(c);

                ////skip the particle if: 1. not avaliable; 2. itself; 3. too far; 
                if((front_particle_states(c)!=AVAILABLE&&front_particle_states(c)!=TERMINATED)||c==i){continue;}
                T length=(pX(p_i)-pX(p_c)).Magnitude();
                if(length>max_segment_length){continue;}

                if(length<nearest_distance){
                    nearest_distance=length;
                    nearest_i=c;
                }
            }
            if(nearest_i==0)continue;

            ////find the third vertex
            j=nearest_i;
            p_j=triangulation_candidate_particles(j);
            //LOG::cout<<"j: "<<j<<std::endl;

            TV_INT cell_index_j=particle_grid.Cell(pX(p_j));
            if(!particle_grid.Inside_Domain(cell_index_j,0))continue;

            for(int nb_c=1;nb_c<=pow<TV::m>(3);nb_c++){
                TV_INT nb_offset=Codimensional::CELL_NEIGHBOR_ITERATOR<TV::m>(nb_c);
                TV_INT nb_cell_idx=cell_index_j+nb_offset;
                for(int nb_k=1;nb_k<=particles_in_cell(nb_cell_idx).m;nb_k++){
                    int nb_i=particles_in_cell(nb_cell_idx)(nb_k);
                    nb_particles.Append_Unique(nb_i);
                }
            }
            T min_circumradius=(T)FLT_MAX;
            int min_i=0;
            for(int nb_k=1;nb_k<=nb_particles.m;nb_k++){
                int c=nb_particles(nb_k);
                int p_c=triangulation_candidate_particles(c);

                ////skip the particle if: 1. not avaliable; 2. ==i or j; 3. too far 4. too big circumradius; 
                if((front_particle_states(c)!=AVAILABLE&&front_particle_states(c)!=TERMINATED)||c==i||c==j)continue;
                if((pX(p_i)-pX(p_c)).Magnitude()>max_segment_length||(pX(p_j)-pX(p_c)).Magnitude()>max_segment_length){continue;}
                TV circumcenter=TRIANGLE_3D<T>::Circumcenter(pX(p_i),pX(p_j),pX(p_c));
                T circumradius=(circumcenter-pX(p_i)).Magnitude();
                if(circumradius>alpha_ball_radius){continue;}

                if(circumradius<min_circumradius){
                    min_circumradius=circumradius;
                    min_i=c;
                }
            }
            if(min_i==0)continue;

            k=min_i;
            //LOG::cout<<"k: "<<k<<std::endl;
            return VECTOR<int,3>(i,j,k);
        }
    }
    return VECTOR<int,3>();
}

template<class T>
FRONT_EDGE* PhysBAM::BALL_PIVOTING_TRIANGULATION<T>::Add_Triangle_Edges_To_Front(const VECTOR<int,3>& v)
{
    FRONT_EDGE* e1=new FRONT_EDGE(v(1),v(2),v(3));
    FRONT_EDGE* e2=new FRONT_EDGE(v(2),v(3),v(1));
    FRONT_EDGE* e3=new FRONT_EDGE(v(3),v(1),v(2));
    Link_Edges(e1,e2);Link_Edges(e2,e3);Link_Edges(e3,e1);
    front_particle_states(v(1))=FRONT;front_particle_states(v(2))=FRONT;front_particle_states(v(3))=FRONT;
    front_edge_hashtable.Set(VECTOR<int,2>(e1->i,e1->j),e1);
    front_edge_hashtable.Set(VECTOR<int,2>(e2->i,e2->j),e2);
    front_edge_hashtable.Set(VECTOR<int,2>(e3->i,e3->j),e3);
    if(front_particle_states(e1->i)==TERMINATED&&front_particle_states(e1->j)==TERMINATED&&front_particle_regions(e1->i)==front_particle_regions(e1->j))e1->active=false;
    if(front_particle_states(e2->i)==TERMINATED&&front_particle_states(e2->j)==TERMINATED&&front_particle_regions(e2->i)==front_particle_regions(e2->j))e2->active=false;
    if(front_particle_states(e3->i)==TERMINATED&&front_particle_states(e3->j)==TERMINATED&&front_particle_regions(e3->i)==front_particle_regions(e3->j))e3->active=false;
    return e1;
}

template<class T>
void PhysBAM::BALL_PIVOTING_TRIANGULATION<T>::Add_Particle_To_Front(FRONT_EDGE* & e,const int k)
{
    FRONT_EDGE* e1=new FRONT_EDGE(e->i,k,e->j);
    FRONT_EDGE* e2=new FRONT_EDGE(k,e->j,e->i);
    Link_Edges(e->pre,e1);Link_Edges(e1,e2);Link_Edges(e2,e->next);
    if(front_particle_states(e1->i)==TERMINATED&&front_particle_states(e1->j)==TERMINATED&&front_particle_regions(e1->i)==front_particle_regions(e1->j))e1->active=false;
    if(front_particle_states(e2->i)==TERMINATED&&front_particle_states(e2->j)==TERMINATED&&front_particle_regions(e2->i)==front_particle_regions(e2->j))e2->active=false;
    if(front_particle_states(k)!=TERMINATED)front_particle_states(k)=FRONT;

    front_edge_hashtable.Delete_If_Present(VECTOR<int,2>(e->i,e->j));
    interior_edge_hashtable.Set(VECTOR<int,2>(e->i,e->j).Sorted());
    //LOG::cout<<"delete "<<VECTOR<int,2>(e->i,e->j)<<std::endl;
    delete e;

    ////add e1, e2 to front
    FRONT_EDGE* re1=0;bool glue_e1=false;
    if(front_edge_hashtable.Get(VECTOR<int,2>(e1->j,e1->i),re1)){
        front_edge_hashtable.Delete_If_Present(VECTOR<int,2>(e1->j,e1->i));
        //LOG::cout<<"delete edge: "<<VECTOR<int,2>(e1->j,e1->i)<<std::endl;
        if(e1->pre==re1)front_particle_states(e1->i)=USED;
        glue_e1=true;
    }
    else{
        front_edge_hashtable.Set(VECTOR<int,2>(e1->i,e1->j),e1);
        //LOG::cout<<"add edge: "<<VECTOR<int,2>(e1->i,e1->j)<<std::endl;
    }
    FRONT_EDGE* re2=0;bool glue_e2=false;
    if(front_edge_hashtable.Get(VECTOR<int,2>(e2->j,e2->i),re2)){
        front_edge_hashtable.Delete_If_Present(VECTOR<int,2>(e2->j,e2->i));
        //LOG::cout<<"delete edge: "<<VECTOR<int,2>(e2->j,e2->i)<<std::endl;
        if(e2->next==re2)front_particle_states(e2->j)=USED;
        glue_e2=true;
    }
    else{
        front_edge_hashtable.Set(VECTOR<int,2>(e2->i,e2->j),e2);
        //LOG::cout<<"add edge: "<<VECTOR<int,2>(e2->i,e2->j)<<std::endl;
    }
    if(re2&&re2->next==re1)front_particle_states(e1->j)=USED;

    if(glue_e1){Link_Edges(e1->pre,re1->next);Link_Edges(re1->pre,e1->next);}
    if(glue_e2){Link_Edges(e2->pre,re2->next);Link_Edges(re2->pre,e2->next);}

    if(glue_e1){interior_edge_hashtable.Set(VECTOR<int,2>(e1->i,e1->j).Sorted());delete e1;e1=0;delete re1;re1=0;}
    if(glue_e2){interior_edge_hashtable.Set(VECTOR<int,2>(e2->i,e2->j).Sorted());delete e2;e2=0;delete re2;re2=0;}
}

template<class T>
BOX<VECTOR<T,3> > PhysBAM::BALL_PIVOTING_TRIANGULATION<T>::Get_Particle_Bounding_Box()
{
    BOX<TV> bounding_box=BOX<TV>(TV((T)FLT_MAX),TV((T)-FLT_MAX));
    for(int i=1;i<=particles.X.m;i++){
        for(int d=1;d<=TV::m;d++){
            if(particles.X(i)(d)<bounding_box.min_corner(d))bounding_box.min_corner(d)=particles.X(i)(d);
            else if(particles.X(i)(d)>bounding_box.max_corner(d))bounding_box.max_corner(d)=particles.X(i)(d);
        }
    }
    bounding_box.Change_Size(TV::Constant_Vector((T)2*alpha_ball_radius));
    return bounding_box;
}

template<class T>
GRID<VECTOR<T,3> > PhysBAM::BALL_PIVOTING_TRIANGULATION<T>::Get_Particle_Grid(const BOX<TV>& bounding_box,const T grid_dX)
{
    TV_INT particle_grid_counts=(TV_INT)(bounding_box.Edge_Lengths()/grid_dX)+1;
    BOX<TV> particle_grid_box;
    particle_grid_box.min_corner=bounding_box.min_corner;
    particle_grid_box.max_corner=bounding_box.min_corner+TV::Constant_Vector(grid_dX)*(TV)particle_grid_counts;
    GRID<TV> p_grid;p_grid.Initialize(particle_grid_counts,particle_grid_box,true);
    return p_grid;
}


template<class T>
void PhysBAM::BALL_PIVOTING_TRIANGULATION<T>::Resize_Elements_In_Cell_With_Grid(ARRAY<ARRAY<int>,TV_INT>& p_in_cell,const GRID<TV>& grid)
{
    p_in_cell.Resize(grid.Domain_Indices(1));
    for(UNIFORM_GRID_ITERATOR_CELL<TV> iter(grid,1);iter.Valid();iter.Next()){
        p_in_cell(iter.Cell_Index()).Remove_All();
    }
}

template class BALL_PIVOTING_TRIANGULATION<float>;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
template class BALL_PIVOTING_TRIANGULATION<double>;
#endif
