//#####################################################################
// Copyright 2013, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __CODIMENSIONAL_MESH__
#define __CODIMENSIONAL_MESH__

#include <PhysBAM_Geometry/Topology/POINT_SIMPLEX_MESH.h>
#include <PhysBAM_Geometry/Topology/SEGMENT_MESH.h>
#include <PhysBAM_Geometry/Topology/TRIANGLE_MESH.h>
#include <PhysBAM_Geometry/Topology/TETRAHEDRON_MESH.h>
#include <PhysBAM_Dynamics/Particles/CODIMENSIONAL_FLUID_PARTICLES.h>
#include <PhysBAM_Tools/Data_Structures/HASHTABLE.h>

namespace PhysBAM
{
template<class TV> class MPI_CODIMENSIONAL;

namespace Codimensional
{
const int PARTICLE_ON_SEGMENT_MESH=0x1;
const int PARTICLE_ON_SEGMENT_MESH_BOUNDARY=0x2;
const int PARTICLE_ON_TRIANGLE_MESH=0x4;
const int PARTICLE_ON_TRIANGLE_MESH_BOUNDARY=0x8;
const int PARTICLE_ON_TETRAHEDRON_MESH=0x10;
const int PARTICLE_ON_TETRAHEDRON_MESH_BOUNDARY=0x20;

inline void Set_On_Segment_Mesh(int& type){type|=PARTICLE_ON_SEGMENT_MESH;}
inline void Set_On_Triangle_Mesh(int& type){type|=PARTICLE_ON_TRIANGLE_MESH;}
inline void Set_On_Tetrahedron_Mesh(int& type){type|=PARTICLE_ON_TETRAHEDRON_MESH;}
inline void Set_On_Segment_Mesh_Boundary(int& type){type|=PARTICLE_ON_SEGMENT_MESH_BOUNDARY;}
inline void Set_On_Triangle_Mesh_Boundary(int& type){type|=PARTICLE_ON_TRIANGLE_MESH_BOUNDARY;}
inline void Set_On_Tetrahedron_Mesh_Boundary(int& type){type|=PARTICLE_ON_TETRAHEDRON_MESH_BOUNDARY;}
inline bool Is_On_Segment_Mesh(const int type){return (type & PARTICLE_ON_SEGMENT_MESH)!=0;}
inline bool Is_On_Triangle_Mesh(const int type){return (type & PARTICLE_ON_TRIANGLE_MESH)!=0;}
inline bool Is_On_Tetrahedron_Mesh(const int type){return (type & PARTICLE_ON_TETRAHEDRON_MESH)!=0;}
inline bool Is_On_Segment_Mesh_Boundary(const int type){return (type & PARTICLE_ON_SEGMENT_MESH_BOUNDARY)!=0;}
inline bool Is_On_Triangle_Mesh_Boundary(const int type){return (type & PARTICLE_ON_TRIANGLE_MESH_BOUNDARY)!=0;}
inline bool Is_On_Tetrahedron_Mesh_Boundary(const int type){return (type & PARTICLE_ON_TETRAHEDRON_MESH_BOUNDARY)!=0;}
inline bool Is_On_Segment_Mesh_Interior(const int type){return Is_On_Segment_Mesh(type) && !Is_On_Segment_Mesh_Boundary(type);}
inline bool Is_On_Triangle_Mesh_Interior(const int type){return Is_On_Triangle_Mesh(type) && !Is_On_Triangle_Mesh_Boundary(type);}
inline bool Is_On_Tetrahedron_Mesh_Interior(const int type){return Is_On_Tetrahedron_Mesh(type) && !Is_On_Tetrahedron_Mesh_Boundary(type);}
inline bool Is_Isolated(const int type){return type==0;}
inline bool Is_On_Codimensional_Mesh_Free_Surface(const int type){return Is_On_Segment_Mesh(type)||Is_On_Triangle_Mesh(type)||Is_On_Tetrahedron_Mesh_Boundary(type);}
inline bool Is_On_Codimensional_Mesh_Boundary(const int type){
    return (Is_On_Segment_Mesh_Boundary(type)&&!Is_On_Triangle_Mesh_Interior(type)&&!Is_On_Tetrahedron_Mesh_Interior(type))||
    (Is_On_Triangle_Mesh_Boundary(type)&&!Is_On_Tetrahedron_Mesh_Interior(type)&&!Is_On_Segment_Mesh_Interior(type))||
    (Is_On_Segment_Mesh_Boundary(type)&&!Is_On_Triangle_Mesh_Interior(type)&&!Is_On_Tetrahedron_Mesh_Interior(type));}

template<int d> struct CODIMENSIONAL_NODE_TYPE_HELPER;
template<> struct CODIMENSIONAL_NODE_TYPE_HELPER<2>
{
    static bool Is_Joint(const int type){return Is_On_Segment_Mesh(type) && Is_On_Triangle_Mesh(type);}
    static bool Is_On_Volume_Mesh(const int type){return Is_On_Triangle_Mesh(type);}
    static bool Is_On_Surface_Mesh(const int type){return Is_On_Triangle_Mesh_Boundary(type);}
};
template<> struct CODIMENSIONAL_NODE_TYPE_HELPER<3>
{
    static bool Is_Joint(const int type){return (Is_On_Segment_Mesh(type)?1:0)+(Is_On_Triangle_Mesh(type)?1:0)+(Is_On_Tetrahedron_Mesh(type)?1:0)>=2;}
    static bool Is_On_Volume_Mesh(const int type){return Is_On_Tetrahedron_Mesh(type);}
    static bool Is_On_Surface_Mesh(const int type){return Is_On_Triangle_Mesh(type)||Is_On_Tetrahedron_Mesh_Boundary(type);}
};

template<int dim> bool Is_Joint(const int type){return CODIMENSIONAL_NODE_TYPE_HELPER<dim>::Is_Joint(type);}
template<int dim> bool Is_On_Volume_Mesh(const int type){return CODIMENSIONAL_NODE_TYPE_HELPER<dim>::Is_On_Volume_Mesh(type);}
template<int dim> bool Is_On_Surface_Mesh(const int type){return CODIMENSIONAL_NODE_TYPE_HELPER<dim>::Is_On_Surface_Mesh(type);}

enum SIMPLEX_TYPE{SIMPLEX_PARTICLE=1,SIMPLEX_SEGMENT,SIMPLEX_TRIANGLE,SIMPLEX_TETRAHEDRON};

class CODIMENSIONAL_ELEMENT
{public:
    int index;int type;VECTOR<int,4> vertices;
    CODIMENSIONAL_ELEMENT():index(0),type(0),vertices(VECTOR<int,4>()){}
    CODIMENSIONAL_ELEMENT(int index,int type,const VECTOR<int,4>& vertices):index(index),type(type),vertices(vertices){}
    CODIMENSIONAL_ELEMENT(const CODIMENSIONAL_ELEMENT& copy){*this=copy;}
    const CODIMENSIONAL_ELEMENT& operator=(const CODIMENSIONAL_ELEMENT& copy){index=copy.index;type=copy.type;vertices=copy.vertices;return *this;}

    VECTOR<int,2> Segment_Element(){PHYSBAM_ASSERT(type==SIMPLEX_SEGMENT);return VECTOR<int,2>(vertices(1),vertices(2));}
    VECTOR<int,3> Triangle_Element(){PHYSBAM_ASSERT(type==SIMPLEX_TRIANGLE);return VECTOR<int,3>(vertices(1),vertices(2),vertices(3));}
    VECTOR<int,4> Tetrahedron_Element(){PHYSBAM_ASSERT(type==SIMPLEX_TETRAHEDRON);return VECTOR<int,4>(vertices(1),vertices(2),vertices(3),vertices(4));}
};
}

using namespace Codimensional;
class CODIMENSIONAL_MESH
{
    typedef VECTOR<int,2> TV_INT2;typedef VECTOR<int,3> TV_INT3;typedef VECTOR<int,4> TV_INT4;
public:
    int number_nodes;
    int number_of_regions;
    int number_of_regions_in_high_dim;

    SEGMENT_MESH& segment_mesh;
    TRIANGLE_MESH& triangle_mesh;
    TETRAHEDRON_MESH& tetrahedron_mesh;
    ARRAY<int> node_connectivities;     
    ARRAY<int> node_connectivities_in_high_dim; ////ignoring the segments for connectivity computation
    ARRAY<int> node_types;

    ARRAY<ARRAY<int> >* neighbor_nodes; // for each node, list of neighboring nodes
    //ARRAY<ARRAY<int> >* incident_elements; // for each node, list of neighboring simplices that contain it
    SEGMENT_MESH rim_segment_mesh;  // rim mesh, for 3d only, rim segments on triangle mesh boundary

    ////auxiliary hashtables
    HASHTABLE<int,ARRAY<int> > vertex_vertex_hashtable;
    HASHTABLE<int,ARRAY<int> > vertex_tetrahedron_hashtable;
    HASHTABLE<int,ARRAY<int> > vertex_triangle_hashtable;
    HASHTABLE<int,ARRAY<int> > vertex_segment_hashtable;
    HASHTABLE<TV_INT2,ARRAY<int> > edge_tetrahedron_hashtable;
    HASHTABLE<TV_INT2,ARRAY<int> > edge_triangle_hashtable;
    HASHTABLE<TV_INT3,ARRAY<int> > face_tetrahedron_hashtable;

    HASHTABLE<TV_INT2> edge_on_tri_mesh_hashtable;
    HASHTABLE<TV_INT2> edge_on_tet_mesh_hashtable;
    HASHTABLE<TV_INT2> edge_on_tet_mesh_boundary_hashtable;
    HASHTABLE<TV_INT3> face_on_tet_mesh_boundary_hashtable;
    HASHTABLE<TV_INT2> rim_segment_hashtable;
    
    HASHTABLE<TV_INT2> connected_segment_hashtable;

    CODIMENSIONAL_MESH(SEGMENT_MESH* segment_mesh_input=0,TRIANGLE_MESH* triangle_mesh_input=0,TETRAHEDRON_MESH* tetrahedron_mesh_input=0);
    CODIMENSIONAL_MESH(const CODIMENSIONAL_MESH& copy);
    CODIMENSIONAL_MESH& operator=(const CODIMENSIONAL_MESH& copy){Initialize_Mesh(copy);return *this;}
    ~CODIMENSIONAL_MESH(){if(neighbor_nodes) delete neighbor_nodes;}

    void Initialize_Mesh(const CODIMENSIONAL_MESH& copy);
    //void Update_Rim_Segment_Mesh();
    void Update_Number_Nodes(int number_nodes_input);
    void Update_Auxiliary_Structures();
    template<class TV> int Update_Node_Connectivities(MPI_CODIMENSIONAL<TV>* mpi_codim);
    template<class TV> int Update_Connectivities_Across_Mpi_Nodes(MPI_CODIMENSIONAL<TV>* mpi_codim);
    bool Common_Hashtable_Entry(const HASHTABLE<int>& lhs,const HASHTABLE<int>& rhs);
    void Get_Region_Correspondences(ARRAY<PAIR<int,int> >& correspondences);

    void Update_Node_Types(ARRAY<int>& node_types); ////needs auxiliary structure: segment_mesh.boundary_mesh, triangle_mesh.boundary_mesh, tetrahedron_mesh.boundary_mesh
    template<class TV> void Update_States(int number_nodes_input,MPI_CODIMENSIONAL<TV>* mpi_codim=0);  ////update everything

    void Update_Segment_Auxiliary_Structures();
    void Update_Triangle_Auxiliary_Structures();
    void Update_Tetrahedron_Auxiliary_Structures();
    void Update_Segment_Boundary_Mesh();
    void Update_Triangle_Boundary_Mesh();
    void Update_Tetrahedron_Boundary_Mesh();
    void Update_Neighbor_Nodes();

    ////topological property functions
    int Particle_Region_Index(const int p) const {return node_connectivities(p);}  ////can also use particles.id
    int Segment_Region_Index(const int s) const {return node_connectivities(segment_mesh.elements(s)(1));}
    int Triangle_Region_Index(const int t) const {return node_connectivities(triangle_mesh.elements(t)(1));}
    int Tetrahedron_Region_Index(const int t) const {return node_connectivities(tetrahedron_mesh.elements(t)(1));}
    int Surface_Region_Index(const int t,const int dim) const {return dim==2?Segment_Region_Index(t):Triangle_Region_Index(t); }
    int Volumetric_Region_Index(const int t,const int dim) const {return dim==2?Triangle_Region_Index(t):Tetrahedron_Region_Index(t);}

    bool Is_Boundary_Edge_On_Triangle_Mesh(const TV_INT2& v) const;   ////needs auxiliary structure: triangle_mesh.boundary_mesh->neighbor_nodes
    bool Is_Boundary_Edge_On_Tetrahedron_Mesh(const TV_INT2& v) const;    ////needs auxiliary structure: tetrahedron_mesh.boundary_mesh->neighbor_nodes
    bool Is_Boundary_Face_On_Tetrahedron_Mesh(const TV_INT3& v) const;
    bool Is_Boundary_Particle_On_Triangle_Mesh(const int p) const;        ////needs auxiliary structure: rim_segment_mesh.incident_nodes
    bool Is_Boundary_Particle_On_Segment_Mesh(const int p) const; ////uses node_types,
    bool Is_Boundary_Particle_On_Tetrahedron_Mesh(const int p) const;
    bool Is_Rim_Particle(const int p) const;  ////on triangle mesh boundary or on segment mesh, uses node_types, needs auxiliary structure: rim_segment_mesh.incident_nodes
};
}

#endif
