//#####################################################################
// Copyright 2013, Bo Zhu, Ed Quigley
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <PhysBAM_Tools/Data_Structures/STACK.h>
#include <PhysBAM_Tools/Data_Structures/QUEUE.h>
#include <PhysBAM_Tools/Data_Structures/UNION_FIND.h>
#include <PhysBAM_Tools/Math_Tools/constants.h>
#include <PhysBAM_Tools/Matrices/MATRIX_MXN.h>
#include <PhysBAM_Tools/Matrices/MATRIX_2X3.h>
#include <PhysBAM_Tools/Matrices/MATRIX_3X2.h>
#include <PhysBAM_Tools/Matrices/MATRIX_3X3.h>
#include <PhysBAM_Tools/Matrices/ROTATION.h>
#include <PhysBAM_Tools/Arrays_Computations/SORT.h>
#include <PhysBAM_Tools/Grids_Uniform/UNIFORM_GRID_ITERATOR_CELL.h>
#include <PhysBAM_Tools/Interpolation/LINEAR_INTERPOLATION.h>
#include <PhysBAM_Tools/Math_Tools/constants.h>
#include <PhysBAM_Geometry/Basic_Geometry/TRIANGLE_2D.h>
#include <PhysBAM_Geometry/Basic_Geometry/TRIANGLE_3D.h>
#include <PhysBAM_Geometry/Basic_Geometry/SPHERE.h>
#include <PhysBAM_Geometry/Topology_Based_Geometry_Computations/TRIANGULATED_SURFACE_SUBDIVISION.h>
#include <PhysBAM_Geometry/Tessellation/SPHERE_TESSELLATION.h>
#include <PhysBAM_Geometry/Basic_Geometry/SPHERE.h>
#include "../Meshing/CODIMENSIONAL_MESHING_BASE.h"
#include "../Meshing/CODIMENSIONAL_MESH_CHECKER.h"

using namespace PhysBAM;
using namespace Codimensional;

template<class TV>
PhysBAM::CODIMENSIONAL_MESHING_BASE<TV>::CODIMENSIONAL_MESHING_BASE(CODIMENSIONAL_MESH_OBJECT<TV>& codimensional_mesh_object_input,MPI_CODIMENSIONAL<TV>* mpi_codim_input/*=0*/)
    :codimensional_mesh_object(codimensional_mesh_object_input),particles(codimensional_mesh_object.particles),
    mesh(codimensional_mesh_object.mesh),segment_mesh(codimensional_mesh_object.mesh.segment_mesh),triangle_mesh(codimensional_mesh_object.mesh.triangle_mesh),
    tetrahedron_mesh(codimensional_mesh_object.mesh.tetrahedron_mesh),rim_segment_mesh(codimensional_mesh_object.mesh.rim_segment_mesh),
    density((T)1000.),epsilon((T)1e-40),smooth_rim_thickness_weight((T).25),smooth_interior_thickness_weight((T).05),smooth_rim_thickness_mode(3),rim_interior_mass_transfer_coefficient((T)0),
    clamp_point_thickness_min((T)-1),clamp_segment_thickness_min((T)-1),clamp_triangle_thickness_min((T)-1),clamp_point_thickness_max((T)-1),clamp_segment_thickness_max((T)-1),clamp_triangle_thickness_max((T)-1),y_in_the_air_min((T)2e-3),break_seg_num(0),allowed_break_seg_num_max(1000),
    use_low_dimensional_incompressibility(true),use_thickness_based_low_dimensional_incompressibility(true),use_distance_to_rim_for_incompressibility(false),min_distance_to_rim_for_incompressibility(5),
    use_thickness_based_dimension_lifting(false),use_crumple_based_dimension_lifting(false),use_crumple_merge(false),use_reseed(true),use_tetrahedron_mesh(true),use_triangle_mesh(true),use_segment_mesh(true),initialize_tetrahedron_mesh(true),initialize_triangle_mesh(true),initialize_segment_mesh(true),
    use_break_thin_segments(true),use_break_thin_triangles(true),use_remove_bad_tetrahedra(false),use_topology_repairing(true),has_tank(false),has_solid(false),has_eulerian_particles(false),use_rim_adaptivity(false),
    use_smooth_rim_thickness(false),use_smooth_interior_thickness(false),use_align_triangle_normals(true),use_rim_mass_redistribution(true),use_break_elements_in_solids(false),use_high_res_volume_mesh(false),use_high_res_segment_mesh(false),
    use_elements_merging(false),use_elements_merging_to_triangle_mesh(false),use_elements_merging_to_tetrahedron_mesh(false),use_remove_low_dim_elements_on_tank_bottom(false),
    use_continuous_meshing_operations(true),use_half_cylindrical_rim(true),has_rim(true),use_adaptive_tetrahedron_remeshing(false),use_adaptive_triangle_remeshing(false),use_adaptive_segment_remeshing(false),preserve_thickness_in_remeshing(true),
    use_thickness_aware_triangle_remeshing(false),use_boundary_vertex_snap(false),use_codim_vertex_snap(false),use_mesh_edge_connect(false),use_contact_aware_triangle_remeshing(false),use_anisotropic_triangle_remeshing(false),use_thin_sheet_aware_tetrahedron_remeshing(false),
    use_constrained_thin_triangle_break(false),use_constrained_thin_segment_break(false),use_enforce_thin_tri_break_constrained(false),use_particle_parents(false),use_clamp_max_thickness(false),
    use_fix_poping(false),use_split_eulerian_particles(false),use_remove_codim_elements_in_the_air(false),use_remove_invalid_elements(false),use_allowed_break_seg_num(false),mpi_codim(mpi_codim_input),fiber(0),water_tank(0),use_dbg(false),dbg_number(0)
{
    random.Set_Seed(1819101);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESHING_BASE<TV>::Initialize_From_External_Mesh_And_Positions(const CODIMENSIONAL_MESH& ext_mesh,const ARRAY_VIEW<TV>& ext_X)
{
    if(mpi_codim){
        mpi_codim->Initialize_From_External_Mesh_And_Positions(ext_mesh,ext_X);
    }
    else{
        for(int i=1;i<=ext_X.m;i++){
            int index=particles.Add_Particle();
            particles.X(index)=ext_X(i);
        }
        
        tetrahedron_mesh.elements.Remove_All();
        if(TV::m==3)for(int i=1;i<=ext_mesh.tetrahedron_mesh.elements.m;i++){
            const TV_INT4& e=ext_mesh.tetrahedron_mesh.elements(i);
            tetrahedron_mesh.elements.Append(e);
        }
        
        triangle_mesh.elements.Remove_All();
        for(int i=1;i<=ext_mesh.triangle_mesh.elements.m;i++){
            const TV_INT3& e=ext_mesh.triangle_mesh.elements(i);
            triangle_mesh.elements.Append(e);
        }

        segment_mesh.elements.Remove_All();
        for(int i=1;i<=ext_mesh.segment_mesh.elements.m;i++){
            const TV_INT2& e=ext_mesh.segment_mesh.elements(i);
            segment_mesh.elements.Append(e);
        }
    }

    mesh.Update_States(particles.X.m,mpi_codim);

    //if(mpi_codim){ARRAY_VIEW<int> node_types_array_view(mesh.node_types);mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(node_types_array_view);}
    Update_Particle_Attributes_After_Having_Mesh_Topology();
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESHING_BASE<TV>::Initialize_Mpi(MPI_CODIMENSIONAL<TV>* mpi_codim_input)
{
    mpi_codim=mpi_codim_input;
    Initialize_Sub_Regions();
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESHING_BASE<TV>::Initialize_Length_Scale(const T length_scale_input,const T adaptive_length_scale_min_input,const T thickness_scale_input/*=(T)0*/,const T rim_thickness_scale_input/*=(T)0*/,
    const T max_thickness_factor/*=(T)1.5*/,const T min_thickness_factor/*=(T).1*/,const T max_rim_thickness_factor/*=(T)1.5*/,const T min_rim_thickness_factor/*=(T).1*/)
{
    length_scale=length_scale_input;
    adaptive_length_scale_min=adaptive_length_scale_min_input;
    thickness_scale=(thickness_scale_input==(T)0?length_scale*(T).02:thickness_scale_input);
    rim_thickness_scale=(rim_thickness_scale_input==(T)0?thickness_scale:rim_thickness_scale_input);
    //max_segment_length=(T)1.5*length_scale;
    //min_segment_length=(T).5*length_scale;
    length_to_split_coef=(T)1.5;
    length_to_collapse_coef=(T).5;
    max_particle_radius=(T).5*length_scale;
    min_particle_radius=(T).25*max_particle_radius;
    max_particle_volume=TV::m==2?(T)pi*pow<2>(max_particle_radius):(T)(4./3.*pi)*pow<3>(max_particle_radius);
    min_particle_volume=TV::m==2?(T)pi*pow<2>(min_particle_radius):(T)(4./3.*pi)*pow<3>(min_particle_radius);
    max_particle_mass=density*max_particle_volume;
    min_particle_mass=density*min_particle_volume;
    max_thickness=max_thickness_factor*thickness_scale;   ////for triangle incompressibility (3d)
    min_thickness=min_thickness_factor*thickness_scale;    ////for triangle break (3d)
    max_rim_thickness=max_rim_thickness_factor*rim_thickness_scale;         ////for segment incompressibility (3d) 
    min_rim_thickness=min_rim_thickness_factor*rim_thickness_scale;////for segment break (3d)
    LOG::cout<<"min_rim: "<<min_rim_thickness<<", "<<min_rim_thickness_factor<<", "<<rim_thickness_scale<<std::endl;

    incompressible_segment_length=length_scale;
    incompressible_triangle_area=(T).1*(T)(.5*sqrt(3.))*pow<2>(length_scale);
    max_thin_simplex_volume=TV::m==2?(T)sqrt(3.)*(T).25*pow<2>(length_scale)*(T)1e-3:(T)sqrt(2.)/(T)12*pow<3>(length_scale)*(T)1e-3;

    //interior_max_segment_length=max_segment_length;
    //boundary_max_segment_length=max_segment_length;
    //diagonal_max_segment_length=max_segment_length;
    //buffer_max_segment_length=max_segment_length;

    //interior_min_segment_length=min_segment_length;
    //boundary_min_segment_length=min_segment_length;
    //diagonal_min_segment_length=min_segment_length;
    //buffer_min_segment_length=min_segment_length;
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESHING_BASE<TV>::Update_Mesh_Incompressibilities()
{
    Update_Segment_Incompressibilities();
    if(TV::m==3)Update_Triangle_Incompressibilities();

    Update_Particle_Incompressibilities();
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESHING_BASE<TV>::Update_Particle_Incompressibilities()
{
    for(int i=1;i<=particles.X.m;i++)particles.incompressibility(i)=false;
    for(int i=1;i<=segment_mesh.elements.m;i++){
        if(segment_incompressibilities(i)){
            const TV_INT2& v=segment_mesh.elements(i);
            for(int d=1;d<=2;d++)particles.incompressibility(v(d))=true;
        }
    }
    for(int i=1;i<=triangle_mesh.elements.m;i++){
        if(triangle_incompressibilities(i)){
            const TV_INT3& v=triangle_mesh.elements(i);
            for(int d=1;d<=3;d++)particles.incompressibility(v(d))=true;
        }
    }

    //particle_incompressibilities.Resize(particles.X.m);particle_incompressibilities.Fill(VECTOR<int,TV::m-1>());
    //if(use_low_dimensional_incompressibility&&use_thickness_based_low_dimensional_incompressibility){
    //    for(int i=1;i<=particles.X.m;i++){
    //        particle_incompressibilities(i)(1)=Is_Incompressible_Segment_Particle(i);
    //        particle_incompressibilities(i)(TV::m-1)=Is_Incompressible_Triangle_Particle(i);
    //    }
    //}

    //for(int i=1;i<=particles.X.m;i++)particles.incompressibilities(i)=particle_incompressibilities(i);
}

//template<class TV>
//bool PhysBAM::CODIMENSIONAL_MESHING_BASE<TV>::Is_Incompressible_Segment_Particle(const int p)
//{
//    if(Is_On_Segment_Mesh(mesh.node_types(p))){
//        if(mesh.Is_Rim_Particle(p)||mesh.Is_Boundary_Particle_On_Segment_Mesh(p))return false;
//        //return codimensional_mesh_object.Nodal_Segment_Length(p)<incompressible_segment_length;
//        return particles.thickness(p)>max_rim_thickness;    ////TODO: in 2D still need to use max_thickness
//    }
//    return false;
//}

//template<class TV>
//bool PhysBAM::CODIMENSIONAL_MESHING_BASE<TV>::Is_Incompressible_Triangle_Particle(const int p)
//{
//    ////for dbg
//    return false;
//    //if(use_distance_to_rim_for_incompressibility&&particles.distance_to_rim(p)!=-1&&particles.distance_to_rim(p)<=min_distance_to_rim_for_incompressibility)return false;
//    //if(Is_On_Triangle_Mesh(mesh.node_types(p))){
//    //    if(mesh.Is_Rim_Particle(p))return false;
//    //    //return codimensional_mesh_object.Nodal_Triangle_Area(p)<incompressible_triangle_area;
//    //    return particles.thickness(p)>max_thickness;
//    //}
//    //return false;
//}

template<class TV>
bool PhysBAM::CODIMENSIONAL_MESHING_BASE<TV>::Is_Incompressible_Segment(const int s)
{
    const TV_INT2& v=mesh.segment_mesh.elements(s);
    //if(mesh.Is_Boundary_Particle_On_Segment_Mesh(v(1))||mesh.Is_Boundary_Particle_On_Segment_Mesh(v(2)))return false;
    if(Is_Contact_With_Fluid(particles.contact(v(1)))&&Is_Contact_With_Fluid(particles.contact(v(2)))){/*LOG::cout<<"incomp contact seg "<<v<<std::endl;*/return true;}
    TV rel_v=particles.V(v(2))-particles.V(v(1));TV dir=particles.X(v(2))-particles.X(v(1));if(TV::Dot_Product(rel_v,dir)>(T)0)return false;    ////no incomp if particles are separating
    T segment_thickness=(T).5*(particles.thickness(v(1))+particles.thickness(v(2)));
    //LOG::cout<<"seg thickness: "<<segment_thickness<<", "<<max_rim_thickness<<", "<<thickness_scale<<std::endl;
    return segment_thickness>max_rim_thickness;
    //return true;
}

template<class TV>
bool PhysBAM::CODIMENSIONAL_MESHING_BASE<TV>::Is_Incompressible_Triangle(const int t)
{
    const TV_INT3& v=mesh.triangle_mesh.elements(t);
    if(mesh.Is_Rim_Particle(v(1))||mesh.Is_Rim_Particle(v(2))||mesh.Is_Rim_Particle(v(3)))return false;
    T triangle_thickness=(T)one_third*(particles.thickness(v(1))+particles.thickness(v(2))+particles.thickness(v(3)));
    return triangle_thickness>max_thickness;

    //return false;   ////dbg
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESHING_BASE<TV>::Update_Segment_Incompressibilities()
{
    segment_incompressibilities.Resize(segment_mesh.elements.m);
    for(int i=1;i<=segment_mesh.elements.m;i++)segment_incompressibilities(i)=Is_Incompressible_Segment(i);

    //segment_incompressibilities.Fill(false);
    //for(int i=1;i<=segment_mesh.elements.m;i++){
    //    const TV_INT2& v=segment_mesh.elements(i);
    //    if(mesh.Is_Boundary_Particle_On_Segment_Mesh(v(1))||mesh.Is_Boundary_Particle_On_Segment_Mesh(v(2))){
    //        segment_incompressibilities(i)=false;
    //    }
    //    else if(particle_incompressibilities(v(1))(1)||particle_incompressibilities(v(2))(1)){
    //        segment_incompressibilities(i)=true;
    //    }
    //}
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESHING_BASE<TV>::Update_Triangle_Incompressibilities()
{
    triangle_incompressibilities.Resize(triangle_mesh.elements.m);
    for(int i=1;i<=triangle_mesh.elements.m;i++){
        triangle_incompressibilities(i)=Is_Incompressible_Triangle(i);
        //if(triangle_incompressibilities(i))LOG::cout<<"incomp tri "<<i<<": "<<triangle_mesh.elements(i)<<std::endl;
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESHING_BASE<TV>::Update_Incompressibility_Mesh(CODIMENSIONAL_MESH& incompressibility_codimensional_mesh)
{
    SEGMENT_MESH& incompressibility_segment_mesh=incompressibility_codimensional_mesh.segment_mesh;
    incompressibility_segment_mesh.elements.Remove_All();
    for(int i=1;i<=segment_incompressibilities.m;i++){
        if(segment_incompressibilities(i)){
            incompressibility_segment_mesh.elements.Append(segment_mesh.elements(i));
            //LOG::cout<<"N"<<(mpi_codim?mpi_codim->node_id:0)<<" add incomp seg "<<segment_mesh.elements(i)<<" guid "<<Element(particles.guid,segment_mesh.elements(i))<<std::endl;
        }
    }

    TRIANGLE_MESH& incompressibility_triangle_mesh=incompressibility_codimensional_mesh.triangle_mesh;
    if(TV::m==2){
        incompressibility_triangle_mesh.elements.Remove_All();
        for(int i=1;i<=triangle_mesh.elements.m;i++)incompressibility_triangle_mesh.elements.Append(triangle_mesh.elements(i));
    }
    else{
        incompressibility_triangle_mesh.elements.Remove_All();
        for(int i=1;i<=triangle_incompressibilities.m;i++){
            if(triangle_incompressibilities(i))incompressibility_triangle_mesh.elements.Append(triangle_mesh.elements(i));
        }
    }

    //LOG::cout<<"N"<<mpi_codim->node_id<<" tri: "<<triangle_mesh.elements.m<<std::endl;
    //for(int i=1;i<=triangle_mesh.elements.m;i++){
    //    const TV_INT3& v=triangle_mesh.elements(i);
    //    TV_INT3 guid=Element(particles.guid,v);
    //    LOG::cout<<guid<<", ";
    //}
    //LOG::cout<<"finish N"<<mpi_codim->node_id<<std::endl;

    incompressibility_codimensional_mesh.Update_States(particles.X.m,mpi_codim);
    //if(mpi_codim){ARRAY_VIEW<int> node_types_array_view(incompressibility_codimensional_mesh.node_types);mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(node_types_array_view);}

    //LOG::cout<<"incomp mesh: "<<incompressibility_codimensional_mesh.tetrahedron_mesh.elements.m<<", "<<incompressibility_codimensional_mesh.triangle_mesh.elements.m<<", "<<incompressibility_codimensional_mesh.segment_mesh.elements.m<<std::endl;
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESHING_BASE<TV>::Update_Viscosity_Mesh(CODIMENSIONAL_MESH& viscosity_codimensional_mesh,bool has_vis_rim_segments/*=false*/,bool has_vis_rim_triangles/*=true*/)
{
    SEGMENT_MESH& viscosity_segment_mesh=viscosity_codimensional_mesh.segment_mesh;
    viscosity_segment_mesh.elements.Remove_All();
    for(int i=1;i<=segment_mesh.elements.m;i++){
        //////dbg
        //const TV_INT2& v=segment_mesh.elements(i);
        //TV_INT2 guid=Element(particles.guid,v);
        //if(!guid.Contains(39800)&&!guid.Contains(45795)&&!guid.Contains(47182)&&!guid.Contains(37884)&&!guid.Contains(18676)){
        //    continue;
        //}

        viscosity_segment_mesh.elements.Append(segment_mesh.elements(i));
    }
    if(has_vis_rim_segments)for(int i=1;i<=mesh.rim_segment_mesh.elements.m;i++){
        //const TV_INT2& v=mesh.rim_segment_mesh.elements(i);
        viscosity_segment_mesh.elements.Append(mesh.rim_segment_mesh.elements(i));
    }

    //if(has_vis_rim_segments)for(int i=1;i<=mesh.rim_segment_mesh.elements.m;i++){
    //    //const TV_INT2& v=mesh.rim_segment_mesh.elements(i);
    //    viscosity_segment_mesh.elements.Append(mesh.rim_segment_mesh.elements(i));
    //}

    TRIANGLE_MESH& viscosity_triangle_mesh=viscosity_codimensional_mesh.triangle_mesh;
    viscosity_triangle_mesh.elements.Remove_All();
    if(has_vis_rim_triangles)for(int i=1;i<=mesh.triangle_mesh.elements.m;i++){
        viscosity_triangle_mesh.elements.Append(mesh.triangle_mesh.elements(i));
    }
    else{
        HASHTABLE<int> rim_particle_hashtable;
        for(int i=1;i<=mesh.rim_segment_mesh.elements.m;i++){
            const TV_INT2& v=mesh.rim_segment_mesh.elements(i);
            rim_particle_hashtable.Set(v(1));rim_particle_hashtable.Set(v(2));
        }
        for(int i=1;i<=mesh.triangle_mesh.elements.m;i++){
            const TV_INT3& v=mesh.triangle_mesh.elements(i);
            bool on_rim=rim_particle_hashtable.Contains(v(1))||rim_particle_hashtable.Contains(v(2))||rim_particle_hashtable.Contains(v(3));
            if(!on_rim)viscosity_triangle_mesh.elements.Append(mesh.triangle_mesh.elements(i));
        }
    }

    viscosity_codimensional_mesh.Update_States(particles.X.m,mpi_codim);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESHING_BASE<TV>::Update_Particle_Geometry_Attributes()
{
    for(int i=1;i<=particles.X.m;i++){
        //particles.type(i)=mesh.node_types(i);
        particles.id(i)=mesh.node_connectivities(i);
        //particles.small_region_id(i)=mesh.node_connectivities_in_high_dim(i);
        particles.on_boundary(i)(TV::m)=Is_On_Tetrahedron_Mesh_Boundary(mesh.node_types(i));
        particles.on_boundary(i)(TV::m-1)=Is_On_Triangle_Mesh_Boundary(mesh.node_types(i));
        particles.on_boundary(i)(1)=Is_On_Segment_Mesh_Boundary(mesh.node_types(i));
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESHING_BASE<TV>::Update_Particle_Attributes_After_Having_Mesh_Topology()
{
    Update_Particle_Geometry_Attributes();
    Update_Particle_Mass_Based_On_Geometry();
    Update_Particle_Thickness(has_rim);
    Update_Mesh_Incompressibilities();  ////needs thickness
    //Update_Particle_Distance_To_Rim();
    Update_Particle_Area_Weighted_Normal();
    Update_Particle_Normal();
}

template<class TV>
BOX<TV> PhysBAM::CODIMENSIONAL_MESHING_BASE<TV>::Get_Particle_Bounding_Box()
{
    if(particles.X.m==0){return BOX<TV>(TV(),TV());}

    BOX<TV> bounding_box=BOX<TV>(TV((T)FLT_MAX),TV((T)-FLT_MAX));
    for(int i=1;i<=particles.X.m;i++){
        for(int d=1;d<=TV::m;d++){
            if(particles.X(i)(d)<bounding_box.min_corner(d))bounding_box.min_corner(d)=particles.X(i)(d);
            else if(particles.X(i)(d)>bounding_box.max_corner(d))bounding_box.max_corner(d)=particles.X(i)(d);
        }
    }
    bounding_box.Change_Size(TV::Constant_Vector((T)2*length_scale));
    return bounding_box;
}

template<class TV>
GRID<TV> PhysBAM::CODIMENSIONAL_MESHING_BASE<TV>::Get_Particle_Grid(const BOX<TV>& bounding_box,const T grid_dX)
{
    TV_INT particle_grid_counts=(TV_INT)(bounding_box.Edge_Lengths()/grid_dX)+1;
    BOX<TV> particle_grid_box;
    particle_grid_box.min_corner=bounding_box.min_corner;
    particle_grid_box.max_corner=bounding_box.min_corner+TV::Constant_Vector(grid_dX)*(TV)particle_grid_counts;
    GRID<TV> p_grid;p_grid.Initialize(particle_grid_counts,particle_grid_box,true);
    return p_grid;
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESHING_BASE<TV>::Resize_Elements_In_Cell_With_Grid(ARRAY<ARRAY<int>,TV_INT>& p_in_cell,const GRID<TV>& grid)
{
    p_in_cell.Resize(grid.Domain_Indices(1));
    for(UNIFORM_GRID_ITERATOR_CELL<TV> iter(grid,1);iter.Valid();iter.Next()){
        p_in_cell(iter.Cell_Index()).Remove_All();
    }
}

//////////////////////////////////////////////////////////////////////////
////mass and thickness

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_MESHING_BASE<TV>::Joint_Thickness(const T V0,const T A1,const T L1)
{
    T t1=(T)0;T t2=(T)0;bool solved=Quadratic_Equation((T).25*(T)pi*L1,A1,-V0,t1,t2);
    if(solved)return t1;else return (T)-1;
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESHING_BASE<TV>::Initialize_Particle_Mass_Based_On_Geometry(T initial_thickness,T initial_rim_thickness,bool initialize_rim/*=true*/,bool initialize_thickness/*=true*/,WATER_SOURCE<TV>* water_source/*=0*/)
{
    LOG::cout<<"ini thickness: "<<initial_thickness<<", "<<initial_rim_thickness<<", ini_rim: "<<initialize_rim<<", init_thk: "<<initialize_thickness<<", half_cyn: "<<use_half_cylindrical_rim<<std::endl;
    if(!initialize_rim)initial_rim_thickness=initial_thickness;

    ////initialize thickness
    if(initialize_thickness)for(int i=1;i<=particles.X.m;i++){
        particles.mass(i)=(T)0;particles.dimensional_mass(i)=TV();

        if(water_source!=0){////eulerian particle
            if(water_source->eulerian_particle_thicknesses.Contains(particles.guid(i))){
                T eulerian_thickness=water_source->eulerian_particle_thicknesses.Get(particles.guid(i));
                particles.thickness(i)=eulerian_thickness;
            }
            else if(water_source->initial_particle_thicknesses.Contains(particles.guid(i))){
                T eulerian_thickness=water_source->initial_particle_thicknesses.Get(particles.guid(i));
                particles.thickness(i)=eulerian_thickness;
            }
        }
        else if(mesh.Is_Rim_Particle(i)){particles.thickness(i)=initial_rim_thickness;}////rim particle
        else{particles.thickness(i)=initial_thickness;}////interior particle
    }

    ////dim-3 vol in 3d and dim-2 vol in 2d
    T volume_fraction=(TV::m==2?(T)one_third:(T).25);
    for(int i=1;i<=codimensional_mesh_object.volume_mesh.elements.m;i++){
        T volume=codimensional_mesh_object.Volumetric_Element_Volume(i);
        for(int d=1;d<=TV::m+1;d++){
            particles.dimensional_mass(codimensional_mesh_object.volume_mesh.elements(i)(d))(TV::m)+=volume_fraction*volume*density;
        }
    }

    if(TV::m==3){
        for(int i=1;i<=particles.X.m;i++){
            if(!initialize_rim||!mesh.Is_Rim_Particle(i)||use_half_cylindrical_rim){
                T triangle_area_sum=codimensional_mesh_object.Nodal_Triangle_Area(i);
                //T extra_triangle_area_sum=codimensional_mesh_object.Nodal_Extra_Triangle_Area_Sum(i);T extra_triangle_area_sum=(T)0;triangle_area_sum+=extra_triangle_area_sum;
                particles.dimensional_mass(i)(TV::m-1)+=triangle_area_sum*particles.thickness(i)*density;
            }
        }
        ////half (or full) cylinder volume on triangle mesh boundary in 3d
        if(initialize_rim)for(int i=1;i<=particles.X.m;i++){
            if(mesh.Is_Boundary_Particle_On_Triangle_Mesh(i)){
                T rim_length_sum=codimensional_mesh_object.Nodal_Segment_Length(i)+codimensional_mesh_object.Nodal_Segment_Length_Sum_On_Rim_Mesh(i);
                T rim_mass=(T)pi*pow<2>((T).5*particles.thickness(i))*rim_length_sum*density;
                particles.dimensional_mass(i)(TV::m-1)+=rim_mass;
            }
        }
    }

    ////cylinder volume on segment mesh in 3d
    for(int i=1;i<=segment_mesh.elements.m;i++){
        TV_INT2 e=segment_mesh.elements(i);T length=codimensional_mesh_object.Segment_Length(i);
        for(int d=1;d<=2;d++){int p=e(d);
        T cross_section_area=codimensional_mesh_object.Segment_Cross_Section_Area(particles.thickness(p));
        particles.dimensional_mass(e(d))(1)+=(T).5*length*cross_section_area*density;
        }
    }

    ////hemisphere volume on segment mesh boundary
    if(initialize_rim)for(int i=1;i<=particles.X.m;i++){
        if(mesh.Is_Boundary_Particle_On_Segment_Mesh(i)){
            T hemisphere_mass=codimensional_mesh_object.Hemisphere_Volume(particles.thickness(i))*density;
            particles.dimensional_mass(i)(1)+=hemisphere_mass;
        }
    }

    ////point volume
    for(int i=1;i<=particles.X.m;i++){
        //particles.effective_mass(i)=particles.dimensional_mass(i)(TV::m);   ////target mass
        particles.mass(i)=particles.dimensional_mass(i).Sum();
    }

    if(mpi_codim)mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(particles.dimensional_mass);
    if(mpi_codim)mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(particles.mass);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESHING_BASE<TV>::Update_Particle_Mass_Based_On_Geometry()
{
    T min_triangle_thickness=min_thickness*(T).5;T max_triangle_thickness=max_thickness*(T)2;
    T min_segment_thickness=min_rim_thickness*(T).5;T max_segment_thickness=max_rim_thickness*(T)2;
    T min_particle_radius=min_rim_thickness*(T).5;T max_particle_radius=max_rim_thickness;

    if(clamp_point_thickness_min>=(T)0)min_particle_radius=clamp_point_thickness_min;
    if(clamp_point_thickness_max>=(T)0)max_particle_radius=clamp_point_thickness_max;
    if(clamp_segment_thickness_min>=(T)0)min_segment_thickness=clamp_segment_thickness_min;
    if(clamp_segment_thickness_max>=(T)0)max_segment_thickness=clamp_segment_thickness_max;
    if(clamp_triangle_thickness_min>=(T)0)min_triangle_thickness=clamp_triangle_thickness_min;
    if(clamp_triangle_thickness_max>=(T)0)max_triangle_thickness=clamp_triangle_thickness_max;
    
    for(int i=1;i<=particles.X.m;i++){
        if(mpi_codim&&!mpi_codim->Inside_Domain(particles.X(i)))continue;
        int p_type=mesh.node_types(i);

        if(Is_On_Triangle_Mesh(p_type)){particles.thickness(i)=clamp_min(particles.thickness(i),min_triangle_thickness);}
        if(Is_On_Segment_Mesh(p_type)){particles.thickness(i)=clamp_min(particles.thickness(i),min_segment_thickness);}
        if(use_clamp_max_thickness){
            if(Is_On_Triangle_Mesh(p_type))particles.thickness(i)=clamp_max(particles.thickness(i),max_triangle_thickness);
            if(Is_On_Segment_Mesh(p_type))particles.thickness(i)=clamp_max(particles.thickness(i),max_segment_thickness);
            if(Is_Isolated(p_type))particles.thickness(i)=clamp_max(particles.thickness(i),max_particle_radius);
        }

        ////for isolated particles transfer all dimensional mass to dim-1
        if(Is_Isolated(p_type)){    ////TODO: use thickness?
            //particles.dimensional_mass(i)=TV();particles.dimensional_mass(i)(1)=particles.mass(i);
            particles.mass(i)=(T)one_sixth*(T)pi*pow<3>(particles.thickness(i))*density;
            particles.dimensional_mass(i)=TV();particles.dimensional_mass(i)(1)=particles.mass(i);
            if(particles.mass(i)==(T)0)particles.mass(i)=particles.dimensional_mass(i)(1)=(T)(4./3.)*(T)pi*pow<3>(min_particle_radius); ////clamp
            continue;
        }

        ////dim 3 mass in 3d and dim 2 mass in 2d
        if(Is_On_Tetrahedron_Mesh(p_type)){
            particles.dimensional_mass(i)(TV::m)=codimensional_mesh_object.Nodal_Tetrahedron_Volume(i)*density;
        }
        else particles.dimensional_mass(i)(TV::m)=(T)0;

        ////triangle and segment particles
        if(preserve_thickness_in_remeshing){    ////TODO: handle the has_rim case
            particles.dimensional_mass(i)(TV::m-1)=(T)0;
            particles.dimensional_mass(i)(1)=(T)0;
            bool p_on_tri_rim=false;
            bool p_on_seg_rim=false;
            if(has_rim){
                if(mesh.Is_Boundary_Particle_On_Segment_Mesh(i)){p_on_seg_rim=true;
                    T hemisphere_mass=(T).5*(T)one_sixth*(T)pi*pow<3>(particles.thickness(i))*density;
                    particles.dimensional_mass(i)(1)+=hemisphere_mass;
                    //if(!use_half_cylindrical_rim){particles.dimensional_mass(i)(1)+=sphere_mass;}
                    //else{particles.dimensional_mass(i)(1)+=(T).5*sphere_mass;}
                }
                else if(mesh.Is_Boundary_Particle_On_Triangle_Mesh(i)){p_on_tri_rim=true;
                    T rim_length_sum=codimensional_mesh_object.Nodal_Segment_Length_Sum_On_Rim_Mesh(i);PHYSBAM_ASSERT(rim_length_sum!=(T)0);
                    T cross_section_area=codimensional_mesh_object.Segment_Cross_Section_Area(particles.thickness(i));
                    T rim_mass=rim_length_sum*cross_section_area*density;
                    particles.dimensional_mass(i)(TV::m-1)+=rim_mass;
                    //if(!use_half_cylindrical_rim){particles.dimensional_mass(i)(2)+=rim_mass;}
                    //else{particles.dimensional_mass(i)(2)+=(T).5*rim_mass;}
                }
            }

            if((use_half_cylindrical_rim||!p_on_seg_rim)&&Is_On_Segment_Mesh(p_type)){
                T segment_length=codimensional_mesh_object.Nodal_Segment_Length(i);
                T cross_section_area=codimensional_mesh_object.Segment_Cross_Section_Area(particles.thickness(i));
                //T thickness=(T).1*length_scale;T cross_section_area=codimensional_mesh_object.Segment_Cross_Section_Area(thickness);    ////for dbg
                particles.dimensional_mass(i)(1)+=segment_length*cross_section_area*density;
            }

            if((use_half_cylindrical_rim||!p_on_tri_rim)&&Is_On_Triangle_Mesh(p_type)){
                T triangle_area_sum=codimensional_mesh_object.Nodal_Triangle_Area(i);
                particles.dimensional_mass(i)(TV::m-1)+=particles.thickness(i)*triangle_area_sum*density;
            }

            if(Is_On_Tetrahedron_Mesh_Boundary(p_type)){
                T tetrahedron_surface_area_sum=codimensional_mesh_object.Nodal_Tetrahedron_Surface_Area(i);
                particles.dimensional_mass(i)(TV::m-1)+=particles.thickness(i)*(T).5*tetrahedron_surface_area_sum*density;
            }
        }
        particles.mass(i)=particles.dimensional_mass(i).Sum();

        //if(particles.guid(i)==216639){
        //    LOG::cout<<"mass based on thickness: "<<particles.guid(i)<<": "<<particles.thickness(i)<<": "<<particles.mass(i)<<", "<<particles.dimensional_mass(i)<<std::endl;
        //    int p_type=mesh.node_types(i);
        //    LOG::cout<<"type: "<<Is_On_Segment_Mesh(p_type)<<", "<<Is_On_Triangle_Mesh(p_type)<<", "<<Is_On_Tetrahedron_Mesh(p_type)<<", "<<Is_Isolated(p_type)<<std::endl;
        //    LOG::cout<<"tet vol: "<<codimensional_mesh_object.Nodal_Tetrahedron_Volume(i)<<std::endl;
        //    LOG::cout<<"inc tet: "<<(*mesh.tetrahedron_mesh.incident_elements)(i).m<<std::endl;
        //    for(int j=1;j<=(*mesh.tetrahedron_mesh.incident_elements)(i).m;j++){
        //        int t=(*mesh.tetrahedron_mesh.incident_elements)(i)(j);
        //        TV_INT4 vtx=mesh.tetrahedron_mesh.elements(t);
        //        LOG::cout<<"tet "<<vtx<<", tet_voL: "<<codimensional_mesh_object.Tetrahedron_Volume(t)<<std::endl;
        //        for(int k=1;k<=4;k++){
        //            LOG::cout<<"X "<<vtx(k)<<" guid "<<particles.guid(vtx(k))<<": "<<particles.X(vtx(k))<<std::endl;
        //        }
        //    }
        //}

        //if(particles.guid(i)==51501){
        //    LOG::cout<<"dbg p mass: "<<i<<" guid "<<particles.guid(i)
        //        <<", mass: "<<particles.mass(i)<<", dim_mass: "<<particles.dimensional_mass(i)<<", vel: "<<particles.V(i)<<", thk: "<<particles.thickness(i)<<std::endl;
        //}
    }

    if(mpi_codim){
        mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(particles.dimensional_mass);
        mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(particles.mass);
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESHING_BASE<TV>::Update_Particle_Thickness(bool has_rim)
{
    for(int i=1;i<=particles.X.m;i++){
        if(mpi_codim&&!mpi_codim->Inside_Domain(particles.X(i)))continue;
        particles.thickness(i)=Discrete_Particle_Thickness(i,has_rim);
    }

    if(use_continuous_meshing_operations){
        ARRAY<T> continuous_thickness(particles.X.m);
        for(int i=1;i<=particles.X.m;i++){
            if(mpi_codim&&!mpi_codim->Inside_Domain(particles.X(i)))continue;
            continuous_thickness(i)=Continuous_Particle_Thickness(i,has_rim);
        }
        for(int i=1;i<=particles.X.m;i++){particles.thickness(i)=continuous_thickness(i);}
    }

    if(mpi_codim){mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(particles.thickness);}
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_MESHING_BASE<TV>::Discrete_Particle_Thickness(const int p,bool has_rim)
{
    int p_type=mesh.node_types(p);
    if(has_rim){
        if(mesh.Is_Boundary_Particle_On_Segment_Mesh(p)){
            if(!use_half_cylindrical_rim){
                return std::pow((T)6*particles.mass(p)/(density*(T)pi),(T)one_third);
            }
            else{
                T segment_length_sum=codimensional_mesh_object.Nodal_Segment_Length(p);
                T r=Cubic_Equation((T)2/(T)3*(T)pi/*two_thirds_pi*/,(T)pi*segment_length_sum,(T)0,-particles.mass(p)/density);
                return (T)2*r;
            }
        }
        else if(mesh.Is_Rim_Particle(p)){
            if(!use_half_cylindrical_rim){
                T rim_length_sum=codimensional_mesh_object.Nodal_Segment_Length(p)+codimensional_mesh_object.Nodal_Segment_Length_Sum_On_Rim_Mesh(p);
                PHYSBAM_ASSERT(rim_length_sum!=(T)0);
                return (T)2*sqrt((particles.dimensional_mass(p)(1)+particles.dimensional_mass(p)(2))/(density*rim_length_sum*(T)pi));
            }
            else{
                //T rim_length_sum=codimensional_mesh_object.Nodal_Segment_Length(p)+codimensional_mesh_object.Nodal_Segment_Length_Sum_On_Rim_Mesh(p);
                //T rim_area_sum=codimensional_mesh_object.Nodal_Triangle_Area(p)+(T).5*codimensional_mesh_object.Nodal_Tetrahedron_Surface_Area(p);
                //T t1=0;T t2=0;bool solved=Quadratic_Equation((T)pi/(T)8*rim_length_sum,rim_area_sum,-(particles.dimensional_mass(p)(1)+particles.dimensional_mass(p)(2))/density,t1,t2);
                //if(solved){/*LOG::cout<<"rim t: "<<p<<": "<<rim_length_sum<<", "<<rim_area_sum<<", "<<(particles.dimensional_mass(p)(1)+particles.dimensional_mass(p)(2))<<": "<<t1<<", "<<t2<<std::endl;*/return t1;}
                //else{LOG::cout<<"cannot solve rim thickness for "<<p<<std::endl;return (T)0;}
                T L=codimensional_mesh_object.Nodal_Segment_Length(p)+codimensional_mesh_object.Nodal_Segment_Length_Sum_On_Rim_Mesh(p);
                T A=codimensional_mesh_object.Nodal_Triangle_Area(p)+(T).5*codimensional_mesh_object.Nodal_Tetrahedron_Surface_Area(p);
                T V0=(particles.dimensional_mass(p)(1)+particles.dimensional_mass(p)(2))/density;
                T t=Joint_Thickness(V0,A,L);

                //if(particles.guid(p)==658){
                //    T area_sum=(T)0;
                //    for(int i=1;i<=(*mesh.triangle_mesh.incident_elements)(p).m;i++){
                //        int t=(*mesh.triangle_mesh.incident_elements)(p)(i);
                //        TV_INT3 v=mesh.triangle_mesh.elements(t);
                //        TV_INT3 guid=Element(particles.guid,v);
                //        LOG::cout<<"v: "<<v<<", "<<guid<<std::endl;
                //        for(int d=1;d<=3;d++)LOG::cout<<particles.X(v(d))<<", ";LOG::cout<<std::endl;
                //    }
                //}

                if(!(t>=(T)0)){LOG::cout<<"error solving rim thickness"<<", p: "<<p<<" guid "<<particles.guid(p)<<", t: "<<t<<", V0: "<<V0<<", A: "<<A<<": "<<codimensional_mesh_object.Nodal_Triangle_Area(p)<<", "<<(T).5*codimensional_mesh_object.Nodal_Tetrahedron_Surface_Area(p)<<", L: "<<L<<", mass: "<<particles.mass(p)
                    <<", dim_mass: "<<particles.dimensional_mass(p)<<", V: "<<particles.V(p)<<std::endl;}
                //PHYSBAM_ASSERT(t>=(T)0);
                if(!(t>=(T)0))t=thickness_scale;
                return t;
            }
        }
    }

    if(Is_Isolated(p_type)){return (T)2*pow((T)3*particles.mass(p)/(density*(T)4*(T)pi),(T)one_third);}
    if(!Is_On_Triangle_Mesh(p_type)&&!Is_On_Segment_Mesh(p_type)){return (T)0;} ////TODO: tet thickness

    T triangle_area_sum=(T)0;T tetrahedron_surface_area_sum=(T)0;T segment_length_sum=(T)0;
    if(Is_On_Triangle_Mesh(p_type))triangle_area_sum=codimensional_mesh_object.Nodal_Triangle_Area(p);
    if(Is_On_Tetrahedron_Mesh(p_type))tetrahedron_surface_area_sum=codimensional_mesh_object.Nodal_Tetrahedron_Surface_Area(p);
    if(Is_On_Segment_Mesh(p_type))segment_length_sum=codimensional_mesh_object.Nodal_Segment_Length(p);
    T A=triangle_area_sum+(T).5*tetrahedron_surface_area_sum;
    T V0=(particles.dimensional_mass(p)(1)+particles.dimensional_mass(p)(2))/density;
    T L=codimensional_mesh_object.Nodal_Segment_Length(p);
    T t=Joint_Thickness(V0,A,L);

    //if(particles.guid(p)==51501){
    //    LOG::cout<<"dbg p thk: "<<p<<" guid "<<particles.guid(p)<<": V0: "<<V0<<", A: "<<A<<": "<<triangle_area_sum<<", "<<tetrahedron_surface_area_sum<<", L: "<<L
    //        <<", mass: "<<particles.mass(p)<<", vel: "<<particles.V(p)<<", thk: "<<particles.thickness(p)<<": t: "<<t<<std::endl;
    //}

    if(!(t>=(T)0)){LOG::cout<<"error solving joint thickness"<<", p: "<<p<<" guid "<<particles.guid(p)<<": V0: "<<V0<<", A: "<<A<<": "<<triangle_area_sum<<", "<<tetrahedron_surface_area_sum<<", L: "<<L<<", mass: "<<particles.mass(p)<<", vel: "<<particles.V(p)<<std::endl;}
    //PHYSBAM_ASSERT(t>=(T)0);
    if(!(t>=(T)0))t=thickness_scale;
    return t;
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_MESHING_BASE<TV>::Continuous_Particle_Thickness(const int p,bool has_rim)
{
    int p_type=mesh.node_types(p);
    if(Is_Isolated(p_type)){return (T)2*pow((T)3*particles.mass(p)/(density*(T)4*(T)pi),(T)one_third);}
    if(!Is_On_Triangle_Mesh(p_type)&&!Is_On_Segment_Mesh(p_type)){return (T)0;} ////TODO: tet thickness

    int number_of_rings=2;T max_factor=(T)3;T min_factor=(T)1/max_factor;
    bool is_p_rim=has_rim&&mesh.Is_Rim_Particle(p);

    T total_node_mass=(T)0;T total_node_area=(T)0;T total_node_length=(T)0;

    if(Is_On_Triangle_Mesh(p_type)){
        ARRAY<int> nbs;Find_Neighbor_Particles_On_Mesh(triangle_mesh,p,nbs,number_of_rings);
        T particle_length_scale=particles.length_scale(p);if(particle_length_scale<=(T)0)particle_length_scale=length_scale;
        T max_distance_squared=pow<2>(particle_length_scale*(T)number_of_rings);
        for(int i=1;i<=nbs.m;i++){int nb_p=nbs(i);
            if(is_p_rim){
                if(!mesh.Is_Rim_Particle(nb_p))continue;    ////skip non-rim particle for has_rim
                T frac=(T)1;
                total_node_length+=codimensional_mesh_object.Nodal_Segment_Length_Sum_On_Rim_Mesh(nb_p)*frac;
                total_node_mass+=particles.dimensional_mass(nb_p)(2)*frac;
            }
            else{
                if(particles.thickness(p)>particles.thickness(nb_p)*max_factor||particles.thickness(p)<particles.thickness(nb_p)*min_factor)continue;   ////skip the particles with big thickness difference
                T distance_squared=(particles.X(p)-particles.X(nb_p)).Magnitude_Squared();
                T normalized_distance_squared=clamp(distance_squared/max_distance_squared,(T)0,(T)1);
                T frac=Gaussian_Kernel(normalized_distance_squared);
                total_node_area+=(codimensional_mesh_object.Nodal_Triangle_Area(nb_p)+(T).5*codimensional_mesh_object.Nodal_Tetrahedron_Surface_Area(nbs(i)))*frac;
                total_node_mass+=particles.dimensional_mass(nb_p)(2)*frac;
            }
        }
    }
    
    if(Is_On_Segment_Mesh(p_type)){
        ARRAY<int> nbs;Find_Neighbor_Particles_On_Mesh(segment_mesh,p,nbs,number_of_rings);
        for(int i=1;i<=nbs.m;i++){int nb_p=nbs(i);
            T frac=(T)1;    ////TODO: use gaussian?
            total_node_length+=codimensional_mesh_object.Nodal_Segment_Length(nb_p)*frac;
            total_node_mass+=particles.dimensional_mass(nb_p)(1)*frac;
        }
    }

    T A=total_node_area;
    T V0=total_node_mass/density;
    T L=total_node_length;
    T t=Joint_Thickness(V0,A,L);
    //if(t<(T)0)LOG::cout<<"error solving joint thickness"<<"p: "<<p<<" guid "<<particles.guid(p)<<", M0: "<<total_node_mass<<": V0: "<<V0<<", A: "<<A<<", L: "<<L<<std::endl;
    //PHYSBAM_ASSERT(t>=(T)0);
    if(!(t>=(T)0))t=thickness_scale;
    return t;
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_MESHING_BASE<TV>::Node_Triangle_Area_Sum(const int p,const HASHTABLE<int,ARRAY<int> >& vtx_tri_hashtable)
{
    const ARRAY<int>* p_tri=vtx_tri_hashtable.Get_Pointer(p);T tri_area_sum=(T)0;
    if(p_tri)for(int k=1;k<=(*p_tri).m;k++){
        int t2=(*p_tri)(k);
        const TV_INT3& v2=triangle_mesh.elements(t2);
        T tri_area=(T)one_third*TRIANGLE_3D<T>::Area(particles.X(v2(1)),particles.X(v2(2)),particles.X(v2(3)));
        tri_area_sum+=tri_area;
    }
    return tri_area_sum;
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_MESHING_BASE<TV>::Nodal_Tetrahedron_Surface_Area_Sum(const int p,const HASHTABLE<int,ARRAY<int> >& vtx_tet_hashtable)
{
    const ARRAY<int>* p_tet=vtx_tet_hashtable.Get_Pointer(p);T tet_area_sum=(T)0;
    if(p_tet)for(int k=1;k<=(*p_tet).m;k++){
        int t2=(*p_tet)(k);
        const TV_INT4& v2=tetrahedron_mesh.elements(t2);
        ARRAY<TV_INT3> faces;Simplex_Faces(v2,faces);
        for(int j=1;j<=4;j++){
            const TV_INT3& face_v=faces(j);
            if(particles.on_boundary(face_v(1))(3)&&particles.on_boundary(face_v(2))(3)&&particles.on_boundary(face_v(3))(3)){  ////approximation
                T tet_area=(T)one_third*TRIANGLE_3D<T>::Area(particles.X(face_v(1)),particles.X(face_v(2)),particles.X(face_v(3)));
                tet_area_sum+=tet_area;
            }
        }
    }
    return tet_area_sum;
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_MESHING_BASE<TV>::Node_Segment_Length_Sum(const int p,const HASHTABLE<int,ARRAY<int> >& vtx_seg_hashtable)
{
    const ARRAY<int>* p_seg=vtx_seg_hashtable.Get_Pointer(p);T seg_length_sum=(T)0;
    if(p_seg)for(int k=1;k<=(*p_seg).m;k++){
        int s2=(*p_seg)(k);
        const TV_INT2& v2=segment_mesh.elements(s2);
        T seg_length=(T).5*(particles.X(v2(1))-particles.X(v2(2))).Magnitude();
        seg_length_sum+=seg_length;
    }
    return seg_length_sum;
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESHING_BASE<TV>::Check_Mass()
{
    for(int i=1;i<=particles.X.m;i++){
        if(particles.mass(i)<=(T)0){
            LOG::cout<<"zero mass: "<<i<<", "<<particles.guid(i)<<", "<<particles.mass(i)<<", "<<particles.dimensional_mass(i)<<", "<<particles.thickness(i)<<std::endl;
        }
    }
}

//template class CODIMENSIONAL_MESHING_BASE<VECTOR<float,2> >;
template class CODIMENSIONAL_MESHING_BASE<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
//template class CODIMENSIONAL_MESHING_BASE<VECTOR<double,2> >;
template class CODIMENSIONAL_MESHING_BASE<VECTOR<double,3> >;
#endif

