//#####################################################################
// Copyright 2013, Bo Zhu, Ed Quigley
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __CODIMENSIONAL_MESHING_BASE__
#define __CODIMENSIONAL_MESHING_BASE__

#include <PhysBAM_Geometry/Topology_Based_Geometry_Computations/DELAUNAY_MESHING_WITH_ALPHA_SHAPE.h>
#include <PhysBAM_Dynamics/Particles/CODIMENSIONAL_FLUID_PARTICLES.h>
#include <PhysBAM_Solids/PhysBAM_Rigids/Rigid_Bodies/RIGID_BODY.h>
#include <PhysBAM_Tools/Data_Structures/HASHTABLE.h>
#include <PhysBAM_Tools/Data_Structures/HASHTABLE_ITERATOR.h>
#include <PhysBAM_Tools/Parallel_Computation/THREADED_UNIFORM_GRID.h>
#include <PhysBAM_Tools/Random_Numbers/RANDOM_NUMBERS.h>
#include "../Meshing/CODIMENSIONAL_MESH.h"
#include "../Meshing/CODIMENSIONAL_MESH_OBJECT_3D.h"
#include "../Misc/WATER_TANK.h"
#include "../Misc/WATER_SOURCE.h"
#include "../Mpi/MPI_CODIMENSIONAL.h"

namespace PhysBAM
{
using namespace Codimensional;
template <class TV> class FIBER;

template<class TV>
class CODIMENSIONAL_MESHING_BASE
{
    typedef typename TV::SCALAR T;
    typedef typename TV::template REBIND<int>::TYPE TV_INT;
    typedef VECTOR<T,2> TV2;typedef VECTOR<T,3> TV3;typedef VECTOR<int,2> TV_INT2;typedef VECTOR<int,3> TV_INT3;typedef VECTOR<int,4> TV_INT4;
    typedef typename IF<TV::m==2,VECTOR<int,1>,VECTOR<int,2> >::TYPE TV_INT_DIM_MINUS_ONE;
    typedef typename IF<TV::m==2,SEGMENTED_CURVE_2D<T>,TRIANGULATED_SURFACE<T> >::TYPE T_SURFACE_MESH_OBJECT;
    typedef typename IF<TV::m==2,SEGMENT_MESH,TRIANGLE_MESH>::TYPE SURFACE_MESH;
    typedef typename IF<TV::m==2,TRIANGULATED_AREA<T>,TETRAHEDRALIZED_VOLUME<T> >::TYPE T_VOLUME_MESH_OBJECT;
    typedef typename IF<TV::m==2,TRIANGLE_MESH,TETRAHEDRON_MESH>::TYPE VOLUME_MESH;
    typedef typename IF<TV::m==2,SEGMENT_2D<T>,TRIANGLE_3D<T> >::TYPE T_SURFACE_SIMPLEX;
    typedef typename IF<TV::m==2,TRIANGLE_2D<T>,TETRAHEDRON<T> >::TYPE T_VOLUME_SIMPLEX;
public:
    CODIMENSIONAL_MESH_OBJECT<TV>& codimensional_mesh_object;
    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles;
    CODIMENSIONAL_MESH& mesh;
    SEGMENT_MESH& segment_mesh;
    TRIANGLE_MESH& triangle_mesh;
    TETRAHEDRON_MESH& tetrahedron_mesh;
    SEGMENT_MESH& rim_segment_mesh;

    ////meshing parameters
    T length_scale;
    T adaptive_length_scale_min;
    T thickness_scale;
    T rim_thickness_scale;
    T length_to_split_coef;
    T length_to_collapse_coef;
    //T max_segment_length;
    //T min_segment_length;

    //T interior_max_segment_length;
    //T interior_min_segment_length;
    //T boundary_max_segment_length;
    //T boundary_min_segment_length;
    //T diagonal_max_segment_length;
    //T diagonal_min_segment_length;
    //T buffer_max_segment_length;
    //T buffer_min_segment_length;

    ////simulation parameters
    T density;
    T max_particle_radius;
    T min_particle_radius;
    T max_particle_volume;
    T min_particle_volume;
    T max_particle_mass;
    T min_particle_mass;
    T incompressible_segment_length;
    T incompressible_triangle_area;
    T max_thickness;
    T min_thickness;
    T max_rim_thickness;
    T min_rim_thickness;
    T max_thin_simplex_volume;
    T epsilon;
    T smooth_rim_thickness_weight;
    T smooth_interior_thickness_weight;
    int smooth_rim_thickness_mode;
    T rim_interior_mass_transfer_coefficient;
    T clamp_point_thickness_min;
    T clamp_segment_thickness_min;
    T clamp_triangle_thickness_min;
    T clamp_point_thickness_max;    
    T clamp_segment_thickness_max;
    T clamp_triangle_thickness_max;
    T y_in_the_air_min;
    int break_seg_num;
    int allowed_break_seg_num_max;

    ////background grid for collision detections
    BOX<TV> particle_bounding_box;
    GRID<TV> particle_grid;
    ARRAY<ARRAY<int>,TV_INT> particles_in_cell; 
    ARRAY<ARRAY<int>,TV_INT> segments_in_cell;
    ARRAY<ARRAY<int>,TV_INT> triangles_in_cell;
    ARRAY<ARRAY<int>,TV_INT> tetrahedra_in_cell;

    ////low dimensional incompressibility
    bool use_low_dimensional_incompressibility;
    bool use_thickness_based_low_dimensional_incompressibility;
    bool use_distance_to_rim_for_incompressibility;
    int min_distance_to_rim_for_incompressibility;
    ARRAY<bool> segment_incompressibilities;
    ARRAY<bool> triangle_incompressibilities;

    ////crumples
    bool use_thickness_based_dimension_lifting;
    bool use_crumple_based_dimension_lifting;
    bool use_crumple_merge;

    ////options and flags
    bool use_reseed;
    bool use_tetrahedron_mesh;
    bool use_triangle_mesh;
    bool use_segment_mesh;
    bool initialize_tetrahedron_mesh;
    bool initialize_triangle_mesh;
    bool initialize_segment_mesh;
    bool use_break_thin_segments;
    bool use_break_thin_triangles;
    bool use_remove_bad_tetrahedra;

    bool use_topology_repairing;
    bool has_tank;
    bool has_solid;
    bool has_eulerian_particles;
    bool use_rim_adaptivity;
    bool use_smooth_rim_thickness;
    bool use_smooth_interior_thickness;
    bool use_align_triangle_normals;
    bool use_rim_mass_redistribution;
    bool use_break_elements_in_solids;  ////for bubble breaking
    bool use_high_res_volume_mesh;
    bool use_high_res_segment_mesh;
    bool use_elements_merging;
    bool use_elements_merging_to_triangle_mesh;
    bool use_elements_merging_to_tetrahedron_mesh;
    bool use_remove_low_dim_elements_on_tank_bottom;
    bool use_continuous_meshing_operations;
    bool use_half_cylindrical_rim;
    bool has_rim;
    bool use_adaptive_tetrahedron_remeshing;
    bool use_adaptive_triangle_remeshing;
    bool use_adaptive_segment_remeshing;
    bool preserve_thickness_in_remeshing;
    bool use_thickness_aware_triangle_remeshing;
    bool use_boundary_vertex_snap;
    bool use_codim_vertex_snap;
    bool use_mesh_edge_connect;
    bool use_contact_aware_triangle_remeshing;
    bool use_anisotropic_triangle_remeshing;
    bool use_thin_sheet_aware_tetrahedron_remeshing;
    bool use_constrained_thin_triangle_break;
    bool use_constrained_thin_segment_break;
    bool use_enforce_thin_tri_break_constrained;
    bool use_particle_parents;
    bool use_clamp_max_thickness;
    bool use_fix_poping;
    bool use_split_eulerian_particles;
    bool use_remove_codim_elements_in_the_air;
    bool use_remove_invalid_elements;
    bool use_allowed_break_seg_num;

    MPI_CODIMENSIONAL<TV>* mpi_codim;
    ARRAY<ARRAY<RANGE<TV> > > sub_regions;

    ////solid boundary
    ARRAY<RIGID_BODY<TV>* > rigid_bodies;
    ARRAY<RIGID_BODY<TV>* > kinematic_rigid_bodies;
    FIBER<TV>* fiber;
    HASHTABLE<int> eulerian_particles_hashtable;        ////storing eulerian particle guid
    HASHTABLE<int,PAIR<TV_INT2,T> > added_eulerian_particle_parents_hashtable;
    //HASHTABLE<int> solid_contacting_particles_hashtable;    ////storing solid contacting particle guid
    WATER_TANK<TV>* water_tank;

    RANDOM_NUMBERS<T> random;
    ////debug
    bool use_dbg;
    int dbg_number;
    TIMER timer;

    CODIMENSIONAL_MESHING_BASE(CODIMENSIONAL_MESH_OBJECT<TV>& codimensional_mesh_object_input,MPI_CODIMENSIONAL<TV>* mpi_codim_input=0);

    virtual void Initialize_Length_Scale(const T length_scale,const T adaptive_length_scale_min,const T thickness_scale=(T)0,const T rim_thickness_scale=(T)0,
        const T max_thickness_factor=(T)1.5,const T min_thickness_factor=(T).1,const T max_rim_thickness_factor=(T)1.5,const T min_rim_thickness_factor=(T).1);
    virtual void Initialize_Mpi(MPI_CODIMENSIONAL<TV>* mpi_codim_input);
    void Initialize_From_External_Mesh_And_Positions(const CODIMENSIONAL_MESH& ext_mesh,const ARRAY_VIEW<TV>& ext_X);

    ////update particle attributes
    virtual void Update_Particle_Attributes_After_Having_Mesh_Topology();
    virtual void Update_Particle_Geometry_Attributes();
    virtual void Update_Particle_Distance_To_Rim(){PHYSBAM_NOT_IMPLEMENTED();}
    virtual void Update_Particle_Thickness(bool has_rim);
    virtual void Update_Particle_Area_Weighted_Normal(){PHYSBAM_NOT_IMPLEMENTED();}
    virtual void Update_Particle_Normal(){PHYSBAM_NOT_IMPLEMENTED();}
    virtual void Initialize_Sub_Regions(){PHYSBAM_NOT_IMPLEMENTED();}

    ////update incompressibility
    virtual void Update_Mesh_Incompressibilities();
    virtual void Update_Particle_Incompressibilities();
    virtual void Update_Segment_Incompressibilities();
    virtual void Update_Triangle_Incompressibilities();
    virtual void Update_Incompressibility_Mesh(CODIMENSIONAL_MESH& incompressibility_codimensional_mesh);
    virtual void Update_Viscosity_Mesh(CODIMENSIONAL_MESH& viscosity_codimensional_mesh,bool has_vis_rim_segments=false,bool has_vis_rim_triangles=true);

    ////mass and thickness
    virtual void Initialize_Particle_Mass_Based_On_Geometry(T initial_thickness,T initial_rim_thickness,bool initialize_rim=true,bool initialize_thickness=true,WATER_SOURCE<TV>* water_source=0);
    virtual void Update_Particle_Mass_Based_On_Geometry();
    virtual T Discrete_Particle_Thickness(const int p,bool has_rim);
    virtual T Continuous_Particle_Thickness(const int p,bool has_rim);

    virtual void Break_Thin_Segments(){PHYSBAM_NOT_IMPLEMENTED();}
    virtual void Smooth_Rim_Thickness(){PHYSBAM_NOT_IMPLEMENTED();}
    virtual void Smooth_Interior_Thickness(){PHYSBAM_NOT_IMPLEMENTED();}

    ////solid boundary
    void Add_Solid(RIGID_BODY<TV>& solid_input){rigid_bodies.Append(&solid_input);has_solid=true;}
    void Initialize_Tank(const BOX<TV>& tank_box_input,const T tank_thickness_input,bool is_open_tank_input=false)
    {has_tank=true;if(water_tank)delete water_tank;water_tank=new WATER_TANK<TV>(tank_box_input,tank_thickness_input,is_open_tank_input,&codimensional_mesh_object);}
    void Add_Eulerian_Particles(HASHTABLE<int,TV>& eulerian_particles_input)
    {has_eulerian_particles=true;for(HASHTABLE_ITERATOR<int,TV> iter(eulerian_particles_input);iter.Valid();iter.Next())eulerian_particles_hashtable.Set(iter.Key());}
    void Add_Eulerian_Particle(const int guid){has_eulerian_particles=true;eulerian_particles_hashtable.Set(guid);}
    void Clear_Eulerian_Particles(){has_eulerian_particles=false;eulerian_particles_hashtable.Remove_All();added_eulerian_particle_parents_hashtable.Remove_All();}

protected:
    BOX<TV> Get_Particle_Bounding_Box();
    GRID<TV> Get_Particle_Grid(const BOX<TV>& bounding_box,const T grid_dX);
    void Resize_Elements_In_Cell_With_Grid(ARRAY<ARRAY<int>,TV_INT>& p_in_cell,const GRID<TV>& grid);
    
    //bool Is_Incompressible_Segment_Particle(const int p);
    //bool Is_Incompressible_Triangle_Particle(const int p);

    bool Is_Incompressible_Segment(const int s);
    bool Is_Incompressible_Triangle(const int t);

    T Joint_Thickness(const T V0,const T A1,const T L1);    ////solve quadratic equation pi*L1/4*x^2+A1*x-V0=0
    T Node_Triangle_Area_Sum(const int p,const HASHTABLE<int,ARRAY<int> >& vtx_tri_hashtable);
    T Node_Segment_Length_Sum(const int p,const HASHTABLE<int,ARRAY<int> >& vtx_seg_hashtable);
    T Nodal_Tetrahedron_Surface_Area_Sum(const int p,const HASHTABLE<int,ARRAY<int> >& vtx_tet_hashtable);

    void Check_Mass();
};
}
#endif
