//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __CODIMENSIONAL_MESH_CHECKER__
#define __CODIMENSIONAL_MESH_CHECKER__
#include "../Meshing/CODIMENSIONAL_HELPER.h"

namespace PhysBAM
{
namespace Codimensional
{
template<class TV_INT> bool Check_Duplicate_Element(const ARRAY<TV_INT>& elements,bool check_reverted_element=true,ARRAY_VIEW<int>* guid=0,int node_id=0,bool show_error_info=true)
{
    bool has_duplicate_element=false;
    HASHTABLE<TV_INT,int> element_hashtable;
    for(int i=1;i<=elements.m;i++){
        const TV_INT& t_key=Unique_Ordered(elements(i));
        
        if(!element_hashtable.Contains(t_key)){element_hashtable.Insert(t_key,i);}
        else{
            has_duplicate_element=true;int j=element_hashtable.Get(t_key);
            if(show_error_info){
                TV_INT t_guid_i=TV_INT();TV_INT t_guid_j=TV_INT();
                if(guid){t_guid_i=Element(*guid,elements(i));t_guid_j=Element(*guid,elements(j));}
                LOG::cout<<" ################## Error: duplicated element "<<j<<": "<<elements(j)<<", "<<i<<": "<<elements(i)<<
                    " guid: "<<t_guid_j<<", "<<t_guid_i<<", N"<<node_id<<" ################## "<<std::endl;
            }
        }

        if(check_reverted_element){
            const TV_INT& reverted_t_key=Unique_Ordered(Reversed(elements(i)));
            if(element_hashtable.Contains(reverted_t_key)){
                int j=element_hashtable.Get(reverted_t_key);
                if(show_error_info){
                    TV_INT t_guid_i=TV_INT();TV_INT t_guid_j=TV_INT();
                    if(guid){t_guid_i=Element(*guid,elements(i));t_guid_j=Element(*guid,elements(j));}
                    LOG::cout<<" ################## Error: reverted element "<<j<<": "<<elements(j)<<", "<<i<<": "<<elements(i)<<
                        " guid: "<<t_guid_j<<", "<<t_guid_i<<", N"<<node_id<<" ################## "<<std::endl;
                }
            }
        }

    }
    return !has_duplicate_element;
}

template<class TV_INT> bool Check_Dupliciate_Vertex(const ARRAY<TV_INT>& elements,ARRAY_VIEW<int>* guid=0,int node_id=0,bool show_error_info=true)
{
    bool has_duplicate_vertex=false;
    for(int i=1;i<=elements.m;i++){
        const TV_INT& t=elements(i);
        for(int d=1;d<=TV_INT::m;d++){
            for(int dd=d+1;dd<=TV_INT::m;dd++){
                if(t(d)==t(dd)){
                    has_duplicate_vertex=true;
                    if(show_error_info){
                        int v_guid=0;if(guid){v_guid=(*guid)(t(d));}
                        LOG::cout<<" ################## Error: duplicate vertex index in element "<<i<<": "<<t<<" guid: "<<v_guid<<", N"<<node_id<<" ################## "<<std::endl;
                    }
                }
            }
        }
    }
    return !has_duplicate_vertex;
}

template<class TV_INT> bool Check_Invalid_Vertex(const ARRAY<TV_INT>& elements,const int idx_min,const int idx_max,ARRAY_VIEW<int>* guid=0,int node_id=0,bool show_error_info=true)
{
    bool has_invalid_vtx=false;
    for(int i=1;i<=elements.m;i++){
        const TV_INT& t=elements(i);bool is_invalid_tet=false;TV_INT t_guid;
        for(int d=1;d<=TV_INT::m;d++){if(t(d)<idx_min||t(d)>idx_max){is_invalid_tet=true;has_invalid_vtx=true;}else if(guid){t_guid(d)=(*guid)(t(d));}}
        if(show_error_info&&is_invalid_tet)LOG::cout<<" ################## Error: invalid vertex index in element "<<i<<": "<<t<<" guid "<<t_guid<<", N"<<node_id<<" ################## "<<std::endl;
    }
    return !has_invalid_vtx;
}

template<class TV_INT> bool Check_Nonmanifold_Face(const ARRAY<TV_INT>& elements,ARRAY_VIEW<int>* guid=0,int node_id=0,bool show_error_info=true)
{
    bool has_nonmanifold_face=false;
    typedef VECTOR<int,TV_INT::m-1> TV_INTF;
    HASHTABLE<TV_INTF,ARRAY<int> > face_e_hashtable;
    for(int i=1;i<=elements.m;i++){
        const TV_INT& v=elements(i);
        ARRAY<TV_INTF> faces;Simplex_Faces(v,faces);
        for(int j=1;j<=Simplex_Face_Number<TV_INT::m>();j++){
            TV_INTF f_key=faces(j).Sorted();TV_INTF f_key_guid=Element(*guid,f_key);
            ARRAY<int>& f_e=face_e_hashtable.Get_Or_Insert(f_key,ARRAY<int>());
            f_e.Append(i);
            if(f_e.m>2){
                for(int k=1;k<=f_e.m;k++){
                    int t=f_e(k);const TV_INT& tv=elements(t);TV_INT tv_guid=Element(*guid,tv);has_nonmanifold_face=true;
                    if(show_error_info)LOG::cout<<" ################## Error: nonmanifold face: "<<f_key<<" guid "<<f_key_guid<<" element: "<<k<<": "<<t<<": "<<tv<<" guid "<<tv_guid<<", N"<<node_id<<" ################## "<<std::endl;
                }
            }
        }
    }
    return !has_nonmanifold_face;
}

template<class T> bool Check_Non_Positive_Volume_Tetrahedron(const ARRAY<VECTOR<int,4> >& elements,ARRAY_VIEW<VECTOR<T,3> >& X,ARRAY_VIEW<int>* guid=0,int node_id=0,bool show_error_info=true)
{
    bool has_non_positive_vol_tet=false;
    for(int i=1;i<=elements.m;i++){
        const VECTOR<int,4>& v=elements(i);
        bool is_non_positive_volume=Is_Non_Positive_Volume_Tetrahedron(X,v);
        if(is_non_positive_volume){
            has_non_positive_vol_tet=true;
            double vol=Accurate_Tetrahedron_Volume(X,v);
            VECTOR<int,4> t_guid;if(guid) for(int d=1;d<=4;d++){t_guid(d)=(*guid)(v(d));}
            if(show_error_info)LOG::cout<<" ################## Error: non positive volume tetrahedron "<<i<<": "<<v<<", guid "<<t_guid<<", vol: "<<vol<<", N"<<node_id<<" ################## "<<std::endl;
        }
    }
    return !has_non_positive_vol_tet;
}

template<class T> bool Check_Negative_Volume_Tetrahedron(const ARRAY<VECTOR<int,4> >& elements,ARRAY_VIEW<VECTOR<T,3> >& X,ARRAY_VIEW<int>* guid=0,int node_id=0,bool show_error_info=true)
{
    bool has_negative_vol_tet=false;
    for(int i=1;i<=elements.m;i++){
        const VECTOR<int,4>& v=elements(i);
        bool is_negative_volume=Is_Non_Positive_Volume_Tetrahedron(X,v);
        if(is_negative_volume){
            has_negative_vol_tet=true;
            double vol=Accurate_Tetrahedron_Volume(X,v);
            VECTOR<int,4> t_guid;if(guid) for(int d=1;d<=4;d++){t_guid(d)=(*guid)(v(d));}
            if(show_error_info)LOG::cout<<" ################## Error: negative volume tetrahedron "<<i<<": "<<v<<", guid "<<t_guid<<", vol: "<<vol<<", N"<<node_id<<" ################## "<<std::endl;
        }
    }
    return !has_negative_vol_tet;
}

template<class T> bool Check_Skinny_Tetrahedron(const ARRAY<VECTOR<int,4> >& elements,ARRAY_VIEW<VECTOR<T,3> >& X,ARRAY_VIEW<T>& length_scale,ARRAY_VIEW<int>* guid=0,int node_id=0,bool show_error_info=true)
{
    bool has_skinny_tet=false;
    //T short_edge_frac=(T).2;
    T short_distance_frac=(T).1;T t_threshold=(T).01;
    for(int i=1;i<=elements.m;i++){
        const VECTOR<int,4>& v=elements(i);
        TETRAHEDRON<T> tet(X(v(1)),X(v(2)),X(v(3)),X(v(4)));
        T particle_length_scale=(T).25*(length_scale(v(1))+length_scale(v(2))+length_scale(v(3))+length_scale(v(4)));     
        T distance_threshold=particle_length_scale*short_distance_frac;
        SKINNY_TETRAHEDRON<T> skinny_tet;int skinny_tet_type=Split_Skinny_Tetrahedron(tet,distance_threshold,t_threshold,skinny_tet);
        if(skinny_tet_type!=0){
            has_skinny_tet=true;
            if(show_error_info){
                std::string name;switch(skinny_tet_type){case SLIVER:name="sliver";break;case SPADE:name="spade";break;case CAP:name="cap";break;}
                VECTOR<int,4> t_guid;if(guid) for(int d=1;d<=4;d++){t_guid(d)=(*guid)(v(d));}
                LOG::cout<<" ################## Skinny tetrahedron "<<i<<": "<<v<<", guid "<<t_guid<<", type: "<<name<<", N"<<node_id<<" ################## "<<std::endl;
            }
        }
    }
    return !has_skinny_tet;
}
}
}

#endif
