//#####################################################################
// Copyright 2013, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __CODIMENSIONAL_MESH_OBJECT__
#define __CODIMENSIONAL_MESH_OBJECT__

#include <PhysBAM_Tools/Vectors/VECTOR.h>
#include <PhysBAM_Tools/Math_Tools/pow.h>
#include <PhysBAM_Geometry/Topology_Based_Geometry/SEGMENTED_CURVE_2D.h>
#include <PhysBAM_Geometry/Topology_Based_Geometry/SEGMENTED_CURVE.h>
#include <PhysBAM_Geometry/Topology_Based_Geometry/TRIANGULATED_AREA.h>
#include <PhysBAM_Geometry/Topology_Based_Geometry/TRIANGULATED_SURFACE.h>
#include <PhysBAM_Geometry/Topology_Based_Geometry/TETRAHEDRALIZED_VOLUME.h>
#include <PhysBAM_Geometry/Topology/SEGMENT_MESH.h>
#include <PhysBAM_Geometry/Topology/TRIANGLE_MESH.h>
#include <PhysBAM_Geometry/Topology/TETRAHEDRON_MESH.h>
#include <PhysBAM_Geometry/Basic_Geometry/TETRAHEDRON.h>
#include <PhysBAM_Geometry/Basic_Geometry/TRIANGLE_2D.h>
#include <PhysBAM_Geometry/Basic_Geometry/TRIANGLE_3D.h>
#include <PhysBAM_Dynamics/Particles/CODIMENSIONAL_FLUID_PARTICLES.h>
#include "../Meshing/CODIMENSIONAL_MESH.h"
#include "../Meshing/CODIMENSIONAL_HELPER.h"

namespace PhysBAM
{
template<class TV>
class CODIMENSIONAL_MESH_OBJECT_BASE
{
    typedef typename TV::SCALAR T;
    typedef typename IF<TV::m==2,TRIANGLE_2D<T>,TRIANGLE_3D<T> >::TYPE T_TRIANGLE;
public:
    CODIMENSIONAL_MESH& mesh;
    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles;

    CODIMENSIONAL_MESH_OBJECT_BASE():mesh(*(new CODIMENSIONAL_MESH())),particles(*(new CODIMENSIONAL_FLUID_PARTICLES<TV>())){}
    CODIMENSIONAL_MESH_OBJECT_BASE(CODIMENSIONAL_MESH& mesh_input,CODIMENSIONAL_FLUID_PARTICLES<TV>& particles_input):mesh(mesh_input),particles(particles_input){}

    TV Total_Momentum(){TV total_momentum=TV();for(int i=1;i<=particles.X.m;i++)total_momentum+=particles.mass(i)*particles.V(i);return total_momentum;}
    T Total_Mass(){T total_mass=(T)0;for(int i=1;i<=particles.X.m;i++)total_mass+=particles.mass(i);return total_mass;}
    TV Total_Dimensional_Mass(){TV total_dimensional_mass=TV();for(int i=1;i<=particles.X.m;i++)total_dimensional_mass+=particles.dimensional_mass(i);return total_dimensional_mass;}

    ////simplex computations
    T Segment_Length(const int e);
    T Triangle_Area(const int e);

    ////nodal simplex computations
    T Nodal_Segment_Length_Sum(const int p);
    T Nodal_Segment_Length_Sum_On_Triangle_Mesh_Boundary(const int p);      ////including segments on triangle mesh boundary only
    T Nodal_Segment_Length_Sum_On_Rim_Mesh(const int p);    ////including segments on both triangle mesh boundary and segment mesh
    T Nodal_Segment_Length_Average(const int p);
    int Nodal_Segment_Degree(const int p);

    T Nodal_Triangle_Area_Sum(const int p);
    int Nodal_Triangle_Degree(const int p);
    T Nodal_Triangle_Area_Average(const int p);
    T Nodal_Extra_Triangle_Area_Sum(const int p);

    static T Nodal_Segment_Length_Sum(ARRAY_VIEW<TV>& X,const int p,const HASHTABLE<int,ARRAY<int> >& rim_vtx_vtx_hashtable);
    static T Nodal_Segment_Length_Fraction(ARRAY_VIEW<TV>& X,const int p,const int p2,const HASHTABLE<int,ARRAY<int> >& rim_vtx_vtx_hashtable);
};

template<class TV> class CODIMENSIONAL_MESH_OBJECT;

//////////////////////////////////////////////////////////////////////////
////codimensional mesh object 2d
template<class T_input>
class CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,2> > : public CODIMENSIONAL_MESH_OBJECT_BASE<VECTOR<T_input,2> >
{
    typedef T_input T;typedef VECTOR<T,2> TV;
    typedef CODIMENSIONAL_MESH_OBJECT_BASE<VECTOR<T,2> > BASE;
public:
    using BASE::mesh; using BASE::particles;
    using BASE::Triangle_Area;using BASE::Segment_Length;using BASE::Nodal_Segment_Length_Sum;using BASE::Nodal_Segment_Degree;
    using BASE::Nodal_Segment_Length_Average;using BASE::Nodal_Triangle_Area_Sum;using BASE::Nodal_Extra_Triangle_Area_Sum;
    using BASE::Nodal_Triangle_Degree;using BASE::Nodal_Triangle_Area_Average;

	TRIANGLE_MESH& volume_mesh;
	SEGMENTED_CURVE_2D<T> segmented_curve;
	TRIANGULATED_AREA<T> triangulated_area;

    CODIMENSIONAL_MESH_OBJECT();
    CODIMENSIONAL_MESH_OBJECT(CODIMENSIONAL_MESH& mesh_input,CODIMENSIONAL_FLUID_PARTICLES<TV>& particles_input);

    ////simplex area and volume calculation
    T Point_Volume(const int p);
    T Segment_Volume(const int e);
    T Triangle_Volume(const int e);
	T Volumetric_Element_Volume(const int e);

    ////nodal fraction and volume calculation
    T Nodal_Segment_Volume(const int p,const int s);
    T Nodal_Segment_Volume(const int p);
    T Nodal_Segment_Volume_Fraction(const int p,const int s);
    
    T Nodal_Triangle_Volume(const int p,const int t);
    T Nodal_Triangle_Volume(const int p);
    T Nodal_Triangle_Volume_Fraction(const int p,const int t);

    T Nodal_Tetrahedron_Volume(const int p,const int t){PHYSBAM_NOT_IMPLEMENTED();}
    T Nodal_Tetrahedron_Volume(const int p){PHYSBAM_NOT_IMPLEMENTED();}
    T Nodal_Tetrahedron_Volume_Sum(const int p){PHYSBAM_NOT_IMPLEMENTED();}

    T Nodal_Volume(const int p);    ////V_seg+V_tri
    T Nodal_Volume_In_Highest_Dimension(const int p);
    T Nodal_Volume_In_Highest_Dimension_Fraction(const int p);  ////V_highest/V_total

    ////nodal normals
    TV Nodal_Triangle_Area_Weighted_Normal(const int p){/*TODO: implement this*/return TV();}
    TV Nodal_Area_Weighted_Mean_Curvature_Normal(const int p){/*TODO: implement this*/return TV();}
    TV Nodal_Mean_Curvature_Normal(const int p){/*TODO: implement this*/return TV();}
    TV Nodal_Volume_Weighted_Normal(const int p){/*TODO: implement this*/return TV();}

    T Total_Volume(){return Total_Triangle_Volume()+Total_Segment_Volume()+Total_Point_Volume();}
    T Total_Triangle_Volume(){T total_area=(T)0;for(int i=1;i<=mesh.triangle_mesh.elements.m;i++){total_area+=Triangle_Volume(i);}return total_area;}
    T Total_Segment_Volume(){T total_area=(T)0;for(int i=1;i<=mesh.segment_mesh.elements.m;i++){total_area+=Segment_Volume(i);}return total_area;}
    T Total_Point_Volume(){T total_area=(T)0;for(int i=1;i<=particles.X.m;i++)if(Is_Isolated(mesh.node_types(i)))total_area+=Point_Volume(i);return total_area;}

	SEGMENTED_CURVE_2D<T>* Get_Mesh_Object_1D(){return &segmented_curve;}
	TRIANGULATED_AREA<T>* Get_Mesh_Object_2D(){return &triangulated_area;}
	TRIANGULATED_AREA<T>* Get_Mesh_Object_3D(){return &triangulated_area;}
	SEGMENTED_CURVE_2D<T>* Get_Mesh_Object_N_Minus_One_D(){return Get_Mesh_Object_1D();}
    
    static T Segment_Cross_Section_Area(const T thickness){return thickness;}
    static T Hemisphere_Volume(const T thickness){return (T).5*(T)pi*pow<2>((T).5*thickness);}
};

//////////////////////////////////////////////////////////////////////////
////codimensional mesh object 3d
template<class T_input>
class CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> > : public CODIMENSIONAL_MESH_OBJECT_BASE<VECTOR<T_input,3> >
{
    typedef T_input T;typedef VECTOR<T,3> TV;
    typedef CODIMENSIONAL_MESH_OBJECT_BASE<VECTOR<T,3> > BASE;
public:
    using BASE::mesh; using BASE::particles;
    using BASE::Triangle_Area;using BASE::Segment_Length;using BASE::Nodal_Segment_Length_Sum;using BASE::Nodal_Segment_Degree;
    using BASE::Nodal_Segment_Length_Average;using BASE::Nodal_Triangle_Area_Sum;using BASE::Nodal_Extra_Triangle_Area_Sum;
    using BASE::Nodal_Triangle_Degree;using BASE::Nodal_Triangle_Area_Average;

	TETRAHEDRON_MESH& volume_mesh;
	SEGMENTED_CURVE<TV> segmented_curve;
	TRIANGULATED_SURFACE<T> triangulated_surface;
	TETRAHEDRALIZED_VOLUME<T> tetrahedralized_volume;

    CODIMENSIONAL_MESH_OBJECT();
    CODIMENSIONAL_MESH_OBJECT(CODIMENSIONAL_MESH& mesh_input,CODIMENSIONAL_FLUID_PARTICLES<TV>& particles_input);

    ////simplex area and volume calculation
    T Point_Volume(const int p);	
    T Segment_Volume(const int e);
    T Triangle_Volume(const int e);
    T Tetrahedron_Volume(const int e);
	T Volumetric_Element_Volume(const int e);

    ////nodal fraction and volume calculation
    T Nodal_Segment_Volume(const int p,const int s);    ////part of the volume of segment s incident to p
    T Nodal_Segment_Volume(const int p);
    T Nodal_Segment_Volume_Fraction(const int p,const int s);   ////V_node/V_seg (V_node is the part of the volume of segment incident to p)
    
    T Nodal_Triangle_Volume(const int p,const int t);   ////part of the volume of triangle t incident to p
    T Nodal_Triangle_Volume(const int p);
    T Nodal_Triangle_Volume_Fraction(const int p,const int t);
    
    T Nodal_Tetrahedron_Volume(const int p,const int t);
    T Nodal_Tetrahedron_Volume(const int p);
    T Nodal_Tetrahedron_Volume_Sum(const int p);
    
    T Nodal_Volume(const int p);    ////V_seg+V_tri+V_tet
    T Nodal_Volume_In_Highest_Dimension(const int p);   ////V_tet
    T Nodal_Volume_In_Highest_Dimension_Fraction(const int p);  ////V_tet/V_total

    ////nodal normals
    TV Nodal_Triangle_Area_Weighted_Normal(const int p);
    TV Nodal_Area_Weighted_Mean_Curvature_Normal(const int p);
    TV Nodal_Mean_Curvature_Normal(const int p);
    TV Nodal_Volume_Weighted_Normal(const int p);

    T Total_Volume(){return Total_Tetrahedron_Volume()+Total_Triangle_Volume()+Total_Segment_Volume()+Total_Point_Volume();}
    T Total_Tetrahedron_Volume(){T total_volume=(T)0;for(int i=1;i<=mesh.tetrahedron_mesh.elements.m;i++){total_volume+=Tetrahedron_Volume(i);}return total_volume;}
    T Total_Triangle_Volume(){T total_volume=(T)0;for(int i=1;i<=mesh.triangle_mesh.elements.m;i++){total_volume+=Triangle_Volume(i);}return total_volume;}
    T Total_Segment_Volume(){T total_volume=(T)0;for(int i=1;i<=mesh.segment_mesh.elements.m;i++){total_volume+=Segment_Volume(i);}return total_volume;}
    T Total_Point_Volume(){T total_volume=(T)0;for(int i=1;i<=particles.X.m;i++)if(Is_Isolated(mesh.node_types(i)))total_volume+=Point_Volume(i);return total_volume;}

	SEGMENTED_CURVE<TV>* Get_Mesh_Object_1D(){return &segmented_curve;}
	TRIANGULATED_SURFACE<T>* Get_Mesh_Object_2D(){return &triangulated_surface;}
	TETRAHEDRALIZED_VOLUME<T>* Get_Mesh_Object_3D(){return &tetrahedralized_volume;}
	TRIANGULATED_SURFACE<T>* Get_Mesh_Object_N_Minus_One_D(){return Get_Mesh_Object_2D();}

    static T Segment_Cross_Section_Area(const T thickness){return (T)pi*pow<2>((T).5*thickness);}
    static T Hemisphere_Volume(const T thickness){return (T)two_thirds_pi*(T)pi*pow<3>((T).5*thickness);}
};
}
   
#endif