//#####################################################################
// Copyright 2013, Bo Zhu, Ed Quigley
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################

#include <PhysBAM_Dynamics/Particles/CODIMENSIONAL_FLUID_PARTICLES.h>
#include <PhysBAM_Geometry/Basic_Geometry/BOX.h>
#include <PhysBAM_Tools/Data_Structures/PAIR.h>
#include <PhysBAM_Tools/Matrices/ROTATION.h>
#include "../Meshing/CODIMENSIONAL_HELPER.h"
#include "../Misc/WATER_ANNULAR_GAP_SOURCE.h"

using namespace PhysBAM;
using namespace Codimensional;

template<class TV>
PhysBAM::WATER_ANNULAR_GAP_SOURCE<TV>::WATER_ANNULAR_GAP_SOURCE(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,const TV& closed_end,const TV& open_end,const T& inner_radius_,const T& outer_radius_,const T& length_scale_,const T& initial_speed_)
    :BASE(closed_end,open_end,length_scale_,(T)0.5*length_scale_,(T)1e-3,initial_speed_),
    inner_radius(inner_radius_),outer_radius(outer_radius_)
{
    TV3 e1(endpoint1),e2(endpoint2);
    TV3 direction3;
    if(e1==e2)direction3=VECTOR<T,3>();
    else direction3=(e2-e1).Normalized();

    inner_cylinder.Set_Endpoints(e1,e2);
    outer_cylinder.Set_Endpoints(e1,e2);
    inner_cylinder.radius=inner_radius;
    outer_cylinder.radius=outer_radius;

    fixed_cylinder.Set_Endpoints(e1-thickness*direction3,e1+thickness*direction3);
    fixed_cylinder.radius=outer_radius;

    Seed_Particles(particles);
}

template<class TV>
int PhysBAM::WATER_ANNULAR_GAP_SOURCE<TV>::Seed_Particles(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles)
{
    int particles_added=0;
    T height=(endpoint1-endpoint2).Magnitude();
    int radial_samples,circular_samples,height_samples;
    radial_samples=(int)ceil((outer_radius-inner_radius)/length_scale);
    height_samples=(int)ceil(height/length_scale);
    TV orthogonal=direction.Orthogonal_Vector().Normalized();

    for(int h=0;h<=height_samples;h++){//includes zero
        T h_factor=((T)h)/height_samples;
        TV center=endpoint1+h_factor*height*direction;

        for(int r=0;r<=radial_samples;r++){
            T r_factor=((T)r)/radial_samples;
            T subradius=inner_radius+r_factor*(outer_radius-inner_radius);
            T circumference=(T)two_pi*subradius;
            circular_samples=(int)ceil(circumference/length_scale);

            for(int c=1;c<=circular_samples;c++){//lol, c++. meta-humor...
                T circular_factor=((T)c)/circular_samples;
                T offset=(h%2==0?((T)0.5/circular_samples):(T)0);
                T theta=(T)two_pi*(circular_factor-offset);
                ROTATION<TV3> rotation(theta,TV3(direction));
                TV3 v=rotation.Rotate(TV3(orthogonal)).Normalized();
                TV3 new_position=TV3(center)+subradius*v;
                if(r!=0&&r!=radial_samples&&h!=0&&h!=height_samples){
                    new_position+=perturb_coef*random.Get_Uniform_Vector(TV3::Constant_Vector((T)-1*length_scale),TV3::Constant_Vector(length_scale));
                }

                int pid=particles.Add_Particle();
                particles.X(pid)=TV_Compatible<T,TV::m>(new_position);
                particles.V(pid)=initial_speed*direction;
                if(h==0)eulerian_particles.Insert(particles.guid(pid),particles.X(pid));
                particles_added++;

            }//circular_samples
        }//radial_samples
    }//height_samples
    return particles_added;
}

template<class TV>
int PhysBAM::WATER_ANNULAR_GAP_SOURCE<TV>::Reseed_Source_Particles(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles)
{
    eulerian_particles.Remove_All();

    int particles_added=0;
    int radial_samples=(int)ceil((outer_radius-inner_radius)/length_scale);
    int circular_samples;
    TV center=endpoint1;
    TV orthogonal=direction.Orthogonal_Vector().Normalized();

    for(int r=0;r<=radial_samples;r++){
        T r_factor=((T)r)/radial_samples;
        T subradius=inner_radius+r_factor*(outer_radius-inner_radius);
        T circumference=(T)two_pi*subradius;
        circular_samples=(int)ceil(circumference/length_scale);

        for(int c=1;c<=circular_samples;c++){
            T circular_factor=((T)c)/circular_samples;
            T offset=(parity?((T)0.5/circular_samples):(T)0);
            T theta=(T)two_pi*(circular_factor-offset);
            ROTATION<TV3> rotation(theta,TV3(direction));
            TV3 v=rotation.Rotate(TV3(orthogonal)).Normalized();
            TV3 new_position=TV3(center)+subradius*v;

            int pid=particles.Add_Particle();
            particles.X(pid)=TV_Compatible<T,TV::m>(new_position);
            particles.V(pid)=initial_speed*direction;
            eulerian_particles.Insert(particles.guid(pid),particles.X(pid));
            particles_added++;
        }
    }
    parity=!parity;
    return particles_added;
}

template<class TV>
bool PhysBAM::WATER_ANNULAR_GAP_SOURCE<TV>::Particle_Is_Inside_Source(const TV& location) const
{
    TV3 l3(location);
    return outer_cylinder.Inside(l3,(T)0)&&!inner_cylinder.Inside(l3,(T)0);
}

template<class TV>
bool PhysBAM::WATER_ANNULAR_GAP_SOURCE<TV>::Particle_Is_Fixed(const TV& location) const
{
    TV3 l3(location);
    return fixed_cylinder.Inside(l3,(T)0);
}

template<class TV>
bool PhysBAM::WATER_ANNULAR_GAP_SOURCE<TV>::Reseed_Necessary(const ARRAY_VIEW<TV>& positions) const
{
    const int threshold=0;
    int particles_in_end=0;
    for(int i=1;i<=positions.m;i++){
        TV3 p3(positions(i));
        if(fixed_cylinder.Inside(p3,(T)0))particles_in_end++;
    }
    return particles_in_end<=threshold;
}

template<class TV>
RANGE<TV> PhysBAM::WATER_ANNULAR_GAP_SOURCE<TV>::Bounding_Box() const
{
    RANGE<TV3> range3=outer_cylinder.Bounding_Box();
    RANGE<TV> range;
    range.min_corner=TV_Compatible<T,TV::m>(range3.min_corner);
    range.max_corner=TV_Compatible<T,TV::m>(range3.max_corner);
    return range;
}


template class WATER_ANNULAR_GAP_SOURCE<VECTOR<float,2> >;
template class WATER_ANNULAR_GAP_SOURCE<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
template class WATER_ANNULAR_GAP_SOURCE<VECTOR<double,2> >;
template class WATER_ANNULAR_GAP_SOURCE<VECTOR<double,3> >;
#endif
