//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <PhysBAM_Tools/Math_Tools/pow.h>
#include "../Misc/WATER_RING_SOURCE.h"

using namespace PhysBAM;

template<class TV_input>
PhysBAM::WATER_RING_SOURCE<TV_input>::WATER_RING_SOURCE(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles_input,CODIMENSIONAL_MESH& mesh_input,
    const TV center_input,const T radius_input,const T length_scale_input,const int source_type_input,const T Q_input,const T sheet_speed_input,const T incoming_jets_angle_input,const int polygon_n_input)
:BASE(TV(),TV(),length_scale_input,(T).5*length_scale_input,(T)1e-3,(T)0),center(center_input),radius(radius_input),
source_type((SOURCE_TYPE)source_type_input),Q(Q_input),sheet_speed(sheet_speed_input),incoming_jets_angle(incoming_jets_angle_input),polygon_n(polygon_n_input)
{
    Initialize_Parameters();
    Seed_Particles(particles_input,mesh_input);
}

template<class TV_input>
void PhysBAM::WATER_RING_SOURCE<TV_input>::Initialize_Parameters()
{
    thickness=Q/((T)two_pi*radius*sheet_speed);
    
    particle_num_per_ring=(int)ceil((T)two_pi*radius/length_scale);
    theta_per_particle=(T)two_pi/(T)particle_num_per_ring;

    miller_coef_a=Q/(T)two_pi*((T)1-pow<2>(cos(incoming_jets_angle*(T).5)))/((T)1+pow<2>(cos(incoming_jets_angle*(T).5)));
    miller_coef_b=(T)2*cos(incoming_jets_angle*(T).5)/((T)1+pow<2>(cos(incoming_jets_angle*(T).5)));

    //T polygon_coef=(T)0;
    //for(int i=0;i<particle_num_per_ring;i++){
    //    polygon_coef+=Flux_Per_Radian_Polygon(theta_per_particle*(T)i);
    //}
    //polygon_coef*=theta_per_particle;
    //polygon_coef=Q/polygon_coef;
    //LOG::cout<<"poly_coef: "<<polygon_coef<<std::endl;
}

template<class TV_input>
typename TV_input::SCALAR PhysBAM::WATER_RING_SOURCE<TV_input>::Flux_Per_Radian_Fishbone(const T theta)
{
    return miller_coef_a/((T)1-miller_coef_b*cos(theta));
}

template<class TV_input>
typename TV_input::SCALAR PhysBAM::WATER_RING_SOURCE<TV_input>::Flux_Per_Radian_Polygon(const T theta)
{
    return Flux_Per_Radian_Fishbone(theta*(T)polygon_n);
}

template<class TV_input>
typename TV_input::SCALAR PhysBAM::WATER_RING_SOURCE<TV_input>::Theta(const TV& pos)
{
    TV dir3=(pos-center).Normalized();
    //T r=(pos-center).Magnitude();
    TV dir;dir.x=dir3.x;dir.y=dir3.y;
    return TV2::Oriented_Angle_Between(TV2((T)0,(T)-1),TV2(dir.x,dir.y));
}

template<class TV_input>
typename TV_input::SCALAR PhysBAM::WATER_RING_SOURCE<TV_input>::Thickness(const T theta)
{
    switch(source_type){
    case FISHBONE:
        //return Flux_Per_Radian_Fishbone(theta)/sheet_speed;
        return (T)two_pi*Flux_Per_Radian_Fishbone(theta)/Q*thickness;
        //return thickness;
    case POLYGON:   ////source thickness
        return thickness;
    case CIRCLE:
        return thickness;
    default:
        return (T)0;
    }
}

template<class TV_input>
TV_input PhysBAM::WATER_RING_SOURCE<TV_input>::Source_Velocity(const TV& pos/*=TV()*/)
{
    TV dir3=(pos-center).Normalized();
    //T r=(pos-center).Magnitude();
    TV dir;dir.x=dir3.x;dir.y=dir3.y;
    T theta=TV2::Oriented_Angle_Between(TV2((T)0,(T)-1),TV2(dir.x,dir.y));
    switch(source_type){
    case FISHBONE:  ////constant sheet speed for fishbone
        {
            //T speed=(T)two_pi*Flux_Per_Radian_Fishbone(theta)/Q*sheet_speed;v=speed*dir;    ////assuming thickness constant
            T speed=sheet_speed;return speed*dir;
        }
    case POLYGON:
        {T speed=(T)two_pi*Flux_Per_Radian_Polygon(theta)/Q*sheet_speed;return speed*dir;}
    case CIRCLE:
        {return sheet_speed*dir;}
    default:return TV();
    }
}

template<class TV_input>
int PhysBAM::WATER_RING_SOURCE<TV_input>::Seed_Particles(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,CODIMENSIONAL_MESH& mesh)
{
    int n=particle_num_per_ring;
    T theta=(T)two_pi/(T)n;
    int ring_num=3;
    int eulerian_ring_num=2;
    int c=0;
    for(int j=0;j<ring_num;j++){
        T r=radius+(T)j*length_scale;
        for(int i=0;i<n;i++){
            T angle=(T)(j%2)*(T).5*theta+(T)i*theta;
            TV dir=(TV::Axis_Vector(1)*cos(angle)+TV::Axis_Vector(2)*sin(angle)).Normalized();
            TV pos=center+r*dir;
            int p=particles.Add_Particle();
            particles.X(p)=pos;particles.V(p)=Source_Velocity(pos);
            if(j<eulerian_ring_num)eulerian_particles.Insert(particles.guid(p),particles.X(p));
            c++;
        }
    }

    ////initialize mesh
    for(int j=0;j<ring_num-1;j++){
        int offset1=j*n;
        int offset2=(j+1)*n;
        for(int i=1;i<n;i++){
            if(j%2==0){
                mesh.triangle_mesh.elements.Append(TV_INT3(offset1+i,offset1+i+1,offset2+i));
                mesh.triangle_mesh.elements.Append(TV_INT3(offset2+i,offset1+i+1,offset2+i+1));
            }
            else{
                mesh.triangle_mesh.elements.Append(TV_INT3(offset1+i,offset2+i+1,offset2+i));
                mesh.triangle_mesh.elements.Append(TV_INT3(offset1+i,offset1+i+1,offset2+i+1));
            }
        }
        if(j%2==0){
            mesh.triangle_mesh.elements.Append(TV_INT3(offset1+n,offset1+1,offset2+n));
            mesh.triangle_mesh.elements.Append(TV_INT3(offset2+n,offset1+1,offset2+1));
        }
        else{
            mesh.triangle_mesh.elements.Append(TV_INT3(offset1+n,offset2+1,offset2+n));
            mesh.triangle_mesh.elements.Append(TV_INT3(offset1+n,offset1+1,offset2+1));
        }
    }
    mesh.Update_States<TV>(particles.X.m);

    for(int i=1;i<=particles.X.m;i++){
        if(eulerian_particles.Contains(particles.guid(i))){
            T thickness=Thickness(Theta(particles.X(i)));
            eulerian_particle_thicknesses.Set(particles.guid(i),thickness);
            //std::cout<<"tk "<<i<<": "<<thickness<<std::endl;
        }
        else{
            T thickness=Thickness(Theta(particles.X(i)));
            initial_particle_thicknesses.Set(particles.guid(i),thickness);
        }
    }    

    return c;
}

//template class WATER_RING_SOURCE<VECTOR<float,2> >;
template class WATER_RING_SOURCE<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
//template class WATER_RING_SOURCE<VECTOR<double,2> >;
template class WATER_RING_SOURCE<VECTOR<double,3> >;
#endif
