//#####################################################################
// Copyright 2013, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __WATER_TANK__
#define __WATER_TANK__

#include <PhysBAM_Geometry/Topology/TRIANGLE_MESH.h>
#include <PhysBAM_Geometry/Topology/TETRAHEDRON_MESH.h>
#include <PhysBAM_Geometry/Basic_Geometry/BOX.h>
#include <PhysBAM_Tools/Data_Structures/HASHTABLE_ITERATOR.h>
#include "../Meshing/CODIMENSIONAL_MESH_OBJECT_3D.h"

namespace PhysBAM
{
template<class TV>
class WATER_TANK
{
    typedef typename TV::SCALAR T;typedef typename TV::template REBIND<int>::TYPE TV_INT; 
    typedef VECTOR<int,2> TV_INT2;typedef VECTOR<int,3> TV_INT3;typedef VECTOR<int,4> TV_INT4;
    typedef typename IF<TV::m==2,VECTOR<int,1>,VECTOR<int,2> >::TYPE TV_INT_DIM_MINUS_ONE;
    typedef typename IF<TV::m==2,SEGMENT_MESH,TRIANGLE_MESH>::TYPE SURFACE_MESH;

public:
    BOX<TV> tank_box;
    T tank_thickness;
    bool is_open_tank;
    T epsilon;
    CODIMENSIONAL_MESH_OBJECT<TV>* codimensional_mesh_object;
    HASHTABLE<int,ARRAY<int> > boundary_neighbor_nodes_hashtable;

    WATER_TANK(const BOX<TV>& tank_box_input,T tank_thickness_input,const bool is_open_tank_input,CODIMENSIONAL_MESH_OBJECT<TV>* codimensional_mesh_object_input)
        :tank_box(tank_box_input),tank_thickness(tank_thickness_input),is_open_tank(is_open_tank_input),epsilon((T)1e-20),codimensional_mesh_object(codimensional_mesh_object_input)
    {}

    TV_INT Tank_Corner(const TV& pos)
    {
        VECTOR<int,TV::m> corner;corner.Fill(0);
        for(int d=1;d<=TV::dimension;d++){
            if(!is_open_tank){
                if(pos(d)<=tank_box.min_corner(d)+epsilon)corner(d)=-1;
                else if(pos(d)>=tank_box.max_corner(d)-epsilon)corner(d)=1;
            }
            else{
                if(d==2){if(pos(d)<=tank_box.min_corner(d)+epsilon)corner(d)=-1;}
                else{
                    if((pos.y>=tank_box.min_corner.y-epsilon&&pos.y<=tank_box.max_corner.y+epsilon)){
                        if((pos(d)<=tank_box.min_corner(d)+epsilon&&pos(d)>=tank_box.min_corner(d)-tank_thickness))corner(d)=-1;
                        else if((pos(d)>=tank_box.max_corner(d)-epsilon&&pos(d)<=tank_box.max_corner(d)+tank_thickness))corner(d)=1;
                    }
                }
            }
        }
        return corner;
    }

    bool On_The_Same_Wall(const ARRAY<TV_INT>& corners)
    {
        TV_INT wall=corners(1);
        for(int i=2;i<=corners.m;i++){
            TV_INT corner=corners(i);
            for(int d=1;d<=TV::m;d++){
                if(corner(d)==0||(wall(d)!=0&&wall(d)!=corner(d)))wall(d)=0;
            }
        }
        for(int d=1;d<=TV::m;d++){if(wall(d)!=0)return true;}return false;
    }

    bool Is_Tank_Contact_Element(const TV_INT& v)
    {
        ARRAY<TV_INT> corners;
        for(int d=1;d<=TV::m;d++){
            TV_INT corner=Tank_Corner(codimensional_mesh_object->particles.X(v(d)));
            corners.Append(corner);
        }
        return On_The_Same_Wall(corners);
    }

    void Update_Boundary_Neighbor_Nodes_Hashtable()
    {
        TRIANGLE_MESH& boundary_mesh=codimensional_mesh_object->mesh.tet_boundary_mesh;
        boundary_neighbor_nodes_hashtable.Remove_All();
        for(int i=1;i<=boundary_mesh.elements.m;i++){
            const TV_INT3& v=boundary_mesh.elements(i);
            ARRAY<TV_INT2> edges;Simplex_Edges(v,edges);
            for(int j=1;j<=3;j++){
                const TV_INT2& e=edges(j);
                if(codimensional_mesh_object->particles.tid(e(1))==codimensional_mesh_object->mesh.tid){
                    ARRAY<int>& v_v=boundary_neighbor_nodes_hashtable.Get_Or_Insert(e(1),ARRAY<int>());
                    v_v.Append_Unique(e(2));
                }
                if(codimensional_mesh_object->particles.tid(e(2))==codimensional_mesh_object->mesh.tid){
                    ARRAY<int>& v_v=boundary_neighbor_nodes_hashtable.Get_Or_Insert(e(2),ARRAY<int>());
                    v_v.Append_Unique(e(1));
                }
            }
        }
    }

    bool Is_Tank_Contact_Sharp_Particle(const int p)
    {
        TV_INT corner=Tank_Corner(codimensional_mesh_object->particles.X(p));
        if(corner==TV_INT()||corner.y!=0)return false;
        //TRIANGLE_MESH& boundary_mesh=codimensional_mesh_object->mesh.tet_boundary_mesh;
        if(boundary_neighbor_nodes_hashtable.Contains(p)){
            ARRAY<int>& nb_nodes=boundary_neighbor_nodes_hashtable.Get(p);
            for(int i=1;i<=nb_nodes.m;i++){
                int nb_p=nb_nodes(i);
                if(Tank_Corner(codimensional_mesh_object->particles.X(nb_p))==TV_INT())return true;    ////if has non-corner neighbor
            }
        }

        //SURFACE_MESH& boundary_mesh=*(codimensional_mesh_object->Get_Mesh_Object_3D()->mesh.boundary_mesh);
        //for(int i=1;i<=(*boundary_mesh.neighbor_nodes)(p).m;i++){
        //    int nb_p=(*boundary_mesh.neighbor_nodes)(p)(i);
        //    if(Tank_Corner(codimensional_mesh_object->particles.X(nb_p))==TV_INT())return true;    ////if has non-corner neighbor
        //}

        return false;
    }

    bool Is_Tank_Contact_Sharp_Edge(const TV_INT_DIM_MINUS_ONE& e,TV_INT& c)
    {
        for(int d=1;d<=TV::m-1;d++)if(!Is_Tank_Contact_Sharp_Particle(e(d)))return false;
        TV_INT corner_i=Tank_Corner(codimensional_mesh_object->particles.X(e(1)));
        if(TV::m==3){////handle the case that two ends are on different walls
            TV_INT corner_j=Tank_Corner(codimensional_mesh_object->particles.X(e(2)));
            TV_INT difference=abs(corner_i)-abs(corner_j);
            if(difference!=TV_INT()){
                bool positive=false;bool negative=false;
                for(int d=1;d<=TV::m;d++)if(difference(d)>0)positive=true;else if(difference(d)<0)negative=true;
                if(positive&&negative){return false;}    
            }
            for(int d=1;d<=TV::m;d++)if(corner_i(d)!=corner_j(d))corner_i(d)=0;
            PHYSBAM_ASSERT(corner_i!=TV_INT());
        }
        c=corner_i;
        return true;
    }

    void Find_Tank_Contact_Sharp_Edges(ARRAY<TV_INT_DIM_MINUS_ONE>& sharp_edges,ARRAY<TV_INT>& corners)
    {
        ////for 3d only
        Update_Boundary_Neighbor_Nodes_Hashtable();
        HASHTABLE<TV_INT2,TV_INT3> sharp_edge_hashtable;
        TRIANGLE_MESH& boundary_mesh=codimensional_mesh_object->mesh.tet_boundary_mesh;
        for(int i=1;i<=boundary_mesh.elements.m;i++){
            const TV_INT3& v=boundary_mesh.elements(i);
            ARRAY<TV_INT2> edges;Simplex_Edges(v,edges);
            for(int j=1;j<=edges.m;j++){
                TV_INT3 corner;if(Is_Tank_Contact_Sharp_Edge(edges(j),corner))sharp_edge_hashtable.Set(edges(j).Sorted(),corner);
            }
        }
        for(HASHTABLE_ITERATOR<TV_INT2,TV_INT3> iter(sharp_edge_hashtable);iter.Valid();iter.Next()){
            VECTOR<int,TV::m-1> edge=iter.Key();
            sharp_edges.Append(iter.Key());corners.Append(iter.Data());
        }

        //HASHTABLE<TV_INT_DIM_MINUS_ONE,TV_INT> sharp_edge_hashtable;
        //SURFACE_MESH& boundary_mesh=*(codimensional_mesh_object->Get_Mesh_Object_3D()->mesh.boundary_mesh);
        //for(int i=1;i<=boundary_mesh.elements.m;i++){
        //    const TV_INT& v=boundary_mesh.elements(i);
        //    ARRAY<TV_INT_DIM_MINUS_ONE> edges;Simplex_Edges(v,edges);
        //    for(int j=1;j<=edges.m;j++){
        //        TV_INT corner;if(Is_Tank_Contact_Sharp_Edge(edges(j),corner))sharp_edge_hashtable.Set(edges(j).Sorted(),corner);
        //    }
        //}
        //for(HASHTABLE_ITERATOR<TV_INT_DIM_MINUS_ONE,TV_INT> iter(sharp_edge_hashtable);iter.Valid();iter.Next()){
        //    VECTOR<int,TV::m-1> edge=iter.Key();
        //    sharp_edges.Append(iter.Key());corners.Append(iter.Data());
        //}
    }
};
}
#endif
