//#####################################################################
// Copyright 2013, Matthew Cong, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <PhysBAM_Tools/Krylov_Solvers/CONJUGATE_GRADIENT.h>
#include <PhysBAM_Tools/Krylov_Solvers/CONJUGATE_RESIDUAL.h>
#include <PhysBAM_Tools/Krylov_Solvers/SYMMQMR.h>
#include <PhysBAM_Tools/Matrices/MATRIX_MXN.h>
#include <PhysBAM_Tools/Read_Write/Octave/OCTAVE_OUTPUT.h>
#include "../Solver/CODIMENSIONAL_INCOMPRESSIBLE.h"
#include "../Solver/CODIMENSIONAL_KRYLOV_SYSTEM.h"
using namespace PhysBAM;
using namespace Codimensional;

template<class TV> PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE<TV>::
CODIMENSIONAL_INCOMPRESSIBLE(CODIMENSIONAL_MESH_OBJECT<TV>& solver_codimensional_mesh_object_input,MPI_CODIMENSIONAL<TV>* mpi_codim_input/*=0*/)
    :codimensional_mesh_object(solver_codimensional_mesh_object_input),particles(codimensional_mesh_object.particles),mesh(codimensional_mesh_object.mesh),
    segment_mesh(mesh.segment_mesh),triangle_mesh(mesh.triangle_mesh),tetrahedron_mesh(mesh.tetrahedron_mesh),
    density((T)1000),length_scale((T)1e-3),thickness_scale((T)1e-5),use_target_div(true),use_particle_thickness(true),use_all_pressure_sample_on_nodes(false),max_iter_number(2000),mpi_codim(mpi_codim_input)
{
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE<TV>::Initialize_Mpi(MPI_CODIMENSIONAL<TV>* mpi_codim_input)
{
    mpi_codim=mpi_codim_input;
    for(int i=1;i<=subsystems.m;i++)subsystems(i).mpi_codim=mpi_codim;
}

template<class TV> 
void CODIMENSIONAL_INCOMPRESSIBLE<TV>::Make_Divergence_Free(const T dt,const T time)
{
    Find_Subsystems();
    Find_Null_Space_Basis();
    Apply_Boundary_Conditions();

    Update_Subsystem_Matrices();
    Solve_Subsystems();
    Apply_Pressures();
    Apply_Boundary_Conditions();

    Check_Divergence();
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE<TV>::Reset_Subsystems()
{
    ////resize subsystems
    subsystems.Resize(number_of_regions);
    for(int i=1;i<=number_of_regions;i++){
        subsystems(i).mpi_codim=mpi_codim;
        subsystems(i).Clear();
        subsystems(i).length_scale=length_scale;
        subsystems(i).thickness_scale=thickness_scale;
        subsystems(i).use_target_div=use_target_div;
        subsystems(i).use_particle_thickness=use_particle_thickness;
        subsystems(i).use_all_pressure_sample_on_nodes=use_all_pressure_sample_on_nodes;

        subsystems(i).codimensional_mesh_object=&codimensional_mesh_object;
        //subsystems(i).element_index_in_subsystem=&element_index_in_subsystem;
        subsystems(i).element_index_in_pressure_samples=&element_index_in_pressure_samples; 
        subsystems(i).particle_index_in_velocity_samples=&particle_index_in_velocity_samples;
        subsystems(i).psi_D_on_mesh_hashtable=&psi_D_on_mesh_hashtable;

        subsystems(i).max_iter_number=max_iter_number;
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE<TV>::Find_Subsystems()
{
    number_of_regions=mesh.number_of_regions;
    Reset_Subsystems();

    CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Initialize(subsystems,&psi_N_on_mesh,&psi_D_on_mesh,&tar_div_on_mesh);
    if(mpi_codim){
        for(int i=1;i<=number_of_regions;i++){
            mpi_codim->Update_Boundary_And_Ghost_Pressure_Indices(subsystems(i).pressure_samples,mesh,subsystems(i).use_all_pressure_sample_on_nodes,subsystems(i).boundary_pressure_indices,subsystems(i).ghost_pressure_indices,subsystems(i).domain_pressure_indices);
        }
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE<TV>::Find_Null_Space_Basis_2D()
{
    ARRAY<bool> node_visited(particles.array_collection->Size(),false);node_visited.Fill(false);
    ARRAY<bool> triangle_visited(triangle_mesh.elements.m,false);triangle_visited.Fill(false);
    QUEUE<int> triangle_queue(triangle_mesh.elements.m+segment_mesh.elements.m);
    ARRAY<ARRAY<VECTOR_ND<int> > >null_space_vectors_array(number_of_regions);

    for(int t=1;t<=triangle_mesh.elements.m;t++){
        if(!triangle_visited(t)){
            int region_idx=mesh.Triangle_Region_Index(t);
            ARRAY<VECTOR_ND<int> >& null_space_vectors=null_space_vectors_array(region_idx);
            triangle_queue.Remove_All();
            triangle_visited(t)=true;
            triangle_queue.Enqueue(t);
            for(int i=1;i<=triangle_mesh.elements(t).m;i++){
                node_visited(triangle_mesh.elements(t)(i))=true;}
            null_space_vectors.Append(VECTOR_ND<int>(subsystems(region_idx).pressure_samples.m,true));
            null_space_vectors(null_space_vectors.Size())(element_index_in_pressure_samples(SIMPLEX_PARTICLE)(triangle_mesh.elements(t)(1)))=1;
            null_space_vectors(null_space_vectors.Size())(element_index_in_pressure_samples(SIMPLEX_PARTICLE)(triangle_mesh.elements(t)(2)))=0;
            null_space_vectors(null_space_vectors.Size())(element_index_in_pressure_samples(SIMPLEX_PARTICLE)(triangle_mesh.elements(t)(3)))=-1;
            null_space_vectors.Append(VECTOR_ND<int>(subsystems(region_idx).pressure_samples.m,true));
            null_space_vectors(null_space_vectors.Size())(element_index_in_pressure_samples(SIMPLEX_PARTICLE)(triangle_mesh.elements(t)(1)))=0;
            null_space_vectors(null_space_vectors.Size())(element_index_in_pressure_samples(SIMPLEX_PARTICLE)(triangle_mesh.elements(t)(2)))=1;
            null_space_vectors(null_space_vectors.Size())(element_index_in_pressure_samples(SIMPLEX_PARTICLE)(triangle_mesh.elements(t)(3)))=-1;
            while(!triangle_queue.Empty()){
                int triangle=triangle_queue.Dequeue();
                ARRAY<int> unvisited_nodes;
                ARRAY<int> accumulated_null_space_vectors(null_space_vectors.Size(),false);
                accumulated_null_space_vectors.Fill(0);
                for(int i=1;i<=triangle_mesh.elements(triangle).m;i++){
                    int node=triangle_mesh.elements(triangle)(i);
                    if(node_visited(node)){
                        for(int d=1;d<=null_space_vectors.Size();d++)
                            accumulated_null_space_vectors(d)-=null_space_vectors(d)(element_index_in_pressure_samples(SIMPLEX_PARTICLE)(node));}
                    else unvisited_nodes.Append(node);}
                if(unvisited_nodes.m>0){
                    for(int i=1;i<=unvisited_nodes.m-1;i++){
                        null_space_vectors.Append(VECTOR_ND<int>(subsystems(region_idx).pressure_samples.m,true));
                        accumulated_null_space_vectors.Append(-1);
                        null_space_vectors(null_space_vectors.Size())(element_index_in_pressure_samples(SIMPLEX_PARTICLE)(unvisited_nodes(i)))=1;}
                    for(int d=1;d<=null_space_vectors.Size();d++)
                        null_space_vectors(d)(element_index_in_pressure_samples(SIMPLEX_PARTICLE)(unvisited_nodes(unvisited_nodes.m)))+=accumulated_null_space_vectors(d);}
                for(int i=1;i<=triangle_mesh.elements(triangle).m;i++){
                    int node=triangle_mesh.elements(triangle)(i);
                    node_visited(node)=true;
                    for(int j=1;j<=(*triangle_mesh.incident_elements)(node).m;j++){
                        int incident_triangle=(*triangle_mesh.incident_elements)(node)(j);
                        if(!triangle_visited(incident_triangle)){
                            triangle_visited(incident_triangle)=true;
                            triangle_queue.Enqueue(incident_triangle);}}}}}}

    for(int region_idx=1;region_idx<=number_of_regions;region_idx++){
        ARRAY<VECTOR_ND<int> >& null_space_vectors=null_space_vectors_array(region_idx);
        for(int i=1;i<=subsystems(region_idx).elements(SIMPLEX_TRIANGLE).m;i++){
            VECTOR_ND<int> average_pressure_constraint(null_space_vectors.Size(),true);
            int t=subsystems(region_idx).elements(SIMPLEX_TRIANGLE)(i);
            for(int j=1;j<=triangle_mesh.elements(t).Size();j++){
                for(int d=1;d<=null_space_vectors.Size();d++){
                    average_pressure_constraint(d)+=null_space_vectors(d)(element_index_in_pressure_samples(SIMPLEX_PARTICLE)(triangle_mesh.elements(t)(j)));}}
            int dependent_pressure_index=0;
            for(int j=1;j<=null_space_vectors.Size();j++){
                if(average_pressure_constraint(j)!=0){
                    dependent_pressure_index=j;break;}}
            if(!dependent_pressure_index) continue;
            VECTOR_ND<int> dependent_pressure=null_space_vectors(dependent_pressure_index);
            for(int j=dependent_pressure_index;j<=null_space_vectors.Size();j++){
                null_space_vectors(j)*=average_pressure_constraint(dependent_pressure_index);
                null_space_vectors(j)-=average_pressure_constraint(j)*dependent_pressure;}}}

    for(int region_idx=1;region_idx<=number_of_regions;region_idx++){
        ARRAY<VECTOR_ND<int> >& null_space_vectors=null_space_vectors_array(region_idx);
        ARRAY<VECTOR_ND<T> >& null_space_basis=subsystems(region_idx).null_space_basis;;
        for(int i=1;i<=null_space_vectors.Size();i++){
            if(null_space_vectors(i).L1_Norm()!=0){
                VECTOR_ND<T> null_space_basis_vector(null_space_vectors(i).Size(),false);
                for(int j=1;j<=null_space_vectors(i).Size();j++){
                    null_space_basis_vector(j)=(T)null_space_vectors(i)(j);}
                null_space_basis.Append(null_space_basis_vector);}}
        for(int i=1;i<=null_space_basis.Size();i++){
            for(int j=1;j<i;j++){
                null_space_basis(i)-=VECTOR_ND<T>::Dot_Product(null_space_basis(i),null_space_basis(j))*null_space_basis(j);}
            null_space_basis(i).Normalize();}}
}

template<class TV> 
void CODIMENSIONAL_INCOMPRESSIBLE<TV>::Find_Null_Space_Basis()
{
    ////TODO: implement mpi
    if(TV::m==2)Find_Null_Space_Basis_2D();
    else /*3d non-threaded version not implemented*/;
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE<TV>::Update_Subsystem_Matrices()
{
    for(int region=1;region<=number_of_regions;region++){subsystems(region).Update_Matrices();}
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE<TV>::Solve_Subsystems()
{
    //LOG::cout<<"solve "<<number_of_regions<<" subsystems"<<std::endl;
    for(int region_idx=1;region_idx<=number_of_regions;region_idx++){
        subsystems(region_idx).Solve();
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE<TV>::Apply_Pressures()
{
    for(int region_idx=1;region_idx<=number_of_regions;region_idx++){
        subsystems(region_idx).Apply_Pressures();
    }
    if(mpi_codim)mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(particles.V);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE<TV>::Apply_Boundary_Conditions()
{
    for(int i=1;i<=psi_N_on_mesh.m;i++){
        int p=psi_N_on_mesh(i).index;
        TV n=psi_N_on_mesh(i).direction;
        TV v_n=particles.V(p);
        T psi_N_value=psi_N_on_mesh(i).v;
        particles.V(p)+=(psi_N_value-TV::Dot_Product(n,v_n))*n;
        //LOG::cout<<"apply bc: "<<n<<", v_n: "<<v_n<<", p.V: "<<particles.V(p)<<std::endl;
    }

    if(mpi_codim){mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(particles.V);}
}

template<class TV> void CODIMENSIONAL_INCOMPRESSIBLE<TV>::
Check_Divergence()
{
    for(int i=1;i<=subsystems.m;i++)subsystems(i).Check_Divergence();
}

//template class CODIMENSIONAL_INCOMPRESSIBLE<VECTOR<float,2> >;
template class CODIMENSIONAL_INCOMPRESSIBLE<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
//template class CODIMENSIONAL_INCOMPRESSIBLE<VECTOR<double,2> >;
template class CODIMENSIONAL_INCOMPRESSIBLE<VECTOR<double,3> >;
#endif
