//#####################################################################
// Copyright 2013, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <PhysBAM_Tools/Krylov_Solvers/CONJUGATE_GRADIENT.h>
#include <PhysBAM_Tools/Krylov_Solvers/CONJUGATE_RESIDUAL.h>
#include <PhysBAM_Tools/Krylov_Solvers/SYMMQMR.h>
#include <PhysBAM_Tools/Read_Write/Octave/OCTAVE_OUTPUT.h>
#include "../Solver/CODIMENSIONAL_INCOMPRESSIBLE.h"
#include "../Solver/CODIMENSIONAL_VISCOSITY_FORCE.h"
#include "../Solver/CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM.h"
#include "../Solver/CODIMENSIONAL_KRYLOV_SYSTEM.h"
#include "../Mpi/MPI_CODIMENSIONAL.h"

using namespace PhysBAM;
using namespace Codimensional;

template<class TV>
PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM(MPI_CODIMENSIONAL<TV>* mpi_codim_input/*=0*/)
    :codimensional_mesh_object(0),particle_index_in_velocity_samples(0),psi_D_on_mesh_hashtable(0),length_scale((T)1e-3),length_scale_min((T)1e-3),thickness_scale((T)1e-4),use_target_div(true),
    use_particle_thickness(false),use_all_pressure_sample_on_nodes(false),use_constant_extrapolation(false),use_precoditioner(false),max_iter_number(2000),mpi_codim(mpi_codim_input)
{}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Clear()
{
    for(int d=1;d<=4;d++)elements(d).Remove_All();

    pressure_samples.Remove_All();
    velocity_samples.Remove_All();
    psi_N.Remove_All();psi_D.Remove_All();
    G.Reset(0);GT.Reset(0);GT_without_Psi_N.Reset(0);A.Reset();
    Gn.Reset(0);GnT.Reset(0);
    b.Resize(0);h.Resize(0);p.Resize(0);
    M_inv.Resize(0);FM_inv.Reset(0);asym_FM_inv.Reset(0);target_div.Resize(0);

    null_space_basis.Resize(0);

    if(mpi_codim){
        mpi_codim->Clear_Arrays(boundary_pressure_indices);
        mpi_codim->Clear_Arrays(ghost_pressure_indices);
        domain_pressure_indices.Remove_All();
    }
}

template<class TV>
int PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Add_Pressure_Sample(const int index,const int type)
{pressure_samples.Append(CODIMENSIONAL_PRESSURE(index,type));return pressure_samples.m;}

template<class TV>
int PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Find_Pressure_Sample(const int index,const int type)
{for(int i=1;i<=pressure_samples.m;i++){if(pressure_samples(i).index==index&&pressure_samples(i).type==type){return i;}}return 0;}

template<class TV>
int PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Add_Velocity_Sample(const int index)
{velocity_samples.Append(index);return velocity_samples.m;}

template<class TV>
int PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Find_Velocity_Sample(const int index)
{return velocity_samples.Find(index);}

template<class TV>
int PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Add_Neumann_Boundary_Sample(const int p,const TV& direction,const T v)
{
    int i=Find_Velocity_Sample(p);if(i==0){PHYSBAM_FATAL_ERROR("invalid index for Neumann boundary");return 0;}
    //T nodal_volume=(T)0;
    //if(use_particle_thickness){nodal_volume=codimensional_mesh_object->Nodal_Volume(p);}
    //else{nodal_volume=codimensional_mesh_object->Nodal_Tetrahedron_Volume(p)+thickness_scale*codimensional_mesh_object->Nodal_Triangle_Area(p)+
    //    codimensional_mesh_object->Segment_Cross_Section_Area(thickness_scale)*codimensional_mesh_object->Nodal_Segment_Length(p);}
    //T volume_weighted_v=nodal_volume*v;
    psi_N.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,direction,/*volume_weighted_v*/v));return psi_N.m;
}

template<class TV>
int PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Add_Dirichlet_Boundary_Sample(const int p,const int type,const T v)
{
    int i=Find_Velocity_Sample(p);if(i==0){PHYSBAM_FATAL_ERROR("invalid index for Dirichlet boundary");return 0;}
    psi_D.Append(CODIMENSIONAL_DIRICHLET_BOUNDARY<T>(i,type,v));return psi_D.m;
}

template<class TV>
int PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Add_Simplex_Index(const int index,const int type)
{elements(type).Append(index);return elements(type).m;}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Update_Matrices()
{
    Update_G(G);
    Update_GT(G,GT_without_Psi_N);
    Modify_G_With_Boundary_Conditions(G,psi_N);
    Update_M_inv(M_inv);
    Update_GT(G,GT);
    Update_A(G,M_inv,A);
    Update_b(b);
    Modify_b_With_Dirichlet_Boundary_Conditions(b,psi_D);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Update_M_inv(VECTOR_ND<T>& M_inv)
{
    T density=(T)1000.;
    M_inv.Resize(velocity_samples.m*TV::m);
    for(int i=1;i<=velocity_samples.m;i++){
        int p=velocity_samples(i);
        T nodal_volume=use_particle_thickness?codimensional_mesh_object->Nodal_Volume(p):codimensional_mesh_object->Nodal_Volume_With_Thickness(p,thickness_scale);
        if(!(nodal_volume>=(T)0))LOG::cout<<"!!!!!!!!!! zero nodal volume "<<codimensional_mesh_object->particles.guid(p)<<nodal_volume<<std::endl;
        T one_over_mass=(T)1/(nodal_volume*density);
        //T one_over_mass=(T)1/codimensional_mesh_object->particles.mass(p);
        for(int d=1;d<=TV::m;d++)M_inv((TV::m*(i-1))+d)=one_over_mass;
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Update_V_inv(VECTOR_ND<T>& V_inv)
{
    V_inv.Resize(velocity_samples.m*TV::m);
    for(int i=1;i<=velocity_samples.m;i++){
        int p=velocity_samples(i);
        T nodal_volume=use_particle_thickness?codimensional_mesh_object->Nodal_Volume(p):codimensional_mesh_object->Nodal_Volume_With_Thickness(p,thickness_scale);
        T one_over_volume=(T)1/nodal_volume;
        if(!(nodal_volume>=(T)0))LOG::cout<<"!!!!!!!!!! zero nodal volume v "<<codimensional_mesh_object->particles.guid(p)<<" "<<nodal_volume<<std::endl;
        for(int d=1;d<=TV::m;d++)M_inv((TV::m*(i-1))+d)=one_over_volume;
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Update_V_inv_With_Dirichlet_Boundary_Conditions(VECTOR_ND<T>& V_inv)
{
    PHYSBAM_ASSERT(psi_D_on_mesh_hashtable!=0);
    V_inv.Resize(velocity_samples.m*TV::m);
    for(int i=1;i<=velocity_samples.m;i++){
        int p=velocity_samples(i);
        T one_over_volume=(T)1/codimensional_mesh_object->Nodal_Partial_Volume(p,(*psi_D_on_mesh_hashtable));
        for(int d=1;d<=TV::m;d++)M_inv((TV::m*(i-1))+d)=one_over_volume;
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Update_FM_inv(const ARRAY_VIEW<T_MATRIX>& FT,const VECTOR_ND<T>& M_inv,SPARSE_MATRIX_FLAT_MXN<T>& FM_inv,bool use_sym_F/*=true*/)
{
    int n=velocity_samples.m*TV::m;
    ARRAY<int> row_lengths;row_lengths.Resize(n);row_lengths.Fill(0);
    for(int i=1;i<=velocity_samples.m;i++){
        for(int d=1;d<=TV::m;d++){
            int r=TV::m*(i-1)+d;
            row_lengths(r)=TV::m;
        }
    }
    FM_inv.Reset(n);
    FM_inv.Set_Row_Lengths(row_lengths);

    for(int i=1;i<=velocity_samples.m;i++){
        int p=velocity_samples(i);
        T_MATRIX mtx=use_sym_F?FT(p).Symmetric_Part():FT(p).Skew_Symmetric_Part();
        int r0=TV::m*(i-1);
        for(int d=1;d<=TV::m;d++){
            int r=r0+d;
            for(int k=1;k<=TV::m;k++){
                int c=r0+k;
                T value=mtx(d,k)*M_inv(c);
                FM_inv.Set_Element(r,c,value);
            }
        }
    }
}

template<class TV>
int PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Neighbor_Element_Number_With_Dirichlet_Boundary_Conditions(const ARRAY<int>& nb_e,const int type)
{
    switch(type){
    case SIMPLEX_PARTICLE:{int c=0;
        for(int i=1;i<=nb_e.m;i++){int p=nb_e(i);
            if(codimensional_mesh_object->particles.inactive(p))continue;
            if(!Full_Ring_Dirichlet_Boundary(codimensional_mesh_object->mesh,(*psi_D_on_mesh_hashtable),p,SIMPLEX_PARTICLE))c++;}
        return c;}
    case SIMPLEX_SEGMENT:{int c=0;
        for(int i=1;i<=nb_e.m;i++){int p=nb_e(i);
            if(Element_Or(codimensional_mesh_object->particles.inactive,codimensional_mesh_object->mesh.segment_mesh.elements(p)))continue;
            if(!Full_Ring_Dirichlet_Boundary(codimensional_mesh_object->mesh,(*psi_D_on_mesh_hashtable),p,SIMPLEX_SEGMENT))c++;}
        return c;}
    case SIMPLEX_TRIANGLE:{int c=0;
        for(int i=1;i<=nb_e.m;i++){int p=nb_e(i);
            if(Element_Or(codimensional_mesh_object->particles.inactive,codimensional_mesh_object->mesh.triangle_mesh.elements(p)))continue;
            if(!Full_Ring_Dirichlet_Boundary(codimensional_mesh_object->mesh,(*psi_D_on_mesh_hashtable),p,SIMPLEX_TRIANGLE))c++;}
        return c;}
    }
    return 0;
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Update_G(SPARSE_MATRIX_FLAT_MXN<T>& G)
{
    int row=velocity_samples.m*TV::m;int column=pressure_samples.m;

    //int dbg_idx=0;
    //resize matrices
    ARRAY<int> row_lengths;row_lengths.Resize(row);row_lengths.Fill(0);
    if(!use_all_pressure_sample_on_nodes){
        for(int i=1;i<=velocity_samples.m;i++){
            int p=velocity_samples(i);
            for(int d=1;d<=TV::m;d++){
                int row_index=TV::m*(i-1)+d;
                if(Is_On_Volume_Mesh<TV::m>(codimensional_mesh_object->mesh.node_types(p))){
                    //row_lengths(row_index)+=(1+(*codimensional_mesh_object->volume_mesh.neighbor_nodes)(p).m);
                    const ARRAY<int>& nbs=(*codimensional_mesh_object->volume_mesh.neighbor_nodes)(p);
                    row_lengths(row_index)+=(1+Neighbor_Element_Number_With_Dirichlet_Boundary_Conditions(nbs,SIMPLEX_PARTICLE));
                    //if(codimensional_mesh_object->particles.guid(p)==dbg_idx){
                    //    LOG::cout<<"p on tet "<<(*codimensional_mesh_object->volume_mesh.neighbor_nodes)(p).m<<": "<<Neighbor_Element_Number_With_Dirichlet_Boundary_Conditions(nbs,SIMPLEX_PARTICLE)
                    //        <<", row_length: "<<row_lengths(row_index)<<std::endl;
                    //}
                }
                if(TV::m==3&&Is_On_Triangle_Mesh(codimensional_mesh_object->mesh.node_types(p))){
                    //row_lengths(row_index)+=(*codimensional_mesh_object->mesh.triangle_mesh.incident_elements)(p).m;
                    const ARRAY<int>& nbs=(*codimensional_mesh_object->mesh.triangle_mesh.incident_elements)(p);
                    row_lengths(row_index)+=Neighbor_Element_Number_With_Dirichlet_Boundary_Conditions(nbs,SIMPLEX_TRIANGLE);

                    //if(codimensional_mesh_object->particles.guid(p)==dbg_idx){
                    //    LOG::cout<<"p on tri "<<(*codimensional_mesh_object->mesh.triangle_mesh.incident_elements)(p).m<<": "<<Neighbor_Element_Number_With_Dirichlet_Boundary_Conditions(nbs,SIMPLEX_TRIANGLE)
                    //        <<", row_length: "<<row_lengths(row_index)<<std::endl;
                    //    for(int k=1;k<=(*codimensional_mesh_object->mesh.triangle_mesh.incident_elements)(p).m;k++){
                    //        int t=(*codimensional_mesh_object->mesh.triangle_mesh.incident_elements)(p)(k);
                    //        TV_INT3 v=codimensional_mesh_object->mesh.triangle_mesh.elements(t);
                    //        TV_INT3 guid=Element(codimensional_mesh_object->particles.guid,v);
                    //        LOG::cout<<"tri "<<v<<" guid "<<guid<<std::endl;
                    //    }
                    //}
                }
                if(Is_On_Segment_Mesh(codimensional_mesh_object->mesh.node_types(p))){
                    //row_lengths(row_index)+=(*codimensional_mesh_object->mesh.segment_mesh.incident_elements)(p).m;
                    const ARRAY<int>& nbs=(*codimensional_mesh_object->mesh.segment_mesh.incident_elements)(p);
                    row_lengths(row_index)+=Neighbor_Element_Number_With_Dirichlet_Boundary_Conditions(nbs,SIMPLEX_SEGMENT);

                    //if(codimensional_mesh_object->particles.guid(p)==dbg_idx){
                    //    LOG::cout<<"p on seg "<<(*codimensional_mesh_object->mesh.segment_mesh.incident_elements)(p).m<<": "<<Neighbor_Element_Number_With_Dirichlet_Boundary_Conditions(nbs,SIMPLEX_SEGMENT)
                    //        <<", row_length: "<<row_lengths(row_index)<<std::endl;
                    //    for(int k=1;k<=(*codimensional_mesh_object->mesh.segment_mesh.incident_elements)(p).m;k++){
                    //        int s=(*codimensional_mesh_object->mesh.segment_mesh.incident_elements)(p)(k);
                    //        TV_INT2 v=codimensional_mesh_object->mesh.segment_mesh.elements(s);
                    //        TV_INT2 guid=Element(codimensional_mesh_object->particles.guid,v);
                    //        LOG::cout<<"seg "<<v<<" guid "<<guid<<std::endl;
                    //    }
                    //}
                }
            }
        }
    }
    else{   ////all pressure samples on nodes
        for(int i=1;i<=velocity_samples.m;i++){
            int p=velocity_samples(i);
            for(int d=1;d<=TV::m;d++){
                int row_index=TV::m*(i-1)+d;
                //row_lengths(row_index)+=(1+(*codimensional_mesh_object->mesh.neighbor_nodes)(p).m);
                const ARRAY<int>& nbs=(*codimensional_mesh_object->mesh.neighbor_nodes)(p);
                row_lengths(row_index)+=(1+Neighbor_Element_Number_With_Dirichlet_Boundary_Conditions(nbs,SIMPLEX_PARTICLE));
                //LOG::cout<<"row "<<p<<", "<<row_index<<": "<<(1+(*codimensional_mesh_object->mesh.neighbor_nodes)(p).m)<<", "<<(1+Neighbor_Element_Number_With_Dirichlet_Boundary_Conditions(nbs,SIMPLEX_PARTICLE))<<std::endl;
            }
        }
    }

    G.Reset(column);G.Set_Row_Lengths(row_lengths);

    ////traverse all pressure samples and construct the matrix
    ARRAY_VIEW<T>& thickness=codimensional_mesh_object->particles.thickness;
    for(int i_ps=1;i_ps<=pressure_samples.m;i_ps++){
        int index=pressure_samples(i_ps).index;
        int type=pressure_samples(i_ps).type;
        switch(type){
        case SIMPLEX_PARTICLE:{
            int p=index;int p_in_ps=i_ps;
            for(int i=1;i<=(*codimensional_mesh_object->mesh.tetrahedron_mesh.incident_elements)(p).m;i++){
                int nb_t=(*codimensional_mesh_object->mesh.tetrahedron_mesh.incident_elements)(p)(i);
                if(Full_Ring_Dirichlet_Boundary(codimensional_mesh_object->mesh,(*psi_D_on_mesh_hashtable),nb_t,SIMPLEX_TETRAHEDRON))continue;
                const VECTOR<int,TV::m+1>& e=codimensional_mesh_object->volume_mesh.elements(nb_t);
                for(int j=1;j<=TV::m+1;j++){
                    int nb_p=e(j);int nb_p_in_vs=(*particle_index_in_velocity_samples)(nb_p);if(nb_p_in_vs==0)continue;
                    TV n=CODIMENSIONAL_MESH_OBJECT<TV>::Face_Area_Weighted_Volumetric_Vertex_Normal(codimensional_mesh_object->particles.X,codimensional_mesh_object->volume_mesh.elements(nb_t),nb_p);
                    T coef=(T).25;
                    Add_Matrix_Vector_Element(G,-coef*n,nb_p_in_vs,p_in_ps,true);
                }
            }
            if(use_all_pressure_sample_on_nodes){
                ////handle triangles
                if(TV::m==3)for(int i=1;i<=(*codimensional_mesh_object->mesh.triangle_mesh.incident_elements)(p).m;i++){
                    int nb_t=(*codimensional_mesh_object->mesh.triangle_mesh.incident_elements)(p)(i);
                    if(Full_Ring_Dirichlet_Boundary(codimensional_mesh_object->mesh,(*psi_D_on_mesh_hashtable),nb_t,SIMPLEX_TRIANGLE))continue;
                    const VECTOR<int,3>& e=codimensional_mesh_object->mesh.triangle_mesh.elements(nb_t);
                    for(int j=1;j<=3;j++){
                        int nb_p=e(j);int nb_p_in_vs=(*particle_index_in_velocity_samples)(nb_p);if(nb_p_in_vs==0)continue;
                        TV n=CODIMENSIONAL_MESH_OBJECT<TV>::Length_Weighted_Triangle_Vertex_Normal(codimensional_mesh_object->particles.X,codimensional_mesh_object->mesh.triangle_mesh.elements(nb_t),nb_p);
                        T coef=(T)one_third;T node_thickness=use_particle_thickness?thickness(nb_p):thickness_scale;
                        Add_Matrix_Vector_Element(G,-coef*n*node_thickness,nb_p_in_vs,p_in_ps,true);
                    }
                }
                ////handle segments
                for(int i=1;i<=(*codimensional_mesh_object->mesh.segment_mesh.incident_elements)(p).m;i++){
                    int nb_t=(*codimensional_mesh_object->mesh.segment_mesh.incident_elements)(p)(i);
                    if(Full_Ring_Dirichlet_Boundary(codimensional_mesh_object->mesh,(*psi_D_on_mesh_hashtable),nb_t,SIMPLEX_SEGMENT))continue;
                    const VECTOR<int,2>& e=codimensional_mesh_object->mesh.segment_mesh.elements(nb_t);
                    for(int j=1;j<=2;j++){
                        int nb_p=e(j);int nb_p_in_vs=(*particle_index_in_velocity_samples)(nb_p);if(nb_p_in_vs==0){LOG::cout<<"nb_p_in_vs "<<nb_p<<", "<<nb_p_in_vs<<std::endl;continue;}
                        TV n=CODIMENSIONAL_MESH_OBJECT<TV>::Segment_Vertex_Normal(codimensional_mesh_object->particles.X,codimensional_mesh_object->mesh.segment_mesh.elements(nb_t),nb_p);
                        T coef=(T).5;T nb_p_area=codimensional_mesh_object->Segment_Cross_Section_Area(use_particle_thickness?thickness(nb_p):thickness_scale);
                        Add_Matrix_Vector_Element(G,-coef*n*nb_p_area,nb_p_in_vs,p_in_ps,true);
                        //LOG::cout<<"add seg "<<e<<": "<<nb_p<<", "<<nb_p_in_vs<<", "<<p_in_ps<<": "<<coef<<", "<<n<<", "<<-coef*n*nb_p_area<<std::endl;
                    }
                }
            }
        }break;
        case SIMPLEX_SEGMENT:{
            int s=index;int s_in_ps=i_ps;
            int p1=codimensional_mesh_object->mesh.segment_mesh.elements(s)(1);int p2=codimensional_mesh_object->mesh.segment_mesh.elements(s)(2);
            int p1_in_vs=(*particle_index_in_velocity_samples)(p1);int p2_in_vs=(*particle_index_in_velocity_samples)(p2);
            T p1_area=(T)pi*pow<2>((T).5*(use_particle_thickness?thickness(p1):thickness_scale));
            T p2_area=(T)pi*pow<2>((T).5*(use_particle_thickness?thickness(p2):thickness_scale));
            TV n=CODIMENSIONAL_MESH_OBJECT<TV>::Segment_Vertex_Normal(codimensional_mesh_object->particles.X,p1,p2);
            Add_Matrix_Vector_Element(G,-n*p1_area,p1_in_vs,s_in_ps,true);
            Add_Matrix_Vector_Element(G,n*p2_area,p2_in_vs,s_in_ps,true);
        }break;
        case SIMPLEX_TRIANGLE:{
            int t=index;int t_in_ps=i_ps;
            const VECTOR<int,3>& p=codimensional_mesh_object->mesh.triangle_mesh.elements(t);
            VECTOR<int,3> p_in_vs;VECTOR<TV,3> n;VECTOR<T,3> node_thickness;
            for(int d=1;d<=3;d++)p_in_vs(d)=(*particle_index_in_velocity_samples)(p(d));
            for(int d=1;d<=3;d++){
                n(d)=CODIMENSIONAL_MESH_OBJECT<TV>::Length_Weighted_Triangle_Vertex_Normal(codimensional_mesh_object->particles.X,/*codimensional_mesh_object->particles.thickness,*/p((d-1)%3+1),p(d%3+1),p((d+1)%3+1));
                node_thickness(d)=use_particle_thickness?thickness(p(d)):thickness_scale;
            }
            for(int d=1;d<=3;d++){Add_Matrix_Vector_Element(G,-n(d)*node_thickness(d),p_in_vs(d),t_in_ps,true);}
        }break;
        }
    }

    ////handle boundary conditions
    if(use_constant_extrapolation)for(int i_ps=1;i_ps<=pressure_samples.m;i_ps++){
        int index=pressure_samples(i_ps).index;
        int type=pressure_samples(i_ps).type;
        if(psi_D.Contains(CODIMENSIONAL_DIRICHLET_BOUNDARY<T>(index))){/*LOG::cout<<"const ext psi_D: "<<index<<std::endl;*/continue;}
        switch(type){
        case SIMPLEX_PARTICLE:{
            int p=index;int p_in_ps=i_ps;
            int p_in_vs=(*particle_index_in_velocity_samples)(p);
            if(codimensional_mesh_object->mesh.Is_Boundary_Particle_On_Tetrahedron_Mesh(p)){
                TV v=Get_Matrix_Vector_Element(G,p_in_vs,p_in_ps,true);
                Set_Matrix_Vector_Element(G,-v,p_in_vs,p_in_ps,true);

                for(int i=1;i<=(*codimensional_mesh_object->mesh.tetrahedron_mesh.incident_elements)(p).m;i++){
                    int nb_t=(*codimensional_mesh_object->mesh.tetrahedron_mesh.incident_elements)(p)(i);
                    const TV_INT4& e=codimensional_mesh_object->mesh.tetrahedron_mesh.elements(nb_t);
                    TV_INT3 other_nodes=Other_Nodes_In_Element(e,p);
                    for(int j=1;j<=3;j++){
                        int nb_p1=other_nodes(j);TV_INT2 nb_p2p3=Other_Nodes_In_Element(other_nodes,nb_p1);  
                        TV_INT3 incident_face(p,nb_p2p3(1),nb_p2p3(2));
                        if(codimensional_mesh_object->mesh.Is_Boundary_Face_On_Tetrahedron_Mesh(incident_face)){
                            for(int k=1;k<=2;k++){
                                int nb_p2_in_vs=(*particle_index_in_velocity_samples)(nb_p2p3(k));
                                //TV n1=CODIMENSIONAL_MESH_OBJECT<TV>::Length_Weighted_Triangle_Vertex_Normal(codimensional_mesh_object->particles.X,codimensional_mesh_object->mesh.triangle_mesh.elements(nb_t),nb_p1);
                                TV n1=CODIMENSIONAL_MESH_OBJECT<TV>::Face_Area_Weighted_Volumetric_Vertex_Normal(codimensional_mesh_object->particles.X,codimensional_mesh_object->mesh.tetrahedron_mesh.elements(nb_t),nb_p1);
                                T coef=(T).25;
                                Add_Matrix_Vector_Element(G,-coef*n1,nb_p2_in_vs,p_in_ps,true);
                            }
                        }
                    }
                }
            }

            if(use_all_pressure_sample_on_nodes){
                ////on segment boundary
                if(codimensional_mesh_object->mesh.Is_Boundary_Particle_On_Segment_Mesh(p)){
                    TV v=Get_Matrix_Vector_Element(G,p_in_vs,p_in_ps,true);
                    Set_Matrix_Vector_Element(G,-v,p_in_vs,p_in_ps,true);
                }
                ////on triangle boundary
                if(codimensional_mesh_object->mesh.Is_Boundary_Particle_On_Triangle_Mesh(p)){
                    TV v=Get_Matrix_Vector_Element(G,p_in_vs,p_in_ps,true);
                    Set_Matrix_Vector_Element(G,-v,p_in_vs,p_in_ps,true);
                    //LOG::cout<<"p: "<<p_in_ps<<": "<<-v<<std::endl;

                    for(int i=1;i<=(*codimensional_mesh_object->mesh.triangle_mesh.incident_elements)(p).m;i++){
                        int nb_t=(*codimensional_mesh_object->mesh.triangle_mesh.incident_elements)(p)(i);
                        const TV_INT3& e=codimensional_mesh_object->mesh.triangle_mesh.elements(nb_t);
                        TV_INT2 other_nodes=Other_Nodes_In_Element(e,p);
                        for(int j=1;j<=2;j++){
                            int nb_p1=other_nodes(j);int nb_p2=Other_Node_In_Element(other_nodes,nb_p1);
                            if(codimensional_mesh_object->mesh.Is_Boundary_Edge_On_Triangle_Mesh(TV_INT2(p,nb_p1))){
                                int nb_p1_in_vs=(*particle_index_in_velocity_samples)(nb_p1);
                                TV n1=CODIMENSIONAL_MESH_OBJECT<TV>::Length_Weighted_Triangle_Vertex_Normal(codimensional_mesh_object->particles.X,codimensional_mesh_object->mesh.triangle_mesh.elements(nb_t),nb_p2);
                                T coef=(T)one_third;
                                T node_thickness=use_particle_thickness?thickness(nb_p1):thickness_scale; ////TODO: thickness?
                                Add_Matrix_Vector_Element(G,-coef*n1*node_thickness,nb_p1_in_vs,p_in_ps,true);
                            }
                        }
                    }
                }
            }
        }break;
        case SIMPLEX_SEGMENT:break;
        case SIMPLEX_TRIANGLE:break;
        }
    }

    for(int r=1;r<=G.m;r++){for(int j=G.offsets(r);j<G.offsets(r+1);j++){if(G.A(j).j==0){
        LOG::cout<<"invalid element in G, row: "<<r<<std::endl;
        int i_vs=(r-1)/3+1;
        int p=velocity_samples(i_vs);
        LOG::cout<<"p: "<<p<<" guid "<<codimensional_mesh_object->particles.guid(p)<<", rowlength: "<<row_lengths(r)<<std::endl;
        LOG::cout<<"G: ";for(int k=G.offsets(r);k<G.offsets(r+1);k++){LOG::cout<<G.A(k).j<<", ";}LOG::cout<<std::endl;
    }}}

    //Check_Matrix_Invalid_Index(G);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Modify_G_With_Boundary_Conditions(SPARSE_MATRIX_FLAT_MXN<T>& G,ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> >& psi_N)
{
    Gn=G;for(int i=1;i<=Gn.A.m;i++)Gn.A(i).a=(T)0;

    ////modify G with Neumann boundary conditions
    /*for(int iter=1;iter<=5;iter++)*/for(int i=1;i<=psi_N.m;i++){	////TODO: fix this
        TV n=psi_N(i).direction;
        int index=psi_N(i).index;
        //T_MATRIX I_minus_nnT=T_MATRIX::Identity_Matrix()-T_MATRIX::Outer_Product(n,n);
        T_MATRIX nnT=T_MATRIX::Outer_Product(n,n);
        int row_idx=(psi_N(i).index-1)*TV::dimension+1;
        for(int k=G.offsets(row_idx);k<G.offsets(row_idx+1);k++){
            int j=G.A(k).j;
            TV v=Get_Matrix_Vector_Element(G,index,j,true);
            //TV v_n=T_MATRIX::Outer_Product(n,n)*v;
            //v=I_minus_nnT*v;
            TV v_proj=nnT*v;
            Set_Matrix_Vector_Element(G,v-v_proj,index,j,true);
            Set_Matrix_Vector_Element(Gn,v_proj,index,j,true);
        }
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Update_GT(const SPARSE_MATRIX_FLAT_MXN<T>& G,SPARSE_MATRIX_FLAT_MXN<T>& GT)
{
    GT.Reset(G.m);
    G.Transpose(GT);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Update_A(SPARSE_MATRIX_FLAT_MXN<T>& G,const VECTOR_ND<T>& M_inv,SPARSE_MATRIX_FLAT_NXN<T>& A)
{
    A.Reset();
    A=G.Transpose_Times_Diagonal_Times(M_inv);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Update_A(const SPARSE_MATRIX_FLAT_MXN<T>& G,const SPARSE_MATRIX_FLAT_MXN<T>& FM_inv,const SPARSE_MATRIX_FLAT_MXN<T>& GT,SPARSE_MATRIX_FLAT_NXN<T>& A)
{
    A.Reset();
    Sparse_Flat_Matrix_MXN_To_NXN(GT*FM_inv*G,A);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Shrink_A_and_b_With_Dirichlet_Boundary_Conditions(SPARSE_MATRIX_FLAT_NXN<T>& B,VECTOR_ND<T>& d,ARRAY<int>& map_A_to_B,ARRAY<int>& map_B_to_A)
{
    Sort(psi_D);
    int B_n=A.n-psi_D.m;

    ARRAY<bool> flags(A.n);flags.Fill(true);
    for(int i=1;i<=psi_D.m;i++)flags(psi_D(i).index)=false;

    map_A_to_B.Resize(A.n);map_A_to_B.Fill(0);int c=1;
    for(int i=1;i<=A.n;i++){if(flags(i)){map_A_to_B(i)=c++;}}

    map_B_to_A.Resize(B_n);c=1;
    for(int i=1;i<=A.n;i++){if(flags(i))map_B_to_A(c++)=i;}

    ////modify d
    d.Resize(B_n);
    for(int i=1;i<=A.n;i++){
        int r_in_B=map_A_to_B(i);
        if(r_in_B==0)continue;
        d(r_in_B)=b(i);
    }
    for(int i_in_s=1;i_in_s<=psi_D.m;i_in_s++){
        int j=psi_D(i_in_s).index;
        T v=psi_D(i_in_s).v;
        for(int i=1;i<=A.n;i++){
            int r_in_B=map_A_to_B(i);
            if(r_in_B==0)continue;
            T A_ij=A.Element_Present(i,j)?A(i,j):(T)0;
            d(r_in_B)-=A_ij*v;
        }
    }

    ////shrink A
    ARRAY<int> row_lengths;row_lengths.Resize(B_n);row_lengths.Fill(0);
    for(int r=1;r<=A.n;r++){
        int r_in_B=map_A_to_B(r);
        if(r_in_B==0)continue;
        for(int k=A.offsets(r);k<A.offsets(r+1);k++){
            int c=A.A(k).j;
            int c_in_B=map_A_to_B(c);
            if(c_in_B!=0)row_lengths(r_in_B)++;
        }
    }
    B.Reset();B.Set_Row_Lengths(row_lengths);

    for(int r=1;r<=A.n;r++){
        int r_in_B=map_A_to_B(r);
        if(r_in_B==0)continue;
        for(int k=A.offsets(r);k<A.offsets(r+1);k++){
            int c=A.A(k).j;
            int c_in_B=map_A_to_B(c);
            T A_rc=A.Element_Present(r,c)?A(r,c):(T)0;
            if(c_in_B!=0)B.Set_Element(r_in_B,c_in_B,A_rc);
        }
    }

    ////mpi remapping
    if(mpi_codim){
        PHYSBAM_ASSERT(use_all_pressure_sample_on_nodes);
        ARRAY<CODIMENSIONAL_PRESSURE> remapped_pressures;
        VECTOR<ARRAY<int>,4> remapped_element_index_in_pressures;
        remapped_element_index_in_pressures(SIMPLEX_PARTICLE).Resize(codimensional_mesh_object->particles.X.m);
        for(int i=1;i<=A.n;i++){
            int i_in_B=map_A_to_B(i);
            if(i_in_B!=0){
                int p=pressure_samples(i).index;
                int pm=remapped_pressures.Append(CODIMENSIONAL_PRESSURE(p,SIMPLEX_PARTICLE));
                remapped_element_index_in_pressures(SIMPLEX_PARTICLE)(p)=pm;
            }
        }

        ////dbg
        //LOG::cout<<"N"<<mpi_codim->node_id<<" bef bdr ";
        //for(int i=1;i<=boundary_pressure_indices.m;i++){
        //    for(int j=1;j<=boundary_pressure_indices(i).m;j++){
        //        LOG::cout<<codimensional_mesh_object->particles.guid(pressure_samples(boundary_pressure_indices(i)(j)).index)<<" ";
        //    }
        //}
        //LOG::cout<<std::endl;

        //LOG::cout<<"N"<<mpi_codim->node_id<<" bef gst ";
        //for(int i=1;i<=ghost_pressure_indices.m;i++){
        //    for(int j=1;j<=ghost_pressure_indices(i).m;j++){
        //        LOG::cout<<codimensional_mesh_object->particles.guid(pressure_samples(ghost_pressure_indices(i)(j)).index)<<" ";
        //    }
        //}
        //LOG::cout<<std::endl;

        mpi_codim->Update_Boundary_And_Ghost_Pressure_Indices(remapped_pressures,codimensional_mesh_object->mesh,use_all_pressure_sample_on_nodes,
            boundary_pressure_indices,ghost_pressure_indices,domain_pressure_indices);

        ////dbg
        //LOG::cout<<"N"<<mpi_codim->node_id<<" bdr ";
        //for(int i=1;i<=boundary_pressure_indices.m;i++){
        //    LOG::cout<<boundary_pressure_indices(i).m<<" ";
        //    //for(int j=1;j<=boundary_pressure_indices(i).m;j++){
        //    //    LOG::cout<<codimensional_mesh_object->particles.guid(remapped_pressures(boundary_pressure_indices(i)(j)).index)<<" ";
        //    //}
        //}
        //LOG::cout<<std::endl;

        //LOG::cout<<"N"<<mpi_codim->node_id<<" gst ";
        //for(int i=1;i<=ghost_pressure_indices.m;i++){
        //    LOG::cout<<ghost_pressure_indices(i).m<<" ";
        //    //for(int j=1;j<=ghost_pressure_indices(i).m;j++){
        //    //    LOG::cout<<codimensional_mesh_object->particles.guid(remapped_pressures(ghost_pressure_indices(i)(j)).index)<<" ";
        //    //}
        //}
        //LOG::cout<<std::endl;

        //for(int i=1;i<=mpi_codim->boundary_regions.m;i++){
        //    ARRAY<int> indices_in_B;
        //    for(int j=1;j<=boundary_pressure_indices(i).m;j++){
        //        int p_in_A=boundary_pressure_indices(i)(j);
        //        int p_in_B=map_A_to_B(p_in_A);
        //        if(p_in_B!=0)indices_in_B.Append(p_in_B);
        //    }
        //    boundary_pressure_indices(i)=indices_in_B;
        //}
        //for(int i=1;i<=mpi_codim->boundary_regions.m;i++){
        //    ARRAY<int> indices_in_B;
        //    for(int j=1;j<=ghost_pressure_indices(i).m;j++){
        //        int p_in_A=ghost_pressure_indices(i)(j);
        //        int p_in_B=map_A_to_B(p_in_A);
        //        if(p_in_B!=0)indices_in_B.Append(p_in_B);
        //    }
        //    ghost_pressure_indices(i)=indices_in_B;
        //}

        //ARRAY<int> indices_in_B;
        //for(int i=1;i<=domain_pressure_indices.m;i++){
        //    int p_in_A=domain_pressure_indices(i);
        //    int p_in_B=map_A_to_B(p_in_A);
        //    if(p_in_B!=0)indices_in_B.Append(p_in_B);
        //}
        //domain_pressure_indices=indices_in_B;
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Set_Zero_Rows_in_A_and_b_With_Dirchlet_Boundary_Conditions()
{
    Sort(psi_D);
    for(int i_in_s=1;i_in_s<=psi_D.m;i_in_s++){
        int j=psi_D(i_in_s).index;
        T v=psi_D(i_in_s).v;
        for(int i=1;i<=A.n;i++){
            if(A.Element_Present(i,j)){
                b(i)-=A(i,j)*v;
                A.Set_Element(i,j,(T)0);
            }
        }
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Update_b(VECTOR_ND<T>& b)
{
    ////fill b
    b.Resize(pressure_samples.m);
    INDIRECT_ARRAY<ARRAY_VIEW<TV> > indirect_v(codimensional_mesh_object->particles.V,velocity_samples);
    VECTOR_ND<T> v(velocity_samples.m*TV::m);ARRAY_TV_to_VECTOR_ND(indirect_v,v);
    GT_without_Psi_N.Times(v,b);

    ////TODO: add non-zero psi_N and non-zero psi_D
    if(use_target_div){for(int i=1;i<=pressure_samples.m;i++)b(i)+=target_div(i);}
}


//template<class TV>
//void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Update_Dirichlet_Boundary_RHS(VECTOR_ND<T>& psi_D_rhs)
//{
//    psi_D_rhs.Resize(pressure_samples.m);
//    VECTOR_ND<T> g(velocity_samples.m*TV::m);
//    CODIMENSIONAL_MESH& mesh=codimensional_mesh_object->mesh;
//    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles=codimensional_mesh_object->particles;
//    for(int i=1;i<=psi_D.m;i++){
//        int e=psi_D(i).index;int type=psi_D(i).type;T v=psi_D(i).v;
//        switch(type){
//        case SIMPLEX_PARTICLE:{
//            const ARRAY<int>& nb_surface_tris=(*mesh.tetrahedron_mesh.boundary_mesh->incident_elements)(e);
//            TV area_weighted_normal=TV();
//            for(int j=1;j<=nb_surface_tris.m;j++){
//                const TV_INT3& v=mesh.tetrahedron_mesh.boundary_mesh->elements(nb_surface_tris(j));
//                area_weighted_normal+=(T)one_third*(T).5*TV::Cross_Product(particles.X(v(2))-particles.X(v(1)),particles.X(v(3))-particles.X(v(1)));
//            }
//            //LOG::cout<<"awn: "<<e<<": "<<area_weighted_normal<<std::endl;
//            int e_in_vs=Find_Velocity_Sample(e);
//            if(e_in_vs){for(int d=1;d<=3;d++)g((e_in_vs-1)*TV::m+d)=area_weighted_normal(d);}
//        }break;
//        case SIMPLEX_SEGMENT:{
//            ////TODO: implement this
//        }break;
//        case SIMPLEX_TRIANGLE:{
//            ////TODO: implement this
//        }break;
//        }
//    }
//    GT.Times(M_inv*g,psi_D_rhs);
//}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Modify_b_With_Dirichlet_Boundary_Conditions(VECTOR_ND<T>& b,ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> >& psi_D)
{
    PHYSBAM_ASSERT(psi_D_on_mesh_hashtable!=0);
    VECTOR_ND<T> psi_D_rhs(pressure_samples.m);
    VECTOR_ND<T> g(velocity_samples.m*TV::m);
    //CODIMENSIONAL_MESH& mesh=codimensional_mesh_object->mesh;
    //CODIMENSIONAL_FLUID_PARTICLES<TV>& particles=codimensional_mesh_object->particles;
    for(int i=1;i<=psi_D.m;i++){
        int p_in_vs=psi_D(i).index;int type=psi_D(i).type;T value=psi_D(i).v;
        switch(type){
        case SIMPLEX_PARTICLE:{
            int p=velocity_samples(p_in_vs);
            //TV area_weighted_normal=codimensional_mesh_object->Nodal_Area_Weighted_Normal(p);
            TV interior_area_weighted_normal;TV exterior_area_weighted_normal;bool has_interior_psi_D_element=Area_Weighted_Normal_With_Dirichlet_Boundary_Conditions(p,(*psi_D_on_mesh_hashtable),interior_area_weighted_normal,exterior_area_weighted_normal);
            TV area_weighted_normal=has_interior_psi_D_element?-interior_area_weighted_normal:exterior_area_weighted_normal;
            //if(has_interior_psi_D_element)LOG::cout<<"int_psi_D: "<<p_in_vs<<", "<<p<<": "<<value<<": "<<area_weighted_normal<<std::endl;
            //LOG::cout<<"psi_D: "<<p_in_vs<<", "<<p<<": "<<particles.thickness(p)<<", "<<value<<", "<<area_weighted_normal<<std::endl;
            for(int d=1;d<=3;d++)g((p_in_vs-1)*TV::m+d)=area_weighted_normal(d)*value;
        }break;
        case SIMPLEX_SEGMENT:break;
        case SIMPLEX_TRIANGLE:break;    ////currently Dirichlet boundaries are on particles only
        }
    }
    if(GT.m==0)Update_GT(G,GT);
    GT.Times(M_inv*g,psi_D_rhs);
    b=b-psi_D_rhs;

    //for(int i=1;i<=psi_D.m;i++){LOG::cout<<psi_D(i).index<<": "<<psi_D(i).v<<", "<<std::endl;}
    //LOG::cout<<psi_D_rhs<<std::endl;
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Modify_b_With_Neumann_Boundary_Conditions(VECTOR_ND<T>& b,ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> >& psi_N)
{
    VECTOR_ND<T> psi_N_rhs(pressure_samples.m);
    VECTOR_ND<T> h(velocity_samples.m*TV::m);
    //CODIMENSIONAL_MESH& mesh=codimensional_mesh_object->mesh;
    //CODIMENSIONAL_FLUID_PARTICLES<TV>& particles=codimensional_mesh_object->particles;
    for(int i=1;i<=psi_N.m;i++){
        int p_in_vs=psi_N(i).index;T value=psi_N(i).v;TV n=psi_N(i).direction;
        //int p=velocity_samples(p_in_vs);
        for(int d=1;d<=3;d++)h((p_in_vs-1)*TV::m+d)=n(d)*value;
        //LOG::cout<<"mpsi_N: "<<p_in_vs<<", "<<n<<", "<<value<<std::endl;
    }
    //if(GT.m==0)Update_GT(G,GT);
    //GT.Times(h,psi_N_rhs);

    if(GnT.m==0)Update_GT(Gn,GnT);
    GnT.Times(h,psi_N_rhs);

    //LOG::cout<<"h: "<<h<<std::endl;
    //LOG::cout<<"rhs: "<<psi_N_rhs<<std::endl;
    b=b-psi_N_rhs;
}

template<class TV>
bool PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Area_Weighted_Normal_With_Dirichlet_Boundary_Conditions(const int p,const HASHTABLE<int>& psi_D_hashtable,TV& interior_area_weighted_normal,TV& exterior_area_weighted_normal)
{
    interior_area_weighted_normal=TV();exterior_area_weighted_normal=TV();bool has_psi_D_element=false;
    CODIMENSIONAL_MESH& mesh=codimensional_mesh_object->mesh;
    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles=codimensional_mesh_object->particles;

    for(int i=1;i<=(*mesh.tetrahedron_mesh.incident_elements)(p).m;i++){
        const int t=(*mesh.tetrahedron_mesh.incident_elements)(p)(i);
        const TV_INT4& v=mesh.tetrahedron_mesh.elements(t);
        //bool is_psi_D_element=true;for(int d=1;d<=4;d++)if(!psi_D_hashtable.Contains(v(d))){is_psi_D_element=false;break;}
        bool is_psi_D_element=Full_Ring_Dirichlet_Boundary(mesh,(*psi_D_on_mesh_hashtable),t,SIMPLEX_TETRAHEDRON);
        TV area_weighted_tet_normal=CODIMENSIONAL_MESH_OBJECT<TV>::Face_Area_Weighted_Volumetric_Vertex_Normal(particles.X,v,p);
        if(is_psi_D_element){has_psi_D_element=true;interior_area_weighted_normal+=area_weighted_tet_normal;}
        else exterior_area_weighted_normal+=area_weighted_tet_normal;
    }

    for(int i=1;i<=(*mesh.triangle_mesh.incident_elements)(p).m;i++){
        const int t=(*mesh.triangle_mesh.incident_elements)(p)(i);
        const TV_INT3& v=mesh.triangle_mesh.elements(t);
        //bool is_psi_D_element=true;for(int d=1;d<=3;d++)if(!psi_D_hashtable.Contains(v(d))){is_psi_D_element=false;break;}
        bool is_psi_D_element=Full_Ring_Dirichlet_Boundary(mesh,(*psi_D_on_mesh_hashtable),t,SIMPLEX_TRIANGLE);
        TV area_weighted_tri_normal=(use_particle_thickness?particles.thickness(p):thickness_scale)*CODIMENSIONAL_MESH_OBJECT<TV>::Length_Weighted_Triangle_Vertex_Normal(particles.X,v,p);
        if(is_psi_D_element){has_psi_D_element=true;interior_area_weighted_normal+=area_weighted_tri_normal;}
        else exterior_area_weighted_normal+=area_weighted_tri_normal;
    }

    for(int i=1;i<=(*mesh.segment_mesh.incident_elements)(p).m;i++){
        const int s=(*mesh.segment_mesh.incident_elements)(p)(i);
        const TV_INT2& v=mesh.segment_mesh.elements(s);
        bool is_psi_D_element=true;for(int d=1;d<=2;d++)if(!psi_D_hashtable.Contains(v(d))){is_psi_D_element=false;break;}
        TV area_weighted_seg_normal=CODIMENSIONAL_MESH_OBJECT<TV>::Segment_Cross_Section_Area((use_particle_thickness?particles.thickness(p):thickness_scale))*CODIMENSIONAL_MESH_OBJECT<TV>::Segment_Vertex_Normal(particles.X,v,p);
        if(is_psi_D_element){has_psi_D_element=true;interior_area_weighted_normal+=area_weighted_seg_normal;}
        else exterior_area_weighted_normal+=area_weighted_seg_normal;
    }

    return has_psi_D_element;
}

//template<class TV>
//TV PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Area_Weighted_Normal_Sum(const int p)
//{
//    TV n_sum=TV();
//    int p_in_s=(*particle_index_in_velocity_samples)(p);
//    int r=(p_in_s-1)*TV::m+1;
//    int row_first_element=G.offsets(r);
//    int column_number=G.offsets(r+1)-row_first_element;
//    for(int i=row_first_element;i<row_first_element+column_number;i++){
//        int c=G.A(i).j;
//        TV n=Get_Matrix_Vector_Element(G,p_in_s,c,true);
//        n_sum+=n;
//    }
//    return n_sum;
//}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Target_One_Ring_Volume(VECTOR<VECTOR<T,2>,3>& p,VECTOR<T,3>& r)
{
    VECTOR<T,2> v2=p(1)+(p(2)-p(1)).Normalized()*(r(1)+r(2));
    VECTOR<T,2> v3=p(1)+(p(3)-p(1)).Normalized()*(r(1)+r(3));
    T area=TRIANGLE_2D<T>::Area(p(1),v2,v3);return area;
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Target_One_Ring_Volume(VECTOR<VECTOR<T,3>,4>& p,VECTOR<T,4>& r)
{
    VECTOR<T,3> v2=p(1)+(p(2)-p(1)).Normalized()*(r(1)+r(2));
    VECTOR<T,3> v3=p(1)+(p(3)-p(1)).Normalized()*(r(1)+r(3));
    VECTOR<T,3> v4=p(1)+(p(4)-p(1)).Normalized()*(r(1)+r(4));
    T volume=TETRAHEDRON<T>::Volume(p(1),v2,v3,v4);return volume;
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Apply_Pressures()
{
    VECTOR_ND<T> gp(velocity_samples.m*TV::m);
    G.Times(p,gp);
    //LOG::cout<<"p: "<<p<<std::endl;
    //LOG::cout<<"grad p: "<<gp<<std::endl;
    //LOG::cout<<"m inv: "<<M_inv<<std::endl;
    gp=M_inv*gp;
    ARRAY<TV> grad_p;VECTOR_ND_to_ARRAY_TV(gp,grad_p);
    INDIRECT_ARRAY<ARRAY_VIEW<TV> > v(codimensional_mesh_object->particles.V,velocity_samples);
    for(int i=1;i<=grad_p.m;i++){
        ////TODO: dbg 
        if(!Is_Valid_Number(grad_p(i)(1))||!Is_Valid_Number(grad_p(i)(2))||!Is_Valid_Number(grad_p(i)(3)))grad_p(i)=TV();

        v(i)-=grad_p(i);
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Set_Matrix_Vector_Element(SPARSE_MATRIX_FLAT_MXN<T>& A,const TV& v,const int i,const int j,bool column_aligned)
{
    if(column_aligned)for(int d=1;d<=TV::dimension;d++){int row=(i-1)*TV::dimension+d;int column=j;A(row,column)=v(d);}
    else for(int d=1;d<=TV::dimension;d++){int row=i;int column=(j-1)*TV::dimension+d;A(row,column)=v(d);}
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Add_Matrix_Vector_Element(SPARSE_MATRIX_FLAT_MXN<T>& A,const TV& v,const int i,const int j,bool column_aligned)
{
    if(column_aligned)for(int d=1;d<=TV::dimension;d++){int row=(i-1)*TV::dimension+d;int column=j;A(row,column)=A(row,column)+v(d);}
    else for(int d=1;d<=TV::dimension;d++){int row=i;int column=(j-1)*TV::dimension+d;A(row,column)=A(row,column)+v(d);}
}

template<class TV>
TV PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Get_Matrix_Vector_Element(SPARSE_MATRIX_FLAT_MXN<T>& A,const int i,const int j,bool column_aligned)
{
    TV v;if(column_aligned)for(int d=1;d<=TV::dimension;d++){int row=(i-1)*TV::dimension+d;int column=j;v(d)=A(row,column);}
    else for(int d=1;d<=TV::dimension;d++){int row=i;int column=(j-1)*TV::dimension+d;v(d)=A(row,column);}return v;
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Check_Divergence()
{
    if(velocity_samples.m==0||pressure_samples.m==0){return;}

    VECTOR_ND<T> div(pressure_samples.m);
    INDIRECT_ARRAY<ARRAY_VIEW<TV> > indirect_v(codimensional_mesh_object->particles.V,velocity_samples);
    VECTOR_ND<T> v(velocity_samples.m*TV::dimension);ARRAY_TV_to_VECTOR_ND(indirect_v,v);
    //GT.Times(v,div);div*=(T)-1; ////div=-GT*v
    GT_without_Psi_N.Times(v,div);div*=(T)-1; ////div=-GT*v
    //LOG::cout<<"div u: "<<div<<std::endl;
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Solve()
{
    //if(psi_D.m>0){
    //    //LOG::cout<<"N"<<mpi_codim->node_id<<" start solve"<<std::endl;
    //    SPARSE_MATRIX_FLAT_NXN<T> B;VECTOR_ND<T> d;ARRAY<int> map_A_to_B;ARRAY<int> map_B_to_A;VECTOR_ND<T> p_B;
    //    Shrink_A_and_b_With_Dirichlet_Boundary_Conditions(B,d,map_A_to_B,map_B_to_A);
    //    Solve_Linear_System(B,p_B,d);
    //    //LOG::cout<<"N"<<mpi_codim->node_id<<" finish solve"<<std::endl;
    //    p.Resize(A.n);
    //    for(int i=1;i<=A.n;i++){int i_B=map_A_to_B(i);if(i_B!=0)p(i)=p_B(i_B);}
    //    for(int i=1;i<=psi_D.m;i++){int i_in_ps=psi_D(i).index;p(i_in_ps)=psi_D(i).v;}
    //    
    //    //LOG::cout<<"N"<<mpi_codim->node_id<<" finish2"<<std::endl;
    //    //////TODO: doesn't work...
    //    //Set_Zero_Rows_in_A_and_b_With_Dirchlet_Boundary_Conditions();
    //    //Solve_Linear_System(A,p,b);
    //    //for(int i=1;i<=psi_D.m;i++){int i_in_ps=psi_D(i).index;p(i_in_ps)=psi_D(i).v;}
    //}
    //else{
        Solve_Linear_System(A,p,b);
    //}
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Solve_Linear_System(SPARSE_MATRIX_FLAT_NXN<T>& A,VECTOR_ND<T>& x,VECTOR_ND<T>& b)
{
    static CONJUGATE_GRADIENT<T> cg;
    static CONJUGATE_RESIDUAL<T> cr;
    static SYMMQMR<T> symmqmr;
    KRYLOV_SOLVER<T>* krylov_solver=0;
    //KRYLOV_SOLVER_TYPE solver_type=krylov_solver_cg;
    KRYLOV_SOLVER_TYPE solver_type=krylov_solver_cr;

    if(solver_type==krylov_solver_cg){krylov_solver=&cg;}
    else if(solver_type==krylov_solver_cr){krylov_solver=&cr;}
    else if(solver_type==krylov_solver_symmqmr){krylov_solver=&symmqmr;}

    bool print_diagnostics=(mpi_codim==0||mpi_codim->node_id==1);
    krylov_solver->print_diagnostics=print_diagnostics;
    krylov_solver->print_residuals=false;
    krylov_solver->restart_iterations=40;
    krylov_solver->nullspace_tolerance=(T)0;
    
    T absolute_tolerance=(T)1e-30;T relative_tolerance=(T)1e-10;
    T b_max_abs=b.Max_Abs();if(mpi_codim)b_max_abs=mpi_codim->Max(b_max_abs);
    T tolerance=max(relative_tolerance*b_max_abs,absolute_tolerance);
    //LOG::cout<<"subsystem tolerance: "<<tolerance<<std::endl;

    x.Resize(A.n);
    VECTOR_ND<T> q(A.n),s(A.n),r(A.n),k(A.n),z(A.n);
    //VECTOR_ND<T> q,s,r,k,z;
    KRYLOV_VECTOR_WRAPPER<T,VECTOR_ND<T>&> kp(x),kb(b),kq(q),ks(s),kr(r),kk(k),kz(z);
    CODIMENSIONAL_KRYLOV_SYSTEM<T> system(A,null_space_basis,mpi_codim,&boundary_pressure_indices,&ghost_pressure_indices,&domain_pressure_indices);
    system.use_preconditioner=use_precoditioner;
    if(use_precoditioner){A.Initialize_Diagonal_Index();}
    krylov_solver->Solve(system,kp,kb,kq,ks,kr,kk,kz,tolerance,1,max_iter_number);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Initialize_Subsystem_Helper(ARRAY<CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>* >& subsystems,
    const ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> >* psi_N_on_mesh/*=0*/,const ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> >* psi_D_on_mesh/*=0*/,const ARRAY<CODIMENSIONAL_TARGET_DIVERGENCE<T> >* tar_div/*=0*/,bool use_one_region/*=false*/)
{
    PHYSBAM_ASSERT(subsystems.m>0);
    PHYSBAM_ASSERT(subsystems(1)->codimensional_mesh_object!=0);
    PHYSBAM_ASSERT(subsystems(1)->element_index_in_pressure_samples!=0);
    PHYSBAM_ASSERT(subsystems(1)->particle_index_in_velocity_samples!=0);
    PHYSBAM_ASSERT(subsystems(1)->psi_D_on_mesh_hashtable!=0);

    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles=subsystems(1)->codimensional_mesh_object->particles;
    CODIMENSIONAL_MESH& mesh=subsystems(1)->codimensional_mesh_object->mesh;
    SEGMENT_MESH& segment_mesh=mesh.segment_mesh;
    TRIANGLE_MESH& triangle_mesh=mesh.triangle_mesh;
    //TETRAHEDRON_MESH& tetrahedron_mesh=mesh.tetrahedron_mesh;

    VECTOR<ARRAY<int>,4>& element_index_in_pressure_samples=(*subsystems(1)->element_index_in_pressure_samples); 
    ARRAY<int>& particle_index_in_velocity_samples=(*subsystems(1)->particle_index_in_velocity_samples);
    bool use_all_pressure_sample_on_nodes=subsystems(1)->use_all_pressure_sample_on_nodes;
    HASHTABLE<int>& psi_D_on_mesh_hashtable=(*subsystems(1)->psi_D_on_mesh_hashtable);

    //////resize global data
    element_index_in_pressure_samples(SIMPLEX_PARTICLE).Resize(particles.X.m);
    element_index_in_pressure_samples(SIMPLEX_SEGMENT).Resize(segment_mesh.elements.m);
    element_index_in_pressure_samples(SIMPLEX_TRIANGLE).Resize(triangle_mesh.elements.m);
    particle_index_in_velocity_samples.Resize(particles.X.m);

    ////initialize subsystem elements
    for(int i=1;i<=particles.X.m;i++){
        int region=mesh.Particle_Region_Index(i);if(region==0)continue;if(use_one_region)region=1;if(particles.inactive(i))continue;
        subsystems(region)->Add_Simplex_Index(i,SIMPLEX_PARTICLE);
    }
    for(int i=1;i<=segment_mesh.elements.m;i++){
        int region=mesh.Segment_Region_Index(i);if(region==0)continue;if(use_one_region)region=1;if(Element_Or(particles.inactive,segment_mesh.elements(i)))continue;
        subsystems(region)->Add_Simplex_Index(i,SIMPLEX_SEGMENT);
    }
    for(int i=1;i<=triangle_mesh.elements.m;i++){
        int region=mesh.Triangle_Region_Index(i);if(region==0)continue;if(use_one_region)region=1;if(Element_Or(particles.inactive,triangle_mesh.elements(i)))continue;
        subsystems(region)->Add_Simplex_Index(i,SIMPLEX_TRIANGLE);
    }

    psi_D_on_mesh_hashtable.Remove_All();
    if(psi_D_on_mesh)for(int i=1;i<=(*psi_D_on_mesh).m;i++)if((*psi_D_on_mesh)(i).type==SIMPLEX_PARTICLE)psi_D_on_mesh_hashtable.Set((*psi_D_on_mesh)(i).index);

    ////initialize pressure samples
    ////pressure on particles
    if(!use_all_pressure_sample_on_nodes){
        for(int i=1;i<=particles.X.m;i++){
            int region=mesh.Particle_Region_Index(i);if(region==0)continue;if(use_one_region)region=1;if(particles.inactive(i))continue;
            if(Full_Ring_Dirichlet_Boundary(mesh,psi_D_on_mesh_hashtable,i,SIMPLEX_PARTICLE))continue;

            if(Is_On_Volume_Mesh<TV::m>(mesh.node_types(i))){
                int pm=subsystems(region)->Add_Pressure_Sample(i,SIMPLEX_PARTICLE);
                element_index_in_pressure_samples(SIMPLEX_PARTICLE)(i)=pm;
            }
        }
        ////pressure on segments
        for(int i=1;i<=segment_mesh.elements.m;i++){
            int region=mesh.Segment_Region_Index(i);if(region==0)continue;if(use_one_region)region=1;if(Element_Or(particles.inactive,segment_mesh.elements(i)))continue;
            if(Full_Ring_Dirichlet_Boundary(mesh,psi_D_on_mesh_hashtable,i,SIMPLEX_SEGMENT))continue;
            int pm=subsystems(region)->Add_Pressure_Sample(i,SIMPLEX_SEGMENT);
            element_index_in_pressure_samples(SIMPLEX_SEGMENT)(i)=pm;
        }
        ////pressure on triangles
        if(TV::m==3)for(int i=1;i<=triangle_mesh.elements.m;i++){
            int region=mesh.Triangle_Region_Index(i);if(region==0)continue;if(use_one_region)region=1;if(Element_Or(particles.inactive,triangle_mesh.elements(i)))continue;
            if(Full_Ring_Dirichlet_Boundary(mesh,psi_D_on_mesh_hashtable,i,SIMPLEX_TRIANGLE))continue;
            int pm=subsystems(region)->Add_Pressure_Sample(i,SIMPLEX_TRIANGLE);
            element_index_in_pressure_samples(SIMPLEX_TRIANGLE)(i)=pm;
        }
    }
    else{   ////all pressure samples on nodes
        for(int i=1;i<=particles.X.m;i++){
            int region=mesh.Particle_Region_Index(i);if(region==0)continue;if(use_one_region)region=1;if(particles.inactive(i))continue;
            if(Full_Ring_Dirichlet_Boundary(mesh,psi_D_on_mesh_hashtable,i,SIMPLEX_PARTICLE)){continue;}
            int pm=subsystems(region)->Add_Pressure_Sample(i,SIMPLEX_PARTICLE);
            element_index_in_pressure_samples(SIMPLEX_PARTICLE)(i)=pm;
        }
    }

    ////initialize velocity samples
    for(int i=1;i<=particles.X.m;i++){
        int region=mesh.Particle_Region_Index(i);if(region==0)continue;if(use_one_region)region=1;if(particles.inactive(i))continue;
        if(Full_Ring_Dirichlet_Boundary(mesh,psi_D_on_mesh_hashtable,i,SIMPLEX_PARTICLE))continue;
        int vm=subsystems(region)->Add_Velocity_Sample(i);
        particle_index_in_velocity_samples(i)=vm;
    }

    //LOG::cout<<"ps: "<<subsystems(1)->pressure_samples.m<<", vs: "<<subsystems(1)->velocity_samples.m<<std::endl;

    ////initialize psi_D samples
    if(psi_D_on_mesh)for(int i=1;i<=(*psi_D_on_mesh).m;i++){
        int p=(*psi_D_on_mesh)(i).index;int type=(*psi_D_on_mesh)(i).type;T v=(*psi_D_on_mesh)(i).v;
        if(type==SIMPLEX_PARTICLE){
            int region=mesh.Particle_Region_Index(p);if(region==0)continue;if(use_one_region)region=1;if(particles.inactive(p))continue;
            if(subsystems(region)->Find_Velocity_Sample(p)!=0){ ////currently all psi_D are on velocity samples
                subsystems(region)->Add_Dirichlet_Boundary_Sample(p,type,v);
            }
            else{/*LOG::cout<<"cannot find psi_D "<<p<<std::endl;*/}
        }
    }

    ////initialize psi_N samples
    if(psi_N_on_mesh)for(int i=1;i<=(*psi_N_on_mesh).m;i++){
        int p=(*psi_N_on_mesh)(i).index;
        int region=mesh.Particle_Region_Index(p);if(region==0)continue;if(use_one_region)region=1;if(particles.inactive(p))continue;
        if(psi_D_on_mesh_hashtable.Contains(p)){
            continue;
            //if(particles.length_scale(p)>=subsystems(region)->length_scale_min*(T)1.5)continue;
            //else{LOG::cout<<"psi_N near fiber "<<p<<std::endl;}
        }   ////if it's near fiber, apply psi_N, else apply psi_D, TODO: fix this by setting better boundary conditions for paint

        TV direction=(*psi_N_on_mesh)(i).direction;T v=(*psi_N_on_mesh)(i).v;
        if(subsystems(region)->Find_Velocity_Sample(p)!=0){
            subsystems(region)->Add_Neumann_Boundary_Sample(p,direction,v);
        }
        else{/*LOG::cout<<"cannot find psi_N "<<p<<std::endl;*/}
    }

    ////initialize tar_div samples
    if(tar_div){
        for(int i=1;i<=subsystems.m;i++){
            subsystems(i)->target_div.Resize(subsystems(i)->pressure_samples.m);
            subsystems(i)->target_div.Fill((T)0);
        }

        for(int i=1;i<=(*tar_div).m;i++){
            int e=(*tar_div)(i).index;int type=(*tar_div)(i).type;int region=0;
            switch(type){
            case SIMPLEX_PARTICLE:region=mesh.Particle_Region_Index(e);break;
            case SIMPLEX_SEGMENT:region=mesh.Segment_Region_Index(e);break;
            case SIMPLEX_TRIANGLE:region=mesh.Triangle_Region_Index(e);break;}
            if(region==0)continue;

            switch(type){
            case SIMPLEX_PARTICLE:if(particles.inactive(e))continue;break;
            case SIMPLEX_SEGMENT:if(Element_Or(particles.inactive,mesh.segment_mesh.elements(e)))continue;break;
            case SIMPLEX_TRIANGLE:if(Element_Or(particles.inactive,mesh.triangle_mesh.elements(e)))continue;break;}

            int ps=subsystems(region)->Find_Pressure_Sample(e,type);
            if(ps!=0)subsystems(region)->target_div(ps)=(*tar_div)(i).v;
        }
        //LOG::cout<<"subsystem tardiv: "<<subsystems(1)->target_div<<std::endl;
    }

    ////dbg
    //LOG::cout<<"subsytems "<<subsystems.m<<": "<<std::endl;
    //for(int i=1;i<=subsystems.m;i++){LOG::cout<<"subsys "<<i<<": "<<subsystems(i)->pressure_samples.m<<", "<<subsystems(i)->velocity_samples.m<<std::endl;}
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Initialize(CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>& subsystem,
    const ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> >* psi_N/*=0*/,const ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> >* psi_D/*=0*/,const ARRAY<CODIMENSIONAL_TARGET_DIVERGENCE<T> >* tar_div/*=0*/)
{
    ARRAY<CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>* > subsystem_ptrs;subsystem_ptrs.Append(&subsystem);
    if(subsystem_ptrs.m>0)Initialize_Subsystem_Helper(subsystem_ptrs,psi_N,psi_D,tar_div,true);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Initialize(ARRAY<CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV> >& subsystems,
    const ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> >* psi_N/*=0*/,const ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> >* psi_D/*=0*/,const ARRAY<CODIMENSIONAL_TARGET_DIVERGENCE<T> >* tar_div/*=0*/)
{
    ARRAY<CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>* > subsystem_ptrs;
    for(int i=1;i<=subsystems.m;i++)subsystem_ptrs.Append(&subsystems(i));
    if(subsystem_ptrs.m>0)Initialize_Subsystem_Helper(subsystem_ptrs,psi_N,psi_D,tar_div,false);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Interpolate_Pressure_To_Particles_Helper(CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>& subsystem,ARRAY<T>& u,ARRAY<T>& w)
{
    CODIMENSIONAL_MESH_OBJECT<TV>& mesh_object=*subsystem.codimensional_mesh_object;
    //CODIMENSIONAL_FLUID_PARTICLES<TV>& particles=mesh_object.particles;
    CODIMENSIONAL_MESH& mesh=mesh_object.mesh;
    SEGMENT_MESH& segment_mesh=mesh.segment_mesh;
    TRIANGLE_MESH& triangle_mesh=mesh.triangle_mesh;
    //TETRAHEDRON_MESH& tetrahedron_mesh=mesh.tetrahedron_mesh;
    VECTOR_ND<T>& pressure=subsystem.p;

    for(int i=1;i<=subsystem.pressure_samples.m;i++){
        int p=subsystem.pressure_samples(i).index;int type=subsystem.pressure_samples(i).type;
        T volume=mesh_object.Element_Partial_Volume(p,type,(*subsystem.psi_D_on_mesh_hashtable));
        switch(type){
        case SIMPLEX_PARTICLE:
            {u(p)+=pressure(i)*volume;w(p)+=volume;}break;
        case SIMPLEX_SEGMENT:
            {
                const TV_INT2& v=segment_mesh.elements(p);
                for(int d=1;d<=2;d++){u(v(d))+=pressure(i)*volume*(T).5;w(v(d))+=volume*(T).5;}
            }break;
        case SIMPLEX_TRIANGLE:
            {
                const TV_INT3& v=triangle_mesh.elements(p);
                for(int d=1;d<=3;d++){u(v(d))+=pressure(i)*volume*(T)one_third;w(v(d))+=volume*(T)one_third;}
            }break;
        }
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Interpolate_Pressure_To_Particles(CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>& subsystem,ARRAY<T>& u)
{
    PHYSBAM_ASSERT(subsystem.codimensional_mesh_object!=0);
    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles=subsystem.codimensional_mesh_object->particles;
    u.Resize(particles.X.m);u.Fill((T)0);ARRAY<T> w(particles.X.m);w.Fill((T)0);
    Interpolate_Pressure_To_Particles_Helper(subsystem,u,w);
    for(int i=1;i<=particles.X.m;i++)if(w(i)!=(T)0)u(i)/=w(i);
}


template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Interpolate_Pressure_To_Particles(ARRAY<CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV> >& subsystems,ARRAY<T>& u)
{
    PHYSBAM_ASSERT(subsystems.m>0);PHYSBAM_ASSERT(subsystems(1).codimensional_mesh_object!=0);
    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles=subsystems(1).codimensional_mesh_object->particles;
    u.Resize(particles.X.m);u.Fill((T)0);ARRAY<T> w(particles.X.m);w.Fill((T)0);
    for(int i=1;i<=subsystems.m;i++)Interpolate_Pressure_To_Particles_Helper(subsystems(i),u,w);
    for(int i=1;i<=particles.X.m;i++)if(w(i)!=(T)0)u(i)/=w(i);
}

template<class TV>
bool PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Full_Ring_Dirichlet_Boundary(CODIMENSIONAL_MESH& mesh,const HASHTABLE<int>& psi_D,const int p,const int type)
{
    ////TODO: fix bug here
    //switch(type){
    //case SIMPLEX_PARTICLE:{
    //    if(!psi_D.Contains(p))return false;
    //    bool has_full_ring_psi_D=true;
    //    //for(int i=1;i<=(*mesh.neighbor_nodes)(p).m;i++){
    //    //    int nb_p=(*mesh.neighbor_nodes)(p)(i);
    //    //    if(!psi_D.Contains(nb_p)){has_full_ring_psi_D=false;break;}
    //    //}
    //    ////TODO: extend this to triangles and tetrahedra
    //    if((*mesh.triangle_mesh.neighbor_nodes)(p).m>0||(*mesh.tetrahedron_mesh.neighbor_nodes)(p).m>0)return false;
    //    for(int i=1;i<=(*mesh.segment_mesh.neighbor_nodes)(p).m;i++){
    //        int nb_p=(*mesh.segment_mesh.neighbor_nodes)(p)(i);
    //        if(!psi_D.Contains(nb_p)){has_full_ring_psi_D=false;break;}
    //    }
    //    return has_full_ring_psi_D;}
    //case SIMPLEX_SEGMENT:{
    //    const TV_INT2& v=mesh.segment_mesh.elements(p);
    //    bool has_full_ring_psi_D=true;
    //    for(int i=1;i<=2;i++)if(!psi_D.Contains(v(i))){has_full_ring_psi_D=false;break;}
    //    return has_full_ring_psi_D;}
    //case SIMPLEX_TRIANGLE:{
    //    //const TV_INT3& v=mesh.triangle_mesh.elements(p);
    //    //bool has_full_ring_psi_D=true;
    //    //for(int i=1;i<=3;i++)if(!psi_D.Contains(v(i))){has_full_ring_psi_D=false;break;}
    //    //return has_full_ring_psi_D;

    //    ////TODO: extend this to triangles and tetrahedra
    //    return false;}
    //case SIMPLEX_TETRAHEDRON:{return false;}
    //}
    return false;
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Update_Implicit_Elasticity_Matrices(const ARRAY_VIEW<T_MATRIX>& FT)
{
    ////assuming G, GT, and M_inv are computed 

    ////Gh, 9nx3n
    int row=G.m*TV::m;int column=G.n*TV::m;
    ARRAY<int> row_lengths;row_lengths.Resize(row);row_lengths.Fill(0);
    for(int i=1;i<=G.m;i++){
        for(int j=1;j<=TV::m;j++){
            row_lengths((j-1)*G.m+i)=G.offsets(i+1)-G.offsets(i);
        }
    }
    Gh.Reset(column);Gh.Set_Row_Lengths(row_lengths);

    for(int r=1;r<=G.m;r++){
        for(int c=G.offsets(r);c<G.offsets(r+1);c++){
            T entry=G.A(c).a;
            int j=G.A(c).j;
            for(int d=1;d<=TV::m;d++){
                int Gh_i=(d-1)*G.m+r;
                int Gh_j=(d-1)*G.n+j;
                Gh(Gh_i,Gh_j)=entry;
            }
        }
    }
    ////GhT, 3nx9n
    GhT.Reset(Gh.m);
    Gh.Transpose(GhT);
    //LOG::cout<<"G:\n"<<Gh<<std::endl;
    //LOG::cout<<"GT:\n"<<GhT<<std::endl;

    ////P, 9nx9n
    row=G.m*TV::m;column=row;
    row_lengths.Resize(row);row_lengths.Fill(1);
    P.Reset(column);P.Set_Row_Lengths(row_lengths);
    for(int i=1;i<=G.m;i++){
        for(int d=1;d<=TV::m;d++){
            int P_i=(i-1)*TV::m+d;
            int P_j=(d-1)*G.m+i;
            P(P_i,P_j)=(T)1;
        }
    }
    //LOG::cout<<"P:\n"<<P<<std::endl;
    ////I+P, 9nx9n
    for(int i=1;i<=P.m;i++){if(!P.Element_Present(i,i)){row_lengths(i)++;}}
    I_plus_P.Reset(column);I_plus_P.Set_Row_Lengths(row_lengths);
    for(int i=1;i<=P.m;i++){
        for(int c=P.offsets(i);c<P.offsets(i+1);c++){
            T entry=P.A(c).a;
            int j=P.A(c).j;
            I_plus_P(i,j)=entry;
        }
        if(!I_plus_P.Element_Present(i,i)){I_plus_P(i,i)=1;}
        else I_plus_P(i,i)+=1;
    }
    //LOG::cout<<"I_plus_P:\n"<<I_plus_P<<std::endl;

    ////Fh, 9nx9n
    row_lengths.Resize(row);row_lengths.Fill(3);
    Fh.Reset(column);Fh.Set_Row_Lengths(row_lengths);
    for(int i=1;i<=velocity_samples.m;i++){
        int p=velocity_samples(i);
        T_MATRIX mtx=FT(p);
        //LOG::cout<<"mtx"<<i<<": "<<mtx<<std::endl;
        for(int q=1;q<=TV::m;q++){
            int r0=((q-1)*velocity_samples.m+(i-1))*TV::m;
            for(int d=1;d<=TV::m;d++){
                int r=r0+d;
                for(int k=1;k<=TV::m;k++){
                    int c=r0+k;
                    T value=mtx(d,k);
                    Fh.Set_Element(r,c,value);
                }
            }
        }
    }
    //LOG::cout<<"Fh:\n"<<Fh<<std::endl;

    ////Vh_inv, 9nx9n, diagonal
    row_lengths.Resize(row);row_lengths.Fill(1);
    Mh_inv.Reset(column);Mh_inv.Set_Row_Lengths(row_lengths);
    for(int d=1;d<=TV::m;d++){
        for(int i=1;i<=M_inv.n;i++){
            T value=M_inv(i);
            int r=(d-1)*M_inv.n+i;
            Mh_inv(r,r)=value;    
        }
    }
    //LOG::cout<<"Mh_inv:\n"<<Mh_inv<<std::endl;

    ////Ah, 9nx9n
    Ah.Reset();
    Sparse_Flat_Matrix_MXN_To_NXN(GhT*P*Fh*P*Fh*I_plus_P*Mh_inv*Gh,Ah);
    //LOG::cout<<"Ah:\n"<<Ah<<std::endl;
}

//template class CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<VECTOR<float,2> >;
template class CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
//template class CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<VECTOR<double,2> >;
template class CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<VECTOR<double,3> >;
#endif
