//#####################################################################
// Copyright 2013, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM__
#define __CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM__
#include <PhysBAM_Tools/Arrays/ARRAY.h>
#include <PhysBAM_Tools/Matrices/MATRIX_2X2.h>
#include <PhysBAM_Tools/Matrices/MATRIX_3X3.h>
#include <PhysBAM_Tools/Matrices/SPARSE_MATRIX_FLAT_MXN.h>
#include <PhysBAM_Tools/Matrices/SPARSE_MATRIX_FLAT_NXN.h>
#include <PhysBAM_Tools/Math_Tools/FACTORIAL.h>
#include <PhysBAM_Tools/Arrays/INDIRECT_ARRAY.h>
#include <PhysBAM_Tools/Parallel_Computation/THREADED_UNIFORM_GRID.h>
#include <PhysBAM_Geometry/Basic_Geometry/TRIANGLE_2D.h>
#include <PhysBAM_Geometry/Basic_Geometry/TETRAHEDRON.h>
#include <PhysBAM_Geometry/Topology/TRIANGLE_MESH.h>
#include <PhysBAM_Geometry/Topology/SEGMENT_MESH.h>
#include <PhysBAM_Geometry/Constitutive_Models/STRAIN_MEASURE.h>
#include <PhysBAM_Dynamics/Particles/CODIMENSIONAL_FLUID_PARTICLES.h>
#include "../Meshing/CODIMENSIONAL_MESH_OBJECT_3D.h"
#include "../Meshing/CODIMENSIONAL_HELPER.h"

namespace PhysBAM
{
class CODIMENSIONAL_PRESSURE;
template <class TV> class CODIMENSIONAL_NEUMANN_BOUNDARY;
template <class T> class CODIMENSIONAL_DIRICHLET_BOUNDARY;
template <class T> class CODIMENSIONAL_TARGET_DIVERGENCE;
template <class TV> class CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM;
template <class TV> class CODIMENSIONAL_INCOMPRESSIBLE;
template <class TV> class CODIMENSIONAL_VISCOSITY_FORCE;
template <class TV> class MPI_CODIMENSIONAL;

class CODIMENSIONAL_PRESSURE
{
public:
    int index;int type;
    CODIMENSIONAL_PRESSURE():index(0),type(0){}
    CODIMENSIONAL_PRESSURE(int index_input,int type_input):index(index_input),type(type_input){}
};

template<class TV>
class CODIMENSIONAL_NEUMANN_BOUNDARY
{
    typedef typename TV::SCALAR T;
public:
    int index;TV direction;T v;
    CODIMENSIONAL_NEUMANN_BOUNDARY(const int index=0,const TV& direction=TV(),const T v=(T)0):index(index),direction(direction),v(v){}
    CODIMENSIONAL_NEUMANN_BOUNDARY<TV>& operator = (const CODIMENSIONAL_NEUMANN_BOUNDARY<TV>& copy){index=copy.index;direction=copy.direction;v=copy.v;return *this;}
    bool operator < (const CODIMENSIONAL_NEUMANN_BOUNDARY<TV>& s){return index<s.index;}
    bool operator < (const CODIMENSIONAL_NEUMANN_BOUNDARY<TV>& s) const {return index<s.index;}
    bool operator == (const CODIMENSIONAL_NEUMANN_BOUNDARY<TV>& s){return index==s.index;}
    bool operator == (const CODIMENSIONAL_NEUMANN_BOUNDARY<TV>& s) const {return index==s.index;}
};

template<class T>
class CODIMENSIONAL_DIRICHLET_BOUNDARY
{public:
    int index;int type;T v;
    CODIMENSIONAL_DIRICHLET_BOUNDARY(const int index=0,const int type=0,const T v=(T)0):index(index),type(type),v(v){}
    CODIMENSIONAL_DIRICHLET_BOUNDARY<T>& operator = (const CODIMENSIONAL_DIRICHLET_BOUNDARY<T>& copy){index=copy.index;type=copy.type;v=copy.v;return *this;}
    bool operator < (const CODIMENSIONAL_DIRICHLET_BOUNDARY<T>& s) {return index<s.index;}
    bool operator < (const CODIMENSIONAL_DIRICHLET_BOUNDARY<T>& s) const {return index<s.index;}
    bool operator == (const CODIMENSIONAL_DIRICHLET_BOUNDARY<T>& s) {return index==s.index;}
    bool operator == (const CODIMENSIONAL_DIRICHLET_BOUNDARY<T>& s) const {return index==s.index;}
};

template<class T>
class CODIMENSIONAL_TARGET_DIVERGENCE
{public:
    int index;int type;T v;
    CODIMENSIONAL_TARGET_DIVERGENCE(const int index=0,const int type=0,const T v=(T)0):index(index),type(type),v(v){}
    CODIMENSIONAL_TARGET_DIVERGENCE<T>& operator = (const CODIMENSIONAL_TARGET_DIVERGENCE& copy){index=copy.index;type=copy.type;v=copy.v;return *this;}
    bool operator < (const CODIMENSIONAL_TARGET_DIVERGENCE<T>& s) {return index<s.index;}
    bool operator < (const CODIMENSIONAL_TARGET_DIVERGENCE<T>& s) const {return index<s.index;}
    bool operator == (const CODIMENSIONAL_TARGET_DIVERGENCE<T>& s) {return index==s.index;}
    bool operator == (const CODIMENSIONAL_TARGET_DIVERGENCE<T>& s) const {return index==s.index;}
};

template<class TV>
class CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM
{
    typedef typename TV::SCALAR T;
    typedef typename TV::template REBIND<int>::TYPE TV_INT;
    typedef VECTOR<int,2> TV_INT2;typedef VECTOR<int,3> TV_INT3;typedef VECTOR<int,4> TV_INT4;
    typedef MATRIX<T,TV::m> T_MATRIX;typedef DIAGONAL_MATRIX<T,TV::m> T_DIAGONAL_MATRIX;typedef SYMMETRIC_MATRIX<T,TV::m> T_SYMMETRIC_MATRIX;
public:
    CODIMENSIONAL_MESH_OBJECT<TV>* codimensional_mesh_object;   ////mesh object for the incompressible solver
    VECTOR<ARRAY<int>,4>* element_index_in_pressure_samples;    ////mesh index -> matrix index
    ARRAY<int>* particle_index_in_velocity_samples;             ////mesh index -> matrix index
    VECTOR<ARRAY<int>,4> elements;  ////index in elements
    ARRAY<CODIMENSIONAL_PRESSURE> pressure_samples;     ////matrix index -> mesh index
    ARRAY<int> velocity_samples;    ////matrix index -> mesh index, assuming all velocity samples are on particles

    ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> > psi_N;   ////psi_N on matrix
    ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> > psi_D;  ////psi_D on matrix
    HASHTABLE<int>* psi_D_on_mesh_hashtable;     ////psi_D indices on mesh

    SPARSE_MATRIX_FLAT_MXN<T> G,GT,GT_without_Psi_N,Gn,GnT;
    SPARSE_MATRIX_FLAT_NXN<T> A;
    VECTOR_ND<T> b,h,p,M_inv;
    SPARSE_MATRIX_FLAT_MXN<T> FM_inv;   ////for elasticity
    SPARSE_MATRIX_FLAT_MXN<T> asym_FM_inv;  ////for elasticity
    VECTOR_ND<T> target_div;

    ////implicit elasticity
    SPARSE_MATRIX_FLAT_MXN<T> Gh,GhT,P,I_plus_P,Fh,Mh_inv;
    VECTOR_ND<T> ph;
    SPARSE_MATRIX_FLAT_NXN<T> Ah;
    VECTOR_ND<T> bh;

    ARRAY<VECTOR_ND<T> > null_space_basis;

    T length_scale;
    T length_scale_min;
    T thickness_scale;
    bool use_target_div;
    bool use_particle_thickness;
    bool use_all_pressure_sample_on_nodes;
    bool use_constant_extrapolation;
    bool use_precoditioner;
    int max_iter_number;

    ////mpi
    MPI_CODIMENSIONAL<TV>* mpi_codim;
    ARRAY<ARRAY<int> > boundary_pressure_indices;
    ARRAY<ARRAY<int> > ghost_pressure_indices;
    ARRAY<int> domain_pressure_indices;

    CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM(MPI_CODIMENSIONAL<TV>* mpi_codim_input=0);

    int Add_Pressure_Sample(const int index,const int type);
    int Find_Pressure_Sample(const int index,const int type);
    int Add_Velocity_Sample(const int index);
    int Find_Velocity_Sample(const int index);
    int Add_Neumann_Boundary_Sample(const int p,const TV& direction,const T v); ////convert mesh index to matrix index
    int Add_Dirichlet_Boundary_Sample(const int p,const int type,const T v);    ////convert mesh index to matrix index
    int Add_Simplex_Index(const int index,const int type);

    void Clear();
    void Update_Matrices();
    void Solve();
    void Solve_Linear_System(SPARSE_MATRIX_FLAT_NXN<T>& A,VECTOR_ND<T>& x,VECTOR_ND<T>& b);
    void Apply_Pressures();

    void Update_G(SPARSE_MATRIX_FLAT_MXN<T>& G);
    void Update_GT(const SPARSE_MATRIX_FLAT_MXN<T>& G,SPARSE_MATRIX_FLAT_MXN<T>& GT);
    void Modify_G_With_Boundary_Conditions(SPARSE_MATRIX_FLAT_MXN<T>& G,ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> >& psi_N);
    void Update_M_inv(VECTOR_ND<T>& M_inv);
    void Update_V_inv(VECTOR_ND<T>& V_inv);
    void Update_V_inv_With_Dirichlet_Boundary_Conditions(VECTOR_ND<T>& V_inv);

    void Update_FM_inv(const ARRAY_VIEW<T_MATRIX>& FT,const VECTOR_ND<T>& M_inv,SPARSE_MATRIX_FLAT_MXN<T>& FM_inv,bool use_sym_F=true);
    void Update_A(SPARSE_MATRIX_FLAT_MXN<T>& G,const VECTOR_ND<T>& M_inv,SPARSE_MATRIX_FLAT_NXN<T>& A);
    void Update_A(const SPARSE_MATRIX_FLAT_MXN<T>& G,const SPARSE_MATRIX_FLAT_MXN<T>& FM_inv,const SPARSE_MATRIX_FLAT_MXN<T>& GT,SPARSE_MATRIX_FLAT_NXN<T>& A);
    void Update_b(VECTOR_ND<T>& b);
    void Shrink_A_and_b_With_Dirichlet_Boundary_Conditions(SPARSE_MATRIX_FLAT_NXN<T>& B,VECTOR_ND<T>& d,ARRAY<int>& map_A_to_B,ARRAY<int>& map_B_to_A);   ////shrink Ax=b to Bx=d
    void Set_Zero_Rows_in_A_and_b_With_Dirchlet_Boundary_Conditions();
    void Modify_b_With_Dirichlet_Boundary_Conditions(VECTOR_ND<T>& b,ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> >& psi_D);
    void Modify_b_With_Neumann_Boundary_Conditions(VECTOR_ND<T>& b,ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> >& psi_N);
    bool Area_Weighted_Normal_With_Dirichlet_Boundary_Conditions(const int p,const HASHTABLE<int>& psi_D_hashtable,TV& interior_area_weighted_normal,TV& exterior_area_weighted_normal);

    void Update_Implicit_Elasticity_Matrices(const ARRAY_VIEW<T_MATRIX>& FT);

    ////for debug
    void Check_Divergence();

    static void Initialize(CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>& subsystem,
        const ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> >* psi_N=0,const ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> >* psi_D=0,const ARRAY<CODIMENSIONAL_TARGET_DIVERGENCE<T> >* tar_div=0);   ////solve all regions in one solver
    static void Initialize(ARRAY<CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV> >& subsystems,
        const ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> >* psi_N=0,const ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> >* psi_D=0,const ARRAY<CODIMENSIONAL_TARGET_DIVERGENCE<T> >* tar_div=0);  ////solve regions in different solvers

    static void Interpolate_Pressure_To_Particles(CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>& subsystem,ARRAY<T>& u);
    static void Interpolate_Pressure_To_Particles(ARRAY<CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV> >& subsystems,ARRAY<T>& u);
    static bool Full_Ring_Dirichlet_Boundary(CODIMENSIONAL_MESH& mesh,const HASHTABLE<int>& psi_D,const int p,const int type);   ////return true if all incident particles are psi_D && itself is psi_D

protected:
    T Target_One_Ring_Volume(VECTOR<VECTOR<T,2>,3>& v,VECTOR<T,3>& r);
    T Target_One_Ring_Volume(VECTOR<VECTOR<T,3>,4>& v,VECTOR<T,4>& r);

    void Set_Matrix_Vector_Element(SPARSE_MATRIX_FLAT_MXN<T>& A,const TV& v,const int i,const int j,bool column_aligned);
    void Add_Matrix_Vector_Element(SPARSE_MATRIX_FLAT_MXN<T>& A,const TV& v,const int i,const int j,bool column_aligned);
    TV Get_Matrix_Vector_Element(SPARSE_MATRIX_FLAT_MXN<T>& A,const int i,const int j,bool column_aligned);

    int Neighbor_Element_Number_With_Dirichlet_Boundary_Conditions(const ARRAY<int>& nb_e,const int type);

    static void Initialize_Subsystem_Helper(ARRAY<CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>* >& subsystems,
        const ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> >* psi_N=0,const ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> >* psi_D=0,const ARRAY<CODIMENSIONAL_TARGET_DIVERGENCE<T> >* tar_div=0,bool use_one_region=false);
    static void Interpolate_Pressure_To_Particles_Helper(CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>& subsystem,ARRAY<T>& u,ARRAY<T>& w);
};
}
#endif
