//#####################################################################
// Copyright 2013, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <PhysBAM_Tools/Grids_Uniform/GRID.h>
#include <PhysBAM_Tools/Grids_Uniform/UNIFORM_GRID_ITERATOR_NODE.h>
#include <PhysBAM_Tools/Grids_Uniform/UNIFORM_GRID_ITERATOR_CELL.h>
#include <PhysBAM_Tools/Read_Write/Grids_Uniform/READ_WRITE_GRID.h>
#include <PhysBAM_Tools/Read_Write/Grids_Uniform_Arrays/READ_WRITE_ARRAYS.h>
#include <PhysBAM_Tools/Read_Write/Grids_Uniform_Arrays/READ_WRITE_FACE_ARRAYS.h>
#include <PhysBAM_Tools/Read_Write/Octave/OCTAVE_OUTPUT.h>
#include <PhysBAM_Tools/Math_Tools/clamp.h>
#include <PhysBAM_Tools/Matrices/ROTATION.h>
#include <PhysBAM_Tools/Random_Numbers/NOISE.h>
#include <PhysBAM_Tools/Interpolation/BSPLINE.h>
#include <PhysBAM_Dynamics/Geometry/GENERAL_GEOMETRY_FORWARD.h>
#include <PhysBAM_Dynamics/Particles/PARTICLES_FORWARD.h>
#include <PhysBAM_Dynamics/Read_Write/Particles/READ_WRITE_PARTICLES.h>
#include <PhysBAM_Geometry/Collisions/COLLISION_GEOMETRY_COLLECTION.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_TRIANGULATED_AREA.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_TRIANGULATED_SURFACE.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_STRUCTURE.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_RIGID_GEOMETRY_COLLECTION.h>
#include <PhysBAM_Geometry/Tessellation/TORUS_TESSELLATION.h>
#include <PhysBAM_Geometry/Tessellation/SPHERE_TESSELLATION.h>
#include <PhysBAM_Geometry/Basic_Geometry/TORUS.h>
#include <PhysBAM_Geometry/Basic_Geometry/TETRAHEDRON.h>
#include <PhysBAM_Solids/PhysBAM_Rigids/Forces_And_Torques/RIGIDS_EXAMPLE_FORCES_AND_VELOCITIES.h>
#include "CODIMENSIONAL_INCOMPRESSIBLE_DRIVER.h"
#include "CODIMENSIONAL_HELPER.h"

using namespace PhysBAM;
using namespace Codimensional;

enum CODIMENSIONAL_INCOMPRESSIBLE_TESTS_ID
{
    ////examples
    fluid_polygon=1,
    water_bell=2,
    milk_crown=3,
    paint=4,
    fishbone=5,
    bubble_break=6,
    bubble_blow=7,
    membrane_square=8,     ////square
    membrane_circle=9,    ////circle
    membrane_oscillation=10,

    ////simple surface tension
    test_surface_tension_ellipsoid=301,
    test_surface_tension_cube=302,
    test_surface_tension_plane=303,
    test_surface_tension_vol_plane=304,
    
    ////simple thin film
    test_thin_film_rectangle=20,
    test_thin_film_rectangle_vol=21,
    test_thin_film_circle_oscillating=22,
    test_thin_film_circle_gravity=23,
    test_thin_film_free_rectangle=24,
    test_thin_film_free_circle=25,

    ////bubble
    test_bubble_single=50,
    test_bubble_blowing=51,
    test_bubble_dumbbell_split=52,

    ////milk crown
    test_milk_crown_hydrostatic=80,
    test_milk_crown=81,
    test_milk_crown_2=82,
    test_milk_crown_3=83,
    test_milk_crown_4=84,
    test_milk_crown_5=85,

    ////fluid polygon
    test_falling_sphere=101,
    test_falling_tet=102,
    test_colliding_spheres=103,
    test_solid_adhesion=104,
    test_fluid_polygon=105,
    test_water_bell=106,

    ////thin sheet on solid surface
    test_thin_sheet=120,
    test_thin_sheet_2=121,
    test_thin_sheet_3=122,

    ////fish bone
    test_impinging_jet_no_gravity=150,
    test_fishbone=151,

    ////water pipe
    test_water_pipe_repair=200,
    test_water_pipe_reseed=201,
    test_water_pipe_reseed_low_res=202,
    test_water_ring_repair=203,
    test_water_ring_repair2=204,

    ////jets
    test_1d_jet = 300
};

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_DRIVER<TV>::Initialize_Data()
{
    global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1));
    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles=example.particles;

    ARRAY<PAIR<TV,TV> > X_and_V;ARRAY<T> R;ARRAY<T> mass;ARRAY<bool> boundary;
    switch(test){
    //////////////////////////////////////////////////////////////////////////
    ////examples
    case fluid_polygon:
        {
            length_scale=(T).002*length_scale_factor;thickness_scale=length_scale*(T).1;
            T box_size=(T).02;T box_top_plane_height=(T)0;

            ////pipe
            T initial_speed=(T)9;
            if(reset_initial_speed_value!=(T)0)initial_speed=reset_initial_speed_value;

            TV pipe_e1=TV::Axis_Vector(2)*((T).02);TV pipe_e2=pipe_e1+TV::Axis_Vector(2)*((T)-.01);T pipe_radius=(T).004;
            water_sources.Append(new WATER_PIPE<TV>(particles,pipe_e1,pipe_e2,pipe_radius,length_scale,initial_speed));

            ////rigid
            RIGID_BODY<TV>* rigid_body=Get_Cylinder_Rigid_Body(rigid_standard_tests,TV::Axis_Vector(1)*((T)half_pi),TV::Axis_Vector(2)*(box_top_plane_height-box_size*(T).5),box_size,/*(T).5**/box_size);
            rigid_bodies.Append(rigid_body);
            has_solid=true;
            mesher.Add_Solid(*rigid_body);

            has_fixed_boundary_condition=true;
            has_gravity=true;
            has_surface_tension=true;
            surface_tension_coefficient*=(T)10;
            surface_tension_coefficient*=surface_tension_rescale_factor;
            
            gravity*=gravity_factor;

            has_viscosity_force=true;
            use_implicit_viscosity_force=true;
            kinematic_viscosity_coefficient*=(T)100;
            kinematic_viscosity_coefficient*=kinematic_viscosity_rescale_factor;

            mesher.use_topology_repairing=true;
            mesher.use_crumple_merge=false;
            mesher.use_break_thin_segments=true;
            mesher.use_small_region_remeshing=false;
            mesher.use_low_dimensional_incompressibility=true;
            mesher.use_distance_to_rim_for_incompressibility=true;
            mesher.min_distance_to_rim_for_incompressibility=5;
            //mesher.use_crumple_based_dimension_lifting=false;
            mesher.use_rim_adaptivity=true;
            if(reset_use_adaptivity!=0)mesher.use_rim_adaptivity=(reset_use_adaptivity>0);
            mesher.use_smooth_rim_thickness=true;
            mesher.smooth_rim_thickness_weight*=smooth_rim_thickness_rescale_factor;
            mesher.smooth_rim_thickness_mode=smooth_rim_thickness_mode;
            mesher.use_smooth_interior_thickness=true;
            mesher.smooth_interior_thickness_weight*=smooth_interior_thickness_rescale_factor;

            if(!set_subdivision_from_args)subdivision=TV_Compatible<int,TV::m>(TV_INT3(1,1,1));
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-.02),TV::Constant_Vector((T).02));

            frame_rate=20000;
            last_frame=1000;

            cfl=(T).5;
            if(reset_cfl_value!=(T)0)cfl=reset_cfl_value;

            //write_substep=false;
            //cfl=(T).25;
            //restart_frame=30;

            //restart_frame=522;
        }break;
    case water_bell:
        {
            length_scale=(T).002*length_scale_factor;thickness_scale=length_scale*(T).1;
            T box_size=(T).01;T box_top_plane_height=(T)0;

            ////pipe
            TV pipe_e1=TV::Axis_Vector(2)*((T).02);TV pipe_e2=pipe_e1+TV::Axis_Vector(2)*((T)-.01);
            T initial_pipe_speed=(T)5;T pipe_radius=(T).004;
            water_sources.Append(new WATER_PIPE<TV>(particles,pipe_e1,pipe_e2,pipe_radius,length_scale,initial_pipe_speed));

            ////tank
            T tank_width=(T).5;T tank_height=(T).01;
            T tank_base_y=(T)-.167273;
            TV tank_size=TV::Constant_Vector(tank_width);tank_size.y=tank_height;
            TV tank_min_corner=TV::Constant_Vector(-(T).5*tank_width);tank_min_corner.y=tank_base_y;
            tank_box=BOX<TV>(tank_min_corner,tank_min_corner+tank_size);
            TV tank_center=tank_box.Center();
            tank=&rigid_standard_tests.Add_Analytic_Box(tank_size);
            tank->X()=tank_center;
            has_tank=true;
            is_open_tank=true;
            mesher.Initialize_Tank(tank_box,is_open_tank);

            //has_water_pipe=true;
            use_fix_pillar_normal=true;
            has_fixed_boundary_condition=true;
            has_gravity=true;
            has_surface_tension=true;
            surface_tension_coefficient*=(T)2;
            surface_tension_coefficient*=surface_tension_rescale_factor;
            //has_viscosity_force=false;
            //kinematic_viscosity_coefficient*=(T)20;
            //kinematic_viscosity_coefficient*=kinematic_viscosity_rescale_factor;

            ////threading
            if(!set_subdivision_from_args)subdivision=TV_Compatible<int,TV::m>(TV_INT3(3,1,3));
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-.075),TV::Constant_Vector((T).075));
            ////threaded io for reading rigid body
            has_solid=true;
            if(has_solid){
                LOG::Initialize_Logging(false,false,1<<30,true,subdivision.Product());
                LOG::Instance()->Copy_Log_To_File(output_directory+"/common/log",restart_frame!=0);
                threaded_io_initialized=true;
            }

            ////mesher
            mesher.use_topology_repairing=true;
            mesher.use_low_dimensional_incompressibility=true;
            mesher.use_distance_to_rim_for_incompressibility=true;
            //mesher.use_adaptivity=true;
            //mesher.use_small_region_remeshing=true;

            //use_initialize_from_sim_data=true;
            //init_sim_frame=480;
            //init_sim_data_directory="../wb_s1_st2/output3";

            frame_rate=10000;
            last_frame=10000;
        }break;
    case milk_crown:
        {
            length_scale=(T).004*length_scale_factor;
            thickness_scale=length_scale*(T).1;

            T tank_width=(T).3;T tank_height=(T).1;
            T water_height=(T).06;
            T water_sphere_radius=(T).012;T water_sphere_height=water_height+water_sphere_radius+(T).5*length_scale;

            T initial_speed=(T)10;
            if(reset_initial_speed_value!=(T)0)initial_speed=reset_initial_speed_value;

            ////tank
            TV tank_size=TV::Constant_Vector(tank_width);tank_size.y=tank_height;
            TV tank_min_corner=TV::Constant_Vector(-(T).5*tank_width);tank_min_corner.y=(T)0;
            tank_box=BOX<TV>(tank_min_corner,tank_min_corner+tank_size);tank_thickness=(T)2*length_scale;
            TV tank_center=tank_box.Center();
            tank=&rigid_standard_tests.Add_Analytic_Box(tank_size);
            tank->X()=tank_center;
            has_tank=true;
            is_open_tank=true;
            
            ////water bulk
            TV water_box_size=TV::Constant_Vector(tank_width);water_box_size.y=water_height;
            BOX<TV> water_box=BOX<TV>(tank_min_corner,tank_min_corner+water_box_size);
            TV_INT particle_counts=TV_INT(ceil(water_box_size/(length_scale*(T).8)));
            T perturb_coef=(T).2;
            ARRAY<TV> X;example.Add_Box_Particles(X,boundary,water_box,particle_counts,perturb_coef);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}
            mesher.Initialize_Tank(tank_box,tank_thickness,is_open_tank);
            
            //water sphere
            TV water_sphere_center=tank_center;water_sphere_center.y=water_sphere_height;
            SPHERE<TV> water_sphere=SPHERE<TV>(water_sphere_center,water_sphere_radius);
            TV_INT sphere_particle_counts=TV_INT(ceil(water_sphere.Bounding_Box().Edge_Lengths()/(length_scale*(T).8)));
            X.Remove_All();example.Add_Sphere_Particles(X,boundary,water_sphere,sphere_particle_counts);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV::Axis_Vector(2)*(-initial_speed)));}

            ////forces
            has_gravity=true;
            has_surface_tension=true;
            surface_tension_coefficient*=(T)1;
            surface_tension_coefficient*=surface_tension_rescale_factor;
            has_adhesion_force=true;
            has_viscosity_force=true;
            use_implicit_viscosity_force=true;
            kinematic_viscosity_coefficient*=(T)20;
            kinematic_viscosity_coefficient*=kinematic_viscosity_rescale_factor;

            mesher.use_topology_repairing=true;
            mesher.use_crumple_merge=false;
            if(use_crumple_merge)mesher.use_crumple_merge=true;

            mesher.use_break_thin_segments=true;
            segment_break_thickness=(T).5*thickness_scale;
            mesher.use_small_region_remeshing=false;
            mesher.use_low_dimensional_incompressibility=true;
            mesher.use_distance_to_rim_for_incompressibility=true;
            mesher.min_distance_to_rim_for_incompressibility=5;
            mesher.use_crumple_based_dimension_lifting=false;
            mesher.use_rim_adaptivity=true;
            mesher.use_smooth_rim_thickness=true;
            mesher.smooth_rim_thickness_weight*=smooth_rim_thickness_rescale_factor;
            mesher.use_smooth_interior_thickness=true;
            mesher.smooth_interior_thickness_weight*=smooth_interior_thickness_rescale_factor;

            if(initial_speed==14){
                if(length_scale==(T).008){
                    use_initialize_from_sim_data=true;
                    init_sim_frame=4;
                    init_sim_data_directory="../init_sim_data/mc_ls008_v14";
                }
                else if(length_scale==(T).004){
                    use_initialize_from_sim_data=true;
                    init_sim_frame=15;
                    init_sim_data_directory="../init_sim_data/mc_ls004_v14";
                }
                else{
                    PHYSBAM_FATAL_ERROR("milkcrown: invalid length scale");
                }
            }
            else if(initial_speed==10){
                if(length_scale==(T).008){
                    use_initialize_from_sim_data=true;
                    init_sim_frame=3;
                    init_sim_data_directory="../init_sim_data/mc_ls008_v10";
                }
                else if(length_scale==(T).004){
                    use_initialize_from_sim_data=true;
                    init_sim_frame=15;
                    init_sim_data_directory="../init_sim_data/mc_ls004_v10";
                }
                else{
                    PHYSBAM_FATAL_ERROR("milkcrown: invalid length scale");
                }
            }
            else{
                PHYSBAM_FATAL_ERROR("milkcrown: invalid initial speed");
            }

            LOG::cout<<"smooth rim thickness weight: "<<mesher.smooth_rim_thickness_weight<<std::endl;
            LOG::cout<<"surface tension coefficient: "<<surface_tension_coefficient<<std::endl;
            LOG::cout<<"viscosity coefficient: "<<kinematic_viscosity_coefficient<<std::endl;
            LOG::cout<<"initial speed: "<<initial_speed<<std::endl;

            if(!set_subdivision_from_args)subdivision=TV_Compatible<int,TV::m>(TV_INT3(1,1,1));
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-.16),TV::Constant_Vector((T).16));
        }break;
    case paint:
        {
            ////parameters
            length_scale=(T).008*length_scale_factor;
            thickness_scale=length_scale*(T).1;
            T tank_width=(T).66;T tank_height=(T).22;
            T water_height=(T).07;
            T half_rock_size_y=(T)0.0381;
            //T half_rock_size_y=(T).02;
            T rock_height=water_height+half_rock_size_y;

            ////tank
            TV tank_size=TV::Constant_Vector(tank_width);tank_size.y=tank_height;
            TV tank_min_corner=TV::Constant_Vector(-(T).5*tank_width);tank_min_corner.y=(T)0;
            tank_box=BOX<TV>(tank_min_corner,tank_min_corner+tank_size);tank_thickness=(T)2*length_scale;
            TV tank_center=tank_box.Center();
            tank=&rigid_standard_tests.Add_Analytic_Box(tank_size);
            tank->X()=tank_center;
            has_tank=true;
            is_open_tank=true;

            ////water bulk
            TV water_box_size=TV::Constant_Vector(tank_width);water_box_size.y=water_height;
            BOX<TV> water_box=BOX<TV>(tank_min_corner,tank_min_corner+water_box_size);
            TV_INT particle_counts=TV_INT(ceil(water_box_size/(length_scale*(T).8)));
            T perturb_coef=(T).2;
            ARRAY<TV> X;example.Add_Box_Particles(X,boundary,water_box,particle_counts,perturb_coef);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}

            ////cfl
            //cfl=(T).2;
            //write_substep=true;

            ////threading
            if(!set_subdivision_from_args)subdivision=TV_Compatible<int,TV::m>(TV_INT3(3,1,3));
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-.33),TV::Constant_Vector((T).33));
            ////threaded io for reading rigid body
            has_solid=true;
            if(has_solid){
                LOG::Initialize_Logging(false,false,1<<30,true,subdivision.Product());
                LOG::Instance()->Copy_Log_To_File(output_directory+"/common/log",restart_frame!=0);
                threaded_io_initialized=true;
            }

            ////rigid body
            STREAM_TYPE float_stream_type(1.0f);
            RIGID_BODY<TV>* rock=&rigid_standard_tests.Add_Rigid_Body("rock_3x5x3",(T)1./*(T)1.2*/,(T)0.0,true,true/*false*/,&float_stream_type);
            //RIGID_BODY<TV>* rock=&rigid_standard_tests.Add_Analytic_Sphere((T).02, (T)1.0);
            rigid_bodies.Append(rock);
            mesher.Add_Solid(*rock);
            if(use_rock_key_frames){
                Initialize_Rock_Key_Frames();}
            else{
                TV rock_center=tank_center;rock_center.y=rock_height;
                rock->X()=rock_center;
                rock->V()=TV::Axis_Vector(2)*initial_rock_speed;}

            ////options for physical forces
            has_gravity=true;
            has_surface_tension=true;
            has_adhesion_force=true;
            surface_tension_coefficient=(T).025;
            surface_tension_coefficient*=surface_tension_rescale_factor;

            ////mesher
            mesher.Initialize_Tank(tank_box,tank_thickness,is_open_tank);
            mesher.use_topology_repairing=true;
            mesher.use_crumple_merge=true;
            //mesher.use_break_thin_segments=true;
            //mesher.use_small_region_remeshing=true;
            mesher.use_low_dimensional_incompressibility=true;
            mesher.use_distance_to_rim_for_incompressibility=true;
            mesher.min_distance_to_rim_for_incompressibility=5;
            mesher.use_crumple_based_dimension_lifting=false;
            mesher.use_rim_adaptivity=true;
        }break;
    case fishbone:
        {
            length_scale=(T)4e-4*length_scale_factor;
            thickness_scale=length_scale*(T).1;
            T pipe_radius=(T)8e-4;

            //T initial_speed=(T)12;
            T initial_speed=(T)10;
            if(reset_initial_speed_value!=(T)0)initial_speed=reset_initial_speed_value;

            //T theta=(T)(one_third*pi);
            T theta=(T)half_pi;
            
            T distance=(T)100e-4;
            T pipe_length=length_scale;
            TV x_normal=TV::Axis_Vector(1);
            TV3 y_normal=TV3::Axis_Vector(2);
            TV p1_end=(T)-0.5*distance*x_normal;
            TV p2_end=(T)0.5*distance*x_normal;

            ROTATION<TV3> rotation_about_z((T)0.5*theta,TV3::Axis_Vector(3));
            TV p1_begin=TV_Compatible<T,TV::m>(rotation_about_z.Rotate(y_normal));
            p1_begin-=(T)0.5*distance*x_normal;
            TV direction=(p1_begin-p1_end).Normalized();
            p1_begin=p1_end+pipe_length*direction;

            ROTATION<TV3> neg_rotation_about_z((T)-0.5*theta,TV3::Axis_Vector(3));
            TV p2_begin=TV_Compatible<T,TV::m>(neg_rotation_about_z.Rotate(y_normal));
            p2_begin+=(T)0.5*distance*x_normal;
            direction=(p2_begin-p2_end).Normalized();
            p2_begin=p2_end+pipe_length*direction;

            water_sources.Append(new WATER_PIPE<TV>(particles,p1_begin,p1_end,pipe_radius,length_scale,initial_speed));
            water_sources.Append(new WATER_PIPE<TV>(particles,p2_begin,p2_end,pipe_radius,length_scale,initial_speed));

            has_gravity=true;
            gravity*=gravity_factor;

            has_surface_tension=true;
            surface_tension_coefficient*=(T)1;
            surface_tension_coefficient*=surface_tension_rescale_factor;

            has_viscosity_force=true;
            use_implicit_viscosity_force=true;
            kinematic_viscosity_coefficient*=(T)20;
            kinematic_viscosity_coefficient*=kinematic_viscosity_rescale_factor;

            triangle_incompressible_thickness=thickness_scale*(T)1;
            triangle_incompressible_thickness*=triangle_incompressible_thickness_factor;

            mesher.use_topology_repairing=true;
            mesher.use_crumple_merge=true;
            mesher.use_break_thin_segments=true;
            mesher.use_small_region_remeshing=false;
            mesher.use_low_dimensional_incompressibility=true;
            mesher.use_distance_to_rim_for_incompressibility=true;
            mesher.min_distance_to_rim_for_incompressibility=5;
            mesher.use_topology_repairing=true;
            mesher.use_rim_adaptivity=true;
            if(reset_use_adaptivity!=0)mesher.use_rim_adaptivity=(reset_use_adaptivity>0);
            mesher.use_smooth_rim_thickness=true;   
            mesher.smooth_rim_thickness_weight=(T).02;
            mesher.smooth_rim_thickness_weight*=smooth_rim_thickness_rescale_factor;
            mesher.use_smooth_interior_thickness=true;
            mesher.smooth_interior_thickness_weight=(T).02;
            mesher.smooth_interior_thickness_weight*=smooth_rim_thickness_rescale_factor;

            LOG::cout<<"smooth rim thickness weight: "<<mesher.smooth_rim_thickness_weight<<std::endl;
            LOG::cout<<"surface tension coefficient: "<<surface_tension_coefficient<<std::endl;
            LOG::cout<<"viscosity coefficient: "<<kinematic_viscosity_coefficient<<std::endl;

            fishbone_center_point=TV::Axis_Vector(2)*(T)-.005;

            if(initial_speed==(T)6){
                use_initialize_from_sim_data=true;
                if(length_scale==(T)4e-4){
                    init_sim_frame=18;
                    init_sim_data_directory="../init_sim_data/fb_tet_a45_v6_ls0004";
                }
                else if(length_scale==(T)2e-4){
                    init_sim_frame=21;
                    init_sim_data_directory="../init_sim_data/fb_tet_a45_v6_ls0002";
                }
                else{
                    PHYSBAM_FATAL_ERROR("fishbone: invalid length scale");
                }
            }
            else if(initial_speed==(T)8){
                use_initialize_from_sim_data=true;
                if(length_scale==(T)4e-4){
                    init_sim_frame=15;
                    init_sim_data_directory="../init_sim_data/fb_tet_a45_v8_ls0004";
                }
                else if(length_scale==(T)2e-4){
                    init_sim_frame=14;
                    init_sim_data_directory="../init_sim_data/fb_tet_a45_v8_ls0002";
                }
                else{
                    PHYSBAM_FATAL_ERROR("fishbone: invalid length scale");
                }
            }
            else if(initial_speed==(T)10){
                use_initialize_from_sim_data=true;
                if(theta==(T)half_pi){
                    if(length_scale==(T)4e-4){
                        init_sim_frame=11;
                        init_sim_data_directory="../init_sim_data/fb_tet_a45_v10_ls0004";
                    }
                    else if(length_scale==(T)2e-4){
                        init_sim_frame=11;
                        init_sim_data_directory="../init_sim_data/fb_tet_a45_v10_ls0002";
                    }
                    else{
                        PHYSBAM_FATAL_ERROR("fishbone: invalid length scale");
                    }
                }
                else if(theta==(T)(one_third*pi)){
                    if(length_scale==(T)4e-4){
                        init_sim_frame=14;
                        init_sim_data_directory="../init_sim_data/fb_tet_a60_v10_ls0004";
                    }
                }
            }
            else if(initial_speed==(T)12){
                use_initialize_from_sim_data=true;
                if(theta==(T)(one_third*pi)){
                    if(length_scale==(T)4e-4){
                        init_sim_frame=13;
                        init_sim_data_directory="../init_sim_data/fb_tet_a60_v12_ls0004";
                    }
                    else{
                        PHYSBAM_FATAL_ERROR("fishbone: invalid length scale");
                    }
                }
            }
            else{
                PHYSBAM_FATAL_ERROR("fishbone: invalid initial speed");
            }

            if(!set_subdivision_from_args)subdivision=TV_Compatible<int,TV::m>(/*TV_INT3(1,4,3)*/TV_INT3(1,1,1));
            TV grid_min_corner=-TV::Constant_Vector((T).02);TV grid_max_corner=TV::Constant_Vector((T).02);
            grid_min_corner.x=(T)-.02;grid_max_corner.x=(T).02;
            grid_min_corner.y=(T)-.08;grid_max_corner.y=(T)0;
            global_grid_domain=RANGE<TV>(grid_min_corner,grid_max_corner);

            //frame_rate=64000;
            frame_rate=24000;
            last_frame=3000;

            cfl=(T)1;
            if(reset_cfl_value!=(T)0)cfl=reset_cfl_value;

            //write_substep=true;
            //restart_frame=30;
        }break;
    case bubble_break:
        {
            TRIANGULATED_SURFACE<T> triangulated_surface(*(new TRIANGLE_MESH()),*(new GEOMETRY_PARTICLES<TV3>()));
            FILE_UTILITIES::Read_From_File(stream_type,"../init_sim_data/bubble3.tri",triangulated_surface);
            external_mesh.triangle_mesh.Initialize_Mesh(triangulated_surface.mesh);
            external_mesh.Update_States(triangulated_surface.particles.X.m);

            ////move mesh to the center
            TV center=TV();for(int i=1;i<=triangulated_surface.particles.X.m;i++){center+=triangulated_surface.particles.X(i);}
            center/=(T)(triangulated_surface.particles.X.m);
            for(int i=1;i<=triangulated_surface.particles.X.m;i++){triangulated_surface.particles.X(i)-=center;}

            ////rescale mesh according to bubble size
            BOX<TV> bounding_box=BOX<TV>(TV((T)FLT_MAX),TV((T)-FLT_MAX));
            for(int i=1;i<=triangulated_surface.particles.X.m;i++){
                for(int d=1;d<=TV::m;d++){
                    if(triangulated_surface.particles.X(i)(d)<bounding_box.min_corner(d))bounding_box.min_corner(d)=triangulated_surface.particles.X(i)(d);
                    else if(triangulated_surface.particles.X(i)(d)>bounding_box.max_corner(d))bounding_box.max_corner(d)=triangulated_surface.particles.X(i)(d);
                }
            }
            T max_length=bounding_box.Edge_Lengths().Max();
            T bubble_size=(T).1;T rescale_factor=bubble_size/max_length;

            ////rescale the mesh
            LOG::cout<<"rescale factor: "<<rescale_factor<<std::endl;
            for(int i=1;i<=triangulated_surface.particles.X.m;i++){triangulated_surface.particles.X(i)*=rescale_factor;}

            ////translate the mesh
            TV translated_center=(TV::Axis_Vector(2)*(T).06+TV::Axis_Vector(1)*(T).02);
            for(int i=1;i<=triangulated_surface.particles.X.m;i++){triangulated_surface.particles.X(i)+=translated_center;}

            T averaged_length=(T)0;int n=0;
            for(int i=1;i<=triangulated_surface.mesh.elements.m;i++){
                const TV_INT3& v=triangulated_surface.mesh.elements(i);
                ARRAY<TV_INT2> edges;Simplex_Edges(v,edges);
                for(int j=1;j<=3;j++){averaged_length+=(triangulated_surface.particles.X(edges(j)(1))-triangulated_surface.particles.X(edges(j)(2))).Magnitude();n++;}
            }
            averaged_length/=(T)n;

            ////stretch the mesh
            for(int i=1;i<=triangulated_surface.particles.X.m;i++){triangulated_surface.particles.X(i).x*=(T)1.1;}

            for(int i=1;i<=triangulated_surface.particles.X.m;i++){
                TV initial_v=TV::Axis_Vector(2)*(T)-0.5;initial_v+=TV::Axis_Vector(1)*(T)-.002;
                X_and_V.Append(PAIR<TV,TV>(TV_Compatible<T,TV::m>(triangulated_surface.particles.X(i)),initial_v));
            }

            length_scale=averaged_length*(T)1.1;thickness_scale=(T)1e-6;
            LOG::cout<<"bubble break length scale: "<<length_scale<<", "<<thickness_scale<<std::endl;

            use_external_mesh=true;
            
            has_bubble=true;
            use_bubble_breaking=true;
            use_constant_thin_film_thickness=true;
            has_surface_tension=true;
            surface_tension_coefficient*=(T)1;

            has_gravity=true;
            gravity*=(T).3;
            gravity*=gravity_factor;

            ////threading
            if(!set_subdivision_from_args)subdivision=TV_Compatible<int,TV::m>(TV_INT3(1,1,1));
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-.2),TV::Constant_Vector((T).2));
            ////threaded io for reading rigid body
            has_solid=true;
            if(has_solid){
                LOG::Initialize_Logging(false,false,1<<30,true,subdivision.Product());
                LOG::Instance()->Copy_Log_To_File(output_directory+"/common/log",restart_frame!=0);
                threaded_io_initialized=true;
            }

            ////rigid body
            STREAM_TYPE float_stream_type(1.0f);
            //RIGID_BODY<TV>* bullet=&rigid_standard_tests.Add_Rigid_Body("bullet",(T).001,(T)0.0,true,true/*false*/,&float_stream_type);
            RIGID_BODY<TV>* bullet=&rigid_standard_tests.Add_Analytic_Sphere((T).001,(T)1);
            //TV bullet_initial_pos=TV((T)-.1,(T).06,(T)0);     ////initialized in update_solid_states
            //bullet->X()=bullet_initial_pos;
            //bullet->V()=TV::Axis_Vector(1)*(T)5;
            //T rotation_angle=(T)-half_pi;
            //TV rotation_spin=TV::Axis_Vector(3);
            //bullet->Rotation()=ROTATION<TV>::From_Rotation_Vector(rotation_angle*rotation_spin)*bullet->Rotation();
            rigid_bodies.Append(bullet);  
            mesher.Add_Solid(*bullet);

            ////mesher settings
            mesher.use_topology_repairing=true;
            //mesher.use_crumple_based_dimension_lifting=true;
            mesher.use_align_triangle_normals=false;
            mesher.use_rim_mass_redistribution=false;
            mesher.use_break_thin_segments=true;
            //segment_break_thickness=(T)1e-5;
            segment_break_thickness=(T)4e-5;
            mesher.use_break_thin_triangles=true;
            triangle_break_thickness=thickness_scale*(T)1e-5;
            mesher.use_break_elements_in_solids=true;
            mesher.use_low_dimensional_incompressibility=true;
            triangle_incompressible_thickness=(T)2*thickness_scale;
            mesher.use_distance_to_rim_for_incompressibility=true;
            mesher.min_distance_to_rim_for_incompressibility=5;
            mesher.use_smooth_rim_thickness=false;
            mesher.use_smooth_interior_thickness=false;
            mesher.use_rim_adaptivity=true;
            mesher.initialize_tetrahedron_mesh=false;
            mesher.initialize_segment_mesh=false;
            mesher.use_high_res_volume_mesh=true;
            mesher.use_high_res_segment_mesh=true;
            mesher.use_small_region_remeshing=false;
            
            initial_particle_dim=TV::m-1;

            use_initialize_from_sim_data=true;
            init_sim_frame=110;
            init_sim_data_directory="../init_sim_data/bbreak";

            frame_rate=40000;
            last_frame=5000;
            //frame_rate=4000;
            //last_frame=4000;
            cfl=(T).1;
            if(reset_cfl_value!=(T)0)cfl=reset_cfl_value;
        }break;
    case bubble_blow:
        {
            length_scale=(T).001*length_scale_factor;thickness_scale=(T)1e-5;
            ARRAY<TV> X;
            T radius=(T).008;
            TV radii=TV_Compatible<T,TV::m>(radius,radius,length_scale);
            BOX<TV> box=BOX<TV>(-radii*(T)1.2,radii*(T)1.2);
            SPHERE<TV2> circle(TV2(),radii.x);
            example.Add_Circle_Particles(X,TV(),circle.radius,TV::Axis_Vector(3),length_scale);
            fixed_boundary_circle=circle;has_fixed_boundary_condition=true;
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}

            //TORUS<T> torus(VECTOR<T,3>(),VECTOR<T,3>::Axis_Vector(3),length_scale*(T).25,radii.x);
            //TRIANGULATED_SURFACE<T>* torus_triangulated_surface=TESSELLATION::Generate_Triangles(torus,10,10);
            //STREAM_TYPE float_stream_type(1.0f);
            //FILE_UTILITIES::Write_To_File(float_stream_type,output_directory+"/common/torus_bubble_blow3.tri",*torus_triangulated_surface);

            has_bubble=true;
            use_constant_thin_film_thickness=true;
            has_surface_tension=true;
            surface_tension_coefficient*=(T).2;
            surface_tension_coefficient*=surface_tension_rescale_factor;

            ////mesher settings
            mesher.use_topology_repairing=true;
            //mesher.use_crumple_based_dimension_lifting=true;
            mesher.use_align_triangle_normals=false;
            mesher.use_rim_mass_redistribution=false;
            mesher.use_break_thin_segments=true;
            segment_break_thickness=thickness_scale*(T).5;
            mesher.use_low_dimensional_incompressibility=true;
            triangle_incompressible_thickness=(T)4*thickness_scale;
            //mesher.use_distance_to_rim_for_incompressibility=true;
            //mesher.min_distance_to_rim_for_incompressibility=5;
            mesher.initialize_tetrahedron_mesh=false;
            mesher.initialize_segment_mesh=false;
            mesher.use_high_res_volume_mesh=true;
            mesher.use_high_res_segment_mesh=true;
            initial_particle_dim=TV::m-1;

            //use_initialize_from_sim_data=true;
            //init_sim_frame=225;
            //init_sim_data_directory="../init_sim_data/bubble_break";

            frame_rate=2000;
            last_frame=10000;
            cfl=(T)1;
            if(reset_cfl_value!=(T)0)cfl=reset_cfl_value;

            ////wind
            has_wind=true;
            wind_speed=(T)2;
            wind_speed*=wind_speed_factor;
            wind_field.solid_force_rescale=(T)4e7;
            wind_field.solid_force_rescale*=bubble_affecting_wind_factor;
            wind_field.pressure_rescale=(T)50;
            wind_field.pressure_rescale*=wind_affecting_bubble_factor;
            wind_field.pressure_rescale_2=(T).75;
            wind_field.pressure_rescale_2*=wind_affecting_bubble_factor_2;
            wind_field.collision_rescale*=bubble_collision_factor;
            wind_field.vorticity_confinement_coef=(T)1;
            wind_field.vorticity_confinement_coef*=vorticity_confinement_factor;
            wind_field.extra_wind_speed*=extra_wind_speed_factor;
            LOG::cout<<"wind speed: "<<wind_speed<<", wind force: "<<wind_field.solid_force_rescale<<std::endl;
            
            TV wind_box_min_corner=TV::Constant_Vector(-radius*(T)2);wind_box_min_corner.x=-radius*(T)2;wind_box_min_corner.y=-radius*(T)2;
            TV wind_box_max_corner=TV::Constant_Vector(radius*(T)16);wind_box_max_corner.x=radius*(T)2;wind_box_max_corner.y=radius*(T)2;
            BOX<TV> wind_box(wind_box_min_corner,wind_box_max_corner);
            ARRAY<T> control_time;ARRAY<T> control_speed;
            ARRAY<PAIR<int,T> > control_frame_speed;

            //timeline 1
            Read_Wind_Speed_From_File(wind_field_directory,control_frame_speed);
            for(int i=1;i<=control_frame_speed.m;i++){
                LOG::cout<<"wind speed: "<<control_frame_speed(i).x<<", "<<control_frame_speed(i).y<<std::endl;
            }

            //control_frame_speed.Append(PAIR<int,T>(0,(T).8));
            //control_frame_speed.Append(PAIR<int,T>(100,(T).9));
            //control_frame_speed.Append(PAIR<int,T>(150,(T).5));
            //control_frame_speed.Append(PAIR<int,T>(180,(T).5));
            //control_frame_speed.Append(PAIR<int,T>(200,(T).5));
            //control_frame_speed.Append(PAIR<int,T>(250,(T)1.2));
            //control_frame_speed.Append(PAIR<int,T>(300,(T).5));

            //control_frame_speed.Append(PAIR<int,T>(350,(T).5));
            //control_frame_speed.Append(PAIR<int,T>(400,(T).9));
            //control_frame_speed.Append(PAIR<int,T>(450,(T)1.2));
            //control_frame_speed.Append(PAIR<int,T>(480,(T).2));

            //control_frame_speed.Append(PAIR<int,T>(500,(T).2));
            //control_frame_speed.Append(PAIR<int,T>(550,(T).25));
            //control_frame_speed.Append(PAIR<int,T>(600,(T).3));

            //control_frame_speed.Append(PAIR<int,T>(650,(T).5));
            //control_frame_speed.Append(PAIR<int,T>(700,(T).8));
            //control_frame_speed.Append(PAIR<int,T>(720,(T).7));
            //control_frame_speed.Append(PAIR<int,T>(750,(T).9));
            //control_frame_speed.Append(PAIR<int,T>(800,(T).5));
            //control_frame_speed.Append(PAIR<int,T>(830,(T)1.2));
            //control_frame_speed.Append(PAIR<int,T>(850,(T).2));
            //control_frame_speed.Append(PAIR<int,T>(900,(T)1.2));
            //control_frame_speed.Append(PAIR<int,T>(950,(T).5));

            //control_frame_speed.Append(PAIR<int,T>(650,(T).5));
            //control_frame_speed.Append(PAIR<int,T>(700,(T).8));
            //control_frame_speed.Append(PAIR<int,T>(720,(T).7));
            //control_frame_speed.Append(PAIR<int,T>(750,(T).9));
            //control_frame_speed.Append(PAIR<int,T>(800,(T).5));
            //control_frame_speed.Append(PAIR<int,T>(830,(T)1.2));
            //control_frame_speed.Append(PAIR<int,T>(850,(T).2));
            //control_frame_speed.Append(PAIR<int,T>(900,(T)1.2));
            //control_frame_speed.Append(PAIR<int,T>(950,(T).5));

            ////timeline 2
            for(int i=1;i<=control_frame_speed.m;i++){
                control_time.Append((T)control_frame_speed(i).x/(T)frame_rate);
                control_speed.Append(control_frame_speed(i).y);
            }
            wind_speed_bspline=new BSPLINE<T,T>(control_time,control_speed,1);

            //for(int i=1;i<=100;i++){
            //    T t=(T)i/(T)300;
            //    LOG::cout<<"t "<<t<<": "<<wind_speed_bspline->Evaluate(t)<<std::endl;
            //}

            if(!set_subdivision_from_args)subdivision=TV_Compatible<int,TV::m>(TV_INT3(1,1,1));
            global_grid_domain=RANGE<TV>(wind_box_min_corner,wind_box_max_corner);

            //restart_frame=100;
            //dbg_restart=true;
            //write_substep=true;
        }break;
    case membrane_square:
        {
            length_scale=(T).001;thickness_scale=(T)1e-5;
            ARRAY<TV> X;
            TV radii=TV::Constant_Vector((T).05);
            T height=length_scale*(T)1;
            BOX<TV> box=BOX<TV>(-radii*(T)1.2,radii*(T)1.2);
            box.min_corner.y=(T)0;box.max_corner.y=height;
            TV_INT counts=TV_INT(ceil(box.Edge_Lengths()/(length_scale*(T).98)));
            TRIANGLE_MESH triangle_mesh;
            example.Add_Square_Particles_And_Mesh(X,triangle_mesh,box.min_corner,box.Edge_Lengths().x,height,length_scale);
            //LOG::cout<<"X: "<<X<<std::endl;
            //LOG::cout<<"tris: "<<triangle_mesh.elements<<std::endl;
            external_mesh.triangle_mesh.Initialize_Mesh(triangle_mesh);
            external_mesh.Update_States(X.m);
            use_external_mesh=true;

            fixed_boundary_squares=box;
            fixed_boundary_min_height=(T)0;
            fixed_boundary_max_height=height;
            has_fixed_boundary_condition=true;
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}

            has_surface_tension=true;
            surface_tension_coefficient*=(T).5;
            surface_tension_coefficient*=surface_tension_rescale_factor;

            has_viscosity_force=true;
            use_implicit_viscosity_force=true;
            kinematic_viscosity_coefficient*=(T)10;
            kinematic_viscosity_coefficient*=kinematic_viscosity_rescale_factor;

            triangle_incompressible_thickness=thickness_scale*(T)6;
            use_constant_thin_film_thickness=true;

            ////mesher settings
            mesher.use_topology_repairing=true;
            mesher.use_break_thin_segments=true;
            mesher.use_align_triangle_normals=false;
            mesher.use_rim_mass_redistribution=false;
            //segment_break_thickness=thickness_scale*(T)2;
            segment_break_thickness=length_scale*(T)2;
            use_constant_thin_film_thickness=true;
            //mesher.use_small_region_remeshing=true;
            mesher.use_low_dimensional_incompressibility=true;
            mesher.use_distance_to_rim_for_incompressibility=true;
            mesher.min_distance_to_rim_for_incompressibility=5;
            mesher.initialize_tetrahedron_mesh=false;
            mesher.use_crumple_based_dimension_lifting=true;
            mesher.initialize_segment_mesh=false;
            mesher.use_high_res_volume_mesh=true;
            mesher.use_high_res_segment_mesh=true;
            mesher.use_elements_merging=true;
            mesher.use_elements_merging_to_triangle_mesh=true;
            initial_particle_dim=TV::m-1;

            if(!set_subdivision_from_args)subdivision=TV_Compatible<int,TV::m>(TV_INT3(1,1,1));
            BOX<TV> grid_box=box;grid_box.min_corner.y-=(T).1;grid_box.max_corner.y+=(T).1;
            global_grid_domain=RANGE<TV>(grid_box.min_corner,grid_box.max_corner);

            cfl=(T).2;
            if(reset_cfl_value!=(T)0)cfl=reset_cfl_value;

            frame_rate=2000;
            last_frame=10000;

            //write_substep=true;
            //restart_frame=1413;
            //dbg_restart=true;

            ////initialize rigid bodies
            //const T torus_minor_radius=(T)0.5*height;
            //const T torus_major_radius=radii.x+torus_minor_radius;
            //const T cylinder_height=(T)5*radii.x;
            //const T cylinder_radius=torus_minor_radius;
            //min_torus=&rigid_standard_tests.Add_Analytic_Torus(torus_minor_radius,torus_major_radius,100,200);
            //max_torus=&rigid_standard_tests.Add_Analytic_Torus(torus_minor_radius,torus_major_radius,100,200);
            //min_torus->Rotation()*=ROTATION<VECTOR<T,3> >::From_Euler_Angles((T)half_pi,(T)0,(T)0);
            //max_torus->Rotation()*=ROTATION<VECTOR<T,3> >::From_Euler_Angles((T)half_pi,(T)0,(T)0);

            //min_cylinder=&rigid_standard_tests.Add_Analytic_Cylinder(cylinder_height,cylinder_radius,100);
            //max_cylinder=&rigid_standard_tests.Add_Analytic_Cylinder(cylinder_height,cylinder_radius,100);
            //min_cylinder->X()(3)-=torus_major_radius+(T)0.5*cylinder_height;
            //max_cylinder->X()(3)-=torus_major_radius+(T)0.5*cylinder_height;
        }break;
    case membrane_circle:
        {
            length_scale=(T).001*length_scale_factor;thickness_scale=(T)1e-4;
            ARRAY<TV> X;
            TV radii=TV::Constant_Vector((T).05);
            T height=length_scale*(T)1;
            BOX<TV> box=BOX<TV>(-radii*(T)1.2,radii*(T)1.2);
            box.min_corner.y=(T)0;box.max_corner.y=radii.y*(T).1;
            TV_INT counts=TV_INT(ceil(box.Edge_Lengths()/(length_scale*(T).98)));
            example.Add_Cylinder_Particles(X,TV(),radii.x,height,length_scale);
            SPHERE<TV2> circle(TV2(),radii.x);
            fixed_boundary_circle=circle;
            fixed_boundary_min_height=(T)0;
            fixed_boundary_max_height=height;
            has_fixed_boundary_condition=true;
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}

            has_surface_tension=true;
            surface_tension_coefficient*=(T).5;
            surface_tension_coefficient*=surface_tension_rescale_factor;

            has_viscosity_force=true;
            use_implicit_viscosity_force=true;
            kinematic_viscosity_coefficient*=(T)10;
            kinematic_viscosity_coefficient*=kinematic_viscosity_rescale_factor;

            triangle_incompressible_thickness=thickness_scale*(T)5;
            use_constant_thin_film_thickness=true;

            ////mesher settings
            mesher.use_topology_repairing=true;
            mesher.use_break_thin_segments=true;
            mesher.use_align_triangle_normals=false;
            mesher.use_rim_mass_redistribution=false;
            segment_break_thickness=thickness_scale*(T)5;
            use_constant_thin_film_thickness=true;
            //mesher.use_small_region_remeshing=true;
            mesher.use_low_dimensional_incompressibility=true;
            mesher.use_distance_to_rim_for_incompressibility=true;
            mesher.min_distance_to_rim_for_incompressibility=5;
            mesher.initialize_tetrahedron_mesh=false;
            mesher.use_crumple_based_dimension_lifting=true;
            mesher.initialize_segment_mesh=false;
            mesher.use_high_res_volume_mesh=true;
            mesher.use_high_res_segment_mesh=true;
            mesher.use_elements_merging=true;
            mesher.use_elements_merging_to_triangle_mesh=true;
            initial_particle_dim=TV::m-1;

            if(!set_subdivision_from_args)subdivision=TV_Compatible<int,TV::m>(TV_INT3(1,1,1));
            BOX<TV> grid_box=box;grid_box.min_corner.y-=(T).1;grid_box.max_corner.y+=(T).1;
            global_grid_domain=RANGE<TV>(grid_box.min_corner,grid_box.max_corner);

            cfl=(T).2;
            if(reset_cfl_value!=(T)0)cfl=reset_cfl_value;

            frame_rate=2000;
            last_frame=10000;

            //write_substep=true;
            //restart_frame=1554;
            ////dbg_restart=true;

            //initialize rigid bodies
            const T torus_minor_radius=(T)0.5*height;
            const T torus_major_radius=radii.x+torus_minor_radius;
            const T cylinder_height=(T)5*radii.x;
            const T cylinder_radius=torus_minor_radius;
            min_torus=&rigid_standard_tests.Add_Analytic_Torus(torus_minor_radius,torus_major_radius,100,200);
            max_torus=&rigid_standard_tests.Add_Analytic_Torus(torus_minor_radius,torus_major_radius,100,200);
            min_torus->Rotation()*=ROTATION<VECTOR<T,3> >::From_Euler_Angles((T)half_pi,(T)0,(T)0);
            max_torus->Rotation()*=ROTATION<VECTOR<T,3> >::From_Euler_Angles((T)half_pi,(T)0,(T)0);

            min_cylinder=&rigid_standard_tests.Add_Analytic_Cylinder(cylinder_height,cylinder_radius,100);
            max_cylinder=&rigid_standard_tests.Add_Analytic_Cylinder(cylinder_height,cylinder_radius,100);
            min_cylinder->X()(3)-=torus_major_radius+(T)0.5*cylinder_height;
            max_cylinder->X()(3)-=torus_major_radius+(T)0.5*cylinder_height;

        }break;
    case membrane_oscillation:
        {
            length_scale=(T).001;thickness_scale=(T)1e-5;
            ARRAY<TV> X;
            T radius=(T).008;
            TV radii=TV_Compatible<T,TV::m>(radius,radius,length_scale);
            BOX<TV> box=BOX<TV>(-radii*(T)1.2,radii*(T)1.2);
            SPHERE<TV2> circle(TV2(),radii.x);
            example.Add_Circle_Particles(X,TV(),circle.radius,TV::Axis_Vector(3),length_scale);
            fixed_boundary_circle=circle;has_fixed_boundary_condition=true;
            
            for(int i=1;i<=X.m;i++){
                T r=sqrt(pow<2>(X(i).x)+pow<2>(X(i).y));
                T w=(circle.radius-r)/circle.radius;
                //T z=(T)10*length_scale*sin((T)half_pi*(circle.radius-r)/circle.radius);
                T z=(T)5*length_scale*w;
                X(i)=TV(X(i).x,X(i).y,z);
            }
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}

            has_surface_tension=true;
            surface_tension_coefficient*=(T).2;
            use_constant_thin_film_thickness=true;

            ////mesher settings
            mesher.use_topology_repairing=true;
            mesher.use_break_thin_segments=true;
            mesher.use_align_triangle_normals=false;
            mesher.use_rim_mass_redistribution=false;
            segment_break_thickness=thickness_scale*(T)2;
            triangle_break_thickness=thickness_scale*(T).0001;
            //mesher.use_small_region_remeshing=false;
            mesher.use_low_dimensional_incompressibility=false;
            //mesher.use_distance_to_rim_for_incompressibility=true;
            //mesher.min_distance_to_rim_for_incompressibility=5;
            //mesher.use_rim_adaptivity=true;
            mesher.initialize_tetrahedron_mesh=false;
            mesher.use_break_elements_in_solids=true;

            initial_particle_dim=TV::m-1;

            if(!set_subdivision_from_args)subdivision=TV_Compatible<int,TV::m>(TV_INT3(1,1,1));
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1));

            frame_rate=2000;
            last_frame=2000;

            cfl=(T).5;
            if(reset_cfl_value!=(T)0)cfl=reset_cfl_value;

            //restart_frame=94;
            write_substep=false;
        }break;
    //////////////////////////////////////////////////////////////////////////
    ////basic surface tension
    case test_surface_tension_ellipsoid: 
        {
            ARRAY<TV> X;
            TV radii=TV::Constant_Vector(length_scale*(T)8);radii.x*=(T)2;
            DIAGONAL_MATRIX<T,TV::m> radii_matrix=DIAGONAL_MATRIX<T,TV::m>(radii);
            T_ELLIPSOID ellipsoid(TV(),radii_matrix);
            TV_INT counts=TV_INT(ellipsoid.Bounding_Box().Edge_Lengths()/length_scale);
            example.Add_Ellipsoid_Particles(X,boundary,ellipsoid,counts);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}
            has_surface_tension=true;

            mesher.use_topology_repairing=true;
        }break;
    case test_surface_tension_cube:
        {
            length_scale=(T).0005;
            ARRAY<TV> X;
            TV radii=TV::Constant_Vector((T).004);
            BOX<TV> box=BOX<TV>(-radii,radii);
            TV_INT counts=TV_INT(box.Edge_Lengths()/length_scale);
            example.Add_Box_Particles(X,boundary,box,counts);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}
            has_surface_tension=true;

            //mesher.use_topology_repairing=true;
        }break;
    case test_surface_tension_plane:
        {
            ARRAY<TV> X;
            TV radii=TV::Constant_Vector(length_scale*(T)2/length_scale_factor);
            BOX<TV> box=BOX<TV>(-radii,radii);  
            TV_INT counts=TV_INT(box.Edge_Lengths()/length_scale);counts.y=1;
            example.Add_Box_Particles(X,boundary,box,counts);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}
            has_surface_tension=true;
        }break;
    case test_surface_tension_vol_plane:
        {
            ARRAY<TV> X;
            TV radii=TV::Constant_Vector(length_scale*(T)2/length_scale_factor);radii.y=length_scale;
            BOX<TV> box=BOX<TV>(-radii,radii);  
            TV_INT counts=TV_INT(box.Edge_Lengths()/length_scale);
            example.Add_Box_Particles(X,boundary,box,counts);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}
            has_surface_tension=true;
        }break;

    //////////////////////////////////////////////////////////////////////////
    ////thin film
    case test_thin_film_rectangle:    ////membrane with two ends fixed (rectangular)
        {
            ARRAY<TV> X;
            TV size=TV::Constant_Vector(length_scale*(T)2/length_scale_factor);size.x=length_scale;size.y*=(T)2;
            BOX<TV> box=BOX<TV>(TV(),size);  
            TV_INT counts=TV_INT(box.Edge_Lengths()/length_scale);counts.x=1;
            example.Add_Box_Particles(X,boundary,box,counts,(T)1e-3);
            fixed_boundary_box=box;has_fixed_boundary_condition=true;
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}
            has_surface_tension=true;

            initial_particle_dim=TV::m-1;
        }break;
    case test_thin_film_rectangle_vol:    ////membrane with two ends fixed (rectangular, volumetric)
        {
            ARRAY<TV> X;
            TV size=TV::Constant_Vector(length_scale*(T)8/length_scale_factor);size.x=length_scale;size.y*=(T)2;
            BOX<TV> box=BOX<TV>(TV(),size);  
            TV_INT counts=TV_INT(box.Edge_Lengths()/length_scale);
            example.Add_Box_Particles(X,boundary,box,counts,(T)1e-3);
            for(int i=1;i<=X.m;i++){
                X(i).x+=(T)4*length_scale*sin((T)pi*(X(i).y-box.min_corner.y)/box.Edge_Lengths().y);
            }
            fixed_boundary_box=box;has_fixed_boundary_condition=true;
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}
            has_surface_tension=true;

            initial_particle_dim=TV::m-1;
        }break;
    case test_thin_film_circle_oscillating:    ////oscillating membrane (ring)
        {
            ARRAY<TV> X;
            TV radii=TV::Constant_Vector(length_scale*(T)4/length_scale_factor);
            radii=TV_Compatible<T,TV::m>(radii.x,radii.y,length_scale);
            BOX<TV> box=BOX<TV>(-radii*(T)1.2,radii*(T)1.2);
            SPHERE<TV2> circle(TV2(),radii.x);
            TV_INT counts=TV_INT(ceil(box.Edge_Lengths()/length_scale));
            counts=TV_Compatible<int,TV::m>(counts.x,counts.y,1);
            example.Add_Circle_Particles(X,boundary,circle,counts,3,(T)1e-3);
            for(int i=1;i<=X.m;i++){
                T r=sqrt(pow<2>(X(i).x)+pow<2>(X(i).y));
                T z=(T)6*length_scale*sin((T)half_pi*(circle.radius-r)/circle.radius);
                X(i)=TV_Compatible<T,TV::m>(X(i).x,X(i).y,z);
            }
            fixed_boundary_circle=circle;has_fixed_boundary_condition=true;
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}
            has_surface_tension=true;surface_tension_coefficient*=(T).005;

            initial_particle_dim=TV::m-1;
        }break;
    case test_thin_film_circle_gravity:    ////membrane in gravity (circular)
        {
            ARRAY<TV> X;
            TV radii=TV::Constant_Vector(length_scale*(T)4/length_scale_factor);
            radii=TV_Compatible<T,TV::m>(radii.x,radii.y,length_scale);
            BOX<TV> box=BOX<TV>(-radii*(T)1.2,radii*(T)1.2);
            SPHERE<TV2> circle(TV2(),radii.x);
            TV_INT counts=TV_INT(ceil(box.Edge_Lengths()/length_scale));
            counts=TV_Compatible<int,TV::m>(counts.x,counts.y,1);
            example.Add_Circle_Particles(X,boundary,circle,counts,3,(T)1e-3);
            fixed_boundary_circle=circle;has_fixed_boundary_condition=true;
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}
            has_surface_tension=true;surface_tension_coefficient*=(T).005;
            has_gravity=true;gravity=TV_Compatible<T,TV::m>((T)0,(T)0,(T)-2);
            use_fixed_mass_on_boundary=true;
            //TORUS<T> torus(TV3(),TV3::Axis_Vector(3),length_scale*(T).25,radii.x);
            //TRIANGULATED_SURFACE<T>* torus_triangulated_surface=TESSELLATION::Generate_Triangles(torus,80,60);
            //FILE_UTILITIES::Write_To_File(stream_type,output_directory+"/common/torus.tri",*torus_triangulated_surface);

            initial_particle_dim=TV::m-1;
        }break;
    case test_thin_film_free_rectangle:    ////shrinking plane
        {
            ARRAY<TV> X;
            TV radii=TV::Constant_Vector(length_scale*(T)2/length_scale_factor);radii.x*=(T)1.6;
            BOX<TV> box=BOX<TV>(-radii,radii);  
            TV_INT counts=TV_INT(box.Edge_Lengths()/length_scale);counts.y=0;
            example.Add_Box_Particles(X,boundary,box,counts);
            for(int i=1;i<=X.m;i++){
                TV v=TV::Axis_Vector(1)*(T)(-0.2)*sin(X(i).x*(T)2)+(TV::Axis_Vector(2)*(T)(0.001)*random.Get_Uniform_Number((T)-1,(T)1));
                X_and_V.Append(PAIR<TV,TV>(X(i),v*(T)10));
            }
            //has_surface_tension=true;//surface_tension_coefficient*=(T).04;

            initial_particle_dim=TV::m-1;

            mesher.use_topology_repairing=true;
            mesher.initialize_tetrahedron_mesh=false;
            mesher.initialize_segment_mesh=false;
            mesher.use_crumple_based_dimension_lifting=true;

            //restart_frame=200;
        }break;
    case test_thin_film_free_circle:    ////shrinking circle
        {
            ARRAY<TV> X;
            TV radii=TV::Constant_Vector(length_scale*(T)4/length_scale_factor);
            radii=TV_Compatible<T,TV::m>(radii.x,radii.y,length_scale);
            BOX<TV> box=BOX<TV>(-radii*(T)1.2,radii*(T)1.2);
            SPHERE<TV2> circle(TV2(),radii.x);
            TV_INT counts=TV_INT(ceil(box.Edge_Lengths()/length_scale));
            counts=TV_Compatible<int,TV::m>(counts.x,counts.y,1);
            example.Add_Circle_Particles(X,boundary,circle,counts,3,(T)1e-3);
            for(int i=1;i<=X.m;i++){
                T r=sqrt(pow<2>(X(i).x)+pow<2>(X(i).y));
                TV v=r>(T).9*radii.x?(-X(i))*(T).5:TV();
                T z=(T).01*length_scale*random.Get_Uniform_Number((T)-1,(T)1);
                X(i)=TV_Compatible<T,TV::m>(X(i).x,X(i).y,z);
                X_and_V.Append(PAIR<TV,TV>(X(i),v));
            }
            has_surface_tension=true;

            initial_particle_dim=TV::m-1;
        }break;

    //////////////////////////////////////////////////////////////////////////
    ////bubbles
    case test_bubble_single:    ////single oscillating bubble
        {
            length_scale=(T).002;thickness_scale=(T)1e-5;
            TRIANGULATED_SURFACE<T> triangulated_surface(*(new TRIANGLE_MESH()),*(new GEOMETRY_PARTICLES<TV3>()));
            //FILE_UTILITIES::Read_From_File(stream_type,"../data/sphere.tri",triangulated_surface);
            FILE_UTILITIES::Read_From_File(stream_type,"../data/ellipsoid4.tri",triangulated_surface);
            //SPHERE<TV3> sphere(TV3(),(T).02);
            //TRIANGULATED_SURFACE<T>& triangulated_surface=*(TESSELLATION::Generate_Triangles(sphere));
            external_mesh.triangle_mesh.Initialize_Mesh(triangulated_surface.mesh);
            external_mesh.Update_States(triangulated_surface.particles.X.m);
            for(int i=1;i<=triangulated_surface.particles.X.m;i++){
                X_and_V.Append(PAIR<TV,TV>(TV_Compatible<T,TV::m>(triangulated_surface.particles.X(i)),TV()));
            }

            use_external_mesh=true;
            has_surface_tension=true;
            has_bubble=true;
            use_constant_thin_film_thickness=true;
            
            initial_particle_dim=TV::m-1;
        }break;
    case test_bubble_blowing:    ////blowing bubble in wind field solved by grid solver
        {
            length_scale=(T).001;thickness_scale=(T)1e-5;
            ARRAY<TV> X;
            TV radii=TV::Constant_Vector(length_scale*(T)8/length_scale_factor);
            radii=TV_Compatible<T,TV::m>(radii.x,radii.y,length_scale);
            BOX<TV> box=BOX<TV>(-radii*(T)1.2,radii*(T)1.2);
            SPHERE<TV2> circle(TV2(),radii.x);
            TV_INT counts=TV_INT(ceil(box.Edge_Lengths()/length_scale));
            counts=TV_Compatible<int,TV::m>(counts.x,counts.y,1);
            example.Add_Circle_Particles(X,boundary,circle,counts,3,(T)1e-3);
            fixed_boundary_circle=circle;has_fixed_boundary_condition=true;
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}
            has_surface_tension=true;surface_tension_coefficient*=(T).2;
            use_fixed_mass_on_boundary=true;

            has_wind=true;wind_field.solid_force_rescale=(T)1e7;
            has_bubble=true;
            use_constant_thin_film_thickness=true;

            ////mesher settings
            mesher.use_low_dimensional_incompressibility=false;mesher.use_tetrahedron_mesh=false;

            initial_particle_dim=TV::m-1;
        }break;
    case test_bubble_dumbbell_split:    ////dumbbell single bubble
        {
            length_scale=(T).002;thickness_scale=(T)1e-5;
            TRIANGULATED_SURFACE<T> triangulated_surface(*(new TRIANGLE_MESH()),*(new GEOMETRY_PARTICLES<TV3>()));
            FILE_UTILITIES::Read_From_File(stream_type,"data/dumbbell2.tri",triangulated_surface);
            //FILE_UTILITIES::Write_To_File(stream_type,"../data/dumbbell3.tri",triangulated_surface);
            external_mesh.triangle_mesh.Initialize_Mesh(triangulated_surface.mesh);
            external_mesh.Update_States(triangulated_surface.particles.X.m);
            for(int i=1;i<=triangulated_surface.particles.X.m;i++){
                X_and_V.Append(PAIR<TV,TV>(TV_Compatible<T,TV::m>(triangulated_surface.particles.X(i)),TV()));
            }

            use_external_mesh=true;
            has_surface_tension=true;surface_tension_coefficient*=(T)1;
            has_bubble=true;
            use_constant_thin_film_thickness=true;

            initial_particle_dim=TV::m-1;
        }break;

        //////////////////////////////////////////////////////////////////////////
        ////milk crown
    case test_milk_crown_hydrostatic:////hydrostatic test
        {
            length_scale=(T).001;thickness_scale=length_scale*(T).1;
            T tank_width=length_scale*(T)4;
            T tank_height=tank_width*(T)2;
            TV tank_size=TV::Constant_Vector(tank_width);tank_size.y=tank_height;
            TV tank_min_corner=TV::Constant_Vector(-(T).5*tank_width);tank_min_corner.y=(T)0;
            tank_box=BOX<TV>(tank_min_corner,tank_min_corner+tank_size);
            TV tank_center=tank_box.Center();
            tank=&rigid_standard_tests.Add_Analytic_Box(tank_size);
            tank->X()=tank_center;
            has_tank=true;

            ////water bulk
            TV water_size=TV::Constant_Vector(tank_width);water_size.y=water_size.x*(T).5;
            BOX<TV> water_box=BOX<TV>(tank_min_corner,tank_min_corner+water_size);
            TV_INT particle_counts=TV_INT(ceil(water_size/length_scale));
            ARRAY<TV> X;example.Add_Box_Particles(X,boundary,water_box,particle_counts);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}

            has_gravity=true;
        }break;
    case test_milk_crown:    ////droplet falling onto water
        {
            length_scale=(T).004;
            thickness_scale=length_scale*(T).1;

            T tank_width=(T).2;T tank_height=tank_width;
            T water_height=(T).04;
            T water_sphere_height=(T).1;T water_sphere_radius=(T).004;
            T initial_water_sphere_speed=(T)-10;

            ////tank
            TV tank_size=TV::Constant_Vector(tank_width);tank_size.y=tank_height;
            TV tank_min_corner=TV::Constant_Vector(-(T).5*tank_width);tank_min_corner.y=(T)0;
            tank_box=BOX<TV>(tank_min_corner,tank_min_corner+tank_size);
            TV tank_center=tank_box.Center();
            tank=&rigid_standard_tests.Add_Analytic_Box(tank_size);
            tank->X()=tank_center;
            has_tank=true;

            ////water bulk
            TV water_box_size=TV::Constant_Vector(tank_width);water_box_size.y=water_height;
            BOX<TV> water_box=BOX<TV>(tank_min_corner,tank_min_corner+water_box_size);
            TV_INT particle_counts=TV_INT(ceil(water_box_size/length_scale));
            ARRAY<TV> X;example.Add_Box_Particles(X,boundary,water_box,particle_counts);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}
            
            ////water sphere
            TV water_sphere_center=tank_center;water_sphere_center.y=water_sphere_height;
            SPHERE<TV> water_sphere=SPHERE<TV>(water_sphere_center,water_sphere_radius);
            TV_INT sphere_particle_counts=TV_INT(ceil(water_sphere.Bounding_Box().Edge_Lengths()/length_scale));
            X.Remove_All();
            example.Add_Sphere_Particles(X,boundary,water_sphere,sphere_particle_counts);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV::Axis_Vector(2)*initial_water_sphere_speed));}

            has_gravity=true;
            has_surface_tension=true;
        }break;
    case test_milk_crown_2:    ////droplet falling onto water
        {
            length_scale=(T).004*length_scale_factor;
            thickness_scale=length_scale*(T).1;
            
            T tank_width=(T).2;T tank_height=(T).06;
            T water_height=(T).04;
            //T water_sphere_radius=(T).012;T water_sphere_height=water_height+water_sphere_radius+(T).5*length_scale;
            //T initial_water_sphere_speed=(T)-16;

            ////tank
            TV tank_size=TV::Constant_Vector(tank_width);tank_size.y=tank_height;
            TV tank_min_corner=TV::Constant_Vector(-(T).5*tank_width);tank_min_corner.y=(T)0;
            tank_box=BOX<TV>(tank_min_corner,tank_min_corner+tank_size);tank_thickness=(T)2*length_scale;
            TV tank_center=tank_box.Center();
            tank=&rigid_standard_tests.Add_Analytic_Box(tank_size);
            tank->X()=tank_center;
            has_tank=true;
            is_open_tank=true;

            ////water bulk
            TV water_box_size=TV::Constant_Vector(tank_width);water_box_size.y=water_height;
            BOX<TV> water_box=BOX<TV>(tank_min_corner,tank_min_corner+water_box_size);
            TV_INT particle_counts=TV_INT(ceil(water_box_size/(length_scale*(T).8)));
            T perturb_coef=(T).2;
            ARRAY<TV> X;example.Add_Box_Particles(X,boundary,water_box,particle_counts,perturb_coef);
            //example.Add_Box_Particles2(water_box.min_corner,water_box.max_corner,length_scale,X,boundary);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}

            ////water sphere
            //TV water_sphere_center=tank_center;water_sphere_center.y=water_sphere_height;
            //SPHERE<TV> water_sphere=SPHERE<TV>(water_sphere_center,water_sphere_radius);
            //TV_INT sphere_particle_counts=TV_INT(ceil(water_sphere.Bounding_Box().Edge_Lengths()/(length_scale*(T).8)));
            //X.Remove_All();example.Add_Sphere_Particles(X,boundary,water_sphere,sphere_particle_counts);
            //for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV::Axis_Vector(2)*initial_water_sphere_speed));}

            has_gravity=true;
            has_surface_tension=true;
            //has_adhesion_force=true;

            //mesher.use_triangle_mesh=false;
            //mesher.use_segment_mesh=false;
            //mesher.use_domain_predicator=true;
            //mesher.use_reseed=true;
            //mesher.analytic_domain_predicator.Add_Geometry(water_box);
            //mesher.analytic_domain_predicator.Add_Geometry(water_sphere);

            mesher.Initialize_Tank(tank_box,tank_thickness,is_open_tank);

            mesher.use_topology_repairing=true;
            mesher.use_crumple_merge=true;
            mesher.use_small_region_remeshing=true;
            mesher.use_low_dimensional_incompressibility=true;
            mesher.use_distance_to_rim_for_incompressibility=true;
            mesher.min_distance_to_rim_for_incompressibility=5;
            mesher.use_crumple_based_dimension_lifting=false;

            //use_initialize_from_sim_data=true;
            //init_sim_frame=30;
            //init_sim_data_directory="init_sim_data/mc_tet_s8";

            //use_initialize_from_sim_data=true;
            //init_sim_frame=30;
            //init_sim_data_directory="init_sim_data/mc_tet_s4";

            //restart_frame=30;

            subdivision=TV_Compatible<int,TV::m>(TV_INT3(3,1,3));
            //subdivision=TV_Compatible<int,TV::m>(TV_INT3(2,1,2));
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-.11),TV::Constant_Vector((T).11));
        }break;
    case test_milk_crown_3:    ////Wen's example
        {
            length_scale=(T).012*length_scale_factor;thickness_scale=length_scale*(T).1;
            T tank_width=(T).3;T tank_height=(T).08;
            T water_height=(T).05;
            T water_sphere_radius=(T).012;T water_sphere_height=water_height+water_sphere_radius+(T)length_scale;
            T initial_water_sphere_speed=(T)-5;

            ////tank
            TV tank_size=TV::Constant_Vector(tank_width);tank_size.y=tank_height;
            TV tank_min_corner=TV::Constant_Vector(-(T).5*tank_width);tank_min_corner.y=(T)0;
            tank_box=BOX<TV>(tank_min_corner,tank_min_corner+tank_size);tank_thickness=(T)2*length_scale;
            TV tank_center=tank_box.Center();
            tank=&rigid_standard_tests.Add_Analytic_Box(tank_size);
            tank->X()=tank_center;
            has_tank=true;
            is_open_tank=true;

            ////water bulk
            TV water_box_size=TV::Constant_Vector(tank_width);water_box_size.y=water_height;
            BOX<TV> water_box=BOX<TV>(tank_min_corner,tank_min_corner+water_box_size);
            TV_INT particle_counts=TV_INT(ceil(water_box_size/length_scale));
            ARRAY<TV> X;example.Add_Box_Particles(X,boundary,water_box,particle_counts);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}

            //water sphere
            TV water_sphere_center=tank_center;water_sphere_center.y=water_sphere_height;
            SPHERE<TV> water_sphere=SPHERE<TV>(water_sphere_center,water_sphere_radius);
            TV_INT sphere_particle_counts=TV_INT(ceil(water_sphere.Bounding_Box().Edge_Lengths()/length_scale));
            X.Remove_All();example.Add_Sphere_Particles(X,boundary,water_sphere,sphere_particle_counts);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV::Axis_Vector(2)*initial_water_sphere_speed));}

            has_gravity=true;
            has_surface_tension=true;
            has_adhesion_force=true;

            if(TV::m==3)mesher.use_triangle_mesh=false;
            mesher.use_domain_predicator=true;
            mesher.use_reseed=true;
            mesher.analytic_domain_predicator.Add_Geometry(water_box);
            mesher.analytic_domain_predicator.Add_Geometry(water_sphere);
            mesher.Initialize_Tank(tank_box,tank_thickness,is_open_tank);
        }break;
    case test_milk_crown_4:    ////more flat water bulk
        {
            length_scale=(T).008*length_scale_factor;
            thickness_scale=length_scale*(T).1;

            T tank_width=(T).2;T tank_height=(T).06;
            T water_height=(T).04;
            T water_sphere_radius=(T).012;T water_sphere_height=water_height+water_sphere_radius+(T).5*length_scale;
            T initial_water_sphere_speed=(T)-20;

            ////tank
            TV tank_size=TV::Constant_Vector(tank_width);tank_size.y=tank_height;
            TV tank_min_corner=TV::Constant_Vector(-(T).5*tank_width);tank_min_corner.y=(T)0;
            tank_box=BOX<TV>(tank_min_corner,tank_min_corner+tank_size);tank_thickness=(T)2*length_scale;
            TV tank_center=tank_box.Center();
            tank=&rigid_standard_tests.Add_Analytic_Box(tank_size);
            tank->X()=tank_center;
            has_tank=true;
            is_open_tank=true;

            ////water bulk
            TV water_box_size=TV::Constant_Vector(tank_width);water_box_size.y=water_height;
            BOX<TV> water_box=BOX<TV>(tank_min_corner,tank_min_corner+water_box_size);
            TV_INT particle_counts=TV_INT(ceil(water_box_size/(length_scale*(T).8)));
            T perturb_coef=(T).2;
            ARRAY<TV> X;example.Add_Box_Particles(X,boundary,water_box,particle_counts,perturb_coef);
            //example.Add_Box_Particles2(water_box.min_corner,water_box.max_corner,length_scale,X,boundary);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}

            //water sphere
            TV water_sphere_center=tank_center;water_sphere_center.y=water_sphere_height;
            SPHERE<TV> water_sphere=SPHERE<TV>(water_sphere_center,water_sphere_radius);
            TV_INT sphere_particle_counts=TV_INT(ceil(water_sphere.Bounding_Box().Edge_Lengths()/(length_scale*(T).8)));
            X.Remove_All();example.Add_Sphere_Particles(X,boundary,water_sphere,sphere_particle_counts);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV::Axis_Vector(2)*initial_water_sphere_speed));}

            has_gravity=true;
            has_surface_tension=true;
            //has_adhesion_force=true;

            //mesher.use_triangle_mesh=false;
            //mesher.use_segment_mesh=false;
            //mesher.use_domain_predicator=true;
            //mesher.use_reseed=true;
            //mesher.analytic_domain_predicator.Add_Geometry(water_box);
            //mesher.analytic_domain_predicator.Add_Geometry(water_sphere);

            mesher.Initialize_Tank(tank_box,tank_thickness,is_open_tank);

            mesher.use_topology_repairing=true;
            mesher.use_crumple_merge=true;
            mesher.use_small_region_remeshing=true;
            mesher.use_low_dimensional_incompressibility=true;
            mesher.use_distance_to_rim_for_incompressibility=true;
            mesher.min_distance_to_rim_for_incompressibility=5;
            mesher.use_crumple_based_dimension_lifting=true;

            use_initialize_from_sim_data=true;
            init_sim_frame=30;
            init_sim_data_directory="init_sim_data/mc_tet_s4";

            //use_initialize_from_sim_data=true;
            //init_sim_frame=380;
            //init_sim_data_directory="mc1";

            //restart_frame=30;
        }break;

    case test_milk_crown_5:
        {
            length_scale=(T).024*length_scale_factor;
            thickness_scale=length_scale*(T).1;

            T tank_width=(T).3;T tank_height=(T).1;
            T water_height=(T).06;
            //T water_sphere_radius=(T).012;T water_sphere_height=water_height+water_sphere_radius+(T).5*length_scale;
            //T initial_water_sphere_speed=(T)-10;

            ////tank
            TV tank_size=TV::Constant_Vector(tank_width);tank_size.y=tank_height;
            TV tank_min_corner=TV::Constant_Vector(-(T).5*tank_width);tank_min_corner.y=(T)0;
            tank_box=BOX<TV>(tank_min_corner,tank_min_corner+tank_size);tank_thickness=(T)2*length_scale;
            TV tank_center=tank_box.Center();
            tank=&rigid_standard_tests.Add_Analytic_Box(tank_size);
            tank->X()=tank_center;
            has_tank=true;
            is_open_tank=true;

            ////water bulk
            TV water_box_size=TV::Constant_Vector(tank_width);water_box_size.y=water_height;
            BOX<TV> water_box=BOX<TV>(tank_min_corner,tank_min_corner+water_box_size);
            TV_INT particle_counts=TV_INT(ceil(water_box_size/(length_scale*(T).8)));
            T perturb_coef=(T).2;
            ARRAY<TV> X;example.Add_Box_Particles(X,boundary,water_box,particle_counts,perturb_coef);
            //example.Add_Box_Particles2(water_box.min_corner,water_box.max_corner,length_scale,X,boundary);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}

            ////water sphere
            //TV water_sphere_center=tank_center;water_sphere_center.y=water_sphere_height;
            //SPHERE<TV> water_sphere=SPHERE<TV>(water_sphere_center,water_sphere_radius);
            //TV_INT sphere_particle_counts=TV_INT(ceil(water_sphere.Bounding_Box().Edge_Lengths()/(length_scale*(T).8)));
            //X.Remove_All();example.Add_Sphere_Particles(X,boundary,water_sphere,sphere_particle_counts);
            //for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV::Axis_Vector(2)*initial_water_sphere_speed));}

            has_gravity=true;
            has_surface_tension=true;
            has_adhesion_force=true;

            //mesher.use_triangle_mesh=false;
            //mesher.use_segment_mesh=false;
            //mesher.use_domain_predicator=true;
            //mesher.use_reseed=true;
            //mesher.analytic_domain_predicator.Add_Geometry(water_box);
            //mesher.analytic_domain_predicator.Add_Geometry(water_sphere);

            mesher.Initialize_Tank(tank_box,tank_thickness,is_open_tank);

            mesher.use_topology_repairing=true;
            mesher.use_crumple_merge=true;
            mesher.use_small_region_remeshing=true;
            mesher.use_low_dimensional_incompressibility=true;
            mesher.use_distance_to_rim_for_incompressibility=true;
            mesher.min_distance_to_rim_for_incompressibility=5;
            mesher.use_crumple_based_dimension_lifting=false;

            //use_initialize_from_sim_data=true;
            //init_sim_frame=30;
            //init_sim_data_directory="init_sim_data/mc_tet_s8";

            //use_initialize_from_sim_data=true;
            //init_sim_frame=30;
            //init_sim_data_directory="init_sim_data/mc_tet_s4";

            //restart_frame=30;

            subdivision=TV_Compatible<int,TV::m>(TV_INT3(3,1,3));
            //subdivision=TV_Compatible<int,TV::m>(TV_INT3(2,1,2));
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-.16),TV::Constant_Vector((T).16));
        }break;

        //////////////////////////////////////////////////////////////////////////
        ////fluid polygon
    case test_falling_sphere:   ////water ball falling onto a cube
        {
            length_scale=(T).004*length_scale_factor;thickness_scale=length_scale*(T).1;
            //T box_size=(T).04;
            T box_top_plane_height=(T)0;
            T water_sphere_radius=(T).012;T water_sphere_height=box_top_plane_height+water_sphere_radius+(T)length_scale;
            T initial_water_sphere_speed=(T)-8;

            //water sphere
            TV water_sphere_center=TV();water_sphere_center.y=water_sphere_height;
            SPHERE<TV> water_sphere=SPHERE<TV>(water_sphere_center,water_sphere_radius);
            TV_INT sphere_particle_counts=TV_INT(ceil(water_sphere.Bounding_Box().Edge_Lengths()/(length_scale*(T).8)));
            ARRAY<TV> X;example.Add_Sphere_Particles(X,boundary,water_sphere,sphere_particle_counts);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV::Axis_Vector(2)*initial_water_sphere_speed));}

            has_gravity=true;

            //rigid_body=&rigid_standard_tests.Add_Analytic_Box(TV::Constant_Vector(box_size));
            //rigid_body->X()=TV();rigid_body->X().y=box_top_plane_height-box_size*(T).5;
            //has_solid=true;

            ////tank
            T tank_width=(T).4;
            TV tank_size=TV::Constant_Vector(tank_width);
            TV tank_min_corner=TV::Constant_Vector(-(T).5*tank_width);tank_min_corner.y=box_top_plane_height;
            tank_box=BOX<TV>(tank_min_corner,tank_min_corner+tank_size);tank_thickness=(T)2*length_scale;
            TV tank_center=tank_box.Center();
            tank=&rigid_standard_tests.Add_Analytic_Box(tank_size);
            tank->X()=tank_center;
            has_tank=true;
            is_open_tank=true;

            //if(TV::m==3)mesher.use_triangle_mesh=false;
            mesher.use_domain_predicator=false;
            mesher.use_reseed=true;
            mesher.use_topology_repairing=true;
            mesher.analytic_domain_predicator.Add_Geometry(water_sphere);
            mesher.Initialize_Tank(tank_box,tank_thickness,is_open_tank);
        }break;
    case test_falling_tet:   ////one tet falling onto a cube
        {
            TV tet_center=TV();
            ARRAY<TV> X;
            example.Add_Regular_Simplex(X,tet_center,length_scale);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}
            has_gravity=true;

            RIGID_BODY<TV>* rigid_body=&rigid_standard_tests.Add_Analytic_Box(TV::Constant_Vector(length_scale*(T)20));
            rigid_body->X()=TV();rigid_body->X().y=-length_scale*(T)30;
            rigid_bodies.Append(rigid_body);
            has_solid=true;
        }break;
    case test_colliding_spheres:	////two colliding water balls
        {
            TV sphere_center=TV();
            SPHERE<TV> sphere=SPHERE<TV>(sphere_center,length_scale*(T)2);
            TV_INT sphere_particle_counts=(TV_INT)(sphere.Bounding_Box().Edge_Lengths()/length_scale);
            ARRAY<TV> X;
            example.Add_Sphere_Particles(X,boundary,sphere,sphere_particle_counts);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV::Axis_Vector(1)*(T)1));}

            X.Remove_All();
            sphere.center.x+=(T)5;
            example.Add_Sphere_Particles(X,boundary,sphere,sphere_particle_counts);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV::Axis_Vector(1)*(T)-1));}
        }break;
    case test_solid_adhesion:    ////adhesion force 2d
        {
            //T s=(T)1;
            //WATER_PIPE<TV>* water_pipe=new WATER_PIPE<TV>();
            //for(int i=1;i<=8*(int)s;i++){water_pipe->Add_Source(TV(TV2((T)length_scale*(T)(i-1),(T)0)),TV());}
            //for(int i=1;i<=water_pipe->locations.m;i++){X_and_V.Append(PAIR<TV,TV>(water_pipe->locations(i),water_pipe->velocities(i)));}
            //has_gravity=true;has_surface_tension=true;

            //water_sources.Append(water_pipe);

            //RIGID_BODY<TV>* rigid_body=&rigid_standard_tests.Add_Analytic_Sphere(length_scale*(T)10*s, (T)1.0);
            //rigid_body->X()=TV(TV2(length_scale*(T)4*s,(T)-1.3*s));
            //rigid_bodies.Append(rigid_body);
            //has_solid=true;has_adhesion_force=true;
        }break;
    case test_fluid_polygon:
        {
            length_scale=(T).002*length_scale_factor;thickness_scale=length_scale*(T).1;
            T box_size=(T).01;T box_top_plane_height=(T)0;

            ////pipe
            TV pipe_e1=TV::Axis_Vector(2)*((T).02);TV pipe_e2=pipe_e1+TV::Axis_Vector(2)*((T)-.01);
            T initial_pipe_speed=(T)9;T pipe_radius=(T).004;
            water_sources.Append(new WATER_PIPE<TV>(particles,pipe_e1,pipe_e2,pipe_radius,length_scale,initial_pipe_speed));

            ////rigid
            //RIGID_BODY<TV>* rigid_body=&rigid_standard_tests.Add_Analytic_Box(TV::Constant_Vector(box_size));
            //rigid_body->X()=TV();rigid_body->X().y=box_top_plane_height-box_size*(T).5;
            RIGID_BODY<TV>* rigid_body=Get_Cylinder_Rigid_Body(rigid_standard_tests,TV::Axis_Vector(1)*((T)half_pi),TV::Axis_Vector(2)*(box_top_plane_height-box_size*(T).5),box_size,/*(T).5**/box_size);
            rigid_bodies.Append(rigid_body);
            has_solid=true;

            //has_water_pipe=true;
            has_fixed_boundary_condition=true;
            has_gravity=true;
            has_surface_tension=true;
            surface_tension_coefficient*=(T)2;
            has_viscosity_force=false;
            //use_implicit_viscosity_force=true;
            //kinematic_viscosity_coefficient*=(T)100;

            mesher.use_topology_repairing=true;
            mesher.use_low_dimensional_incompressibility=false;
            mesher.Add_Solid(*rigid_body);

            subdivision=TV_Compatible<int,TV::m>(TV_INT3(3,1,3));
            //subdivision=TV_Compatible<int,TV::m>(TV_INT3(2,1,2));
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-.04),TV::Constant_Vector((T).04));
        }break;
    case test_water_bell:
        {
            length_scale=(T).002*length_scale_factor;thickness_scale=length_scale*(T).1;
            T box_size=(T).01;T box_top_plane_height=(T)0;

            ////pipe
            TV pipe_e1=TV::Axis_Vector(2)*((T).02);TV pipe_e2=pipe_e1+TV::Axis_Vector(2)*((T)-.01);
            T initial_pipe_speed=(T)5;T pipe_radius=(T).004;
            water_sources.Append(new WATER_PIPE<TV>(particles,pipe_e1,pipe_e2,pipe_radius,length_scale,initial_pipe_speed));

            ////rigid
            RIGID_BODY<TV>* rigid_body=Get_Cylinder_Rigid_Body(rigid_standard_tests,TV::Axis_Vector(1)*((T)half_pi),TV::Axis_Vector(2)*(box_top_plane_height-box_size*(T).5),box_size,(T).5*box_size);
            rigid_bodies.Append(rigid_body);
            has_solid=true;

            //has_water_pipe=true;
            has_fixed_boundary_condition=true;
            has_gravity=true;
            has_surface_tension=true;
            surface_tension_coefficient*=(T)2;
            has_viscosity_force=false;
            kinematic_viscosity_coefficient*=(T)20;

            mesher.use_topology_repairing=true;
            mesher.use_low_dimensional_incompressibility=true;
            mesher.use_distance_to_rim_for_incompressibility=true;
            //mesher.use_thickness_based_dimension_lifting=true;
            mesher.Add_Solid(*rigid_body);
        }break;

    //////////////////////////////////////////////////////////////////////////
    ////thin fluid on ground (or solid surface)
    case test_thin_sheet:
        {
            length_scale=(T).004*length_scale_factor;thickness_scale=length_scale*(T).1;
            T water_box_width=(T).012;T water_box_height=length_scale*(T).5;
            T tank_bottom_plane_height=(T)0;
            ////water bulk
            TV water_box_size=TV::Constant_Vector(water_box_width);water_box_size.y=water_box_height;
            TV water_box_min_corner=TV::Constant_Vector(-(T).5*water_box_width);water_box_min_corner.y=tank_bottom_plane_height;
            BOX<TV> water_box=BOX<TV>(water_box_min_corner,water_box_min_corner+water_box_size);
            TV_INT particle_counts=TV_INT(ceil(water_box_size/length_scale));particle_counts.y=1;
            ARRAY<TV> X;example.Add_Box_Particles(X,boundary,water_box,particle_counts);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}

            has_gravity=true;
            has_surface_tension=true;
            surface_tension_coefficient*=(T).001;
            ////tank
            T tank_width=(T).4;
            TV tank_size=TV::Constant_Vector(tank_width);
            TV tank_min_corner=TV::Constant_Vector(-(T).5*tank_width);tank_min_corner.y=tank_bottom_plane_height;
            tank_box=BOX<TV>(tank_min_corner,tank_min_corner+tank_size);tank_thickness=(T)2*length_scale;
            TV tank_center=tank_box.Center();
            tank=&rigid_standard_tests.Add_Analytic_Box(tank_size);
            tank->X()=tank_center;
            has_tank=true;
            is_open_tank=true;

            //if(TV::m==3)mesher.use_triangle_mesh=false;
            mesher.use_topology_repairing=true;
            mesher.Initialize_Tank(tank_box,tank_thickness,is_open_tank);
        }break;
    case test_thin_sheet_2:
        {
            length_scale=(T).004*length_scale_factor;thickness_scale=length_scale*(T).1;
            T water_box_width=(T).012;T water_box_height=water_box_width*(T)2;
            T tank_bottom_plane_height=(T)0;
            ////water bulk
            TV water_box_size=TV::Constant_Vector(water_box_width);water_box_size.y=water_box_height;
            TV water_box_min_corner=TV::Constant_Vector(-(T).5*water_box_width);water_box_min_corner.y=tank_bottom_plane_height;
            BOX<TV> water_box=BOX<TV>(water_box_min_corner,water_box_min_corner+water_box_size);
            TV_INT particle_counts=TV_INT(ceil(water_box_size/length_scale));
            T perturb_coef=(T).2;
            ARRAY<TV> X;example.Add_Box_Particles(X,boundary,water_box,particle_counts,perturb_coef);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}

            has_gravity=true;
            has_surface_tension=true;
            surface_tension_coefficient*=(T).001;
            ////tank
            T tank_width=(T).4;
            TV tank_size=TV::Constant_Vector(tank_width);
            TV tank_min_corner=TV::Constant_Vector(-(T).5*tank_width);tank_min_corner.y=tank_bottom_plane_height;
            tank_box=BOX<TV>(tank_min_corner,tank_min_corner+tank_size);tank_thickness=(T)2*length_scale;
            TV tank_center=tank_box.Center();
            tank=&rigid_standard_tests.Add_Analytic_Box(tank_size);
            tank->X()=tank_center;
            has_tank=true;
            is_open_tank=true;

            //if(TV::m==3)mesher.use_triangle_mesh=false;
            mesher.use_topology_repairing=true;
            mesher.Initialize_Tank(tank_box,tank_thickness,is_open_tank);
        }break;
    case test_thin_sheet_3:
        {
            length_scale=(T).001*length_scale_factor;thickness_scale=(T)1e-5;
            ARRAY<TV> X;
            TV radii=TV::Constant_Vector((T).01);
            radii=TV_Compatible<T,TV::m>(radii.x,radii.y,length_scale);
            BOX<TV> box=BOX<TV>(-radii*(T)1.2,radii*(T)1.2);
            SPHERE<TV2> circle(TV2(),radii.x);
            TV_INT counts=TV_INT(ceil(box.Edge_Lengths()/length_scale));
            counts=TV_Compatible<int,TV::m>(counts.x,counts.y,1);
            //example.Add_Circle_Particles(X,boundary,circle,counts,3,(T)1e-3);
            example.Add_Circle_Particles(X,TV(),radii.x,TV::Axis_Vector(1),length_scale);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}
            has_surface_tension=true;
            has_viscosity_force=true;
            has_gravity=false;
            use_constant_thin_film_thickness=true;

            ////mesher settings
            mesher.use_topology_repairing=true;
            mesher.use_tetrahedron_mesh=false;
            mesher.use_low_dimensional_incompressibility=false;
            initial_particle_dim=TV::m-1;
        }break;

    //////////////////////////////////////////////////////////////////////////
    ////fishbone
    case test_impinging_jet_no_gravity:
        {
            length_scale=(T)5e-4;
            thickness_scale=length_scale*(T).1;
            T radius=(T)8e-4;
            T initial_speed=(T)5;
            T theta=(T)pi;
            T distance=(T)10*length_scale;
            T pipe_length=length_scale;
            TV x_normal=TV::Axis_Vector(1);
            TV3 y_normal=TV3::Axis_Vector(2);
            TV p1_end=(T)-0.5*distance*x_normal;
            TV p2_end=(T)0.5*distance*x_normal;

            ROTATION<TV3> rotation_about_z((T)0.5*theta,TV3::Axis_Vector(3));
            TV p1_begin=TV_Compatible<T,TV::m>(rotation_about_z.Rotate(y_normal));
            p1_begin-=(T)0.5*distance*x_normal;
            TV direction=(p1_begin-p1_end).Normalized();
            p1_begin=p1_end+pipe_length*direction;

            ROTATION<TV3> neg_rotation_about_z((T)-0.5*theta,TV3::Axis_Vector(3));
            TV p2_begin=TV_Compatible<T,TV::m>(neg_rotation_about_z.Rotate(y_normal));
            p2_begin+=(T)0.5*distance*x_normal;
            direction=(p2_begin-p2_end).Normalized();
            p2_begin=p2_end+pipe_length*direction;

            water_sources.Append(new WATER_PIPE<TV>(particles,p1_begin,p1_end,radius,length_scale,initial_speed));
            water_sources.Append(new WATER_PIPE<TV>(particles,p2_begin,p2_end,radius,length_scale,initial_speed));

            has_gravity=true;
            has_surface_tension=true;

            mesher.use_low_dimensional_incompressibility=false;
            //mesher.use_triangle_mesh=false;
            //mesher.use_segment_mesh=false;
            mesher.use_topology_repairing=true;
            restart_frame=31;
            //restart_frame=77;
        }break;
    case test_fishbone:
        {
            length_scale=(T)5e-4;
            thickness_scale=length_scale*(T).1;
            T radius=(T)8e-4;
            T initial_speed=(T)6;
            //T theta=(T)(one_third*pi);
            T theta=(T)(.5*pi);
            T distance=(T)100e-4;
            T pipe_length=length_scale;
            TV x_normal=TV::Axis_Vector(1);
            TV3 y_normal=TV3::Axis_Vector(2);
            TV p1_end=(T)-0.5*distance*x_normal;
            TV p2_end=(T)0.5*distance*x_normal;

            ROTATION<TV3> rotation_about_z((T)0.5*theta,TV3::Axis_Vector(3));
            TV p1_begin=TV_Compatible<T,TV::m>(rotation_about_z.Rotate(y_normal));
            p1_begin-=(T)0.5*distance*x_normal;
            TV direction=(p1_begin-p1_end).Normalized();
            p1_begin=p1_end+pipe_length*direction;

            ROTATION<TV3> neg_rotation_about_z((T)-0.5*theta,TV3::Axis_Vector(3));
            TV p2_begin=TV_Compatible<T,TV::m>(neg_rotation_about_z.Rotate(y_normal));
            p2_begin+=(T)0.5*distance*x_normal;
            direction=(p2_begin-p2_end).Normalized();
            p2_begin=p2_end+pipe_length*direction;

            water_sources.Append(new WATER_PIPE<TV>(particles,p1_begin,p1_end,radius,length_scale,initial_speed));
            water_sources.Append(new WATER_PIPE<TV>(particles,p2_begin,p2_end,radius,length_scale,initial_speed));

            has_gravity=true;
            has_surface_tension=true;
            //surface_tension_coefficient*=(T).5;
            has_viscosity_force=true;
            kinematic_viscosity_coefficient*=(T)20;

            mesher.use_low_dimensional_incompressibility=true;
            mesher.use_distance_to_rim_for_incompressibility=true;
            mesher.min_distance_to_rim_for_incompressibility=5;
            //mesher.use_triangle_mesh=false;
            //mesher.use_segment_mesh=false;
            mesher.use_topology_repairing=true;
            mesher.use_crumple_merge=true;
            mesher.use_break_thin_segments=true;
            mesher.use_small_region_remeshing=true;

            //use_initialize_from_sim_data=true;
            //init_sim_frame=36;
            //init_sim_data_directory="init_sim_data/fb_tet_a45_v5_s10";

            //use_initialize_from_sim_data=true;
            //init_sim_frame=85;
            //init_sim_data_directory="init_sim_data/fb_tet_a45_v5_s5";

            use_initialize_from_sim_data=true;
            init_sim_frame=85;
            init_sim_data_directory="init_sim_data/fb_tet_a45_v6_s10";
            
            //restart_frame=280;
        }break;

    //////////////////////////////////////////////////////////////////////////
    ////water pipe
    case test_water_pipe_repair:  ////case 200
        {
            TV e1=TV();TV e2=TV::Axis_Vector(2)*((T)-10*length_scale);
            T initial_speed=(T)50*length_scale;T radius=(T)5*length_scale;
            water_sources.Append(new WATER_PIPE<TV>(particles,e1,e2,radius,length_scale,initial_speed));

            has_gravity=true;
            has_surface_tension=false;

            mesher.use_topology_repairing=true;
            has_fixed_boundary_condition=true;
        }break;
    case test_water_pipe_reseed:  ////case 201
        {
            TV e1=TV();TV e2=TV::Axis_Vector(2)*((T)-2*length_scale);
            T initial_speed=(T)50*length_scale;T radius=(T)5*length_scale;
            water_sources.Append(new WATER_PIPE<TV>(particles,e1,e2,radius,length_scale,initial_speed));

            has_gravity=true;
            has_surface_tension=false;

            mesher.use_topology_repairing=false;
            has_fixed_boundary_condition=true;
        }break;
    case test_water_pipe_reseed_low_res:  ////case 202
        {
            TV e1=TV();TV e2=TV::Axis_Vector(2)*((T)-2*length_scale);
            T initial_speed=(T)10;T radius=(T)2*length_scale;
            water_sources.Append(new WATER_PIPE<TV>(particles,e1,e2,radius,length_scale,initial_speed));

            has_gravity=true;
            has_surface_tension=false;

            mesher.use_topology_repairing=true;
            has_fixed_boundary_condition=true;
        }break;
    case test_water_ring_repair: ////case 300
        {
            //length_scale=(T).01*length_scale_factor;thickness_scale=length_scale*(T).1;
            length_scale=(T).005*length_scale_factor;thickness_scale=length_scale*(T).1;
            //T box_size=(T)0.07;//T box_size=(T).05;
            //T box_top_plane_height=(T)0;

            ////pipe
            TV pipe_e1=TV::Axis_Vector(2)*((T).1);TV pipe_e2=pipe_e1+TV::Axis_Vector(2)*((T)-.02);
            T initial_pipe_speed=(T)2;
            T pipe_radius=(T).03;
            T inner_radius=(T).02;

            water_sources.Append(new WATER_RING<TV>(particles,pipe_e1,pipe_e2,inner_radius,pipe_radius,length_scale,initial_pipe_speed));

            ////rigid
            //RIGID_BODY<TV>* rigid_body=&rigid_standard_tests.Add_Analytic_Box(TV::Constant_Vector(box_size));
            //rigid_body->X()=TV();rigid_body->X().y=box_top_plane_height-box_size*(T).5;
            TV base_euler_angles=TV();
            base_euler_angles.x=(T)half_pi;
            T base_height=(T)0.05;
            T base_radius=(T)0.04;
            TV base_translation=TV();
            base_translation.y-=(T)0.5*base_height;
            RIGID_BODY<TV>* rigid_body=Get_Cylinder_Rigid_Body<T>(rigid_standard_tests,base_euler_angles,base_translation,base_height,base_radius);
            rigid_bodies.Append(rigid_body);
            has_solid=true;

            has_fixed_boundary_condition=true;
            has_gravity=true;
            has_surface_tension=true;
            has_viscosity_force=false;

            //use_constant_thin_film_thickness=true;
            mesher.use_topology_repairing=true;
            mesher.use_low_dimensional_incompressibility=true;
            mesher.Add_Solid(*rigid_body);

            //water ring rigid body
            T cylinder_height=pipe_e1.y;
            const T epsilon=(T)5e-4;
            TV euler_angles=TV();
            euler_angles.x=(T)half_pi;
            TV translation=TV::Axis_Vector(2)*(T)0.5*cylinder_height;
            RIGID_BODY<TV>* cylinder=Get_Cylinder_Rigid_Body<T>(rigid_standard_tests,euler_angles,translation-TV::Axis_Vector(2)*(T).005,cylinder_height+(T).01,inner_radius-epsilon);
            if(cylinder!=0){
                mesher.Add_Solid(*cylinder);
                rigid_bodies.Append(cylinder);
            }
        }break;
    case test_water_ring_repair2: //case 301
        {
            //length_scale=(T).01*length_scale_factor;thickness_scale=length_scale*(T).1;
            length_scale=(T).0005*length_scale_factor;thickness_scale=length_scale*(T).1;
            //T box_size=(T)0.07;//T box_size=(T).05;
            //T box_top_plane_height=(T)0;

            ////pipe
            TV pipe_e1=TV::Axis_Vector(2)*((T).01);TV pipe_e2=pipe_e1+TV::Axis_Vector(2)*((T)-.002);
            T initial_pipe_speed=(T)1.27324;
            T pipe_radius=(T).003;
            T inner_radius=(T).002;

            water_sources.Append(new WATER_RING<TV>(particles,pipe_e1,pipe_e2,inner_radius,pipe_radius,length_scale,initial_pipe_speed));

            ////rigid
            //RIGID_BODY<TV>* rigid_body=&rigid_standard_tests.Add_Analytic_Box(TV::Constant_Vector(box_size));
            //rigid_body->X()=TV();rigid_body->X().y=box_top_plane_height-box_size*(T).5;
            TV base_euler_angles=TV();
            base_euler_angles.x=(T)half_pi;
            T base_height=(T)0.005;
            T base_radius=(T)0.004;
            TV base_translation=TV();
            base_translation.y-=(T)0.5*base_height;
            RIGID_BODY<TV>* rigid_body=Get_Cylinder_Rigid_Body<T>(rigid_standard_tests,base_euler_angles,base_translation,base_height,base_radius);
            rigid_bodies.Append(rigid_body);
            has_solid=true;

            has_fixed_boundary_condition=true;
            has_gravity=true;
            has_surface_tension=true;
            has_viscosity_force=false;

            //use_constant_thin_film_thickness=true;
            mesher.use_topology_repairing=true;
            mesher.use_low_dimensional_incompressibility=true;
            mesher.Add_Solid(*rigid_body);

            //water ring rigid body
            T cylinder_height=pipe_e1.y;
            const T epsilon=(T)5e-5;
            TV euler_angles=TV();
            euler_angles.x=(T)half_pi;
            TV translation=TV::Axis_Vector(2)*(T)0.5*cylinder_height;
            RIGID_BODY<TV>* cylinder=Get_Cylinder_Rigid_Body<T>(rigid_standard_tests,euler_angles,translation-TV::Axis_Vector(2)*(T).0005,cylinder_height+(T).001,inner_radius-epsilon);
            if(cylinder!=0){
                mesher.Add_Solid(*cylinder);
                rigid_bodies.Append(cylinder);
            }
        }break;

    case test_1d_jet:
        {

        }break;
        //////////////////////////////////////////////////////////////////////////
    default:PHYSBAM_FATAL_ERROR("invalid test id");break;
    }

    for(int i=1;i<=X_and_V.m;i++){
        int index=particles.Add_Particle();
        particles.X(index)=X_and_V(i).x;particles.V(index)=X_and_V(i).y;
    }
    if(R.m>0)for(int i=1;i<=particles.X.m;i++)particles.radius(i)=R(i);
    else for(int i=1;i<=particles.X.m;i++)particles.radius(i)=(T).5*length_scale;
    if(mass.m>0)for(int i=1;i<=particles.X.m;i++){particles.mass(i)=mass(i);}
    else for(int i=1;i<=particles.X.m;i++){particles.mass(i)=(T)1;}
    if(boundary.m>0)for(int i=1;i<=particles.X.m;i++)particles.on_boundary(i)(initial_particle_dim)=boundary(i);
    //else for(int i=1;i<=particles.X.m;i++)particles.on_boundary(i)=false;
    for(int i=1;i<=particles.X.m;i++)particles.dim(i)=initial_particle_dim;

    for(int w=1;w<=water_sources.m;w++){
        mesher.Add_Eulerian_Particles(water_sources(w)->eulerian_particles);
    }

    example.incompressible.length_scale=length_scale;
    example.incompressible.thickness_scale=thickness_scale;
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_DRIVER<TV>::Initialize_Wind_Force()
{
    switch(test){
    case bubble_break:
        {}break;
    case bubble_blow:
        {
            T radius=fixed_boundary_circle.radius;
            TV wind_box_min_corner=TV::Constant_Vector(-radius*(T)2);wind_box_min_corner.x=-radius*(T)4;wind_box_min_corner.y=-radius*(T)4;
            TV wind_box_max_corner=TV::Constant_Vector(radius*(T)30);wind_box_max_corner.x=radius*(T)4;wind_box_max_corner.y=radius*(T)4;
            BOX<TV> wind_box(wind_box_min_corner,wind_box_max_corner);
            T cell_size=radius*(T).2;
            TV_INT wind_counts=TV_INT(ceil(wind_box.Edge_Lengths()/cell_size));
            GRID<TV> wind_grid;wind_grid.Initialize(wind_counts,wind_box,true);
            wind_field.Initialize(example.geometrical_codimensional_mesh_object,wind_grid,length_scale);
            TV source_dx=wind_grid.dX*(T).4;
            TV3 source_box_min_corner=TV3(wind_box.min_corner);TV3 source_box_max_corner=TV3(wind_box.max_corner);
            source_box_min_corner.x*=(T).25;source_box_min_corner.x*=wind_source_size_factor;
            source_box_min_corner.y*=(T).25;source_box_min_corner.y*=wind_source_size_factor;
            source_box_max_corner.x*=(T).25;source_box_max_corner.x*=wind_source_size_factor;
            source_box_max_corner.y*=(T).25;source_box_max_corner.y*=wind_source_size_factor;
            source_box_max_corner.z=source_box_min_corner.z+cell_size;
            BOX<TV> source_box=BOX<TV>(TV_Compatible<T,TV::m>(source_box_min_corner)-source_dx,TV_Compatible<T,TV::m>(source_box_max_corner)+source_dx);
            
            //T radius=fixed_boundary_circle.radius;
            //TV wind_box_min_corner=TV::Constant_Vector(-radius*(T)2);wind_box_min_corner.x=-radius*(T)2;wind_box_min_corner.y=-radius*(T)2;
            //TV wind_box_max_corner=TV::Constant_Vector(radius*(T)16);wind_box_max_corner.x=radius*(T)2;wind_box_max_corner.y=radius*(T)2;
            //BOX<TV> wind_box(wind_box_min_corner,wind_box_max_corner);
            //T cell_size=radius*(T).2;
            //TV_INT wind_counts=TV_INT(ceil(wind_box.Edge_Lengths()/cell_size));
            //GRID<TV> wind_grid;wind_grid.Initialize(wind_counts,wind_box,true);
            //wind_field.Initialize(example.geometrical_codimensional_mesh_object,wind_grid,length_scale);
            //TV source_dx=wind_grid.dX*(T).4;
            //TV3 source_box_min_corner=TV3(wind_box.min_corner);TV3 source_box_max_corner=TV3(wind_box.max_corner);
            //source_box_min_corner.x*=(T).5;source_box_min_corner.x*=wind_source_size_factor;
            //source_box_min_corner.y*=(T).5;source_box_min_corner.y*=wind_source_size_factor;
            //source_box_max_corner.x*=(T).5;source_box_max_corner.x*=wind_source_size_factor;
            //source_box_max_corner.y*=(T).5;source_box_max_corner.y*=wind_source_size_factor;
            //source_box_max_corner.z=source_box_min_corner.z+cell_size;
            //BOX<TV> source_box=BOX<TV>(TV_Compatible<T,TV::m>(source_box_min_corner)-source_dx,TV_Compatible<T,TV::m>(source_box_max_corner)+source_dx);

            ////constant source
            //TV3 source_velocity=TV3::Axis_Vector(3)*wind_speed;
            //wind_field.Initialize_Source(source_box,TV_Compatible<T,TV::m>(source_velocity));
            ////variable source
            TV3 source_velocity_min=TV3::Axis_Vector(3)*wind_speed*(T).5;TV3 source_velocity_max=TV3::Axis_Vector(3)*wind_speed*(T)1.5;
            wind_field.Initialize_Source(source_box,RANGE<TV>(TV_Compatible<T,TV::m>(source_velocity_min),TV_Compatible<T,TV::m>(source_velocity_max)));
            wind_field.Initialize_Blowing_Speed(*wind_speed_bspline);
            wind_field.Use_Bubbles(&bubbles);
        }break;
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_DRIVER<TV>::Update_Fixed_Boundary_Conditions(const T time,const int tid)
{
    ARRAY<int>& particle_indices=threaded_examples(tid)->particle_indices;

    switch(test){
    case bubble_break:
    case bubble_blow:
    case membrane_oscillation:
        {
            for(int local_i=1;local_i<=particle_indices.m;local_i++){
                int i=particle_indices(local_i);
                TV3 pos=TV3(example.particles.X(i));
                if(abs(TV2(pos.x,pos.y).Magnitude()-fixed_boundary_circle.radius)<length_scale*(T).1&&abs(pos.z)<length_scale*(T).1){
                    for(int d=1;d<=TV::m;d++)threaded_examples(tid)->incompressible.psi_N.Append(CODIMENSIONAL_INCOMPRESSIBLE_NEUMANN_BOUNDARY_SAMPLE<TV>(i,TV::Axis_Vector(d),(T)0));
                    example.particles.F(i)=TV();example.particles.surface_tension_force(i)=TV();example.particles.pressure_force(i)=TV();
                    if(use_fixed_mass_on_boundary){
                        example.particles.dimensional_mass(i)(TV::m-1)=example.geometrical_codimensional_mesh_object.Nodal_Triangle_Area_Sum(i)*thickness_scale*density;
                        example.particles.mass(i)=example.particles.dimensional_mass(i)(TV::m-1);
                    }
                }
            }
        }break;
    case membrane_square:
        {
            static bool first_time=true;
            static T original_min_height=(T)0;
            static T original_max_height=(T)0;
            T v_y=(T).05;
            if(first_time){original_min_height=fixed_boundary_min_height;original_max_height=fixed_boundary_max_height;first_time=false;}

            const T threshold=(T).044;

            fixed_boundary_max_height=original_max_height+v_y*time;if(fixed_boundary_max_height>threshold)fixed_boundary_max_height=threshold;
            fixed_boundary_min_height=original_min_height-v_y*time;if(fixed_boundary_min_height<-threshold)fixed_boundary_min_height=-threshold;

            fixed_boundary_squares.min_corner.y=fixed_boundary_min_height;
            fixed_boundary_squares.max_corner.y=fixed_boundary_max_height;

            for(int local_i=1;local_i<=particle_indices.m;local_i++){
                int i=particle_indices(local_i);
                TV3 pos=TV3(example.particles.X(i));
                bool on_boundary=false;T boundary_y=(T)0;
                if(abs(pos.y-fixed_boundary_squares.min_corner.y)<length_scale*(T).1&&
                    ((abs(pos.x-fixed_boundary_squares.min_corner.x)<length_scale*(T).1||abs(pos.x-fixed_boundary_squares.max_corner.x)<length_scale*(T).1)||
                    (abs(pos.z-fixed_boundary_squares.min_corner.z)<length_scale*(T).1||abs(pos.z-fixed_boundary_squares.max_corner.z)<length_scale*(T).1))){
                    boundary_y=fixed_boundary_squares.min_corner.y;on_boundary=true;
                }
                else if(abs(pos.y-fixed_boundary_squares.max_corner.y)<length_scale*(T).1&&
                    ((abs(pos.x-fixed_boundary_squares.min_corner.x)<length_scale*(T).1||abs(pos.x-fixed_boundary_squares.max_corner.x)<length_scale*(T).1)||
                    (abs(pos.z-fixed_boundary_squares.min_corner.z)<length_scale*(T).1||abs(pos.z-fixed_boundary_squares.max_corner.z)<length_scale*(T).1))){
                    boundary_y=fixed_boundary_squares.max_corner.y;on_boundary=true;
                }

                if(on_boundary){
                    for(int d=1;d<=TV::m;d++)threaded_examples(tid)->incompressible.psi_N.Append(CODIMENSIONAL_INCOMPRESSIBLE_NEUMANN_BOUNDARY_SAMPLE<TV>(i,TV::Axis_Vector(d),(T)0));
                    example.particles.F(i)=TV();example.particles.surface_tension_force(i)=TV();example.particles.pressure_force(i)=TV();
                    example.particles.X(i).y=boundary_y;
                    if(use_fixed_mass_on_boundary){
                        example.particles.dimensional_mass(i)(TV::m-1)=example.geometrical_codimensional_mesh_object.Nodal_Triangle_Area_Sum(i)*thickness_scale*density;
                        example.particles.mass(i)=example.particles.dimensional_mass(i)(TV::m-1);
                    }
                }
            }
        }break;
    case membrane_circle:
        {
            static bool first_time=true;
            static T original_min_height=(T)0;
            static T original_max_height=(T)0;
            T v_y=(T).05;
            if(first_time){
                original_min_height=fixed_boundary_min_height;original_max_height=fixed_boundary_max_height;
                min_torus->X().y=fixed_boundary_min_height;
                max_torus->X().y=fixed_boundary_max_height;
                min_cylinder->X().y=fixed_boundary_min_height;
                max_cylinder->X().y=fixed_boundary_max_height;
                first_time=false;
            }

            const T threshold=(T).0365;

            fixed_boundary_max_height=original_max_height+v_y*time;if(fixed_boundary_max_height>threshold)fixed_boundary_max_height=threshold;
            fixed_boundary_min_height=original_min_height-v_y*time;if(fixed_boundary_min_height<-threshold)fixed_boundary_min_height=-threshold;

            min_torus->X().y=original_min_height-v_y*time;if(min_torus->X().y<-threshold)min_torus->X().y=-threshold;
            max_torus->X().y=original_max_height+v_y*time;if(max_torus->X().y>threshold)max_torus->X().y=threshold;

            min_cylinder->X().y=original_min_height-v_y*time;if(min_cylinder->X().y<-threshold)min_cylinder->X().y=-threshold;
            max_cylinder->X().y=original_max_height+v_y*time;if(max_cylinder->X().y>threshold)max_cylinder->X().y=threshold;

            for(int local_i=1;local_i<=particle_indices.m;local_i++){
                int i=particle_indices(local_i);
                TV3 pos=TV3(example.particles.X(i));
                bool on_boundary=false;T boundary_y=(T)0;
                if(abs(pos.y-fixed_boundary_min_height)<length_scale*(T).1&&
                    abs(TV2(pos.x,pos.z).Magnitude()-fixed_boundary_circle.radius)<length_scale*(T).1){
                    boundary_y=fixed_boundary_min_height;on_boundary=true;
                }
                else if(abs(pos.y-fixed_boundary_max_height)<length_scale*(T).1&&
                    abs(TV2(pos.x,pos.z).Magnitude()-fixed_boundary_circle.radius)<length_scale*(T).1){
                    boundary_y=fixed_boundary_max_height;on_boundary=true;
                }

                if(on_boundary){
                    for(int d=1;d<=TV::m;d++)threaded_examples(tid)->incompressible.psi_N.Append(CODIMENSIONAL_INCOMPRESSIBLE_NEUMANN_BOUNDARY_SAMPLE<TV>(i,TV::Axis_Vector(d),(T)0));
                    example.particles.F(i)=TV();example.particles.surface_tension_force(i)=TV();example.particles.pressure_force(i)=TV();
                    example.particles.X(i).y=boundary_y;
                    if(use_fixed_mass_on_boundary){
                        example.particles.dimensional_mass(i)(TV::m-1)=example.geometrical_codimensional_mesh_object.Nodal_Triangle_Area_Sum(i)*thickness_scale*density;
                        example.particles.mass(i)=example.particles.dimensional_mass(i)(TV::m-1);
                    }
                }
            }
        }break;
    }

    Thread_Barrier_Wait(tid);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_DRIVER<TV>::Update_Solid_States(const T dt,const T time)
{
    switch(test){
        case paint:{
            RIGID_BODY<TV>* rock=rigid_bodies(1);
            if(use_rock_key_frames){
                RIGID_BODY_STATE<TV> interpolated_rock_state;
                T key_frame_fraction=((time-rock_key_frames(1).time)/(rock_key_frames(rock_key_frames.m).time-rock_key_frames(1).time));
                clamp(key_frame_fraction,(T)0.,(T)1.);
                int current_key_frame=(int)(key_frame_fraction*(rock_key_frames.m-1)+1);
                rock->Interpolate_Between_States(rock_key_frames(current_key_frame),rock_key_frames(current_key_frame+1),time,interpolated_rock_state);
                rock->Restore_State(interpolated_rock_state);}
            else{
                rock->X()+=rock->V()*dt;
                T half_rock_height=(T)0.0381;
                if(rock->X().y<tank_box.min_corner.y+half_rock_height){
                    rock->X().y=tank_box.min_corner.y+half_rock_height;
                    rock->V()=TV();}}
        }break;
        
        case bubble_break:{
            RIGID_BODY<TV>* bullet=rigid_bodies(1);
            static bool is_first_time=true;
            if(is_first_time){
                //TV bullet_initial_pos=TV((T)-.1,(T).05,(T)0);
                //bullet->X()=bullet_initial_pos;
                //bullet->V()=TV::Axis_Vector(1)*(T)5;
                TV bullet_initial_pos=TV((T)-.0335,(T).0491267,(T)0);
                bullet->X()=bullet_initial_pos;
                bullet->V()=TV((T)5*(T)1.5,(T)-.13034*(T)2,(T)0);
                is_first_time=false;
            }

            TV g=TV::Axis_Vector(2)*(T)(-9.8)*(T)2;
            bullet->V()+=g*dt;
            bullet->X()+=bullet->V()*dt;
            LOG::cout<<"bullet: "<<bullet->X()<<", "<<bullet->V()<<std::endl;
        }break;
    }
}

