//#####################################################################
// Copyright 2013, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __CODIMENSIONAL_KRYLOV_SYSTEM__
#define __CODIMENSIONAL_KRYLOV_SYSTEM__
#include <PhysBAM_Tools/Krylov_Solvers/KRYLOV_SYSTEM_BASE.h>
#include <PhysBAM_Tools/Krylov_Solvers/KRYLOV_VECTOR_WRAPPER.h>
#include <PhysBAM_Tools/Vectors/Dot_Product.h>
#include "../Mpi/MPI_CODIMENSIONAL.h"

namespace PhysBAM{
template<class T> class SPARSE_MATRIX_FLAT_NXN;
template<class T> class VECTOR_ND;

template<class T>
class CODIMENSIONAL_KRYLOV_SYSTEM:public KRYLOV_SYSTEM_BASE<T>
{
    typedef VECTOR_ND<T> VECTOR_T;typedef SPARSE_MATRIX_FLAT_NXN<T> T_MATRIX;typedef KRYLOV_VECTOR_WRAPPER<T,VECTOR_T&> KRYLOV_VECTOR_T;

    const T_MATRIX& A;
    const ARRAY<VECTOR_ND<T> >& null_space_basis;

    mutable VECTOR_T temp;
    MPI_CODIMENSIONAL<VECTOR<T,3> >* mpi_codim;
    ARRAY<ARRAY<int> >* boundary_pressure_indices;
    ARRAY<ARRAY<int> >* ghost_pressure_indices;
    ARRAY<int>* domain_pressure_indices;

public:

    CODIMENSIONAL_KRYLOV_SYSTEM(const T_MATRIX& A,const ARRAY<VECTOR_ND<T> >& null_space_basis,MPI_CODIMENSIONAL<VECTOR<T,3> >* mpi_codim_input=0,
        ARRAY<ARRAY<int> >* boundary_pressure_indices_input=0,ARRAY<ARRAY<int> >* ghost_pressure_indices_input=0,ARRAY<int>* domain_pressure_indices_input=0)
        :KRYLOV_SYSTEM_BASE<T>(false,true),A(A),null_space_basis(null_space_basis),mpi_codim(mpi_codim_input),
        boundary_pressure_indices(boundary_pressure_indices_input),ghost_pressure_indices(ghost_pressure_indices_input),domain_pressure_indices(domain_pressure_indices_input)
    {}

    void Multiply(const KRYLOV_VECTOR_BASE<T>& bx,KRYLOV_VECTOR_BASE<T>& bresult) const PHYSBAM_OVERRIDE
    {
        const KRYLOV_VECTOR_T& x=debug_cast<const KRYLOV_VECTOR_T&>(bx);
        KRYLOV_VECTOR_T& result=debug_cast<KRYLOV_VECTOR_T&>(bresult);
        result.v.Resize(A.n);
        if(mpi_codim)mpi_codim->Exchange_Boundary_And_Ghost_Pressure_Values(x.v,*boundary_pressure_indices,*ghost_pressure_indices);
        A.Times(x.v,result.v);
    }

    double Inner_Product(const KRYLOV_VECTOR_BASE<T>& bx,const KRYLOV_VECTOR_BASE<T>& by) const PHYSBAM_OVERRIDE
    {
        const KRYLOV_VECTOR_T& x=debug_cast<const KRYLOV_VECTOR_T&>(bx),&y=debug_cast<const KRYLOV_VECTOR_T&>(by);
        if(mpi_codim){
            double inner_product=0.;
            for(int i=1;i<=(*domain_pressure_indices).m;i++){
                int p=(*domain_pressure_indices)(i);
                inner_product+=(double)x.v(p)*(double)y.v(p);
            }
            return mpi_codim->Sum(inner_product);
        }
        else return Dot_Product_Double_Precision(x.v,y.v);
    }

    T Convergence_Norm(const KRYLOV_VECTOR_BASE<T>& bx) const PHYSBAM_OVERRIDE
    {
        const KRYLOV_VECTOR_T& x=debug_cast<const KRYLOV_VECTOR_T&>(bx);
        if(mpi_codim){
            T norm=(T)0;
            for(int i=1;i<=(*domain_pressure_indices).m;i++){
                int p=(*domain_pressure_indices)(i);
                norm=PhysBAM::max(norm,abs(x.v(p)));
            }
            return mpi_codim->Max(norm);
        }
        else return x.v.Maximum_Magnitude();
    }

    void Project_Nullspace(KRYLOV_VECTOR_BASE<T>& bx) const PHYSBAM_OVERRIDE
    {
        KRYLOV_VECTOR_T& x=debug_cast<KRYLOV_VECTOR_T&>(bx);
        if(mpi_codim){
            for(int i=1;i<=null_space_basis.m;i++){
                double dot_product=0.;
                //for(int j=1;j<=(*domain_pressure_indices).m;j++){
                    //int p=(*domain_pressure_indices)(i);
                    //local_dot_product+=x.v(p)*null_space_basis(i)(pressure_samples(p).index);
                //}

                dot_product=mpi_codim->Sum(dot_product);
                //for(int j=1;j<=(*domain_pressure_indices).m;j++){
                    //int p=(*domain_pressure_indices)(i);
                    //x.v(p)-=dot_product*null_space_basis(i)(pressure_samples(p).index);
                //}
            }
        }
        else{
            for(int i=1;i<=null_space_basis.m;i++){x.v-=VECTOR_ND<T>::Dot_Product(x.v,null_space_basis(i))*null_space_basis(i);}
        }
    }

    void Set_Boundary_Conditions(KRYLOV_VECTOR_BASE<T>& x) const PHYSBAM_OVERRIDE
    {Project_Nullspace(x);}

    void Apply_Preconditioner(const KRYLOV_VECTOR_BASE<T>& br,KRYLOV_VECTOR_BASE<T>& bz) const PHYSBAM_OVERRIDE // solve Mz=r
    //{const KRYLOV_VECTOR_T& r=debug_cast<const KRYLOV_VECTOR_T&>(br);KRYLOV_VECTOR_T& z=debug_cast<KRYLOV_VECTOR_T&>(bz);
    //temp.Resize(A.n);z.v.Resize(A.n);
    //A.C->Solve_Forward_Substitution(r.v,temp,true); // diagonal should be treated as the identity
    //A.C->Solve_Backward_Substitution(temp,z.v,false,true);} // diagonal is inverted to save on divides
    {   
        ////diagonal preconditioner
        const KRYLOV_VECTOR_T& rr=debug_cast<const KRYLOV_VECTOR_T&>(br);
        KRYLOV_VECTOR_T& zz=debug_cast<KRYLOV_VECTOR_T&>(bz);
        //LOG::cout<<"here: "<<rr.v.n<<", "<<zz.v.n<<std::endl;
        for(int i=1;i<=rr.v.n;i++){
            //LOG::cout<<i<<": "<<A.diagonal_index(i)<<": "<<A.A(A.diagonal_index(i)).a<<", rr: "<<rr.v(i)<<std::endl;
            PHYSBAM_ASSERT(A.A(A.diagonal_index(i)).a!=(T)0);
            zz.v(i)=rr.v(i)/A.A(A.diagonal_index(i)).a;
        }
        if(mpi_codim)mpi_codim->Exchange_Boundary_And_Ghost_Pressure_Values(zz.v,*boundary_pressure_indices,*ghost_pressure_indices);
    }

    void Project(KRYLOV_VECTOR_BASE<T>& x) const PHYSBAM_OVERRIDE {}
    //#####################################################################
};
}
#endif
