//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <PhysBAM_Tools/Grids_Uniform/UNIFORM_GRID_ITERATOR_CELL.h>
#include <PhysBAM_Tools/Nonlinear_Equations/ITERATIVE_SOLVER.h>
#include <PhysBAM_Dynamics/Geometry/GENERAL_GEOMETRY_FORWARD.h>
#include <PhysBAM_Dynamics/Particles/PARTICLES_FORWARD.h>
#include <PhysBAM_Dynamics/Read_Write/Particles/READ_WRITE_PARTICLES.h>
#include <PhysBAM_Solids/PhysBAM_Deformables/Forces/LINEAR_SPRINGS.h>
#include <PhysBAM_Solids/PhysBAM_Deformables/Forces/SEGMENT_BENDING_SPRINGS.h>
#include <PhysBAM_Solids/PhysBAM_Deformables/Forces/DEFORMABLE_GRAVITY.h>
#include <PhysBAM_Solids/PhysBAM_Deformables/Collisions_And_Interactions/DEFORMABLE_OBJECT_COLLISIONS.h>
#include <PhysBAM_Solids/PhysBAM_Deformables/Collisions_And_Interactions/DEFORMABLE_OBJECT_COLLISION_PARAMETERS.h>
#include <PhysBAM_Solids/PhysBAM_Deformables/Collisions_And_Interactions/TRIANGLE_REPULSIONS_AND_COLLISIONS_GEOMETRY.h>
#include <PhysBAM_Solids/PhysBAM_Rigids/Forces_And_Torques/RIGIDS_EXAMPLE_FORCES_AND_VELOCITIES.h>
#include <PhysBAM_Geometry/Solids_Geometry/DEFORMABLE_GEOMETRY_COLLECTION.h>
#include <PhysBAM_Geometry/Basic_Geometry/TETRAHEDRON.h>
#include <PhysBAM_Geometry/Basic_Geometry/BOX.h>
#include <PhysBAM_Geometry/Basic_Geometry/CYLINDER.h>
#include <PhysBAM_Geometry/Level_Sets/IMPLICIT_OBJECT_ON_A_RAY.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_RIGID_GEOMETRY_COLLECTION.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_STRUCTURE.h>
#include <PhysBAM_Tools/Interpolation/BSPLINE.h>
#include <PhysBAM_Tools/Interpolation/LINEAR_INTERPOLATION.h>
#include <PhysBAM_Tools/Read_Write/Utilities/FILE_UTILITIES.h>
#include <fstream>
#include "../Meshing/CODIMENSIONAL_MESH_GENERATOR.h"
#include "../Viscoelastic/FIBER.h"
#include "../Meshing/CODIMENSIONAL_HELPER.h"

using namespace PhysBAM;

template<class TV>
PhysBAM::FIBER<TV>::FIBER()
    :fiber_geometry(fiber_mesh,fiber_particles),fiber_deformables_example(this),
    deformable_body_collection(fiber_deformables_example.deformable_body_collection),deformable_geometry_collection(deformable_body_collection.deformable_geometry),rigid_geometry_collection(fiber_deformables_example.rigid_geometry_collection),
    collision_body_list(fiber_deformables_example.collision_body_list),deformables_parameters(fiber_deformables_example.deformables_parameters),rigid_body_type(NONE),
    density((T)1e4),fiber_bending_start_frac((T).5),fiber_bending_perturb_frac(TV()),fiber_particle_thickness(0,0),fiber_particle_forces(0,0),implicit_fiber_segment_number(3),fiber_deformable_model(RIGID),fiber_evolution(deformables_parameters,deformable_body_collection,rigid_geometry_collection,fiber_geometry),use_gravity(false),
    linear_stiffness_coefficient((T)10.),linear_damping_coefficient((T)1.0),bending_stiffness_coefficient((T)10.),bending_damping_coefficient((T)1.0),
    use_guided_fibers(false),guided_deformable_geometry_collection(guided_fiber_particles,false),guided_fiber_geometry(guided_fiber_mesh,guided_fiber_particles),
    spatial_partition(0),use_prescribed_motion(false),file_frame_count(0),use_virtual_springs(false),virtual_linear_stiffness_coefficient((T)1),virtual_linear_damping_coefficient((T)1),virtual_spring_force_id(0),virtual_linear_springs(0),mpi_codim(0)
{
    fiber_particles.Store_Mass();fiber_particles.Store_One_Over_Mass();fiber_particles.Store_Effective_Mass();fiber_particles.Store_One_Over_Effective_Mass();
    fiber_particles.array_collection->Add_Array(ATTRIBUTE_ID_THICKNESS,&fiber_particle_thickness);
    fiber_particles.array_collection->Add_Array(ATTRIBUTE_ID_FORCE,&fiber_particle_forces);
    guided_fiber_particles.Store_Mass();guided_fiber_particles.Store_One_Over_Mass();guided_fiber_particles.Store_Effective_Mass();guided_fiber_particles.Store_One_Over_Effective_Mass();
    fiber_evolution.Set_Deformables_Evolution_Callbacks(deformables_evolution_callbacks);
    random.Set_Seed(12309191);
    use_fiber_length_variation_on_x_axis=true;
}

template<class TV>
void PhysBAM::FIBER<TV>::Initialize_Structure()
{
    deformable_body_collection.Update_Simulated_Particles();
    rigid_geometry_collection.Update_Kinematic_Particles();

    ////initialize collisions
    DEFORMABLE_OBJECT_COLLISION_PARAMETERS<TV>& deformable_object_collision_parameters=deformables_parameters.deformable_object_collision_parameters;
    deformable_body_collection.collisions.Initialize_Object_Collisions(deformable_object_collision_parameters.collide_with_interior,deformable_object_collision_parameters.collision_tolerance,
        deformable_object_collision_parameters.use_spatial_partition_for_levelset_collision_objects,deformable_object_collision_parameters.disable_multiple_levelset_collisions,
        deformable_object_collision_parameters.maximum_levelset_collision_projection_velocity);
    deformable_body_collection.collisions.collision_body_list.Update_Spatial_Partition(deformable_object_collision_parameters.spatial_partition_voxel_size_heuristic,
        deformable_object_collision_parameters.spatial_partition_number_of_cells,deformable_object_collision_parameters.spatial_partition_voxel_size_scale_factor);

    ////initialize cfl
    deformable_body_collection.Set_CFL_Number(deformables_parameters.cfl);
}

template<class TV>
void PhysBAM::FIBER<TV>::Initialize_Forces()
{
    if(use_gravity)LOG::cout<<"fiber has gravity"<<std::endl;

    if(fiber_deformable_model==MASS_SPRING)Initialize_Mass_Spring_Forces();
    if(use_gravity)deformable_body_collection.Add_Force(new DEFORMABLE_GRAVITY<TV>(deformable_body_collection.particles,fiber_mesh));
    Initialize_Structure();
}

template<class TV>
void PhysBAM::FIBER<TV>::Initialize_Mass_Spring_Forces()
{
    fiber_deformable_model=MASS_SPRING;

    ////springs parameters
    T linear_stiffness=(T)1*linear_stiffness_coefficient;
    //T keep_constant_damping_fraction=(T)1./sqrt(linear_stiffness);
    //T linear_overdamping_fraction=(T)1e3*keep_constant_damping_fraction*linear_damping_coefficient;
    T linear_overdamping_fraction=linear_damping_coefficient;
    T bending_stiffness=bending_stiffness_coefficient;
    //T keep_constant_bending_damping_fraction=(T)1./sqrt(bending_stiffness);
    //T bending_overdamping_fraction=(T)1e3*keep_constant_bending_damping_fraction*bending_damping_coefficient;
    T bending_overdamping_fraction=(T)2*bending_damping_coefficient;
    bool limit_time_step_by_strain_rate=true;
    T max_strain_per_time_step=(T).1;
    bool use_rest_state_for_strain_rate=true;
    T restlength_enlargement_fraction=(T)0;
    bool verbose=false;
    bool implicit=true;
    bool fully_implicit=false;

    LOG::cout<<"linear: "<<linear_stiffness<<", "<<linear_overdamping_fraction<<", bending: "<<bending_stiffness<<", "<<bending_overdamping_fraction<<std::endl;

    LINEAR_SPRINGS<TV>* linear_springs=Create_Edge_Springs(fiber_geometry,linear_stiffness,linear_overdamping_fraction,
        limit_time_step_by_strain_rate,max_strain_per_time_step,use_rest_state_for_strain_rate,restlength_enlargement_fraction,verbose,implicit);
    deformable_body_collection.Add_Force(linear_springs);

    SEGMENT_BENDING_SPRINGS<TV>* bending_springs=Create_Bending_Springs(fiber_geometry,bending_stiffness,bending_overdamping_fraction,
        limit_time_step_by_strain_rate,max_strain_per_time_step,use_rest_state_for_strain_rate,restlength_enlargement_fraction,verbose,implicit);
    deformable_body_collection.Add_Force(bending_springs);

    bool use_variable_bending_coefficient=true;
    if(use_variable_bending_coefficient){
        T max_bending_stiffness_coef=(T)1;T min_bending_stiffness_coef=(T)1;
        T max_thickness=(T)-FLT_MAX;T min_thickness=(T)FLT_MAX;
        for(int i=1;i<=fiber_particle_thickness.m;i++){
            if(fiber_particle_thickness(i)<min_thickness)min_thickness=fiber_particle_thickness(i);
            if(fiber_particle_thickness(i)>max_thickness)max_thickness=fiber_particle_thickness(i);
        }

        ARRAY<T> variable_bending_stiffness(fiber_mesh.elements.m);
        for(int i=1;i<=fiber_mesh.elements.m;i++){
            const TV_INT2& v=fiber_mesh.elements(i);T thickness=(T).5*(fiber_particle_thickness(v(1))+fiber_particle_thickness(v(2)));
            T frac=(thickness-min_thickness)/(max_thickness-min_thickness);
            T local_bending_stiffness_coef=((T)1-frac)*min_bending_stiffness_coef+frac*max_bending_stiffness_coef;
            variable_bending_stiffness(i)=local_bending_stiffness_coef*bending_stiffness;
        }
        ARRAY_VIEW<T> variable_bending_stiffness_arrayview(variable_bending_stiffness);
        bending_springs->Set_Stiffness(variable_bending_stiffness_arrayview);
    }

    if(use_virtual_springs){
        T_GEOMETRY virtual_geometry(virtual_mesh,fiber_particles);
        T virtual_linear_stiffness=(T)1*virtual_linear_stiffness_coefficient;
        T virtual_linear_overdamping_fraction=(T)2*virtual_linear_damping_coefficient;
        virtual_linear_springs=Create_Edge_Springs(virtual_geometry,virtual_linear_stiffness,virtual_linear_overdamping_fraction,
            limit_time_step_by_strain_rate,max_strain_per_time_step,use_rest_state_for_strain_rate,restlength_enlargement_fraction,verbose,implicit);
        virtual_spring_restlength=virtual_linear_springs->restlength;
        virtual_spring_force_id=virtual_linear_springs->unique_id;
        deformable_body_collection.Add_Force(virtual_linear_springs);
    }

    ////full implicit
    if(fully_implicit)for(int i=1;i<=deformable_body_collection.deformables_forces.m;i++){
        deformable_body_collection.deformables_forces(i)->use_implicit_velocity_independent_forces=true;
    }
}

template<class TV>
void PhysBAM::FIBER<TV>::Span_Fibers(const T span_coef/*=(T)1.*/)
{
    if(virtual_linear_springs==0)return;
    ARRAY<T> current_restlength;current_restlength.Resize(virtual_spring_restlength.m);ARRAY_VIEW<T> current_restlength_arrayview(current_restlength);
    for(int i=1;i<=virtual_spring_restlength.m;i++){current_restlength(i)=virtual_spring_restlength(i)*span_coef;}
    virtual_linear_springs->Set_Restlength(current_restlength);
}

template<class TV>
void PhysBAM::FIBER<TV>::Use_Virtual_Springs(bool use/*=true*/)
{
    use_virtual_springs=use;
    if(!use_virtual_springs){
        int id_to_rmv=0;
        for(int i=1;i<=deformable_body_collection.deformables_forces.m;i++){
            if(deformable_body_collection.deformables_forces(i)->unique_id==virtual_spring_force_id){id_to_rmv=i;break;}
        }
        if(id_to_rmv)deformable_body_collection.deformables_forces.Remove_Index(id_to_rmv);
        delete virtual_linear_springs;virtual_linear_springs=0;
    }
}

template<class TV>
void PhysBAM::FIBER<TV>::Initialize_Geometry(const ARRAY<TV>& positions,const T length_scale_input,const T fiber_radius_input,const T fiber_length_input,const T guided_fiber_radius_input/*=(T)0*/,const T guided_fiber_distance_input/*=(T)0*/)
{
    length_scale=length_scale_input;
    fiber_radius=fiber_radius_input;
    fiber_length=fiber_length_input;
    if(guided_fiber_radius_input!=(T)0||guided_fiber_distance_input!=(T)0){
        use_guided_fibers=true;
        guided_fiber_radius=guided_fiber_radius_input;
        guided_fiber_distance=guided_fiber_distance_input;
    }

    TV dir=TV::Axis_Vector(2)*(T)-1;
    LOG::cout<<"fiber rotation: "<<object_frame.Rotation()<<std::endl;
    TV center=TV();for(int i=1;i<=positions.m;i++)center+=positions(i);if(positions.m>0)center/=(T)positions.m;T epsilon=(T)1e-20;
    T farthest_distance=(T)0;T min_z=(T)FLT_MAX;T max_z=(T)-FLT_MAX;T min_x=(T)FLT_MAX;T max_x=(T)-FLT_MAX;
    for(int i=1;i<=positions.m;i++){
        TV object_X=(positions(i)-center);
        T distance=object_X.Magnitude();if(distance>farthest_distance)farthest_distance=distance;
        if(object_X.z<min_z)min_z=object_X.z;if(object_X.z>max_z)max_z=object_X.z;
        if(object_X.x<min_x)min_x=object_X.x;if(object_X.x>max_x)max_x=object_X.x;
    }

    for(int i=1;i<=positions.m;i++){
        TV object_X=positions(i)-center;T ratio=object_X.Magnitude()/(farthest_distance+epsilon);TV dir2=object_X.Normalized();T max_angle=(T)pi*(T).0;
        TV axis=TV::Cross_Product(dir,dir2);TV perturbed_dir=ROTATION<TV>(max_angle*ratio,axis).Rotate(dir);

        //int periodic_fiber_num=10;
        //T freq=(max_z-min_z)/(periodic_fiber_num*(T)2*fiber_radius);/*T freq=(T)1;*/
        //T period=(T)1/freq;T frac=(object_X.z-min_z)/(max_z-min_z+epsilon);T fracfrac=(frac-floor(frac/period)*period)/period;fracfrac=(T)2*abs(fracfrac-(T).5);
        //if(fracfrac<(T).4)fracfrac=random.Get_Uniform_Number((T).95,(T)1);
        //else fracfrac=random.Get_Uniform_Number((T).1,(T).4);
        //T length_perturb_frac=(T).1;T target_fiber_length_frac=((T)1-length_perturb_frac)+(T)length_perturb_frac*fracfrac;

        ////T target_fiber_length_frac=(T).8+(T).2*random.Get_Uniform_Number((T)0,(T)1);
        //int type=random.Get_Uniform_Number((T)0,(T)1)>(T).5?0:1;
        //T target_fiber_length_frac=(T)1;
        //if(type==0)target_fiber_length_frac=(T).95+(T).02*random.Get_Uniform_Number((T)0,(T)1);
        //else target_fiber_length_frac=(T).98+(T).02*random.Get_Uniform_Number((T)0,(T)1);
        //T target_fiber_length=target_fiber_length_frac*fiber_length;
        T target_fiber_length=fiber_length;

        ////vary length based on x coord
        
        T x_frac=(T)0;
        if(use_fiber_length_variation_on_x_axis){
            x_frac=(object_X.x-min_x)/(max_x-min_x);T length_perturb=(T).15*((T)1-x_frac);
            target_fiber_length=((T).85+length_perturb)*fiber_length;
        }

        Add_Fiber(fiber_particles,positions(i),perturbed_dir,length_scale,fiber_radius,target_fiber_length,fiber_bending_start_frac,fiber_bending_perturb_frac,
            fiber_start_indices,fiber_end_indices,fiber_mesh,&near_end_segment_hashtable,&kinematic_particle_object_space_positions);

        ////set longer fibers to be trace fiber
        if(x_frac<(T).5){
            int end_index=fiber_end_indices(fiber_end_indices.m);
            Initialize_Trace_Segment_Helper(fiber_trace_segment_hashtable,end_index,implicit_fiber_segment_number);
        }
    }

    if(use_guided_fibers){
        for(int i=1;i<=fiber_start_indices.m;i++){
            T guiding_square_length=fiber_radius*(T)2;
            int guided_fiber_n=(int)floor(guiding_square_length/guided_fiber_distance);T half_length=(T)(guided_fiber_n-1)*guided_fiber_distance*(T).5;
            ARRAY<TV> guided_positions;TV start_pos=TV::Axis_Vector(1)*(T)-half_length+TV::Axis_Vector(3)*(T)-half_length;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles(guided_positions,guided_fiber_n,guided_fiber_n,start_pos,guided_fiber_distance,TV::Axis_Vector(1),TV::Axis_Vector(3));
            for(int j=1;j<=guided_positions.m;j++){
                TV pos=guided_positions(j)+object_frame.Inverse_Times(fiber_particles.X(fiber_start_indices(i)));
                Add_Fiber(guided_fiber_particles,pos,dir,length_scale,guided_fiber_radius,fiber_length,fiber_bending_start_frac,fiber_bending_perturb_frac,
                    guided_fiber_start_indices,guided_fiber_end_indices,guided_fiber_mesh);
            }
        }
        guided_deformable_geometry_collection.Add_Structure(&guided_fiber_geometry);
        guided_deformable_geometry_collection.structures.Last()->Update_Number_Nodes();
        Initialize_Guided_Fiber_Weights();
    }

    if(use_virtual_springs){
        int offset=6;//T distance_threshold=fiber_distance*(T)2;
        for(int i=1;i<fiber_start_indices.m;i++){   ////assuming particles are located in a 2D array
            int p1=fiber_start_indices(i)+offset;
            int start2=fiber_start_indices(i+1);
            int p2=0;T nearest_distance=(T)FLT_MAX;
            for(int j=start2;j<=fiber_end_indices(i+1);j++){
                T distance=(fiber_particles.X(p1)-fiber_particles.X(j)).Magnitude();
                if(distance<nearest_distance){p2=j;nearest_distance=distance;}
            }
            virtual_mesh.elements.Append(TV_INT2(p1,p2));
        }
        LOG::cout<<"virtual springs: "<<virtual_mesh.elements<<std::endl;
    }

    deformable_geometry_collection.Add_Structure(&fiber_geometry);
    deformable_geometry_collection.structures.Last()->Update_Number_Nodes();

    deformable_body_collection.collisions.collision_structures.Append_Elements(deformable_body_collection.deformable_geometry.structures);
    deformable_body_collection.triangle_repulsions_and_collisions_geometry.structures.Append_Elements(deformable_body_collection.deformable_geometry.structures);

    Initialize_Mass_Based_On_Geometry();
    Initialize_Implicit_Geometry();
}

template<class TV>
void PhysBAM::FIBER<TV>::Add_Fiber(PARTICLES<TV>& particles,const TV& start_pos,const TV& dir,const T length_scale,const T radius,const T target_fiber_length,const T& perturb_start_frac,const TV& perturb_frac,
    ARRAY<int>& start_indices,ARRAY<int>& end_indices,T_MESH& mesh,HASHTABLE<int>* near_end_segment_hashtable/*=0*/,ARRAY<PAIR<int,TV> >* kinematic_positions/*=0*/)
{
    bool use_adpative_fiber_particles=true;T min_length_scale=(T).5*length_scale;T adaptive_start_frac=(T).6;T adaptive_end_frac=(T).8;
    bool use_variable_thickness=true;T max_thichness=radius*(T)2;T min_thickness=radius*(T)1;T variable_thickness_start_frac=(T).6;T variable_thickness_end_frac=(T).8;

    ARRAY_VIEW<T>* particle_thickness=particles.array_collection->template Get_Array<T>(ATTRIBUTE_ID_THICKNESS);
    int first=0;int last=0;T current_fiber_length=(T)0;TV current_fiber_endpoint=start_pos;
    int k=0;
    while(current_fiber_length<target_fiber_length){
        if(current_fiber_length>=target_fiber_length)break;
        int idx=particles.array_collection->Add_Element();
        if(k==0)first=idx;last=idx;
        T segment_length=length_scale;
        
        if(use_adpative_fiber_particles){
            T frac=(T)current_fiber_length/target_fiber_length;
            T fracfrac=(frac-adaptive_start_frac)/(adaptive_end_frac-adaptive_start_frac);fracfrac=clamp(fracfrac,(T)0,(T)1);
            segment_length=((T)1-fracfrac)*length_scale+fracfrac*min_length_scale;
        }

        T length_dif=target_fiber_length-current_fiber_length;
        if(length_dif>segment_length*(T).5&&length_dif<segment_length*(T)1.5)segment_length=length_dif*(T)1.01;

        TV pos=current_fiber_endpoint+dir*segment_length;
        //LOG::cout<<"dir: "<<dir<<", seg_length: "<<segment_length<<std::endl;
        particles.X(idx)=object_frame*pos;
        current_fiber_length+=segment_length;
        current_fiber_endpoint=pos;

        T thickness=radius*(T)2;

        if(use_variable_thickness){
            T frac=(T)current_fiber_length/target_fiber_length;
            T fracfrac=(frac-variable_thickness_start_frac)/(variable_thickness_end_frac-variable_thickness_start_frac);fracfrac=clamp(fracfrac,(T)0,(T)1);
            thickness=((T)1-fracfrac)*max_thichness+fracfrac*min_thickness;
        }

        if(particle_thickness)(*particle_thickness)(idx)=thickness;
        k++;
    }
    start_indices.Append(first);end_indices.Append(last);
    TV perturb_scale=target_fiber_length*perturb_frac;int bspline_control_point_n=6;
    Randomize_Fiber(particles,first,last,bspline_control_point_n,perturb_start_frac,perturb_scale);

    int near_end_segment_num=5;
    for(int k=first;k<last;k++){int s=mesh.elements.Append(TV_INT2(k,k+1));if(near_end_segment_hashtable&&last-k<=near_end_segment_num){near_end_segment_hashtable->Set(s);/*LOG::cout<<"near seg: "<<s<<std::endl;*/}}

    if(kinematic_positions)for(int k=first;k<=last;k++){if(k==first||k==first+1||k==first+2||fiber_deformable_model==RIGID)(*kinematic_positions).Append(PAIR<int,TV>(k,object_frame.Inverse_Times(particles.X(k))));}
}

template<class TV>
void PhysBAM::FIBER<TV>::Randomize_Fiber(PARTICLES<TV>& particles,const int first,const int last,const int control_point_n,const T& perturb_start_frac,const TV& perturb)
{
    TV dir=(particles.X(last)-particles.X(first)).Normalized();
    ARRAY<T> control_times;ARRAY<TV> control_points;int order=2;
    for(int i=1;i<=control_point_n;i++){
        T frac=(T)(i-1)/(T)(control_point_n-1);//T frac_perturb=random.Get_Uniform_Number((T)-.1,(T).1);frac+=frac_perturb;frac=clamp(frac,(T)0,(T)1);
        TV point=((T)1-frac)*particles.X(first)+frac*particles.X(last);
        if(i>control_point_n*perturb_start_frac){TV perturb_vec=random.Get_Uniform_Vector(-perturb,perturb);perturb_vec.y=(T)0;point+=perturb_vec;}
        if(i==1)for(int j=1;j<=order;j++){control_times.Append((T)0);control_points.Append(point);}
        control_times.Append((T)frac);control_points.Append(point);
        if(i==control_point_n)for(int j=1;j<=order;j++){control_times.Append((T)(i+j-1)/(T)(control_point_n-1));control_points.Append(point);}
    }
    BSPLINE<T,TV> fiber_bspline(control_times,control_points,order);
    T fiber_length=(particles.X(last)-particles.X(first)).Magnitude();
    TV first_pos=particles.X(first);
    for(int i=first;i<=last;i++){
        //T frac=(T)(i-first)/(T)(last-first);
        T length=(particles.X(i)-first_pos).Magnitude();T frac=length/fiber_length;
        TV pos=fiber_bspline.Clamped_Evaluate(frac);particles.X(i)=pos;
    }
}

template<class TV>
void PhysBAM::FIBER<TV>::Initialize_Implicit_Geometry_Helper(HASHTABLE<int,IMPLICIT_FIBER_GEOMETRY<TV> >& implicit_geometry_hashtable,ARRAY<int>& implicit_fiber_elements,
    const ARRAY<int>& end_indices,const PARTICLES<TV>& particles,const SEGMENT_MESH& segment_mesh,const FIBER_SHAPE fiber_shape,int segment_number_to_initialize/*=1*/)
{
    HASHTABLE<TV_INT2,int> segment_index_hashtable;
    for(int i=1;i<=segment_mesh.elements.m;i++)segment_index_hashtable.Set(Unique_Ordered(segment_mesh.elements(i)),i);
    //implicit_geometry.Resize(end_indices.m*segment_number_to_initialize);
    //implicit_frames.Resize(end_indices.m*segment_number_to_initialize);
    for(int i=1;i<=end_indices.m;i++){
        for(int s=1;s<=segment_number_to_initialize;s++){
            int p=end_indices(i)+1-s;
            //int k=(i-1)*segment_number_to_initialize+s;

            int seg_index=0;if(!segment_index_hashtable.Get(Unique_Ordered(TV_INT2(p-1,p)),seg_index))continue;
            implicit_fiber_elements.Append(seg_index);
            T radius=(T).25*(fiber_particle_thickness(p-1)+fiber_particle_thickness(p));
            IMPLICIT_OBJECT<TV>* implicit=0;
            T height=(particles.X(p-1)-particles.X(p)).Magnitude();
            if(fiber_shape==GUIDE_BOX){
                BOX<TV> box(TV(-radius,(T)0,-radius),TV(radius,height,radius));
                implicit=new ANALYTIC_IMPLICIT_OBJECT<BOX<TV> >(box);
            }
            else if(fiber_shape==GUIDE_CYLINDER){
                CYLINDER<T> cylinder(TV(),TV::Axis_Vector(2)*height,radius);
                implicit=new ANALYTIC_IMPLICIT_OBJECT<CYLINDER<T> >(cylinder);
            }

            TV translation=particles.X(p);
            TV vector=(particles.X(p-1)-particles.X(p)).Normalized();
            ROTATION<TV> rotation=ROTATION<TV>::From_Rotated_Vector(TV::Axis_Vector(2),vector);
            
            //implicit_geometry(k)=new IMPLICIT_OBJECT_TRANSFORMED<TV,FRAME<TV> >(implicit,true,frame);
            //implicit_frames(k)=frame;
            FRAME<TV>* frame=new FRAME<TV>(translation,rotation);
            IMPLICIT_OBJECT_TRANSFORMED<TV,FRAME<TV> >* implicit_object_transformed=new IMPLICIT_OBJECT_TRANSFORMED<TV,FRAME<TV> >(implicit,true,frame);
            
            implicit_geometry_hashtable.Set(seg_index,IMPLICIT_FIBER_GEOMETRY<TV>(implicit_object_transformed,frame));
        }
    }
}

template<class TV>
void PhysBAM::FIBER<TV>::Initialize_Trace_Segment_Helper(HASHTABLE<int>& fiber_trace_segment_hashtable,const int end_index,int segment_number_to_initialize/*=1*/)
{
    HASHTABLE<TV_INT2,int> segment_index_hashtable;
    for(int i=1;i<=fiber_mesh.elements.m;i++)segment_index_hashtable.Set(Unique_Ordered(fiber_mesh.elements(i)),i);

    for(int i=1;i<=segment_number_to_initialize;i++){
        TV_INT2 seg_v(end_index-i+1,end_index-i);
        int seg_index=0;if(!segment_index_hashtable.Get(Unique_Ordered(seg_v),seg_index))continue;
        fiber_trace_segment_hashtable.Set(seg_index);
    }
}

template<class TV>
void PhysBAM::FIBER<TV>::Update_Implicit_Geometry_Helper(HASHTABLE<int,IMPLICIT_FIBER_GEOMETRY<TV> >& implicit_geometry_hashtable,const ARRAY<int>& end_indices,const PARTICLES<TV>& particles,const SEGMENT_MESH& segment_mesh,const FIBER_SHAPE fiber_shape)
{
    HASHTABLE<TV_INT2,int> segment_index_hashtable;
    for(int i=1;i<=segment_mesh.elements.m;i++)segment_index_hashtable.Set(Unique_Ordered(segment_mesh.elements(i)),i);

    for(int i=1;i<=end_indices.m;i++){
        for(int s=0;s<implicit_fiber_segment_number;s++){
            int p=end_indices(i)-s;if(p-1<=0||p<=0)continue;
            T height=(particles.X(p-1)-particles.X(p)).Magnitude();
            int seg_index=0;if(!segment_index_hashtable.Get(Unique_Ordered(TV_INT2(p-1,p)),seg_index))continue;
            IMPLICIT_FIBER_GEOMETRY<TV>* implicit_fiber_geometry=implicit_geometry_hashtable.Get_Pointer(seg_index);if(!implicit_fiber_geometry)continue;
            IMPLICIT_OBJECT_TRANSFORMED<TV,FRAME<TV> >* implicit_object_transformed=implicit_fiber_geometry->implicit_object_transformed;
            FRAME<TV>* frame=implicit_fiber_geometry->frame;

            ////update object_space_implicit_object
            if(fiber_shape==GUIDE_BOX){
                ANALYTIC_IMPLICIT_OBJECT<BOX<TV> >* implicit=dynamic_cast<ANALYTIC_IMPLICIT_OBJECT<BOX<TV> >* >(implicit_object_transformed->object_space_implicit_object);
                implicit->analytic.max_corner.y=height;
            }
            else if(fiber_shape==GUIDE_CYLINDER){
                ANALYTIC_IMPLICIT_OBJECT<CYLINDER<T> >* implicit=dynamic_cast<ANALYTIC_IMPLICIT_OBJECT<CYLINDER<T> >* >(implicit_object_transformed->object_space_implicit_object);
                implicit->analytic.height=height;
            }

            ////update frame
            TV translation=particles.X(p);
            TV vector=(particles.X(p-1)-particles.X(p)).Normalized();
            ROTATION<TV> rotation=ROTATION<TV>::From_Rotated_Vector(TV::Axis_Vector(2),vector);
            frame->t=translation;
            frame->r=rotation;
        }
    }
}

template<class TV>
void PhysBAM::FIBER<TV>::Initialize_Implicit_Geometry()
{
    Initialize_Implicit_Geometry_Helper(implicit_fiber_geometry_hashtable,implicit_fiber_elements,fiber_end_indices,fiber_particles,fiber_mesh,/*GUIDE_BOX*/GUIDE_CYLINDER,implicit_fiber_segment_number);
    if(use_guided_fibers)Initialize_Implicit_Geometry_Helper(implicit_guided_fiber_geometry_hashtable,implicit_guided_fiber_elements,guided_fiber_end_indices,guided_fiber_particles,guided_fiber_mesh,GUIDE_CYLINDER,implicit_fiber_segment_number);
    Update_Bounding_Box();
}

template<class TV>
void PhysBAM::FIBER<TV>::Update_Implicit_Geometry()
{
    Update_Implicit_Geometry_Helper(implicit_fiber_geometry_hashtable,fiber_end_indices,fiber_particles,fiber_mesh,/*GUIDE_BOX*/GUIDE_CYLINDER);
    if(use_guided_fibers)Update_Implicit_Geometry_Helper(implicit_guided_fiber_geometry_hashtable,guided_fiber_end_indices,guided_fiber_particles,guided_fiber_mesh,GUIDE_CYLINDER);
    Update_Bounding_Box();
}

template<class TV>
void PhysBAM::FIBER<TV>::Update_Bounding_Box()
{
    T ext=fiber_length*(T).1;bounding_box=BOX<TV>::Bounding_Box(fiber_particles.X);bounding_box=bounding_box.Thickened(ext);
}

template<class TV>
void PhysBAM::FIBER<TV>::Initialize_Mass_Based_On_Geometry_Helper(const T_MESH& mesh,PARTICLES<TV>& particles,const T radius)
{
    for(int i=1;i<=mesh.elements.m;i++){
        const TV_INT2& s=mesh.elements(i);
        T length=(particles.X(s(1))-particles.X(s(2))).Magnitude();
        T cylinder_mass=(T)pi*pow<2>(radius)*length*density;
        particles.mass(s(1))+=(T).5*cylinder_mass;
        particles.mass(s(2))+=(T).5*cylinder_mass;
    }
    for(int i=1;i<=particles.X.m;i++){assert(particles.mass(i)!=(T)0);
        particles.one_over_mass(i)=(T)1/particles.mass(i);
        particles.effective_mass(i)=particles.mass(i);
        particles.one_over_effective_mass(i)=(T)1/particles.effective_mass(i);
    }
}

template<class TV>
void PhysBAM::FIBER<TV>::Initialize_Mass_Based_On_Geometry()
{
    Initialize_Mass_Based_On_Geometry_Helper(fiber_mesh,fiber_particles,fiber_radius);
    if(use_guided_fibers)Initialize_Mass_Based_On_Geometry_Helper(guided_fiber_mesh,guided_fiber_particles,guided_fiber_radius);
}

template<class TV>
void PhysBAM::FIBER<TV>::Initialize_Guided_Fiber_Weights()
{
    ARRAY<T> values;values.Resize(fiber_particles.X.m);values.Fill((T)1);ARRAY_VIEW<T> values_array_view(values);
    T range=fiber_radius*(T)2;
    guided_p_indices.Resize(guided_fiber_particles.X.m);
    guided_p_weights.Resize(guided_fiber_particles.X.m);
    for(int i=1;i<=guided_fiber_particles.X.m;i++){
        T interpolated_value=(T)0;/*bool affected=*/Interpolate_Value(guided_fiber_particles.X(i),range,values_array_view,guided_p_indices(i),guided_p_weights(i),interpolated_value);
    }
}

template<class TV>
void PhysBAM::FIBER<TV>::Update_Guided_Fibers()
{
    ARRAY<TV> fiber_particle_displacements;fiber_particle_displacements.Resize(fiber_particles.X.m);
    for(int i=1;i<=fiber_particles.X.m;i++)fiber_particle_displacements(i)=fiber_particles.X(i)-fiber_particle_X_old(i);
    for(int i=1;i<=guided_fiber_particles.X.m;i++){
        TV displacement=TV();
        for(int j=1;j<=guided_p_indices(i).m;j++){
            int p=guided_p_indices(i)(j);
            T weight=guided_p_weights(i)(j);
            displacement+=fiber_particle_displacements(p)*weight;
        }
        guided_fiber_particles.X(i)+=displacement;
    }
}

template<class TV>
void PhysBAM::FIBER<TV>::Store_Old_States()
{
    fiber_particle_X_old.Resize(fiber_particles.X.m);
    for(int i=1;i<=fiber_particles.X.m;i++)fiber_particle_X_old(i)=fiber_particles.X(i);
}

template<class TV>
void PhysBAM::FIBER<TV>::Advance(const T dt,const T time)
{
    if(fiber_deformable_model==READ_FROM_FILE){
        Advance_Using_File_Data(time);
    }
    else{
        fiber_evolution.Advance(dt,time);
    }
}

template<class TV> template<class T2>
bool PhysBAM::FIBER<TV>::Interpolate_Value(const TV& position,const T range,ARRAY_VIEW<T2>& values,ARRAY<int>& indices,ARRAY<T>& weights,/*result*/T2& interpolated_value)
{
    for(int i=1;i<=fiber_particles.X.m;i++){
        T distance=(fiber_particles.X(i)-position).Magnitude();
        if(distance<=range){indices.Append(i);weights.Append((T)1-distance/range);}
    }
    if(indices.m>0){
        T weights_sum=Sum(weights);
        interpolated_value=T2();
        for(int i=1;i<=weights.m;i++){
            int index=indices(i);
            weights(i)=weights(i)/weights_sum;
            interpolated_value+=weights(i)*values(index);
        }
    }
    return indices.m>0;
}

template<class TV> template<class T2>
bool PhysBAM::FIBER<TV>::Interpolate_Value(const TV& position,const T range,ARRAY_VIEW<T2>& values,/*result*/T2& interpolated_value)
{
    ARRAY<int> indices;ARRAY<T> weights;
    return Interpolate_Value(position,range,values,indices,weights,interpolated_value);
}

template<class TV> template<class T2>
bool PhysBAM::FIBER<TV>::Interpolate_Value(const int p,const T range,const CODIMENSIONAL_MESH& mesh,ARRAY_VIEW<TV>& X,ARRAY_VIEW<T2>& values,/*result*/T2& interpolated_value)
{
    ARRAY<int> indices;ARRAY<T> weights;
    TV fluid_position=X(p);
    for(int i=1;i<=fiber_particles.X.m;i++){
        const TV& fiber_position=fiber_particles.X(i);
        T distance=(fiber_position-fluid_position).Magnitude();
        if(distance<=range){
            bool inside_tet=false;
            for(int j=1;j<=(*mesh.tetrahedron_mesh.incident_elements)(p).m;j++){
                int t=(*mesh.tetrahedron_mesh.incident_elements)(p)(j);
                TV_INT4 vtx=mesh.tetrahedron_mesh.elements(t);
                INDIRECT_ARRAY<ARRAY_VIEW<TV>,TV_INT4& > indir_X(X,vtx);
                if(TETRAHEDRON<T>(indir_X).Inside(fiber_position)){inside_tet=true;break;}
            }
            if(inside_tet){indices.Append(i);weights.Append((T)1-distance/range);}
        }
    }
    if(indices.m>0){
        T weights_sum=Sum(weights);
        interpolated_value=T2();
        for(int i=1;i<=weights.m;i++){
            int index=indices(i);
            T weight=weights(i)/weights_sum;
            interpolated_value+=weight*values(index);
        }
    }
    return indices.m>0;
}

template<class TV>
typename TV::SCALAR PhysBAM::FIBER<TV>::Signed_Distance(const TV& pos) const
{
    int unused=0;return Signed_Distance(pos,unused);
}

template<class TV>
typename TV::SCALAR PhysBAM::FIBER<TV>::Signed_Distance(const TV& pos,/*rst*/int& nearest_fiber_segment_index) const
{
    if(!use_guided_fibers)return Signed_Distance_Helper(pos,nearest_fiber_segment_index,implicit_fiber_geometry_hashtable,spatial_partition);
    else return Signed_Distance_Helper(pos,nearest_fiber_segment_index,implicit_guided_fiber_geometry_hashtable,spatial_partition);
}

template<class TV>
typename TV::SCALAR PhysBAM::FIBER<TV>::Signed_Distance_To_Segment(const TV& pos,const int segment_index) const
{
    const IMPLICIT_FIBER_GEOMETRY<TV>* implicit_fiber_geometry=implicit_fiber_geometry_hashtable.Get_Pointer(segment_index);
    if(implicit_fiber_geometry)return implicit_fiber_geometry->implicit_object_transformed->Signed_Distance(pos);
    else return (T)FLT_MAX;
}

template<class TV>
int PhysBAM::FIBER<TV>::Nearest_Fiber_Segment_Index(const TV& pos) const
{
    int segment_index=0;Signed_Distance(pos,segment_index);return segment_index;
}

template<class TV>
bool PhysBAM::FIBER<TV>::Fiber_Segment_Intersect(const TV& p1,const TV& p2,const int segment_index,/*rst*/T& t) const
{
    const IMPLICIT_FIBER_GEOMETRY<TV>* implicit_fiber_geometry=implicit_fiber_geometry_hashtable.Get_Pointer(segment_index);if(implicit_fiber_geometry==0)return false;
    bool p1_inside=implicit_fiber_geometry->implicit_object_transformed->Inside(p1);bool p2_inside=implicit_fiber_geometry->implicit_object_transformed->Inside(p2);
    if(p1_inside&&p2_inside)return true;

    TV p_i=p2;
    if(!p1_inside&&!p2_inside){
        TV mid=(T).5*(p1+p2);
        bool inside=false;
        for(int iter=1;iter<=3;iter++){
            bool mid_inside=implicit_fiber_geometry->implicit_object_transformed->Inside(mid);
            if(mid_inside){inside=true;break;}
            mid=(T).5*(p1+mid);
        }
        if(!inside)return false;else p_i=mid;
    }

    RAY<TV> ray(SEGMENT_3D<T>(p1,p_i));
    IMPLICIT_OBJECT_ON_A_RAY<IMPLICIT_OBJECT_TRANSFORMED<TV,FRAME<TV> > > implicit_surface_on_a_ray(*implicit_fiber_geometry->implicit_object_transformed,ray); 
    T tolerance=(p1-p_i).Magnitude()*(T)1e-4;
    ITERATIVE_SOLVER<T> iterative_solver;iterative_solver.tolerance=tolerance;
    t=iterative_solver.Bisection_Secant_Root(implicit_surface_on_a_ray,(T)0,ray.t_max)/(p1-p_i).Magnitude();return true;
}

template<class TV>
bool PhysBAM::FIBER<TV>::Fiber_Segment_Inside(const TV& pos,const int segment_index) const
{
    const IMPLICIT_FIBER_GEOMETRY<TV>* implicit_fiber_geometry=implicit_fiber_geometry_hashtable.Get_Pointer(segment_index);if(implicit_fiber_geometry==0)return false;
    return implicit_fiber_geometry->implicit_object_transformed->Inside(pos);
}

template<class TV>
bool PhysBAM::FIBER<TV>::Fiber_Inside(const TV& pos,/*rst*/int& segment_index) const
{
    int nearest_fiber_segment_index=Nearest_Fiber_Segment_Index(pos);
    if(!near_end_segment_hashtable.Contains(nearest_fiber_segment_index))return false;
    if(Fiber_Segment_Inside(pos,nearest_fiber_segment_index)){segment_index=nearest_fiber_segment_index;return true;}
    else return false;
}

template<class TV>
bool PhysBAM::FIBER<TV>::Is_End_Segment(const int seg_id)
{
    return near_end_segment_hashtable.Contains(seg_id)&&fiber_trace_segment_hashtable.Contains(seg_id);
}

template<class TV>
typename TV::SCALAR PhysBAM::FIBER<TV>::Signed_Distance_Helper(const TV& pos,/*rst*/int& nearest_fiber_segment_index,const HASHTABLE<int,IMPLICIT_FIBER_GEOMETRY<TV> >& implicit_geometry_hashtable,CODIMENSIONAL_GEOMETRY_SPATIAL_PARTITION<TV>* spatial_partition/*=0*/) const
{
    T min_distance=(T)FLT_MAX;
    if(spatial_partition){
        ARRAY<CODIMENSIONAL_ELEMENT_INDEX> element_indices;
        spatial_partition->Get_From_Cell_With_One_Ring_Neighbors(pos,2,element_indices);
        for(int i=1;i<=element_indices.m;i++){
            int s=element_indices(i).index;
            const IMPLICIT_FIBER_GEOMETRY<TV>* implicit_fiber_geometry=implicit_geometry_hashtable.Get_Pointer(s);
            if(implicit_fiber_geometry){
                T distance=implicit_fiber_geometry->implicit_object_transformed->Signed_Distance(pos);
                if(distance<min_distance){nearest_fiber_segment_index=s;min_distance=distance;}
            }
        }
    }
    else{
        for(int current_h=1;current_h<=implicit_geometry_hashtable.table.m;current_h++){
            if(implicit_geometry_hashtable.table(current_h).state==ENTRY_ACTIVE){
                const IMPLICIT_FIBER_GEOMETRY<TV>& implicit_fiber_geometry=implicit_fiber_geometry_hashtable.table(current_h).data;
                int key=implicit_fiber_geometry_hashtable.table(current_h).key;
                T distance=implicit_fiber_geometry.implicit_object_transformed->Signed_Distance(pos);
                if(distance<min_distance){nearest_fiber_segment_index=key;min_distance=distance;}
            }
        }

        //for(HASHTABLE_ITERATOR<int,IMPLICIT_FIBER_GEOMETRY<TV> > iter(implicit_geometry_hashtable);iter.Valid();iter.Next()){
        //    const IMPLICIT_FIBER_GEOMETRY<TV>& implicit_fiber_geometry=iter.Data();
        //    T distance=implicit_fiber_geometry.implicit_object_transformed->Signed_Distance(pos);
        //    if(distance<min_distance){nearest_fiber_segment_index=iter.Key();min_distance=distance;}
        //}
    }

    return min_distance;
}

template<class TV>
void PhysBAM::FIBER<TV>::Update_Kinematic_Velocities(ARRAY_VIEW<TV>& V)
{
    for(int i=1;i<=kinematic_particle_object_space_positions.m;i++){
        int p=kinematic_particle_object_space_positions(i).x;
        const TV& pos=fiber_particles.X(p);
        V(p)=RIGID_GEOMETRY<TV>::Pointwise_Object_Velocity(object_twist,object_frame.t,pos);
    }
}

template<class TV>
void PhysBAM::FIBER<TV>::Update_Kinematic_Positions(ARRAY_VIEW<TV>& X)  
{
    for(int i=1;i<=kinematic_particle_object_space_positions.m;i++){
        int p=kinematic_particle_object_space_positions(i).x;
        TV& object_pos=kinematic_particle_object_space_positions(i).y;
        X(p)=object_frame*object_pos;
    }
    Update_Rigid_Body_Positions();
}

template<class TV>
void PhysBAM::FIBER<TV>::Update_Rigid_Body_Positions()
{
    if(handle_rigid_bodies.m>0)for(int i=1;i<=handle_rigid_bodies.m;i++){
        RIGID_BODY<TV>* handle_rigid_body=handle_rigid_bodies(i);
        handle_rigid_body->X()=object_frame.t;
        ROTATION<TV> rot1=ROTATION<TV>((T)half_pi,TV::Axis_Vector(1));
        handle_rigid_body->Rotation()=object_frame.r*rot1;
    }
}

template<class TV>
void PhysBAM::PRESCRIBED_MOTION<TV>::Initialize_Prescribed_Motion()
{
    motion_X_bspline=new BSPLINE<T,TV>(motion_t,motion_X,motion_bspline_order);motion_Q_bspline=new BSPLINE<T,QUATERNION<T> >(motion_t,motion_Q,motion_bspline_order);
}

template<class TV>
void PhysBAM::PRESCRIBED_MOTION<TV>::Add_Prescribed_Motion_Key_Frame(const T t,const TV& x,const ROTATION<TV>& r)
{
    if(motion_t.m==0)for(int j=1;j<=motion_bspline_order;j++){motion_t.Append(t);motion_X.Append(x);motion_Q.Append(r.Quaternion());}
    else for(int j=1;j<=motion_bspline_order;j++){motion_t.Pop();motion_X.Pop();motion_Q.Pop();}////first node
    motion_t.Append(t);motion_X.Append(x);motion_Q.Append(r.Quaternion());
    for(int j=1;j<=motion_bspline_order;j++){motion_t.Append(t);motion_X.Append(x);motion_Q.Append(r.Quaternion());}////last node
}

template<class TV>
FRAME<TV> PhysBAM::PRESCRIBED_MOTION<TV>::Prescribed_Frame(const T time)
{
    TV pos=motion_X_bspline->Clamped_Evaluate(time);
    ROTATION<TV> rot=ROTATION<TV>::From_Quaternion(motion_Q_bspline->Clamped_Evaluate(time));
    return FRAME<TV>(pos,rot);
}

template<class TV>
TWIST<TV> PhysBAM::PRESCRIBED_MOTION<TV>::Prescribed_Twist(const T dt,const T time)
{
    FRAME<TV> frame1=Prescribed_Frame(time);FRAME<TV> frame2=Prescribed_Frame(time+dt);
    return (frame2.Delta()-frame1.Delta())/dt;
}

template<class TV>
void PhysBAM::PRESCRIBED_MOTION<TV>::Initialize_Prescribed_Motion_From_File_With_Constant_Rotation(const std::string file_name,const ROTATION<TV>& r,const T s0)
{
    std::string str;STREAM_TYPE stream_type((T()));std::fstream in(file_name.c_str(),std::ios::in);
    if(!in){LOG::cout<<"cannot open prescribed motion file "<<file_name<<std::endl;}

    T t=0;TV x=TV();
    while(in>>t>>x.x>>x.y>>x.z){Add_Prescribed_Motion_Key_Frame(t,x*s0,r);}
    Initialize_Prescribed_Motion();
}

template<class TV>
void PhysBAM::FIBER<TV>::Initialize_Velocity(const TV& v0)
{
    for(int i=1;i<=fiber_particles.X.m;i++){fiber_particles.V(i)=v0;}
}

template<class TV>
void PhysBAM::FIBER<TV>::Initialize_Attributes_From_File(const std::string& directory,const T t0,const T t1,const T motion_t0/*=(T)0*/)
{
    file_start_time=t0;
    file_end_time=t1;
    motion_start_time=motion_t0;

    PARTICLES<TV> file_particles;
    file_particles.Store_Mass();
    file_particles.Store_Velocity();
    STREAM_TYPE stream_type((T()));

    int i=1;//ignoring frame 0
    while(true){
        std::string f=STRING_UTILITIES::string_sprintf("%d",i);
        std::string file_name=directory+"/"+f+"/particles2";
        if(!FILE_UTILITIES::File_Exists(file_name))break;
        FILE_UTILITIES::Read_From_File(stream_type,file_name,file_particles);
        file_position_data.Append(ARRAY<TV>(file_particles.X.m));
        file_velocity_data.Append(ARRAY<TV>(file_particles.V.m));
        for(int j=1;j<=file_particles.X.m;j++){
            file_position_data(i)(j)=file_particles.X(j);
            file_velocity_data(i)(j)=file_particles.V(j);
        }
        i++;
    }
    if(i==1)LOG::cout<<"WARNING -- no fiber data read from file."<<std::endl;
    file_frame_count=file_position_data.m;
}

template<class TV>
void PhysBAM::FIBER<TV>::Advance_Using_File_Data(const T& time)
{
    T weight=(time+motion_start_time-file_start_time)/(file_end_time-file_start_time);
    T scaled_weight=weight*(file_frame_count);
    int i0=(int)(std::floor(scaled_weight));
    int i1=i0+1;
    T frac=scaled_weight-(T)i0;

    i0++;i1++;//one-based indexing

    if(i0<1){
        for(int j=1;j<=fiber_particles.X.m;j++){
            fiber_particles.X(j)=file_position_data(1)(j);
            fiber_particles.V(j)=file_velocity_data(1)(j);
        }
    }
    else if(i0>=file_frame_count){
        for(int j=1;j<=fiber_particles.X.m;j++){
            fiber_particles.X(j)=file_position_data(file_frame_count)(j);
            fiber_particles.V(j)=file_velocity_data(file_frame_count)(j);
        }
    }
    else{
        for(int j=1;j<=fiber_particles.X.m;j++){
            fiber_particles.X(j)=LINEAR_INTERPOLATION<T,TV>::Linear(file_position_data(i0)(j),file_position_data(i1)(j),frac);
            fiber_particles.V(j)=LINEAR_INTERPOLATION<T,TV>::Linear(file_velocity_data(i0)(j),file_velocity_data(i1)(j),frac);
        }
    }

    Update_Rigid_Body_Positions();
}

#define INSTANTIATION_HELPER(T,d)           \
    template class PRESCRIBED_MOTION<VECTOR<T,d> >;     \
    template class FIBER<VECTOR<T,d> >;     \
    template bool FIBER<VECTOR<T,d> >::Interpolate_Value(const VECTOR<T,d>& position,const T range,ARRAY_VIEW<VECTOR<T,d> >& values,ARRAY<int>& indices,ARRAY<T>& weights,/*result*/VECTOR<T,d> & interpolated_value);   \
    template bool FIBER<VECTOR<T,d> >::Interpolate_Value(const VECTOR<T,d>& position,const T range,ARRAY_VIEW<T>& values,ARRAY<int>& indices,ARRAY<T>& weights,/*result*/T & interpolated_value);   \
    template bool FIBER<VECTOR<T,d> >::Interpolate_Value(const VECTOR<T,d>& position,const T range,ARRAY_VIEW<VECTOR<T,d> >& values,/*result*/VECTOR<T,d> & interpolated_value);   \
    template bool FIBER<VECTOR<T,d> >::Interpolate_Value(const VECTOR<T,d>& position,const T range,ARRAY_VIEW<T>& values,/*result*/T & interpolated_value);   \
    template bool FIBER<VECTOR<T,d> >::Interpolate_Value(const int p,const T range,const CODIMENSIONAL_MESH& mesh,ARRAY_VIEW<VECTOR<T,d> >& X,ARRAY_VIEW<VECTOR<T,d> >& values,/*result*/VECTOR<T,d> & interpolated_value);   \
    template bool FIBER<VECTOR<T,d> >::Interpolate_Value(const int p,const T range,const CODIMENSIONAL_MESH& mesh,ARRAY_VIEW<VECTOR<T,d> >& X,ARRAY_VIEW<T>& values,/*result*/T & interpolated_value);

INSTANTIATION_HELPER(float,3)
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
INSTANTIATION_HELPER(double,3)
#endif
