//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <PhysBAM_Geometry/Basic_Geometry/BOX.h>
#include <PhysBAM_Geometry/Level_Sets/IMPLICIT_OBJECT_ON_A_RAY.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_TRIANGULATED_AREA.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_STRUCTURE.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_RIGID_GEOMETRY_COLLECTION.h>
#include <PhysBAM_Tools/Read_Write/Matrices_And_Vectors/READ_WRITE_FRAME.h>
#include <PhysBAM_Tools/Read_Write/Matrices_And_Vectors/READ_WRITE_TWIST.h>
#include <PhysBAM_Tools/Nonlinear_Equations/ITERATIVE_SOLVER.h>
#include <PhysBAM_Dynamics/Geometry/GENERAL_GEOMETRY_FORWARD.h>
#include <PhysBAM_Dynamics/Particles/PARTICLES_FORWARD.h>
#include <PhysBAM_Dynamics/Read_Write/Particles/READ_WRITE_PARTICLES.h>
#include <fstream>
#include "../Meshing/CODIMENSIONAL_MESH_GENERATOR.h"
#include "../Viscoelastic/PAINT_DRIVER.h"
#include "../Viscoelastic/PAINT_DRIVER_TESTS.h"
#include "../Viscoelastic/PAINT_TUBE.h"

using namespace PhysBAM;

template<class TV>
PhysBAM::PAINT_DRIVER<TV>::PAINT_DRIVER():BASE(),coupled_evolution(fiber,example.mesher,example.incompressible,example.viscosity_force,example.elasticity_force,rigid_bodies,example.spatial_partition),
    knife(0),ladle(0),rock(0),scraper(0),use_levelset(true),levelset(example.geometrical_codimensional_mesh,example.particles),levelset_region_number(4),update_slow_interface_phis(true),
    guided_fiber_radius((T)0),guided_fiber_distance((T)0),viscosity_factor((T)1),viscosity_min_factor((T)1),elastic_modulus_factor((T)1),band_width_min((T)0),band_width_max((T)3e-3),
    min_tri_thickness_factor((T)1),min_seg_thickness_factor((T)1),adaptive_length_scale_min_factor((T)1),adaptive_length_scale_min_factor2((T)1),lambda_cy((T)0),alpha_cy((T)0),n_cy((T)0),
    particle_inactive_mode(SMALL_VELOCITY),use_particle_incompressibility_for_solids(false),use_particle_incompressibility_for_fiber(false),paint_tube_test(1),bunny_test(1),paint_height_factor((T)1),use_1d_tube(false),
    use_tube_rigid(false),use_overlapping_tube(false),use_full_letters(false),use_letter_r(false),use_low_fric_near_fiber(false),piz_thickness_min_factor((T)1),piz_thickness_max_factor((T)1),ring_source_peak_number(5),use_ring_source_randomness(false),
    cylinder_angular_velocity((T)200),paint_trace_test(1),piz_init_frame(3),piz_init_angle((T)0),piz_mesh_noise_scale((T)400),piz_mesh_noise_offset(TV::Constant_Vector((T)10)),piz_mesh_filename("./piz_mesh"),piz_test(1),
    spray_init_frame(1080),spray_filename("./j5_cy_s1"),trace_init_frame(600),trace_filename("./paint_simple_2_restart"),use_sim_box(false)
{
    Use_Rim(false);has_fiber_fluid_coupling_force=false;has_friction=true;use_spatial_partition=true;
}

template<class TV>
PhysBAM::PAINT_DRIVER<TV>::~PAINT_DRIVER()
{
    if(scraper!=0)delete scraper;
    if(rock!=0)delete rock;
    if(ladle!=0)delete ladle;
    if(knife!=0)delete knife;
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Convection(const T dt,const T time)
{
    BASE::Convection(dt,time);
    coupled_evolution.Advance_One_Timestep(dt,time);
    //Squash_Fluid();
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Update_Mesh()
{
    if(remesh){
        Update_Particle_Thickness();    ////discrete thickness
        Update_Particle_Contact();  /////particles.contact will be used in the mesher
        Update_Particle_Inactive();
        Update_Particle_Meshing_Direction();
        example.mesher.Update_Mesh_Topology();
        Add_Mesh_Objects();
        example.mesher.Update_Mesh_States();

        Correct_Particle_Normals();
        if(use_spatial_partition){  ////up to now geometry and fiber are all updated
            Update_Spatial_Partition();
            coupled_evolution.collision_pairs.Remove_All();
            coupled_evolution.Detect_Fiber_Fluid_Collisions(coupled_evolution.collision_pairs);
        }
        Update_Particle_Incompressibility();
        Update_Particle_Length_Scale();
        example.mesher.Update_Incompressibility_Mesh(example.incompressibility_codimensional_mesh);
        example.mesher.Update_Viscosity_Mesh(example.viscosity_codimensional_mesh,has_rim_segments_for_viscosity,has_rim_triangles_for_viscosity);
        if(example.mesher.has_eulerian_particles&&example.mesher.use_split_eulerian_particles&&tubes.m>0){
            for(int i=1;i<=tubes.m;i++){tubes(i)->Add_Eulerian_Particles_From_Mesher(example.mesher.eulerian_particles_hashtable,example.mesher.added_eulerian_particle_parents_hashtable,example.particles);}
        }
    }
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Write_Output_Files(const int frame)
{
    BASE::Write_Output_Files(frame);

    fiber.deformable_geometry_collection.Write(stream_type,output_directory,frame,frame,true,std::string("2"));
    std::string f=STRING_UTILITIES::string_sprintf("%d",frame);
    FILE_UTILITIES::Write_To_File(stream_type,output_directory+"/"+f+"/particles2",fiber.fiber_particles);
    FILE_UTILITIES::Write_To_File(stream_type,output_directory+"/"+f+"/fiber_frame",fiber.object_frame);
    FILE_UTILITIES::Write_To_File(stream_type,output_directory+"/"+f+"/fiber_twist",fiber.object_twist);

    if(tubes.m>0){
        for(int i=1;i<=tubes.m;i++){    ////TODO: write to binary file
            PAINT_TUBE<TV>* tube=tubes(i);std::string t=STRING_UTILITIES::string_sprintf("%d",i);
            FILE_UTILITIES::Write_To_File(stream_type,output_directory+"/"+f+"/tube_frame_"+t,tube->object_frame);
            FILE_UTILITIES::Write_To_File(stream_type,output_directory+"/"+f+"/tube_twist_"+t,tube->object_twist);
            Write_Tube_Eulerian_Particles_To_File(*tubes(i),output_directory+"/"+f+"/tube_ep_"+t);
        }
    }
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Read_Output_Files(const int frame)
{
    BASE::Read_Output_Files(frame);

    fiber.deformable_geometry_collection.Read(stream_type,output_directory,frame,frame,true,std::string("2"));
    std::string f=STRING_UTILITIES::string_sprintf("%d",frame);
    std::string file_name=output_directory+"/"+f+"/particles2";
    if(FILE_UTILITIES::File_Exists(file_name)){FILE_UTILITIES::Read_From_File(stream_type,file_name,fiber.fiber_particles);}
    file_name=output_directory+"/"+f+"/fiber_frame";
    if(FILE_UTILITIES::File_Exists(file_name)){FILE_UTILITIES::Read_From_File(stream_type,file_name,fiber.object_frame);}
    file_name=output_directory+"/"+f+"/fiber_twist";
    if(FILE_UTILITIES::File_Exists(file_name)){FILE_UTILITIES::Read_From_File(stream_type,file_name,fiber.object_twist);}
    
    if(tubes.m>0){
        for(int i=1;i<=tubes.m;i++){
            PAINT_TUBE<TV>* tube=tubes(i);std::string t=STRING_UTILITIES::string_sprintf("%d",i);
            file_name=output_directory+"/"+f+"/tube_frame_"+t;
            if(FILE_UTILITIES::File_Exists(file_name)){FILE_UTILITIES::Read_From_File(stream_type,file_name,tube->object_frame);}
            file_name=output_directory+"/"+f+"/tube_twist_"+t;
            if(FILE_UTILITIES::File_Exists(file_name)){FILE_UTILITIES::Read_From_File(stream_type,file_name,tube->object_twist);}
            file_name=output_directory+"/"+f+"/tube_ep_"+t;
            if(FILE_UTILITIES::File_Exists(file_name)){Read_Tube_Eulerian_Particles_From_File(*tubes(i),file_name);}
        }
    }
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Initialize_Solver()
{
    BASE::Initialize_Solver();
    if(fiber.fiber_particles.X.m>0){
        fiber.Initialize_Forces();
        coupled_evolution.Initialize_Forces();
        if(use_spatial_partition)fiber.spatial_partition=&example.spatial_partition;
    }
    Initialize_Paint_Phis();
    Initialize_Paint_Color();
    Update_Position_Based_State((T)0,true);
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Update_Boundary_Conditions()
{
    BASE::Update_Boundary_Conditions();
    //coupled_evolution.Update_Boundary_Conditions();
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Update_Boundary_Conditions(CODIMENSIONAL_INCOMPRESSIBLE<TV>& incompressible)
{
    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles=incompressible.particles;
    ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> >& psi_N_on_mesh=incompressible.psi_N_on_mesh;
    ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> >& psi_D_on_mesh=incompressible.psi_D_on_mesh;
    psi_N_on_mesh.Remove_All();psi_D_on_mesh.Remove_All();
    for(int i=1;i<=particles.X.m;i++){
        //////kin rigid body
        //{int id=0;T phi=(T)0;TV rigid_normal;TV rigid_v;
        //if(SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Tall_Particle_Intersect_Solid_And_Value(i,particles,kinematic_rigid_bodies,id,phi,rigid_normal,rigid_v)){
        //    T v_in_normal_direction=TV::Dot_Product(rigid_normal,rigid_v);psi_N_on_mesh.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,rigid_normal,v_in_normal_direction));   
        //    //LOG::cout<<"tall particle psi_N: "<<i<<": "<<v_in_normal_direction<<", "<<rigid_normal<<std::endl;
        //}}
        ////rigid body
        {int id=0;T phi=(T)0;TV rigid_normal;TV rigid_v;
        if(SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Solid_Inside_And_Value(particles.X(i),rigid_bodies,id,phi,rigid_normal,rigid_v)){
            Process_Rigid_V_And_Normal(rigid_v,rigid_normal);
            if(use_friction_boundary)rigid_v=particles.V(i);
            bool non_slip_solid=use_non_slip_on_solids||(solid_non_slip_hashtable.Contains(id)&&/*nonslip*/solid_non_slip_hashtable.Get(id).x);
            if(non_slip_solid||use_friction_boundary){for(int d=1;d<=TV::m;d++)psi_N_on_mesh.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,TV::Axis_Vector(d),rigid_v(d)));}
            else{T v_in_normal_direction=TV::Dot_Product(rigid_normal,rigid_v);psi_N_on_mesh.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,rigid_normal,v_in_normal_direction));}
            //LOG::cout<<"incomp rigid "<<particles.guid(i)<<": "<<rigid_normal<<", "<<rigid_v<<", twist: "<<rigid_bodies(id)->Twist()<<", "<<rigid_bodies(id)->V()<<std::endl;
        }}

        ////kinematic particle
        if(kinematic_particle_velocity_hashtable.Contains(particles.guid(i))){
            const TV& kinematic_v=kinematic_particle_velocity_hashtable.Get(particles.guid(i));
            for(int d=1;d<=TV::m;d++)psi_N_on_mesh.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,TV::Axis_Vector(d),kinematic_v(d)));
        }

        //////connect particle
        //if(Is_Contact_With_Fluid(particles.contact(i))){
        //    TV connect_v=TV();
        //    for(int d=1;d<=TV::m;d++)psi_N_on_mesh.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,TV::Axis_Vector(d),connect_v(d)));
        //}

        ////tank
        if(tank!=0&&Is_Contact_With_Tank(particles.contact(i))){
            TV wall_pos=TV();TV wall_normal=SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Tank_Normal_And_Position(particles.X(i),*tank,wall_pos);
            if(use_non_slip_on_solids) for(int d=1;d<=TV::m;d++)psi_N_on_mesh.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,TV::Axis_Vector(d),(T)0));
            else for(int d=1;d<=TV::m;d++)if(wall_normal(d)!=(T)0)psi_N_on_mesh.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,TV::Axis_Vector(d),(T)0));
        }
    }

    ////free surface: 0 psi_D
    for(int i=1;i<=particles.X.m;i++){particles.psi_N(i)=false;particles.psi_D(i)=false;}
    {Update_Particle_Psi_N(particles,psi_N_on_mesh);Update_Particle_Psi_D(particles,psi_D_on_mesh);}
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Update_Position_Based_State(const T time,const bool is_position_update)
{
    BASE::Update_Position_Based_State(time,is_position_update);

    ////update length scale from levelset
    if(use_levelset){
        if(tubes.m>0)for(int i=1;i<=tubes.m;i++)tubes(i)->Update_Eulerian_Particle_Phis(example.particles);
        levelset.Reinitialize(update_slow_interface_phis);
        if(use_adaptive_length_scale_on_levelset)Update_Particle_Length_Scale_From_Levelset(length_scale*(T).2,length_scale,length_scale*(T).5,length_scale*(T)1);
    }
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Update_Particle_Length_Scale_From_Levelset(const T length_scale_min,const T length_scale_max,const T band_width_min,const T band_width_max)
{
    for(int i=1;i<=example.particles.X.m;i++){
        T phi=(T)FLT_MAX;/*int r=*/levelset.Inside_Region(i,phi);
        T frac=clamp((abs(phi)-band_width_min)/(band_width_max-band_width_min),(T)0,(T)1);
        T particle_length_scale=((T)1-frac)*length_scale_min+frac*length_scale_max;
        if(example.particles.length_scale(i)>particle_length_scale)example.particles.length_scale(i)=particle_length_scale;
    }
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Add_Canvas(const T size/*=(T).2*/,const TV& dir/*=TV::Axis_Vector(2)*/,const TV center/*=TV()*/)
{
    TV rigid_body_size=TV::Constant_Vector(size);rigid_body_size.y*=(T).1;
    {RIGID_BODY<TV>* rigid_body=&rigid_standard_tests.Add_Analytic_Box(rigid_body_size);
    rigid_body->X()=center;if(center==TV())rigid_body->X().y=(T)-.5*rigid_body_size.y;
    ROTATION<TV> rot=ROTATION<TV>::From_Rotated_Vector(TV::Axis_Vector(2),dir.Normalized());rigid_body->Rotation()=rot;
    rigid_bodies.Append(rigid_body);rigid_body->name="canvas";canvas_normal=TV::Axis_Vector(2);}

    {RIGID_GEOMETRY<TV>* rigid_body=&fiber.fiber_deformables_example.rigid_standard_tests.Add_Analytic_Box(rigid_body_size);
    rigid_body->X()=TV();rigid_body->X().y=(T)-.5*rigid_body_size.y+fiber_radius*(T).24;ROTATION<TV> rot=ROTATION<TV>::From_Rotated_Vector(TV::Axis_Vector(2),dir.Normalized());rigid_body->Rotation()=rot;}
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Add_Bread_Slices(const T size/*=(T).1*/)
{
    TV rigid_body_size=TV::Constant_Vector(size);rigid_body_size.y*=(T).1;
    //RIGID_BODY<TV>* slice1=&rigid_standard_tests.Add_Analytic_Box(rigid_body_size);

    STREAM_TYPE float_stream_type(1.f);
    //RIGID_BODY<TV>* slice1=&rigid_standard_tests.Add_Rigid_Body("bread_attempt_dbl",(T).015,(T)0.0,true,true,&double_stream_type);
    RIGID_BODY<TV>* slice1=&rigid_standard_tests.Add_Rigid_Body("bread_attempt_scaled_flt",(T)1.0,(T)0.0,true,true,&float_stream_type);

    slice1->X()=TV();
    //slice1->X().y=(T)-.5*rigid_body_size.y;
    rigid_bodies.Append(slice1);slice1->name="bread_slice1";

    //TV center2((T)-1.5*size,(T)0,(T)0);
    TV center2((T)-1.3*size,(T)0,(T).3*size);
    //T rotation2=(T)pi/(T)6;
    T rotation2=(T)pi/(T)9;
    //RIGID_BODY<TV>* slice2=&rigid_standard_tests.Add_Analytic_Box(rigid_body_size);
    RIGID_BODY<TV>* slice2=&rigid_standard_tests.Add_Rigid_Body("bread_attempt_scaled_flt",(T)1.0,(T)0.0,true,true,&float_stream_type);
    slice2->X()=center2;//slice2->X().y=(T)-.5*rigid_body_size.y;
    ROTATION<TV> rot(rotation2,TV::Axis_Vector(2));slice2->Rotation()=rot;
    rigid_bodies.Append(slice2);slice2->name="bread_slice2";

    TV plate_size((T)5*rigid_body_size(1),(T)3*rigid_body_size(2),(T)3*rigid_body_size(3));
    RIGID_BODY<TV>* plate=&rigid_standard_tests.Add_Analytic_Box(plate_size);
    plate->X()=TV();plate->X().y=(T)-1*rigid_body_size.y-(T)1.5*rigid_body_size.y;
    rigid_bodies.Append(plate);plate->name="plate";

    canvas_normal=TV::Axis_Vector(2);//??
}

template<class TV>
typename TV::SCALAR PhysBAM::PAINT_DRIVER<TV>::Distance_Fiber_To_Canvas()
{
    T min_distance=(T)FLT_MAX;
    for(int i=1;i<=fiber.fiber_end_indices.m;i++){
        int p=fiber.fiber_end_indices(i);
        if(fiber.fiber_particles.X(p).y<(T)min_distance){
            min_distance=fiber.fiber_particles.X(p).y;
        }
    }
    return min_distance;
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Update_Kinematic_Positions(const T dt,const T time)
{
    BASE::Update_Kinematic_Positions(dt,time);
    Update_Kinematic_Object_Velocities_And_Positions(dt,time);
    Update_Kinematic_Paint_Velocities_And_Positions(dt,time);
    if(use_correcting_particle_position_on_solids)Correct_Particle_Positions();    
    if(tubes.m>(T)0)for(int i=1;i<=tubes.m;i++)tubes(i)->Advance(dt,time);
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Initialize_Kinematic_Positions()
{
    for(int i=1;i<=tubes.m;i++)tubes(i)->Initialzie_Eulerian_Particles(example.particles,tubes(i)->eulerian_frac);
    Update_Kinematic_Object_Velocities_And_Positions((T)0,(T)0);
    fiber.Initialize_Velocity(fiber.object_twist.linear);
    Update_Kinematic_Paint_Velocities_And_Positions((T)0,(T)0);
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Add_Tube_Rigid_Body(PAINT_TUBE<TV>& tube)
{
    if(tube.rigid_body_type==PAINT_TUBE<TV>::PLACEHOLDER){
        tube.rigid_body_height=tube.tube_height;
        tube.rigid_body_radius=tube.radius*(T)1.2;
        RIGID_BODY<TV>* rigid_body=&rigid_standard_tests.Add_Analytic_Cylinder(tube.rigid_body_height,tube.rigid_body_radius,100);
        rigid_body->name="tube_cylinder";tube.rigid_bodies.Append(rigid_body);rigid_bodies.Append(rigid_body);kinematic_rigid_bodies.Append(rigid_body);
    }
    else if(tube.rigid_body_type==PAINT_TUBE<TV>::RENDERABLE){
        STREAM_TYPE double_stream_type(1.0);
        //tube body
        {RIGID_BODY<TV>* tube_rigid_body=&rigid_standard_tests.Add_Rigid_Body("tube_body_double",(T).0055,(T)0.,true,true,&double_stream_type);
        tube_rigid_body->name="tube_body";tube.rigid_bodies.Append(tube_rigid_body);rigid_bodies.Append(tube_rigid_body);}
        //tube metal
        {RIGID_BODY<TV>* tube_rigid_body=&rigid_standard_tests.Add_Rigid_Body("tube_metal_double",(T).0055,(T)0.,true,true,&double_stream_type);
        tube_rigid_body->name="tube_metal";tube.rigid_bodies.Append(tube_rigid_body);rigid_bodies.Append(tube_rigid_body);}

        ////tube body
        //{RIGID_BODY<TV>* tube_rigid_body=&rigid_standard_tests.Add_Rigid_Body("tube_body_double",(T).0055*T(100),(T)0.,true,true,&double_stream_type);
        //tube_rigid_body->name="tube_body";tube.rigid_bodies.Append(tube_rigid_body);rigid_bodies.Append(tube_rigid_body);}
        ////tube metal
        //{RIGID_BODY<TV>* tube_rigid_body=&rigid_standard_tests.Add_Rigid_Body("tube_metal_double",(T).0055*T(100),(T)0.,true,true,&double_stream_type);
        //tube_rigid_body->name="tube_metal";tube.rigid_bodies.Append(tube_rigid_body);rigid_bodies.Append(tube_rigid_body);}
    }

    tube.Update_Position_Based_States();
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Add_Fiber_Rigid_Body(const ARRAY<TV>& fiber_positions,FIBER<TV>& fiber)
{
    RANGE<TV> bounding_box;
    for(int i=1;i<=fiber_positions.m;i++){bounding_box.Enlarge_To_Include_Point(fiber_positions(i));}
    
    ////place holder
    if(fiber.rigid_body_type==FIBER<TV>::PLACEHOLDER){
        RIGID_BODY<TV>* handle_rigid_body=&rigid_standard_tests.Add_Analytic_Box(TV((T).002,(T).004,(T).008));
        handle_rigid_body->X()=fiber.object_frame.t;    
        handle_rigid_body->Rotation()=fiber.object_frame.r;
        fiber.handle_rigid_bodies.Append(handle_rigid_body);
    }
    else if(fiber.rigid_body_type==FIBER<TV>::RENDERABLE){
        STREAM_TYPE double_stream_type(1.0);
        ////handle
        {RIGID_BODY<TV>* handle_rigid_body=&rigid_standard_tests.Add_Rigid_Body("aligned_thinbrush_body_double",(T).037,(T)0.,true,true,&double_stream_type);
        ROTATION<TV> rot1=ROTATION<TV>((T)half_pi,TV::Axis_Vector(1));
        handle_rigid_body->X()=fiber.object_frame.t;
        handle_rigid_body->Rotation()=fiber.object_frame.r*rot1;
        handle_rigid_body->name="brush_body";
        fiber.handle_rigid_bodies.Append(handle_rigid_body);}
        ////metal
        {RIGID_BODY<TV>* handle_rigid_body=&rigid_standard_tests.Add_Rigid_Body("aligned_thinbrush_metal_double",(T).037,(T)0.,true,true,&double_stream_type);
        ROTATION<TV> rot1=ROTATION<TV>((T)half_pi,TV::Axis_Vector(1));
        handle_rigid_body->X()=fiber.object_frame.t;
        handle_rigid_body->Rotation()=fiber.object_frame.r*rot1;
        handle_rigid_body->name="brush_metal";
        fiber.handle_rigid_bodies.Append(handle_rigid_body);}
    }    
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Add_Toothbrush_Rigid_Body(const ARRAY<TV>& fiber_positions,FIBER<TV>& fiber)
{
    STREAM_TYPE double_stream_type(1.0);
    ////handle
    {RIGID_BODY<TV>* handle_rigid_body=&rigid_standard_tests.Add_Rigid_Body("toothbrush_body_double",(T).015,(T)0.,true,true,&double_stream_type);
//    ROTATION<TV> rot1=ROTATION<TV>((T)half_pi,TV::Axis_Vector(1));
    handle_rigid_body->X()=fiber.object_frame.t;
    handle_rigid_body->Rotation()=fiber.object_frame.r;//*rot1;
    handle_rigid_body->name="toothbrush_body";
    fiber.handle_rigid_bodies.Append(handle_rigid_body);}
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Update_Spatial_Partition()
{
    BASE::Update_Spatial_Partition();

    CODIMENSIONAL_GEOMETRY_SPATIAL_PARTITION<TV>& spatial_partition=example.spatial_partition;
    ////insert fiber geometry into partition
    fiber.fiber_element_voxel_ranges.Remove_All();int object_id=2;
    for(int i=1;i<=fiber.fiber_mesh.elements.m;i++){
        const TV_INT2& e=fiber.fiber_mesh.elements(i);RANGE<TV_INT> voxel_range;T thickness=(T).5*(fiber.fiber_particle_thickness(e(1))+fiber.fiber_particle_thickness(e(2)));
        spatial_partition.Insert_Into_Hashtable(fiber.fiber_particles.X,e,i,object_id,voxel_range,thickness);
        fiber.fiber_element_voxel_ranges.Append(voxel_range);
    }
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Correct_Particle_Positions()
{
    BASE::Correct_Particle_Positions();
    
    ////TODO: inside fiber test
}


template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Initialize_Mesh()
{
    BASE::Initialize_Mesh();
    if(kinematic_rigid_bodies.m>0)example.mesher.kinematic_rigid_bodies=kinematic_rigid_bodies;
    if(fiber.fiber_particles.X.m>0)example.mesher.fiber=&fiber;
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Initialize_Mpi()
{
    BASE::Initialize_Mpi();
    fiber.Initialize_Mpi(example.mpi_codim);
    coupled_evolution.Initialize_Mpi(example.mpi_codim);
}

//template<class TV>
//typename TV::SCALAR PhysBAM::PAINT_DRIVER<TV>::Prism_Lost_Volume(const TV_INT3& v,const RIGID_BODY<TV>& rigid_body)
//{
//    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles=example.particles;
//    TV center=(T)one_third*(particles.X(v(1))+particles.X(v(2))+particles.X(v(3)));
//    TV triangle_normal=canvas_normal;   ////TODO: handle non-contact case
//    T averaged_thickness=(T)one_third*(particles.thickness(v(1))+particles.thickness(v(2))+particles.thickness(v(3)));
//    T particles_length_scale=(T)one_third*(particles.length_scale(v(1))+particles.length_scale(v(2))+particles.length_scale(v(3)));
//    //T tolerance=particles_length_scale*(T)1e-3;
//    T triangle_area=TRIANGLE_3D<T>::Area(particles.X(v(1)),particles.X(v(2)),particles.X(v(3)));
//    //RAY<TV> ray(SEGMENT_3D<T>(center,center+triangle_normal*averaged_thickness));
//    //IMPLICIT_OBJECT_ON_A_RAY<IMPLICIT_OBJECT_TRANSFORMED<TV,RIGID_GEOMETRY<TV> > > implicit_surface_on_a_ray(*rigid_body.implicit_object,ray); 
//    //ITERATIVE_SOLVER<T> iterative_solver;iterative_solver.tolerance=tolerance;
//    //T t_intersect=iterative_solver.Bisection_Secant_Root(implicit_surface_on_a_ray,(T)0,ray.t_max);
//    T t_intersect=Segment_Intersection_Fraction_With_Solid(center,center+triangle_normal*averaged_thickness,rigid_body);
//    return ((T)1-t_intersect)*averaged_thickness*triangle_area;
//}
//
//template<class TV>
//typename TV::SCALAR PhysBAM::PAINT_DRIVER<TV>::Segment_Intersection_Fraction_With_Solid(const TV& p1,const TV& p2,const RIGID_BODY<TV>& rigid_body)
//{
//    RAY<TV> ray(SEGMENT_3D<T>(p1,p2));T tolerance=thickness_scale*(T)1e-3;
//    IMPLICIT_OBJECT_ON_A_RAY<IMPLICIT_OBJECT_TRANSFORMED<TV,RIGID_GEOMETRY<TV> > > implicit_surface_on_a_ray(*rigid_body.implicit_object,ray); 
//    ITERATIVE_SOLVER<T> iterative_solver;iterative_solver.tolerance=tolerance;
//    T t_intersect=iterative_solver.Bisection_Secant_Root(implicit_surface_on_a_ray,(T)0,ray.t_max)/(p1-p2).Magnitude();;return t_intersect;
//}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Update_Target_Divergence()
{
    ARRAY<CODIMENSIONAL_TARGET_DIVERGENCE<T> >& tar_div_on_mesh=example.incompressible.tar_div_on_mesh;
    CODIMENSIONAL_MESH& solver_mesh=example.incompressible.mesh;
    tar_div_on_mesh.Remove_All();

    //////update target div due to kinematic rigid body
    if(use_particle_incompressibility_for_solids)for(int i=1;i<=solver_mesh.triangle_mesh.elements.m;i++){
        const TV_INT3& v=solver_mesh.triangle_mesh.elements(i);int kinematic_rigid_body_id=0;
        if(SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Prism_Intersect_Solid(v,example.particles,kinematic_rigid_bodies,kinematic_rigid_body_id)){
            T lost_volume=SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Prism_Lost_Volume(v,example.particles,*kinematic_rigid_bodies(kinematic_rigid_body_id));T relax=(T)1e3;
            T tri_tar_div=lost_volume*relax;
            tar_div_on_mesh.Append(CODIMENSIONAL_TARGET_DIVERGENCE<T>(i,SIMPLEX_TRIANGLE,tri_tar_div));
        }
    }

    ////update target div due to fiber
    if(use_particle_incompressibility_for_fiber)coupled_evolution.Update_Target_Divergence();
}

//template<class TV>
//bool PhysBAM::PAINT_DRIVER<TV>::Prism_Intersect_Solid(const TV_INT3& v,const ARRAY<RIGID_BODY<TV>* >& rigid_bodies,/*rst*/int& id)
//{
//    //CODIMENSIONAL_FLUID_PARTICLES<TV>& particles=example.particles;
//    //TV center=(T)one_third*(particles.X(v(1))+particles.X(v(2))+particles.X(v(3)));
//    //TV triangle_normal=canvas_normal;   ////TODO: handle non-contact case
//    //T averaged_thickness=(T)one_third*(particles.thickness(v(1))+particles.thickness(v(2))+particles.thickness(v(3)));
//    //return Inside_Kinematic_Rigid_Body(center+triangle_normal*averaged_thickness,kinematic_rigid_body_id);
//    for(int i=1;i<=3;i++){
//        TV p2=example.particles.X(v(i))+canvas_normal*example.particles.thickness(v(i));
//        if(SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Solid_Inside(p2,rigid_bodies,id))return true;
//    }
//    return false;
//}

//template<class TV>
//bool PhysBAM::PAINT_DRIVER<TV>::Tall_Particle_Intersect_Solid(const int p,const ARRAY<RIGID_BODY<TV>* >& rigid_bodies,/*rst*/int& id)
//{
//    TV pos=example.particles.X(p)+canvas_normal*example.particles.thickness(p);
//    return SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Solid_Inside(pos,rigid_bodies,id);
//}
//
//template<class TV>
//bool PhysBAM::PAINT_DRIVER<TV>::Tall_Particle_Intersect_Solid_And_Value(const int p,const ARRAY<RIGID_BODY<TV>* >& rigid_bodies,/*rst*/int id,/*rst*/T& phi,/*rst*/TV& rigid_normal,/*rst*/TV& rigid_v,const T contour/*=(T)1e-10*/)
//{
//    TV pos=example.particles.X(p)+canvas_normal*example.particles.thickness(p);
//    return SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Solid_Inside_And_Value(pos,rigid_bodies,id,phi,rigid_normal,rigid_v,contour);
//}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Update_Particle_Incompressibility()
{
    /////update particle incompressibility due to kinematic object
    if(use_particle_incompressibility_for_solids){
        ARRAY<bool>& triangle_incompressibilities=example.mesher.triangle_incompressibilities;
        TRIANGLE_MESH& triangle_mesh=example.geometrical_codimensional_mesh.triangle_mesh;
        CODIMENSIONAL_FLUID_PARTICLES<TV>& particles=example.particles;
        for(int i=1;i<=triangle_mesh.elements.m;i++){
            const TV_INT3& v=triangle_mesh.elements(i);int kinematic_rigid_body_id=0;
            if(SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Prism_Intersect_Solid(v,particles,kinematic_rigid_bodies,kinematic_rigid_body_id)){
                triangle_incompressibilities(i)=true;for(int d=1;d<=3;d++)particles.incompressibility(v(d))=true;
            }
        }
    }
    ////TODO: segments

    //////update particle incompressibility due to fiber
    if(use_particle_incompressibility_for_fiber)coupled_evolution.Update_Particle_Incompressibility();
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Update_Particle_Length_Scale()
{
    if(use_adaptive_length_scale_on_codim)Update_Particle_Length_Scale_From_Codimensions();
    if(use_adaptive_length_scale_on_rim)Update_Particle_Length_Scale_From_Rim();
    if(use_adaptive_length_scale_based_on_test)Update_Particle_Length_Scale_Based_On_Tests();
    if(use_adaptive_length_scale_on_fiber)coupled_evolution.Update_Particle_Length_Scale();
    if(knife)coupled_evolution.Update_Particle_Length_Scale_Using_Rigid_Body(knife->rigid_body);
    if(ladle)coupled_evolution.Update_Particle_Length_Scale_Using_Rigid_Body(ladle->rigid_body);
    if(rock)coupled_evolution.Update_Particle_Length_Scale_Using_Rigid_Body(rock->rigid_body);
    if(scraper)coupled_evolution.Update_Particle_Length_Scale_Using_Rigid_Body(scraper->rigid_body);

    if(example.mpi_codim)example.mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(example.particles.length_scale);
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Update_Particle_Length_Scale_From_Codimensions()
{
    T tet_length_scale=length_scale;
    T tri_length_scale=adaptive_length_scale_min;
    T seg_length_scale=adaptive_length_scale_min;
    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles=example.particles;
    CODIMENSIONAL_MESH& mesh=example.geometrical_codimensional_mesh;

    ARRAY<bool> flag_changeable_low_dim(particles.X.m);flag_changeable_low_dim.Fill(true);
    ARRAY<bool> flag_changeable_high_dim(particles.X.m);flag_changeable_high_dim.Fill(true);
    for(int i=1;i<=particles.X.m;i++){
        TV_INT3 on_dim;
        int p_type=mesh.node_types(i);
        on_dim(1)=Is_On_Segment_Mesh(p_type);
        on_dim(2)=Is_On_Triangle_Mesh(p_type);
        on_dim(3)=Is_On_Tetrahedron_Mesh(p_type);
        if(on_dim(3)/*&&tet_length_scale<particles.length_scale(i)*/){particles.length_scale(i)=tet_length_scale;}
        else if(on_dim(2)&&tri_length_scale<particles.length_scale(i)){particles.length_scale(i)=tri_length_scale;}
        else if(on_dim(1)&&seg_length_scale<particles.length_scale(i)){particles.length_scale(i)=seg_length_scale;}

        if(on_dim(1)||on_dim(2))flag_changeable_low_dim(i)=false;
        if(on_dim(3)){flag_changeable_high_dim(i)=false;flag_changeable_low_dim(i)=false;}
    }
    
    SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL<TV>::Smooth_Particle_Length_Scale(mesh,particles,2,/*refine*/false,&flag_changeable_low_dim);
    SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL<TV>::Smooth_Particle_Length_Scale(mesh,particles,2,/*coarsen*/true,&flag_changeable_high_dim);
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Update_Particle_Length_Scale_From_Rim()
{
    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles=example.particles;
    CODIMENSIONAL_MESH& mesh=example.geometrical_codimensional_mesh;
    T rim_length_scale=adaptive_length_scale_min;
    ARRAY<bool> flag_changeable_rim(particles.X.m);flag_changeable_rim.Fill(true);
    ARRAY<bool> flag_changeable_high_dim(particles.X.m);flag_changeable_high_dim.Fill(true);
    for(int i=1;i<=particles.X.m;i++){
        int p_type=mesh.node_types(i);
        bool is_eulerian_particle=example.mesher.eulerian_particles_hashtable.Contains(particles.guid(i));
        if(mesh.Is_Rim_Particle(i)){
            particles.length_scale(i)=rim_length_scale;
            flag_changeable_rim(i)=false;
        }
        else if(is_eulerian_particle){
            particles.length_scale(i)=length_scale;
            flag_changeable_rim(i)=false;
        }
        else{particles.length_scale(i)=length_scale;}

        if(Is_On_Tetrahedron_Mesh(p_type)){flag_changeable_high_dim(i)=false;}
    }

    SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL<TV>::Smooth_Particle_Length_Scale(mesh,particles,3,/*refine*/false,&flag_changeable_rim);
    SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL<TV>::Smooth_Particle_Length_Scale(mesh,particles,3,/*coarsen*/true,&flag_changeable_high_dim);
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Add_Velocity_Independent_Forces(const T dt,const T time)
{
    BASE::Add_Velocity_Independent_Forces(dt,time);
    if(has_friction&&!use_damped_friction){
        for(int i=1;i<=example.particles.X.m;i++)example.particles.rim_surface_tension_force(i)=TV(); ////TODO: change attribute name to friction force
        Add_Friction_Forces(example.particles.rim_surface_tension_force,dt,time);
        Apply_Forces(example.particles.rim_surface_tension_force,dt,time);
        for(int i=1;i<=example.particles.X.m;i++)example.particles.F(i)+=example.particles.rim_surface_tension_force(i);
    } ////friction force
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Add_Velocity_Dependent_Forces(const T dt,const T time)
{
    BASE::Add_Velocity_Dependent_Forces(dt,time);

    if(has_fiber_fluid_coupling_force){
        for(int i=1;i<=example.particles.X.m;i++)example.particles.body_force(i)=TV();
        coupled_evolution.Add_Fiber_Fluid_Coupling_Forces(example.particles.body_force,dt,time);
        Apply_Forces(example.particles.body_force,dt,time);
        for(int i=1;i<=example.particles.X.m;i++)example.particles.F(i)+=example.particles.body_force(i);
    }
    
    if(has_friction&&use_damped_friction){
        ARRAY_VIEW<TV>& friction_force=example.particles.film_surface_tension_force;    ////dbg
        for(int i=1;i<=example.particles.X.m;i++)friction_force(i)=TV();
        Add_Friction_Forces(friction_force,dt,time);
        
        if(use_low_fric_near_fiber)for(int i=1;i<=example.particles.X.m;i++)if(has_fiber_fluid_coupling_force){
            if(Is_Contact_With_Solid3(example.particles.contact(i))){
                friction_force(i)=TV(); ////let particles contact fiber endpoints move with the fiber
            }
        }

        Apply_Forces(friction_force,dt,time);
        for(int i=1;i<=example.particles.X.m;i++)example.particles.F(i)+=friction_force(i);
    } ////damped friction force
    
    //Enforce_Solid_Boundary_Velocities();
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Update_Particle_Contact()
{
    BASE::Update_Particle_Contact();

    ////update particle contact solid2
    TV canvas_normal=TV::Axis_Vector(2);
    if(kinematic_rigid_bodies.m>0)for(int i=1;i<=example.particles.X.m;i++){int kin_id=0;
        if(example.particles.thickness(i)>(T)0&&SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Tall_Particle_Intersect_Solid(i,example.particles,kinematic_rigid_bodies,kin_id)){
            Set_Contact_With_Solid2(example.particles.contact(i));
        }
    }
    if(fiber.fiber_particles.X.m>0)for(int i=1;i<=example.particles.X.m;i++){int seg_id=0;
        if(example.particles.thickness(i)>(T)0&&SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Tall_Particle_Intersect_Solid(i,example.particles,fiber,seg_id)){
            //LOG::cout<<"seg "<<i<<": "<<seg_id<<std::endl;
            Set_Contact_With_Solid2(example.particles.contact(i));
            if(fiber.Is_End_Segment(seg_id)){
                Set_Contact_With_Solid3(example.particles.contact(i));
                //LOG::cout<<"set contact3: "<<seg_id<<std::endl;
            }
        }
    }
    
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Update_Particle_Inactive()
{
    if(use_inactive_particle){
        //switch(particle_inactive_mode){
        //case SMALL_VELOCITY:{

        //}break;
        //case FAR_FROM_FOCUS:{
        //    coupled_evolution.Update_Particle_Inactive();
        //}break;
        //}
        Update_Inactive_Particle_Based_On_Tests();
    }
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Update_Particle_Meshing_Direction()
{
    if(example.mesher.use_anisotropic_triangle_remeshing){
        //for(int i=1;i<=example.particles.X.m;i++){  ////for test
        //    example.particles.meshing_direction(i)=TV((T)1,(T)0,(T).3).Normalized();
        //}
        coupled_evolution.Update_Particle_Meshing_Direction();
    }
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Squash_Fluid()
{
    ARRAY<T> target_thickness;target_thickness.Resize(example.particles.X.m);
    for(int i=1;i<=example.particles.X.m;i++)target_thickness(i)=example.particles.thickness(i);

    T min_triangle_thickness=example.mesher.thickness_scale*(T)1e-3;
    for(int i=1;i<=example.particles.X.m;i++){
        int kin_rigid_body_id=0;
        if(SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Tall_Particle_Intersect_Solid(i,example.particles,kinematic_rigid_bodies,kin_rigid_body_id)){
            TV p1=example.particles.X(i);TV p2=example.particles.X(i)+example.particles.thickness(i)*canvas_normal;
            T t=SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Segment_Intersection_Fraction_With_Solid(p1,p2,*kinematic_rigid_bodies(kin_rigid_body_id));
            target_thickness(i)=clamp_min(example.particles.thickness(i)*t,min_triangle_thickness);
            //LOG::cout<<"squash t "<<i<<": "<<t<<", "<<p1<<", "<<p2<<", "<<kin_rigid_body_id<<std::endl;
        }
    }

    ARRAY<T> delta_mass;delta_mass.Resize(example.particles.X.m);delta_mass.Fill((T)0);
    //int number_of_rings=3;
    /*for(int iter=1;iter<=3;iter++)*/for(int i=1;i<=example.particles.X.m;i++){
        if(target_thickness(i)==example.particles.thickness(i)||example.particles.thickness(i)==(T)0)continue;
        T particle_delta_frac=clamp_min((example.particles.thickness(i)-target_thickness(i))/example.particles.thickness(i),(T)0);
        T particle_delta_mass=example.particles.dimensional_mass(i)(TV::m-1)*particle_delta_frac;
        delta_mass(i)-=particle_delta_mass;

        //T max_distance_squared=pow<2>(example.particles.length_scale(i)*(T)number_of_rings);
        //ARRAY<int> nbs;Find_Neighbor_Particles_On_Mesh(fluid_mesh.triangle_mesh,i,nbs,number_of_rings);
        //ARRAY<T> weights(nbs.m);T total_weight=(T)0;
        //for(int j=1;j<=nbs.m;j++){int nb_p=nbs(j);if(nb_p==i)continue;
        //T distance_squared=(example.particles.X(i)-example.particles.X(nb_p)).Magnitude_Squared();
        //T normalized_distance_squared=clamp(distance_squared/max_distance_squared,(T)0,(T)1);
        //T frac=Gaussian_Kernel(normalized_distance_squared);
        //weights(j)=frac;total_weight+=frac;
        //}
        //for(int j=1;j<=nbs.m;j++){weights(j)/=total_weight;}
        //for(int j=1;j<=nbs.m;j++){int nb_p=nbs(j);delta_mass(nb_p)+=particle_delta_mass*weights(j);}
    }

    for(int i=1;i<=example.particles.X.m;i++){
        example.particles.dimensional_mass(i)(TV::m-1)+=(T).99*delta_mass(i);
        example.particles.mass(i)=example.particles.dimensional_mass(i).Sum();
    }   
    example.mesher.Update_Particle_Thickness(example.mesher.has_rim);
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Add_Particles_And_Mesh(const ARRAY<TV>& added_X,const CODIMENSIONAL_MESH& added_mesh,const ARRAY<TV>* added_V/*=0*/,const ARRAY<T>* added_thickness/*=0*/,const ARRAY<ARRAY<T> >* added_phis/*=0*/)
{
    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles=example.particles;
    CODIMENSIONAL_MESH& mesh=example.geometrical_codimensional_mesh;
    int offset=example.particles.X.m;
    for(int i=1;i<=added_X.m;i++){
        int v=particles.Add_Particle();
        particles.X(v)=added_X(i);
        particles.length_scale(v)=length_scale;

        if(added_V)particles.V(v)=(*added_V)(i);
        if(added_thickness){particles.thickness(v)=(*added_thickness)(i);}
        if(added_phis)for(int j=1;j<=levelset_region_number;j++)particles.phis(j)(v)=(*added_phis)(j)(i);

        if(example.mesher.use_particle_parents){
            particles.parents_number(v)=1;particles.parents_guid(1)(v)=particles.guid(v);particles.parents_weight(1)(v)=(T)1;
            for(int j=2;j<=particles.parents_size;j++){
                particles.parents_guid(j)(v)=0;particles.parents_weight(j)(v)=(T)0;
            }
        }
    }
    for(int i=1;i<=added_mesh.tetrahedron_mesh.elements.m;i++){
        TV_INT4 v=added_mesh.tetrahedron_mesh.elements(i);
        for(int d=1;d<=4;d++)v(d)+=offset;
        mesh.tetrahedron_mesh.elements.Append(v);
    }
    for(int i=1;i<=added_mesh.triangle_mesh.elements.m;i++){
        TV_INT3 v=added_mesh.triangle_mesh.elements(i);
        for(int d=1;d<=3;d++)v(d)+=offset;
        mesh.triangle_mesh.elements.Append(v);
    }
    for(int i=1;i<=added_mesh.segment_mesh.elements.m;i++){
        TV_INT2 v=added_mesh.segment_mesh.elements(i);
        for(int d=1;d<=2;d++)v(d)+=offset;
        mesh.segment_mesh.elements.Append(v);
    }

    LOG::cout<<"add particles: "<<example.particles.X.m-offset<<std::endl;
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Add_Cheese_Filaments()
{
    ARRAY<TV> added_X;CODIMENSIONAL_MESH added_mesh;ARRAY<T> added_thickness;
    Add_Cheese_Filaments_Helper(added_X,added_mesh,0,&added_thickness,0);
    Add_Particles_And_Mesh(added_X,added_mesh,0,&added_thickness);
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Add_Cheese_Filaments_Helper(ARRAY<TV>& added_X,CODIMENSIONAL_MESH& added_mesh,ARRAY<TV>* added_V/*=0*/,ARRAY<T>* added_thickness/*=0*/,ARRAY<ARRAY<T> >* added_phis/*=0*/)
{
    T period=(T).02;static int c=0;static T next_t0=(T)0;
    bool flag=false;if(time>=next_t0){c++;next_t0=period*(T)c;flag=true;}if(!flag)return;

    PAINT_TUBE<TV>* tube=tubes(1);
    //TV pos=tube->e1;TV pos2=pos+TV::Axis_Vector(1)*length_scale*(T)5;
    //CODIMENSIONAL_MESH_GENERATOR<TV>::Filament(added_mesh,added_X,pos,pos2,length_scale);
    
    int n_min=1;int n_max=(int)((T)20/scale);
    int n=n_min+(int)((T)n_max*tube->source_speed);

    for(int k=1;k<=n;k++){
        T h=(T)tube->height*random.Get_Uniform_Number((T)3,(T)4);
        TV r=tube->radius*random.template Get_Vector_In_Unit_Sphere<TV>();r.y=(T)0;
        TV start_pos=TV(tube->e1.x+r.x,tube->e1.y-h,tube->e1.z+r.z);

        T length_min=length_scale*(T)5;T length_max=length_scale*(T)10;
        T length=random.Get_Uniform_Number(length_min,length_max);

        ARRAY<PAIR<T,TV> > time_value;BSPLINE<T,TV>* bspline;
        TV dir1=random.template Get_Direction<TV>();dir1.y*=(T).6;dir1.Normalize();
        TV dir2=random.template Get_Direction<TV>();dir2.y*=(T).6;dir2.Normalize();
        Add_Bspline_Time_Value_Helper((T)0,dir1,time_value);
        T frac2=(T)1;bool has_dir3=random.Get_Uniform_Number((T)0,(T)1)>(T).75;if(has_dir3)frac2=(T).5;
        Add_Bspline_Time_Value_Helper(frac2,dir2,time_value);
        if(has_dir3){
            TV dir3=random.template Get_Direction<TV>();dir3.y*=(T).6;dir3.Normalize();
            Add_Bspline_Time_Value_Helper((T)1,dir3,time_value);
        }
        Initialize_Bspline_Helper(time_value,bspline);

        int start=added_X.m+1;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Bended_Filament(added_mesh,added_X,start_pos,length,length_scale,*bspline);
        int seg_p_n=added_X.m-start+1;
        delete bspline;

        if(added_thickness){
            ARRAY<VECTOR<T,2> > time_value;BSPLINE<T,T>* thickness_bspline;
            bool has_sharp_end1=random.Get_Uniform_Number((T)0,(T)1)>(T).75;
            bool has_sharp_end2=random.Get_Uniform_Number((T)0,(T)1)>(T).75;
            T frac_sharp=(T).3;
            if(has_sharp_end1){
                Add_Bspline_Time_Value_Helper((T)0,(T).2,time_value);
                Add_Bspline_Time_Value_Helper(frac_sharp*(T).5,(T).4,time_value);
                Add_Bspline_Time_Value_Helper(frac_sharp,(T)1,time_value);
            }
            else Add_Bspline_Time_Value_Helper((T)0,(T)1,time_value);

            if(has_sharp_end2){
                Add_Bspline_Time_Value_Helper((T)1-frac_sharp,(T)1,time_value);
                Add_Bspline_Time_Value_Helper((T)1-frac_sharp*(T).5,(T).4,time_value);
                Add_Bspline_Time_Value_Helper((T)1,(T).2,time_value);
            }
            else Add_Bspline_Time_Value_Helper((T)1,(T)1,time_value);
            Add_Bspline_Time_Value_Helper((T)2,(T)1,time_value);

            Initialize_Bspline_Helper(time_value,thickness_bspline);

            for(int i=1;i<=seg_p_n;i++){
                T frac=(T)(i-1)/(T)(seg_p_n-1);
                T thickness_frac=thickness_bspline->Evaluate(frac);
                T thickness=thickness_frac*thickness_scale;
                (*added_thickness).Append(thickness);
            }
            delete thickness_bspline;
        }
    }
}

template<class TV>
TV PhysBAM::PAINT_DRIVER<TV>::Read_Tube_Initial_Position_From_File(const std::string& file_name)
{
    std::string str;STREAM_TYPE stream_type((T()));std::fstream in(file_name.c_str(),std::ios::in);
    if(!in){LOG::cout<<"cannot open prescribed motion file "<<file_name<<std::endl;}

    T t=0;TV x=TV();
    if(in>>t>>x.x>>x.y>>x.z){return x;}else{LOG::cout<<"fail reading prescribed motion file "<<file_name<<std::endl;return TV();}
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Write_Tube_Eulerian_Particles_To_File(PAINT_TUBE<TV>& tube,const std::string& file_name)
{
    std::string str;STREAM_TYPE stream_type((T()));std::fstream out(file_name.c_str(),std::ios::out);
    if(!out){LOG::cout<<"cannot open eulerian particle file to write "<<file_name<<std::endl;return;}

    out<<tube.eulerian_particles.Size()<<std::endl;
    for(HASHTABLE_ITERATOR<int,typename PAINT_TUBE<TV>::EULERIAN_PARTICLE> iter(tube.eulerian_particles);iter.Valid();iter.Next()){
        int guid=iter.Key();typename PAINT_TUBE<TV>::EULERIAN_PARTICLE ep=iter.Data();
        out<<guid<<" "<<ep.r<<" "<<ep.angle<<" "<<ep.object_space_pos.x<<" "<<ep.object_space_pos.y<<" "<<ep.object_space_pos.z<<" "<<
            ep.object_space_pos0.x<<" "<<ep.object_space_pos0.y<<" "<<ep.object_space_pos0.z<<" "<<ep.has_constant_thickness<<" "<<ep.thickness<<" "<<
            thickness_scale<<" "<<ep.texture_coordinates.x<<" "<<ep.texture_coordinates.y<<" "<<ep.texture_coordinates.z<<std::endl;
    }
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Read_Tube_Eulerian_Particles_From_File(PAINT_TUBE<TV>& tube,const std::string& file_name)
{
    std::string str;STREAM_TYPE stream_type((T()));std::fstream in(file_name.c_str(),std::ios::in);
    if(!in){LOG::cout<<"cannot open eulerian particle file to read "<<file_name<<std::endl;return;}

    tube.eulerian_particles.Remove_All();
    int size=0;in>>size;
    for(int i=1;i<=size;i++){
        int guid;typename PAINT_TUBE<TV>::EULERIAN_PARTICLE ep;
        in>>guid>>ep.r>>ep.angle>>ep.object_space_pos.x>>ep.object_space_pos.y>>ep.object_space_pos.z>>
            ep.object_space_pos0.x>>ep.object_space_pos0.y>>ep.object_space_pos0.z>>ep.has_constant_thickness>>ep.thickness>>
            thickness_scale>>ep.texture_coordinates.x>>ep.texture_coordinates.y>>ep.texture_coordinates.z;
        tube.eulerian_particles.Set(guid,ep);
    }
}

//template class PAINT_DRIVER<VECTOR<float,2> >;
template class PAINT_DRIVER<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
//template class PAINT_DRIVER<VECTOR<double,2> >;
template class PAINT_DRIVER<VECTOR<double,3> >;
#endif
