//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __PAINT_TUBE_DRIVER__
#define __PAINT_TUBE_DRIVER__
#include <PhysBAM_Geometry/Basic_Geometry/CYLINDER.h>
#include <PhysBAM_Tools/Interpolation/BSPLINE.h>
#include <PhysBAM_Tools/Matrices/FRAME.h>
#include <PhysBAM_Tools/Vectors/TWIST.h>
#include <PhysBAM_Tools/Random_Numbers/RANDOM_NUMBERS.h>
#include <PhysBAM_Dynamics/Particles/CODIMENSIONAL_FLUID_PARTICLES.h>
#include <PhysBAM_Solids/PhysBAM_Rigids/Rigid_Bodies/RIGID_BODY.h>
#include "../Meshing/CODIMENSIONAL_MESH.h"
#include "../Meshing/CODIMENSIONAL_MESH_GENERATOR.h"
#include "../Viscoelastic/FIBER.h"

namespace PhysBAM
{
template<class TV>
class PAINT_TUBE
{
    typedef typename TV::SCALAR T;typedef typename TV::template REBIND<int>::TYPE TV_INT;typedef VECTOR<T,2> TV2;
public:
    class EULERIAN_PARTICLE 
    {
        public:int guid;T r;T angle;TV object_space_pos;TV object_space_pos0;bool has_constant_thickness;T thickness;TV texture_coordinates;ARRAY<T> phis;
        EULERIAN_PARTICLE(const int guid_input=0,const T r_input=(T)0,const T angle_input=(T)0,const TV& object_space_pos_input=TV(),const int levelset_region_number=4)
            :guid(guid_input),r(r_input),angle(angle_input),object_space_pos(object_space_pos_input),object_space_pos0(object_space_pos),has_constant_thickness(false){phis.Resize(levelset_region_number);}
        EULERIAN_PARTICLE& operator=(const EULERIAN_PARTICLE& copy)
        {guid=copy.guid;r=copy.r;angle=copy.angle;object_space_pos=copy.object_space_pos;object_space_pos0=copy.object_space_pos0;
        has_constant_thickness=copy.has_constant_thickness;thickness=copy.thickness;texture_coordinates=copy.texture_coordinates;phis=copy.phis;return *this;}

        EULERIAN_PARTICLE(const EULERIAN_PARTICLE& copy){*this=copy;}
        void Set_Constant_Thickness(const T thickness_input){has_constant_thickness=true;thickness=thickness_input;}
    };

    TV e0;
    TV e1;
    TV dir;
    TV axis1;
    T tube_height;  ////tube height
    T height;   ////paint height
    T radius;
    CYLINDER<T> cylinder;
    T length_scale;
    T source_speed;
    T source_thickness_factor;
    FRAME<TV> object_frame;     ////translation and rotation
    TWIST<TV> object_twist;     ////linear and angular velocity
    TV object_space_dir;
    TV object_space_axis1;
    TV object_space_e0;
    TV object_space_e1;
    HASHTABLE<int,TV> eulerian_particle_object_space_positions;    
    T eulerian_frac;
    HASHTABLE<int,EULERIAN_PARTICLE> eulerian_particles;    ////hashtable<guid,radical_distance>

    enum PAINT_TUBE_RIGID_BODY_TYPE{NONE=0,PLACEHOLDER,RENDERABLE} rigid_body_type;
    ARRAY<RIGID_BODY<TV>*> rigid_bodies;
    T rigid_body_height;
    T rigid_body_radius;

    bool use_prescribed_motion;
    PRESCRIBED_MOTION<TV>& prescribed_motion;
    bool enabled;
    bool use_ring_source;
    bool use_texture_coordinates;
    bool initialize_texture_coordinates_box;
    bool has_rigid_body;
    bool use_reinitialize_eulerian_particles;

    ////ring source
    T ring_source_radius_max;
    T ring_source_angle;
    int ring_source_peak_num;
    bool use_ring_source_randomness;

    ////bsplines
    ARRAY<VECTOR<T,2> > source_speed_bspline_time_value; 
    BSPLINE<T,T>* source_speed_bspline;
    ARRAY<VECTOR<T,2> > source_thickness_factor_bspline_time_value;
    BSPLINE<T,T>* source_thickness_factor_bspline;
    ARRAY<VECTOR<T,2> > ring_source_speed_bspline_time_value; 
    BSPLINE<T,T>* ring_source_speed_bspline;
    ARRAY<VECTOR<T,2> > source_thickness_bspline_time_value;
    BSPLINE<T,T>* ring_source_thickness_bspline;

    ////texture
    TV texture_coordinate_space_scale;

    ////bounding box
    T object_space_x_max;
    T object_space_r;
    T object_space_z_rescale;
    T object_space_x_rescale;

    RANDOM_NUMBERS<T> random;

    PAINT_TUBE(const TV& e1_input,const TV& dir_input,const T tube_height_input,const T paint_height_input,const T radius_input,const T length_scale_input,const T source_speed_input,const T eulerian_frac_input,
        const TV object_space_dir_input=TV::Axis_Vector(2)*(T)-1,PRESCRIBED_MOTION<TV>* prescribed_motion_input=0)
        :e1(e1_input),dir(dir_input.Normalized()),tube_height(tube_height_input),height(paint_height_input),radius(radius_input),cylinder(e1,e1+dir*height,radius),length_scale(length_scale_input),
        source_speed(source_speed_input),source_thickness_factor((T)1),object_space_dir(object_space_dir_input),object_space_axis1(TV::Axis_Vector(1)),object_space_e0(TV()),object_space_e1(object_space_dir*tube_height),eulerian_frac(eulerian_frac_input),
        rigid_body_type(NONE),rigid_body_height((T)0),rigid_body_radius((T)0),use_prescribed_motion(false),prescribed_motion(prescribed_motion_input==0?*(new PRESCRIBED_MOTION<TV>()):*prescribed_motion_input),enabled(true),use_ring_source(false),use_texture_coordinates(false),initialize_texture_coordinates_box(false),
        has_rigid_body(true),use_reinitialize_eulerian_particles(false),ring_source_angle((T)0),ring_source_peak_num(5),use_ring_source_randomness(false),source_speed_bspline(0),source_thickness_factor_bspline(0),ring_source_speed_bspline(0),ring_source_thickness_bspline(0),texture_coordinate_space_scale(TV::Constant_Vector(1))
    {
        //if(dir.y>0)object_space_dir=TV::Axis_Vector(2);
        e0=e1-dir*tube_height;
        object_frame.t=e0-object_space_e0;
        object_frame.r=ROTATION<TV>::From_Rotated_Vector(object_space_dir,dir);
        axis1=object_frame.Rotation().Rotate(object_space_axis1);
        random.Set_Seed(120111021);

        object_space_x_max=radius*(T)1.11;
        object_space_r=radius*(T)1.11;
        object_space_z_rescale=(T)1;
        object_space_x_rescale=(T)1;
    }

    void Initialize_Volumetric_Mesh_And_X_And_V(CODIMENSIONAL_MESH& mesh,ARRAY<TV>& X,ARRAY<TV>& V,const T hemisphere_frac=(T).8);
    void Initialize_Volumetric_Thin_Sheet_Mesh_And_X_And_V(CODIMENSIONAL_MESH& mesh,ARRAY<TV>& X,ARRAY<TV>& V,const T thickness);
    void Initialize_Volumetric_Rounded_Thin_Sheet_Mesh_And_X_And_V(CODIMENSIONAL_MESH& mesh,ARRAY<TV>& X,ARRAY<TV>& V,const T thickness);
    void Initialize_Thin_Sheet_Mesh_And_X_And_V(CODIMENSIONAL_MESH& mesh,ARRAY<TV>& X,ARRAY<TV>& V);
    void Initialize_Filament_Mesh_And_X_And_V(CODIMENSIONAL_MESH& mesh,ARRAY<TV>& X,ARRAY<TV>& V);
    void Initialize_Ring_Mesh_And_X_And_V(CODIMENSIONAL_MESH& mesh,ARRAY<TV>& X,ARRAY<TV>& V);
    void Initialzie_Eulerian_Particles(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,T frac);
    void Initialize_Source_Speed_Bspline();
    void Initialize_Source_Speed_Bspline_From_File(const std::string& file_name);
    void Initialize_Source_Thickness_Factor_Bspline();
    void Initialize_Texture_Coordinates(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles);
    void Initialize_Ring_Phis(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,const int number_of_regions);
    
    TV Source_Velocity(const T r=(T)0,const T a=(T)0,const T time=(T)0);
    T Source_Thickness_Factor(const T r=(T)0,const T a=(T)0,const T time=(T)0);
    void Advance(const T dt,const T time);
    void Advance_Texture_Coordinates(const T dt,const T time);

    void Update_Position_Based_States();
    void Update_Eulerian_Particle_Velocities(HASHTABLE<int,TV>& particle_velocities);
    void Update_Eulerian_Particle_Positions_And_Mass(CODIMENSIONAL_MESH_OBJECT<TV>& mesh_object,T min_tri_thickness,T min_seg_thickness,T time=(T)0);
    void Update_Eulerian_Particle_Phis(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles);
    void Update_Eulerian_Particle_Texture_Coordinates(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles);
    void Reinitialize_Eulerian_Particles(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles);
    void Add_Eulerian_Particles_From_Mesher(const HASHTABLE<int>& eulerian_particles_hashtable,HASHTABLE<int,PAIR<VECTOR<int,2>,T> > added_eulerian_particle_parents_hashtable,const CODIMENSIONAL_FLUID_PARTICLES<TV>& particles);

    void Add_Source_Speed_Bspline_Time_Value(const T time,const T value);
    void Add_Source_Thickness_Factor_Bspline_Time_Value(const T time,const T value);
    void Enable(bool enable,CODIMENSIONAL_FLUID_PARTICLES<TV>* particles=0);

    TV Cylindrical_Normal(const TV& pos);
    void Cylindrical_Coordinates(const TV& pos,/*rst*/T& axial_distance,/*rst*/T& radical_distance,/*rst*/T& angle);

    void Initialize_Ring_Source_Speed_Bspline();
    void Add_Ring_Source_Speed_Bspline_Time_Value(const T time,const T value);
    void Set_Constant_Thickness(const T thickness);

    T Angular_Noise(const T angle);

    void Remove_All_Eulerian_Particles();

protected:
    void Initialize_V_Helper(const ARRAY<TV>& X,/*rst*/ARRAY<TV>& V);
};
}
#endif
